import numpy as np
from sklearn.metrics import silhouette_samples
from sklearn.utils import check_random_state
import utils

class AffinityMatrix(object):
    PARTITION_SIMILARITY_METRICS = {
        'sil': "Silhouette score",
        'avg_cl_sim': "Average cluster similarity",
    }
    def __init__(self, matrix, objects_names, is_distance_matrix=False):
        self._matrix = matrix
        self._objects_names = objects_names
        self._is_distance_matrix = is_distance_matrix

    def get_matrix(self):
        return self._matrix

    def get_objects_names(self):
        return self._objects_names

    def get_is_distance_matrix(self):
        return self._is_distance_matrix

    def get_objects_count(self):
        return len(self.get_objects_names())

    @staticmethod
    def get_similarity_metric_limits(metric=None):
        return (-1, 1) if metric=='sil' else (0,None)

    @staticmethod
    def from_affinities_list(affinities_list, **kwargs):
        objects_names = set()
        for o_1,o_2,aff in affinities_list:
            objects_names.add(o_1)
            objects_names.add(o_2)
        objects_names = sorted(objects_names)
        objects_indices = dict([(o,i) for i,o in enumerate(objects_names)])

        matrix = np.zeros((len(objects_names), len(objects_names)))
        for o_1,o_2,aff in affinities_list:
            matrix[objects_indices[o_1], objects_indices[o_2]] = aff
            if kwargs.get('is_symmetric', True):
                matrix[objects_indices[o_2], objects_indices[o_1]] = aff

        return AffinityMatrix(matrix, objects_names, is_distance_matrix=kwargs.get('is_distance_matrix', False))
    @staticmethod
    def from_string(input_string, **kwargs):
        lines = input_string.split(kwargs.get('record_separator'))
        return AffinityMatrix.from_lines(lines, 0, **kwargs)
    @staticmethod
    def from_lines(lines, i_start, **kwargs):
        n = kwargs.get('n_records')
        val_separator = kwargs.get('value_separator', ',')
        is_distance_matrix =  kwargs.get('is_distance_matrix', False)
        i_next = 0
        if n is None:
            if val_separator in lines[i_start]:
                values = lines[i_start].strip().split(val_separator)
                n = int(values[0])
                is_distance_matrix = values[1] == "is_distance_matrix"
            else:
                n = int(lines[i_start])
            i_next += 1

        convert_objects_to_ints = kwargs.get('convert_objects_to_ints', False)
        affinities = []
        for i in xrange(i_start + i_next, i_start + n):
            values = lines[i].strip().split(val_separator)
            object_id_1 = values[0] if not convert_objects_to_ints else int(values[0])
            object_id_2 = values[1] if not convert_objects_to_ints else int(values[1])
            affinity = float(values[2])
            affinities.append((object_id_1, object_id_2, affinity))

        return AffinityMatrix.from_affinities_list(affinities, is_distance_matrix=is_distance_matrix)

    def to_string(self, record_separator="\n"):
        res = "%s" % (self.get_objects_count())
        if self.get_is_distance_matrix():
            res += ",is_distance_matrix"
        res += record_separator

        for i,obj_1 in enumerate(self.get_objects_names()):
            for j,obj_2 in enumerate(self.get_objects_names()):
                if self.get_matrix()[i,j] > 0:
                    res += "%s,%s,%s%s" % (obj_1, obj_2, self.get_matrix()[i,j], record_separator)
        return res

    def get_submatrix(self, objects_sub_list, as_distance_matrix=False):
        if objects_sub_list is not None:
            indices = []
            for o in objects_sub_list:
                indices.append(self.get_objects_names().index(o))
            res = self.get_matrix()[indices,:][:,indices]
            if self.get_is_distance_matrix() == as_distance_matrix:
                return res
        else:
            res = self.get_matrix().copy()
        if self.get_is_distance_matrix() != as_distance_matrix:
            res = np.max(self.get_matrix()) - res
            if as_distance_matrix:
                for i in xrange(res.shape[0]):
                    res[i, i] = 0.0
        return res
    @staticmethod
    def compute_partition_internal_quality(partition, affinity_matrix, metric='sil', **kwargs):
        if metric not in AffinityMatrix.PARTITION_SIMILARITY_METRICS:
            raise Exception("Unknown metric {metric}, should be one of {metrics}".format(
                    metric=metric,
                    metrics=AffinityMatrix.PARTITION_SIMILARITY_METRICS.keys()
                )
            )

        if len(affinity_matrix.get_objects_names()) == 0:
            return None

        common_objects = utils.get_intersection(partition.get_objects_names(), affinity_matrix.get_objects_names())
        sub_partition = partition.get_subpartition(common_objects)

        common_partition_array, common_objects = sub_partition.to_array_form()

        if metric == 'sil':
            common_matrix = affinity_matrix.get_submatrix(common_objects, as_distance_matrix=True)
            return AffinityMatrix._silh_with_nan(
                X=common_matrix,
                labels=np.array(common_partition_array),
                metric='precomputed',
            )
        elif metric =='avg_cl_sim':
            common_matrix = affinity_matrix.get_submatrix(common_objects, as_distance_matrix=False)
            res = 0.0
            for cl,objs in sub_partition.get_clusters_objects().iteritems():
                for o1 in objs:
                    i = common_objects.index(o1)
                    o_res = 0.0
                    for o2 in objs:
                        if o1==o2:
                            continue
                        j = common_objects.index(o2)
                        o_res += common_matrix[i,j]
                    if len(objs) > 1:
                        o_res /= (len(objs) - 1)
                    res += o_res
            res /= (sub_partition.get_objects_count())
            return res

    # from sklearn - to handle nan values
    @staticmethod
    def _silh_with_nan(X, labels, metric='euclidean', sample_size=None,
                     random_state=None, **kwds):
        n_labels = len(np.unique(labels))
        n_samples = X.shape[0]
        if not 1 < n_labels < n_samples:
            return np.nan
            # raise ValueError("Number of labels is %d. Valid values are 2 "
            #                  "to n_samples - 1 (inclusive)" % n_labels)

        if sample_size is not None:
            random_state = check_random_state(random_state)
            indices = random_state.permutation(X.shape[0])[:sample_size]
            if metric == "precomputed":
                X, labels = X[indices].T[indices].T, labels[indices]
            else:
                X, labels = X[indices], labels[indices]
        return np.mean(map(np.nan_to_num, silhouette_samples(X, labels, metric=metric, **kwds)))

