import os
import sys
from log import logger
import cPickle
import numpy as np
from scipy.stats import hmean
from sklearn.metrics import (
    precision_recall_curve,
    roc_curve,
    average_precision_score,
    roc_auc_score,
)

class ClassificationResult:
    def __init__(self, ground_truth_labels, predicted_scores):
        self._ground_truth_labels = ground_truth_labels
        self._predicted_scores = predicted_scores

        self._clear_cache()

    def _clear_cache(self):
        self._precisions = None
        self._recalls = None
        self._tprs = None
        self._fprs = None
        self._f1scores = None
        self._max_f1score = None
        self._aupr = None
        self._auroc = None

    def load(self, input_dir, input_file):
        pickle_path = input_dir + input_file + '.cpickle'
        if not os.path.exists(pickle_path):
            logger.warning("No saved classificaiton result in %s" % pickle_path)
        with open(pickle_path, 'r') as pickle_file:
            return cPickle.load(pickle_file)

    def save(self, output_dir, output_file):
        output_path = output_dir + output_file + '.cpickle'
        logger.debug("Saving classification result to %s" % output_path)
        if not os.path.exists(os.path.dirname(output_path)):
            os.makedirs(os.path.dirname(output_path))
        with open(output_path, 'wb') as pickle_file:
            self_without_cache = ClassificationResult(
                self.get_ground_truth_labels(),
                self.get_predicted_scores()
            )
            cPickle.dump(self_without_cache, pickle_file)

    def get_summary(self):
        return "AUPR={aupr:.5f}, AUROC={auroc:.5f}, Max F1-score={max_f1:.5f} ".format(
            aupr=self.get_aupr(),
            auroc=self.get_auroc(),
            max_f1=self.get_max_f1score(),
        )

    def get_ground_truth_labels(self):
        return self._ground_truth_labels

    def get_predicted_scores(self):
        return self._predicted_scores

    def _compute_precision_recall(self):
        self._precisions, self._recalls, _ = precision_recall_curve(
            y_true=self.get_ground_truth_labels(),
            probas_pred=self.get_predicted_scores(),
        )

    def get_precisions(self):
        if self._precisions is None:
            self._compute_precision_recall()
        return self._precisions

    def get_recalls(self):
        if self._recalls is None:
            self._compute_precision_recall()
        return self._recalls

    def _compute_roc(self):
        self._fprs, self._tprs, _ = roc_curve(
            y_true=self.get_ground_truth_labels(),
            y_score=self.get_predicted_scores(),
        )

    def get_tprs(self):
        if self._tprs is None:
            self._compute_roc()
        return self._tprs

    def get_fprs(self):
        if self._fprs is None:
            self._compute_roc()
        return self._fprs

    def get_f1scores(self):
        if self._f1scores is None:
            self._f1scores = [
                hmean([self.get_precisions()[i], self.get_recalls()[i]])
                if (self.get_precisions()[i] * self.get_recalls()[i]) > 0 else 0.0
                for i in xrange(len(self.get_precisions()))
            ]
        return self._f1scores

    def get_max_f1score(self):
        if self._max_f1score is None:
            self._max_f1score = max(self.get_f1scores())
        return self._max_f1score

    def get_aupr(self):
        if self._aupr is None:
            self._aupr = average_precision_score(
                y_true=self.get_ground_truth_labels(),
                y_score=self.get_predicted_scores(),
            )
        return self._aupr

    def get_auroc(self):
        if self._auroc is None:
            self._auroc = roc_auc_score(
                y_true=self.get_ground_truth_labels(),
                y_score=self.get_predicted_scores(),
            )
        return self._auroc


def compute_pr_roc(test_scores, ground_truth_scores, output_path, verbose=True):
    """
    Give two lists of scores - a test one and a ground truth one - compute the corresponding P/R values
    :param test_dict: Test scores; format: value => score
    :param ground_truth_dict: Ground truth scores; format: value => score
    :param output_path:
    :return:
    """

    p1, r1, _ = precision_recall_curve(
        y_true = ground_truth_scores,
        probas_pred=test_scores,
    )
    print average_precision_score(ground_truth_scores, test_scores)
    print roc_auc_score(ground_truth_scores, test_scores)
   # return p1, r1
    n = len(test_scores)
    #get list of values sorted by their score (from highest to lowest)
    test_args = np.argsort(test_scores)[::-1]
    gt_args = [ground_truth_scores[i] for i in test_args]

    #calculate precision/recall
    #for area under the curve
    precisions = []
    recalls = []
    tp_rates = []
    fp_rates = []

    if not os.path.exists(os.path.dirname(output_path)):
        os.makedirs(os.path.dirname(output_path))

    with open(output_path, 'wb') as outfile:
        outfile.write("k\tTP\tFP\tTN\tFN\tRecall\tPrecision\tTP rate\tFP rate\tF-score\tAccuracy\n")
        for i in _get_range_breakpoints(n, 100):
            k  = 1.0 * i / n
            tp = gt_args[0:i].count(1)
            fp = i - tp
            fn = gt_args[i:].count(1)
            tn = n - i - fn

            if tp + fn == 0:
                recall = 0
            else:
                recall = float(tp)/(tp+fn)
            if tp + fp == 0:
                precision = 1.0
            else:
                precision = float(tp)/(tp+fp)

            tp_rate = recall
            if fp + tn == 0:
                fp_rate = 0
            else:
                fp_rate = float(fp)/(fp+tn)

            if precision + recall == 0:
                f_score = 0
            else:
                f_score = (2.0 * precision * recall) / (precision + recall)

            if tp + fp + tn + fn == 0:
                accuracy = 0
            else:
                accuracy = float(tp + tn) / (tp + fp + tn + fn)

            outfile.write(str(k) + "\t" + str(tp) + "\t" + str(fp) + "\t" + str(tn) + "\t" + str(fn) + "\t" + str(recall) + "\t" + str(precision)
                          + "\t" + str(tp_rate) + "\t" + str(fp_rate)
                          + "\t" + str(f_score) + "\t" + str(accuracy) + "\n")

            if verbose:
                logger.warning(
                    "k=%.6f: tp=%d, fp=%d, tn=%d, fn=%d, precision=%.3f, "
                    "recall=%.3f, tp-rate=%.3f, fp-rate=%.3f, "
                    "f-score=%.3f, accuracy=%.3f"
                    % (k,tp,fp,tn,fn,precision, recall, tp_rate, fp_rate, f_score,accuracy)
                )
            else:
                sys.stdout.write('\rk=%.6f...' % k)
                sys.stdout.flush()

            precisions.append(precision)
            recalls.append(recall)
            tp_rates.append(tp_rate)
            fp_rates.append(fp_rate)

    # to remove progress text
    if not verbose:
        sys.stdout.write("\r")
        sys.stdout.flush()

    logger.info('AUPR  = %.3f' % _area_under_the_curve(recalls, precisions))
    logger.info('AUROC = %.3f' % _area_under_the_curve(fp_rates, tp_rates))

    return precisions, recalls

