import os
import bisect
from log import logger
from partition import Partition
from affinity_matrix import AffinityMatrix
import method
import dynamic_network
import utils

class CommunityEvolution(object):
    _CACHE = {}
    IO_DATA_FILENAME = 'data.txt'

    def __init__(self, segments_partitions, changepoints,
                 score=None,
                 method_id=None,
                 network_id=None, dynamic_network=None, **kwargs):
        self.score = score
        self._segments_partitions = segments_partitions
        self._changepoints = sorted(changepoints or [])
        self._changepoints_with_start_end = None
        self._segments_intervals = None
        self._method_id = method_id

        self._network_id = None
        self._dynamic_network = None
        if dynamic_network is not None:
            self._dynamic_network = dynamic_network
            self._network_id = dynamic_network.get_id()
        elif network_id is not None:
            self._network_id = network_id

        self._start_timepoint = kwargs.get('start_timepoint')
        self._end_timepoint = kwargs.get('end_timepoint')

        if kwargs.get('scores'):
            self._scores = kwargs.get('scores')

    def get_precomputed_score(self, scoring_method_id):
        if not hasattr(self, '_scores'):
            return None
        return self._scores.get(scoring_method_id)

    def get_network_method_description(self):
        return 'community_evolution-{network_id}-{method_id}'.format(
            network_id=self.get_network_id() or 'none',
            method_id = self.get_method_id(),
        )

    def has_partitions(self):
        return len(self.get_segments_partitions())>0 and isinstance(self.get_segments_partitions()[0], Partition)

    def get_segments_partitions(self):
        return self._segments_partitions

    def get_changepoints(self):
        return self._changepoints

    def get_start_timepoint(self):
        if self._start_timepoint is None:
            self._start_timepoint = 0
        return self._start_timepoint

    def get_end_timepoint(self):
        if self._end_timepoint is None:
            self._end_timepoint = None if self.get_network_id() is None \
                else len(self.get_dynamic_network().get_timepoints())-1
        return self._end_timepoint

    def get_changepoints_with_start_end(self):
        """
        Get changepoints + start of the first segment + end of the last segment
        :return:
        """
        if self._changepoints_with_start_end is None:
            self._changepoints_with_start_end = [self.get_start_timepoint()] + self.get_changepoints() +\
                [self.get_end_timepoint()]
        return self._changepoints_with_start_end

    def get_segments_intervals(self):
        """
        Get list of each pairs (segment_start, segment_end)
        :return:
        """
        if self._segments_intervals is None:
            changepoints_with_start_end = self.get_changepoints_with_start_end()[:]
            # last time point is inclusive
            changepoints_with_start_end[-1] += 1
            self._segments_intervals = []
            for i in xrange(1, len(changepoints_with_start_end)):
                self._segments_intervals.append((changepoints_with_start_end[i-1], changepoints_with_start_end[i] - 1))
        return self._segments_intervals

    def get_segments_intervals_labels(self):
        return [
            ('%s-%s' % (s[0], s[1]) if s[0] != s[1] else str(s[0])) for s in self.get_segments_intervals()
        ]

    def get_method_id(self):
        return self._method_id

    def set_method_id(self, method_id):
        self._method_id = method_id

    def get_method_title(self):
        from methods.methods_utils import get_method_title
        return get_method_title(self.get_method_id())

    def get_network_id(self):
        return self._network_id

    def set_network_id(self, network_id):
        self._network_id = network_id

    def get_dynamic_network(self):
        if self._dynamic_network is None and self.get_network_id() is not None:
            self._dynamic_network = dynamic_network.DynamicNetwork.load(self.get_network_id())
        return self._dynamic_network

    def set_dynamic_network(self, dynamic_network):
        self._dynamic_network = dynamic_network
        self._network_id = self._dynamic_network.get_id()

    def get_description(self):
        from methods.methods_utils import get_method_description
        method_description = get_method_description(self.get_method_id())
        return "This community evolution was produced by '{method_title}' method. " \
               "Method description: {method_description} " \
               "The produced community evolution contains {n_segments} segment{plural} " \
               "for [{t_start}, {t_end}] time interval: {segments}. " \
               "Each segment has the following partition sizes: [{partitions_sizes}].".format(
            method_title=self.get_method_title(),
            method_description=method_description,
            n_segments=len(self.get_segments_partitions()),
            plural='s' if len(self.get_segments_partitions())!=1 else '',
            t_start=self.get_start_timepoint(),
            t_end=self.get_end_timepoint(),
            segments=self.get_segments_intervals_labels(),
            partitions_sizes=", ".join(
                map(
                    str,
                    [
                        p.get_clusters_count() for p in self.get_segments_partitions()
                    ]
                )
            ),
        )

    def get_summary(self):
        return "Community evolution - method id: {method_id}; " \
               "network id: {network_id}; " \
               "changepoints: [{changepoints}]; " \
               "partitions sizes [{communities_sizes}]".format(
            method_id=self._method_id,
            network_id=self.get_network_id() or "NONE",
            changepoints=", ".join(map(str, self._changepoints)),
            communities_sizes=", ".join(
                map(
                    str,
                    [
                        p.get_clusters_count() for p in self._segments_partitions
                    ]
                )
            ),
        )

    def get_dir(self):
        if self.get_network_id() is None:
            return None
        return method.Method.get_method_output_dir(self.get_network_id(), self.get_method_id())

    def get_images_dir(self):
        if self.get_network_id() is None:
            return None
        return method.Method.get_method_images_dir(self.get_network_id(), self.get_method_id())

    def get_comparison_types_list(self, only_pairwise=False):
        """
        Get list of comparison methods hashes involving current community evolution
        Comparison type - list of methods for comparison
        (each output type will involve comparison of the current community evolution vs one or more other ones)
        :return:
        """
        comp_dir = dynamic_network.DynamicNetwork.get_network_methods_comparison_dir(self.get_network_id())
        res = []
        for d in utils.get_subdir_list_from_dir(comp_dir, full_path=False):
            with open(comp_dir + d + '/methods.txt', 'r') as infile:
                m_list = [m.replace('\n','') for m in infile.readlines()]
                if only_pairwise and len(m_list) != 2:
                    continue
                if self.get_method_id() in m_list:
                    res.append(d)
        return res

    def save(self, **kwargs):
        if kwargs.get('output_dir') is not None:
            output_dir = kwargs.get('output_dir')
        elif kwargs.get('as_ground_truth', False):
            output_dir = dynamic_network.DynamicNetwork.get_network_ground_truth_dir(self.get_network_id()) + self.get_method_id() + '/'
        else:
            output_dir = self.get_dir()
        output_filename = kwargs.get('output_filename', CommunityEvolution.IO_DATA_FILENAME)
        logger.debug(
            "Saving community evolution to {output_dir}".format(
                output_dir=output_dir or 'NONE',
            )
        )
        if output_dir is None:
            raise Exception("Cannot save community evolution without network_id or output_dir")

        CommunityEvolution._save_data_to_file(
            output_dir=output_dir,
            output_filename=output_filename,
            segments_partitions=self._segments_partitions,
            changepoints=self.get_changepoints(),
            params={
                'score': self.score,
                'start_timepoint': self.get_start_timepoint(),
                'end_timepoint': kwargs.get('end_timepoint', self.get_end_timepoint()),
            }
        )


    @staticmethod
    def _save_data_to_file(output_dir, output_filename,
                           segments_partitions, changepoints, params=None):
        output_path = output_dir + output_filename
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)

        with open(output_path, "wb") as outfile:
            params_count = len(params.keys()) if params is not None else 0
            outfile.write("params=%s\n" % params_count)
            if params_count>0:
                for p, v in params.iteritems():
                    outfile.write("{p}={v}\n".format(p=p,v=v))

            outfile.write(
                "changepoints=%s\n" % (",".join(map(str, changepoints))
                                       if len(changepoints or []) > 0 else 'None')
            )

            start_timepoint = params.get('start_timepoint', 'start') if params is not None else 'start'
            end_timepoint = params.get('end_timepoint', 'end') if params is not None else 'end'
            changepoints_with_start_end = [start_timepoint] + changepoints + [end_timepoint]
            outfile.write("partitions_count=%s\n" % (len(segments_partitions) if segments_partitions else 0))
            if segments_partitions is not None:
                for i in xrange(len(segments_partitions)):
                    outfile.write("{cl} [{start}, {end}]\n".format(
                        cl="partition" if isinstance(segments_partitions[i], Partition) else "affinity_matrix",
                        start=changepoints_with_start_end[i],
                        end=changepoints_with_start_end[i+1] - (0 if i == len(segments_partitions)-1 else 1),
                    ))
                    outfile.write(segments_partitions[i].to_string())

    @staticmethod
    def check_exists(network_id, method_id):
        input_path = method.Method.get_method_output_dir(network_id, method_id) + \
            CommunityEvolution.IO_DATA_FILENAME
        return os.path.exists(input_path)

    @staticmethod
    def load(network_id, method_id, **kwargs):
        """
        Load community evolution by id
        :param network_id:
        :param method_id:
        :param kwargs:
        :return: CommunityEvolution
        """
        com_id = '%s-%s' % (network_id, method_id)
        if com_id in CommunityEvolution._CACHE:
            logger.debug("Obtaining community evolution '%s' from cache" % com_id)
        else:
            input_path = method.Method.get_method_output_dir(network_id, method_id) + \
                         CommunityEvolution.IO_DATA_FILENAME
            CommunityEvolution._CACHE[com_id] = CommunityEvolution.load_from_path(
                input_path,
                network_id=network_id,
                method_id=method_id,
                **kwargs
            )
        return CommunityEvolution._CACHE[com_id]

    @staticmethod
    def load_from_path(input_path, **kwargs):
        """
        Load community evolution by id
        :param network_id:
        :param method_id:
        :param kwargs:
        :return: CommunityEvolution
        """

        logger.debug(
            "Loading community evolution from '{input_path}'".format(
                input_path=input_path,
            )
        )

        if os.path.isdir(input_path):
            input_path = os.path.join(input_path, CommunityEvolution.IO_DATA_FILENAME)

        changepoints = []
        segments_partitions = []
        params = dict()
        precomputed_scores = dict()
        if not os.path.exists(input_path):
            if kwargs.get('check_exists'):
                logger.warning("File doesn't exists: %s" % input_path)
            return None
        with open(input_path, 'r') as infile:
            lines = infile.readlines()
            n_params = int(lines[0].split("=")[1])
            for i in xrange(1, n_params + 1):
                values = lines[i].replace("\n","").split("=")
                k = values[0]
                v = values[1]
                if v == 'None':
                    v = None
                else:
                    if k in ['score', 'avg_mod']:
                        v = float(v)
                    elif k in ['start_timepoint', 'end_timepoint']:
                        v = int(v)

                if k in ['avg_mod']:
                    precomputed_scores['(avg_sn)(mod)'] = v
                else:
                    params[k] = v

            i = n_params + 1
            values = lines[i].replace(",\n","").replace("\n","").split("=")
            if values[0] != 'changepoints':
                raise Exception("Incorrect file format")
            if values[1] == 'None' or values[1] == '':
                changepoints = []
            else:
                changepoints = map(int, values[1].split(','))

            i += 1
            n_partitions = int(lines[i].split("=")[1])
            i += 1
            n_partitions_processed = 0
            while i < len(lines):
                if lines[i].startswith("partition"):
                    cl = Partition
                elif lines[i].startswith("affinity_matrix"):
                    cl = AffinityMatrix
                else:
                    i += 1
                    continue
                    # raise Exception("Unknown line format %s" % lines[i_line])
                temp_partition = cl.from_lines(
                    lines,
                    i+1,
                    convert_objects_to_ints=kwargs.get('convert_objects_to_ints'),
                    convert_clusters_to_ints=kwargs.get('convert_clusters_to_ints'),
                    convert_objects_names_dict=kwargs.get('convert_objects_names_dict'),
                )
                segments_partitions.append(temp_partition)
                i += 2
                n_partitions_processed += 1
            if n_partitions_processed != n_partitions:
                raise Exception("Processed %s partitions instead of %s" % (n_partitions_processed, n_partitions))

        return CommunityEvolution(
            segments_partitions=segments_partitions,
            changepoints=changepoints,
            score=params.get('score'),
            method_id=kwargs.get('method_id'),
            network_id=kwargs.get('network_id'),
            start_timepoint=params.get('start_timepoint'),
            end_timepoint=params.get('end_timepoint'),
            scores=precomputed_scores if len(precomputed_scores)>0 else None,
        )

    def get_snapshot_segment(self, i_snapshot):
        return bisect.bisect(self.get_changepoints(), i_snapshot)

    def get_snapshot_partition(self, i_snapshot):
        if len(self.get_segments_partitions()) == 0:
            return None
        return self.get_segments_partitions()[self.get_snapshot_segment(i_snapshot)]

    def generate_snapshot_partition_pairs(self):
        """
        Generate pairs (snapshot, snapshot_partition)
        :return:
        """
        for i, s in enumerate(self.get_dynamic_network().get_snapshots()):
            yield (s, self.get_snapshot_partition(i))

    @staticmethod
    def rename_clusters_as_ints_per_unique_adjacent_cluster(partitions_list):
        """
        Rename clusters in partitions list so that they are all integers starting from 0,
        and if two consecutive partitions have clusters with exactly the same elements, they will have the same names
        E.g., This is useful when displaying Sankey flow - adjacent clusters that are unchanged will have the same colors
        :param partitions_list:
        :return:
        """
        res = []
        i_cluster = -1
        for i, p in enumerate(partitions_list):
            # use objects_clusters dict to avoid issue, e.g., when we rename 1 -> 2 in a partition,
            #  but there is already cluster 2 in this partition
            p_objects_clusters = dict()
            if i==0:
                for c in p.get_clusters_names():
                    i_cluster += 1
                    p_objects_clusters[i_cluster] = p.get_cluster_objects(c)
            else:
                for c in p.get_clusters_names():
                    i_temp = None
                    for c_prev in res[-1].get_clusters_names():
                        if p.get_cluster_objects(c) == res[-1].get_cluster_objects(c_prev):
                            i_temp = c_prev
                            break
                    if i_temp is None:
                        i_cluster += 1
                        i_temp = i_cluster
                    p_objects_clusters[i_temp] = p.get_cluster_objects(c)
            res.append(Partition.from_clusters_objects(p_objects_clusters))

        return res