import random
from ..partition import Partition
from .. import settings

class CommunityEvolutionRecord(object):
    def apply(self, P_in):
        """
        Apply community evolution modification to P_in and return the updated partition
        :param P_in: Initial partition
        :return: Updated partition
        """
        raise NotImplementedError()

    def to_string(self):
        raise NotImplementedError()


class InitCommunityEvolutionRecord(CommunityEvolutionRecord):
    def __init__(self, n, c, cluster_sizes = None):
        self.n = n
        self.c = c
        self.cluster_sizes = cluster_sizes

    def apply(self, P_in = None):
        if self.n is None:
            raise Exception("Cannot apply InitCommunityEvolutionRecord without specified n")
        return Partition.generate_random_partition(self.n, self.c, self.cluster_sizes)

    def to_string(self):
        return 'init{sep}{c}'.format(
            sep=settings.COMMUNITY_EVOLUTION_RECORD_SEPARATOR,
            c=self.c
        )

class KMergeCommunityEvolutionRecord(CommunityEvolutionRecord):
    def __init__(self, c_in_list, c_out):
        self.c_in_list = c_in_list
        self.c_out = c_out

    def apply(self, P_in):
        data_out = P_in.get_objects_clusters().copy()
        for c_in in self.c_in_list:
            for v in P_in.get_cluster_objects(c_in):
                data_out[v] = self.c_out
        return Partition(data_out)

    def to_string(self):
        return 'kmerge{sep}{c_in_list}{sep}{c_out}'.format(
            sep=settings.COMMUNITY_EVOLUTION_RECORD_SEPARATOR,
            c_in_list=settings.COMMUNITY_EVOLUTION_RECORD_CLUSTERS_LIST_SEPARATOR.join(
                map(str, self.c_in_list)
            ),
            c_out=self.c_out,
        )

class KSplitCommunityEvolutionRecord(CommunityEvolutionRecord):
    def __init__(self, c_in, c_out_list):
        self.c_in = c_in
        self.c_out_list = c_out_list

    def apply(self, P_in):
        data_out = P_in.get_objects_clusters().copy()
        k = len(self.c_out_list)
        split_fractions = [1.0 / k] * k

        c_in_size = len(P_in.get_cluster_objects(self.c_in))
        cluster_sizes = [c_in_size * x for x in split_fractions]

        curr_c_out_ind = 0
        curr_added = 0
        nodes_shuffled = list(P_in.get_cluster_objects(self.c_in))
        random.shuffle(nodes_shuffled)
        for v in nodes_shuffled:
            data_out[v] = self.c_out_list[curr_c_out_ind]
            curr_added += 1
            if curr_added >= cluster_sizes[curr_c_out_ind]:
                curr_added = 0
                curr_c_out_ind += 1

        return Partition(data_out)

    def to_string(self):
        return 'ksplit{sep}{c_in}{sep}{c_out_list}'.format(
            sep=settings.COMMUNITY_EVOLUTION_RECORD_SEPARATOR,
            c_in=self.c_in,
            c_out_list=settings.COMMUNITY_EVOLUTION_RECORD_CLUSTERS_LIST_SEPARATOR.join(
                map(str, self.c_out_list)
            ),
        )