import os
import math
import random
from synthetic_network_evolution import (
    SyntheticNetworkEvolution,
    CommunityEvolutionStep,
)
from community_evolution_record import (
    InitCommunityEvolutionRecord,
    KSplitCommunityEvolutionRecord,
    KMergeCommunityEvolutionRecord,
)
from ..community_evolution import CommunityEvolution
from.. import settings
from .. log import logger
from ..partition import Partition
from scipy.stats import rv_discrete
from networkx.algorithms.bipartite.generators import gnmk_random_graph
import sys
from ..draw import com_ev_draw

MIN_CLUSTER_SIZE = 50

def load_community_evolution_record_from_string(s, **kwargs):
    """
    Load community evolution record from a string (e.g., split record)
    :param s:
    :param kwargs:
    :return:
    """
    values = s.split(settings.COMMUNITY_EVOLUTION_RECORD_SEPARATOR)
    if values[0] == 'init':
        c = int(values[1])
        return InitCommunityEvolutionRecord(kwargs.get('n'), c)
    elif values[0] == 'kmerge':
        c_in_list = map(int, values[1].split(settings.COMMUNITY_EVOLUTION_RECORD_CLUSTERS_LIST_SEPARATOR))
        c_out = int(values[2])
        return KMergeCommunityEvolutionRecord(c_in_list, c_out)
    elif values[0] == 'ksplit':
        c_in = int(values[1])
        c_out_list = map(int, values[2].split(settings.COMMUNITY_EVOLUTION_RECORD_CLUSTERS_LIST_SEPARATOR))
        return KSplitCommunityEvolutionRecord(c_in, c_out_list)
    raise Exception("Cannot load community evolution record from string '%s'" % s)

def load_segments_partitions_history(history_type, **kwargs):
    """
    Load list of steps describing how partitions evolve (i.e., history):
    For each segment, its duration and the corresponding transformation based on the previous one (e.g., merge partitions)
    :param history_type:
    :param kwargs:
    :return:
    """
    evolution_history = []

    input_path = settings.SYNTHETIC_SEGMENTS_PARTITIONS_HISTORIES_DIR + history_type + '.txt'
    if not os.path.exists(input_path):
        if history_type.startswith('rand'):
            values = history_type.split(settings.SYNTHETIC_COMMUNITY_EVOLUTION_HISTORY_TYPE_SEPARATOR)
            random_instance = values[-1]
            evolution_history = generate_segments_partitions_history(
                n_objects=kwargs.get('n'),
                n_segments=int(values[1]),
                n_timesteps=int(values[2]),
            )

            # Note: even though evolution_history is generated with a given value of n, it is saved/loaded without it
            save_segments_partitions_history(evolution_history, history_type)
        else:
            raise Exception("Unknown segments partitions history '%s'" % history_type)
    else:
        with open(input_path, 'r') as infile:
            lines = infile.readlines()
            i = 0
            values = lines[i].split('=')
            if values[0]!='segments_count':
                raise Exception('Incorrect line format')
            segments_count=int(values[1])

            i += 1
            for i_step in xrange(segments_count):
                values = lines[i].split(';')
                if not values[0].startswith('segment='):
                    raise Exception('Incorrect line format')

                if not values[1].startswith('duration='):
                    raise Exception('Incorrect line format')
                duration=int(values[1].split('=')[1])

                if not values[2].startswith('records_count='):
                    raise Exception('Incorrect line format')
                records_count=int(values[2].split('=')[1])

                i += 1
                records = []
                for i_record in xrange(records_count):
                    records.append(
                        load_community_evolution_record_from_string(
                            s=lines[i],
                            n=kwargs.get('n'),
                        )
                    )
                    i += 1
                evolution_history.append(
                    CommunityEvolutionStep(
                        community_evolution_records=records,
                        duration=duration,
                    )
                )
    return evolution_history


def save_segments_partitions_history(evolution_history, history_name):
    output_dir = settings.SYNTHETIC_SEGMENTS_PARTITIONS_HISTORIES_DIR
    output_path = output_dir + history_name + '.txt'
    logger.debug("Saving segments partitions history to %s" % output_path)

    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    with open(output_path, 'wb') as outfile:
        outfile.write('segments_count=%s\n' % len(evolution_history))
        for i_step, step in enumerate(evolution_history):
            outfile.write('segment={i_step};duration={duration};records_count={records_count}\n'.format(
                i_step=i_step,
                duration=step.get_duration(),
                records_count=len(step.get_records())
            ))
            for i_record, record in enumerate(step.get_records()):
                outfile.write('%s\n' % record.to_string())


def load_community_evolution_by_history_id(history_id):
    """
    Get community evolution based on segments partitions history (by id). If a given random instance exists, load it.
    Otherwise, generate a new one: get a corresponding history (e.g., sequence of steps descirbing evolution),
    and then generate segments partitions based on them
    :param history_id:
    :return:
    """
    values = history_id.split(settings.SYNTHETIC_COMMUNITY_EVOLUTION_HISTORY_ID_SEPARATOR)
    # Note: as each community evolution is generated randomly from history, it has a random instance
    random_instance = values[-1]

    io_dir = settings.SYNTHETIC_SEGMENTS_PARTITIONS_DIR
    io_filename = history_id + '.txt'
    if os.path.exists(io_dir + io_filename):
        logger.info("Loading existing community evolution instance by history id '%s'" % history_id)
        com_ev = CommunityEvolution.load_from_path(
            input_path=io_dir + io_filename,
            convert_objects_to_ints=True,
            convert_clusters_to_ints=True,
        )
    else:
        logger.info("Generating community evolution instance by history id '%s'" % history_id)
        n = int(values[1])
        evolution_history = load_segments_partitions_history(
            history_type=values[0],
            n=n,
        )
        com_ev = create_community_evolution_from_history_records(evolution_history)

        com_ev_draw.draw_segments_community_evolution(
            com_ev=com_ev,
            output_dir=io_dir + 'images/',
            output_filename=history_id + '.html'
        )

        com_ev.save(
            output_dir = io_dir,
            output_filename=io_filename,
        )

    return com_ev


def create_community_evolution_from_history_records(evolution_history):
    """
    Create community evolution from a sequence of CommunityEvolutionStep objects
    :param evolution_history:
    :return:
    """
    segments_partitions = []
    changepoints = []

    partition = None
    segment_start = 0

    for i_segment, s in enumerate(evolution_history):
        partition = s.apply(partition)
        if min(partition.get_clusters_sizes()) < MIN_CLUSTER_SIZE:
            raise Exception("Minimum allowed cluster size {min_c}, got {c} at segment {i_s}".format(
                min_c=MIN_CLUSTER_SIZE,
                c=min(partition.get_clusters_sizes()),
                i_s=i_segment,
            ))
        segments_partitions.append(partition)
        if segment_start != 0:
            changepoints.append(segment_start)
        segment_start += s.get_duration()

    return CommunityEvolution(
        segments_partitions=segments_partitions,
        changepoints=changepoints,
        start_timepoint=0,
        end_timepoint=segment_start - 1,
    )


def create_synthetic_network_evolution_by_id(synthetic_network_id):
    values = synthetic_network_id.split(settings.SYNTHETIC_NETWORK_ID_SEPARATOR)
    if values[0] == SyntheticNetworkEvolution.get_id_prefix():
        return SyntheticNetworkEvolution(synthetic_network_id)
    raise "Unknown synthetic network evolution %s" % synthetic_network_id


def generate_segments_partitions_history(n_objects, n_timesteps, n_segments):
    """
    Generate a random community evolution history
    :param n_objects:
    :param n_timesteps:
    :param n_segments:
    :return:
    """
    logger.debug("Generating random segments partitions history with {n_timesteps} timesteps "
                 "and {n_segments} segments".format(
        n_timesteps=n_timesteps,
        n_segments=n_segments,
    ))

    # possible number of clusters at any time point
    # NOTE: we divide by 2 max cluster count, because if i.e., n=50, MIN_CLUSTER_SIZE=10 and
    # prev_clusters_count = 5 (after init - so they are roughly 10 nodes each) and curr_clusters_count = 5
    # we need to divide one of the current clusters to create new segment,
    # which is not possible as they are already small
    n_clusters_values = [i for i in xrange(2, int(math.floor(n_objects / (2.0 * MIN_CLUSTER_SIZE))))]
    if len(n_clusters_values) == 1 and n_clusters_values[0] == 2:
        raise Exception("Too little different cluster counts: %s" % n_clusters_values)

    n_clusters_probs = [1.0 / len(n_clusters_values) for x in range(len(n_clusters_values))]
    dist_clusters_count = rv_discrete(values=(n_clusters_values, n_clusters_probs))

    # initial clusters
    prev_clusters_count = dist_clusters_count.rvs()
    prev_clusters = range(prev_clusters_count)

    curr_max_cluster_index = prev_clusters[-1]

    community_evolution_records = []
    community_evolution_records.append(
        [InitCommunityEvolutionRecord(n_objects, prev_clusters_count)]
    )

    prev_clusters_approx_sizes = [(1.0 * n_objects) / prev_clusters_count for i in xrange(prev_clusters_count)]

    for i_segment in xrange(1, n_segments):
        # random.shuffle(prev_clusters)
        prev_clusters_count = len(prev_clusters)

        # current number of clusters
        curr_clusters_count = dist_clusters_count.rvs()
        # cannot be two cluster to two clusters - no split/merge changes
        if prev_clusters_count == 2:
            while curr_clusters_count == 2:
                curr_clusters_count = dist_clusters_count.rvs()
        curr_clusters = []
        curr_evolution_records = []

        # create bipartite graph describing how communities change from previous to current step
        # Note: each edge should have at least one endpoint of degree one
        # both with one - community survived, left with one - merge, right with one - split
        # k - number of edges. If left and right sizes (i.e. n=m) are equal - it should be [n+1, (n-1)*2],
        # otherwise - [max(n,m), m+n-2]
        if prev_clusters_count == curr_clusters_count:
            k_min = prev_clusters_count + 1
            k_max = (prev_clusters_count - 1) * 2
        else:
            k_min = max(prev_clusters_count, curr_clusters_count)
            k_max = prev_clusters_count + curr_clusters_count - 2

        i_g = 0
        MAX_RETRIES = 10000
        while True:
            i_g += 1
            sys.stdout.write("\rGenerating bipartite partitions change networks: attempt %s" % i_g)
            sys.stdout.flush()

            if i_g > MAX_RETRIES:
                raise Exception(
                    "Cannot generate appropriate bipartite graph (%s, %s) with [%s,%s] edges in %s retries"
                    % (prev_clusters_count, curr_clusters_count, k_min, k_max, MAX_RETRIES)
                )

            g = gnmk_random_graph(
                n=prev_clusters_count,
                m=curr_clusters_count,
                k=random.choice(range(k_min, k_max + 1))
            )

            # repeat until there are no isolated nodes
            isolated_nodes_ok = False
            if min(g.degree().values()) > 0:
                # make sure there is no, e.g., (1, a), (1, b) and (2, a), (2, b)
                # i.e. each edge should have at least one endpoint of degree one
                for u,v in g.edges():
                    if g.degree(u) > 1 and g.degree(v) > 1:
                        g.remove_edge(u, v)
                # check if there is no perfect matching - in this case, no splits/merges, so try again
                if max(g.degree().values()) > 1:
                    isolated_nodes_ok = True
            if not isolated_nodes_ok:
                continue

            # make sure that no cluster is too small
            # this is a copy of the loop below for generating steps,
            # we repeat everything in the same way here to make sure that the sizes are ordered correctly
            # TODO: consider simplifying this
            processed_nodes = set()
            clusters_sizes_ok = True
            i_prev_cluster = 0
            for v, v_data in g.nodes(data=True):
                if v in processed_nodes:
                    continue
                if v_data['bipartite'] != 0:
                    continue
                v_neighbors = g.neighbors(v)
                cl_sizes = []
                if len(v_neighbors) == 1: # survive or merge
                    u = v_neighbors[0]
                    u_neighbors = g.neighbors(u)
                    if len(u_neighbors) == 1: # survive
                        cl_sizes.append(prev_clusters_approx_sizes[i_prev_cluster])
                        i_prev_cluster += 1
                    else: # merge
                        cl_sizes.append(sum(prev_clusters_approx_sizes[x] for x in xrange(i_prev_cluster, i_prev_cluster + len(u_neighbors))))
                        i_prev_cluster += len(u_neighbors)
                        curr_max_cluster_index += 1
                        for w in u_neighbors:
                            processed_nodes.add(w)
                else: # split
                        c_out_list = range(curr_max_cluster_index+1, curr_max_cluster_index + len(v_neighbors) + 1)
                        cl_sizes += [prev_clusters_approx_sizes[i_prev_cluster] / len(c_out_list)] * len(c_out_list)
                        i_prev_cluster += 1
                        curr_max_cluster_index += len(v_neighbors) + 1
                        for u in v_neighbors:
                            processed_nodes.add(u)
                if min(cl_sizes) < MIN_CLUSTER_SIZE:
                    clusters_sizes_ok = False
                    break
            if clusters_sizes_ok:
                break
        # to remove progress text
        sys.stdout.write("\r")
        sys.stdout.flush()

        # print "i_segment = %s" % i_segment
        # print (prev_clusters_count, curr_clusters_count)
        # print [x for x in g.nodes(data=True)]
        # print g.edges()

        curr_clusters_approx_sizes = []
        processed_nodes = set()
        i_prev_cluster = 0
        for v, v_data in g.nodes(data=True):
            if v in processed_nodes:
                continue
            if v_data['bipartite'] != 0:
                continue
            v_neighbors = g.neighbors(v)
            if len(v_neighbors) == 1: # survive or merge
                u = v_neighbors[0]
                u_neighbors = g.neighbors(u)
                if len(u_neighbors) == 1: # survive
                    curr_clusters.append(prev_clusters[i_prev_cluster])
                    curr_clusters_approx_sizes.append(prev_clusters_approx_sizes[i_prev_cluster])
                    i_prev_cluster += 1
                else: # merge
                    curr_evolution_records.append(
                        KMergeCommunityEvolutionRecord(
                            c_in_list=prev_clusters[i_prev_cluster:i_prev_cluster + len(u_neighbors)],
                            c_out=curr_max_cluster_index + 1
                        )
                    )
                    curr_clusters.append(curr_max_cluster_index + 1)
                    curr_clusters_approx_sizes.append(sum(prev_clusters_approx_sizes[x] for x in xrange(i_prev_cluster, i_prev_cluster + len(u_neighbors))))
                    i_prev_cluster += len(u_neighbors)
                    curr_max_cluster_index += 1
                    for w in u_neighbors:
                        processed_nodes.add(w)
            else: # split
                    c_out_list = range(curr_max_cluster_index+1, curr_max_cluster_index + len(v_neighbors) + 1)
                    curr_evolution_records.append(
                        KSplitCommunityEvolutionRecord(
                            c_in=prev_clusters[i_prev_cluster],
                            c_out_list=c_out_list,
                        )
                    )
                    curr_clusters += c_out_list
                    curr_clusters_approx_sizes += [prev_clusters_approx_sizes[i_prev_cluster] / len(c_out_list)] * len(c_out_list)
                    i_prev_cluster += 1
                    curr_max_cluster_index += len(v_neighbors) + 1
                    for u in v_neighbors:
                        processed_nodes.add(u)

            processed_nodes.add(v)

        community_evolution_records.append(curr_evolution_records)
        prev_clusters = curr_clusters
        prev_clusters_approx_sizes = curr_clusters_approx_sizes
        logger.debug("Approximate cluster sizes at segment %s: %s" % (i_segment, curr_clusters_approx_sizes))
    segments_durations = Partition._constrained_sum_sample_pos(n_segments, n_timesteps)
    evolution_history = []
    for i_segment in range(n_segments):
        evolution_history.append(
            CommunityEvolutionStep(
                community_evolution_records=community_evolution_records[i_segment],
                duration=segments_durations[i_segment],
            )
        )

    return evolution_history

