import os
import random
import networkx as nx
from .. import network
from ..dynamic_network import DynamicNetwork
from ..log import logger
from .. import static_clustering
import generator_utils
from .. import utils
from .. import settings
from ..methods import methods_utils

class CommunityEvolutionStep:
    def __init__(self, community_evolution_records, duration=1):
        if not isinstance(community_evolution_records, list):
            community_evolution_records = [community_evolution_records]
        self._community_evolution_records = community_evolution_records
        self._duration = duration

    def get_duration(self):
        return self._duration

    def get_records(self):
        return self._community_evolution_records

    def apply(self, previous_partition):
        res_partition = previous_partition
        for r in self._community_evolution_records:
            res_partition = r.apply(res_partition)
        return res_partition

class SyntheticNetworkEvolution:

    def __init__(self, synthetic_network_id):
        values = synthetic_network_id.split(settings.SYNTHETIC_NETWORK_ID_SEPARATOR)
        if values[0] != SyntheticNetworkEvolution.get_id_prefix():
            raise Exception("Cannot create synthetic network with id '%s'" % synthetic_network_id)
        history_id = values[1]
        self._c_in = int(values[2])
        self._c_out = int(values[3])
        self._noise_level = int(values[4]) # can use it to randomly alter each snapshot's partition
        self._random_instance = int(values[5])

        self.network_id = settings.SYNTHETIC_NETWORK_ID_GLOBAL_PREFIX + synthetic_network_id
        if os.path.exists(DynamicNetwork.get_network_dir(self.network_id)):
            del_prev = utils.user_yes_no_prompt(
                "Synthetic network '%s' already exists. "
                "Do you want to overwrite it (all previous results will be lost)?" % self.network_id
            )
            if del_prev:
                DynamicNetwork.delete(self.network_id)
            else:
                raise Exception("Aborting generating synthetic network '%s'" % self.network_id)

        self._community_evolution = generator_utils.load_community_evolution_by_history_id(
            history_id
        )
        self._generate_evolution()

        network_description = "This is a synthetically generated network. " \
            "Its community evolution ('{history_id}') contains {n_segments} segment{plural} " \
            "for [{t_start}, {t_end}] time interval: {segments}. " \
            "Each segment has the following partition sizes: [{partitions_sizes}]. " \
            "Given a community structure at a given timepoint (determined by the corresponding segment), " \
            "the corresponding network snapshot was generated independently via a stochastic block model (SBM) " \
            "(each edge is generated randomly and independently " \
            "so that average intra cluster degree is {c_in} and average inter cluster degree is {c_out}) " \
            "with noise level '{noise_level}' " \
            "({noise_level}% of edges were randomly rewired). " \
            "Each network snapshot has the following number of nodes and edges: [{nodes_edges}].".format(
            history_id=history_id,
            n_segments=len(self._community_evolution.get_segments_partitions()),
            plural='s' if len(self._community_evolution.get_segments_partitions())!=1 else '',
            t_start=self._community_evolution.get_start_timepoint(),
            t_end=self._community_evolution.get_end_timepoint(),
            segments=self._community_evolution.get_segments_intervals_labels(),
            partitions_sizes=", ".join(
                map(
                    str,
                    [
                        p.get_clusters_count() for p in self._community_evolution.get_segments_partitions()
                    ]
                )
            ),
            c_in=self._c_in,
            c_out=self._c_out,
            noise_level=self._noise_level,
            nodes_edges=", ". join(
               "(%s, %s)" % (s.get_nodes_count(), s.get_edges_count()) for s in self._dynamic_network.get_snapshots()
            )
        )
        self._dynamic_network.set_description(network_description)

    @staticmethod
    def get_id_prefix():
        return 'sbm'

    def _generate_evolution(self):
        """
        Generate community evolution - sequence of network snapshots for each time point
        :return:
        """
        snapshots = []

        for i in xrange(self._community_evolution.get_start_timepoint(), self._community_evolution.get_end_timepoint() + 1):
            snapshots.append(
                SyntheticSnapshot(
                    c_in=self._c_in,
                    c_out=self._c_out,
                    P=self._community_evolution.get_snapshot_partition(i),
                    name="%s_network-%s_segment-" % (i, self._community_evolution.get_snapshot_segment(i))
                )
            )

        self._community_evolution.set_method_id(methods_utils.SEGMENT_GROUND_TRUTH_METHOD_ID)
        self._community_evolution.set_network_id(self.network_id)
        ground_truths = [self._community_evolution]

        self._dynamic_network = DynamicNetwork(
            network_id=self.network_id,
            snapshots=[s.get_network() for s in snapshots],
            ground_truths=ground_truths
        )

    def save(self):
        logger.debug("Saving network evolution for network '%s'" % self._dynamic_network._network_id)
        self._dynamic_network.save(save_data=False)

class SyntheticSnapshot:

    def __init__(self, c_in, c_out, P=None, name=None):
        """
        Initialize synthetic snapshot
        :param n: Number of nodes
        :param p_in: Probability of an edge between nodes in the same community
        :param p_out: Probability of an edge between nodes in different communities
        :param P: community partition
        :return:
        """
        self._c_in = c_in
        self._c_out = c_out
        self._P = P
        self.name = name

        self._generate_G()

    def _generate_G(self):
        """
        Generate graph based on community partition and p_in/p_out probabilities
        :return:
        """
        #number of nodes
        n = self.get_partition().get_objects_count()
        p_in = 1.0 * self._c_in / n
        p_out = 1.0 * self._c_out / n

        #create empty graph
        nx_g = nx.Graph()
        nx_g.add_nodes_from(range(n))

        #add edges based on community partition
        for u in xrange(n):
            for v in xrange(u+1, n):
                r  = random.random()
                if (self.get_partition().get_object_cluster(u) == self.get_partition().get_object_cluster(v)
                         and r<=p_in) or (self.get_partition().get_object_cluster(u) != self.get_partition().get_object_cluster(v)
                         and r<=p_out):
                    nx_g.add_edge(u,v)

        self._network = network.Network(nx_g)

        # check for isolated nodes
        while True:
            isolates = nx.isolates(nx_g)
            if len(isolates) > 0:
                logger.warning("Network %s has %s isolated node(s) - adding one intra-community edge for each" % (self.name, len(isolates)))
                for u in isolates:
                    u_cl = self.get_partition().get_object_cluster(u)
                    u_cl_objs = [v for v in self.get_partition().get_cluster_objects(u_cl) if v != u]
                    nx_g.add_edge(u, random.choice(u_cl_objs))
            else:
                break

        partition_quality_description = []
        for metric, metric_title in static_clustering.PARTITION_QUALITY_METRICS.iteritems():
            partition_quality_description.append(
                "{metric} - {score}".format(
                    metric=metric_title,
                    score=static_clustering.compute_partition_quality(self._network, self._P, metric)
                )
            )
        logger.debug(
            "Synthetic snapshot{name} partition quality: {descr}".format(
                name=' ' + self.name if self.name is not None else '',
                descr=', '.join(partition_quality_description)
            )
        )

    def get_network(self):
        return self._network

    def get_partition(self):
        return self._P

    def save(self, output_dir, output_filename, write_method = 'edgelist'):
        logger.debug(
            "Saving synthetic snapshot to {output_path} using {write_method} method".format(
                output_path=output_dir + output_filename,
                write_method=write_method,
            )
        )
        self.get_network().save(output_dir,output_filename, write_method, save_data=False)