import subprocess
import os
import utils
from log import logger
import numpy as np
from network import Network
import settings
import static_clustering
from methods.other_methods.multistep import MultistepMethod

CONS_MATR_EXEC = settings.BASE_DIR + 'dynamic_communities_cpp/other/consensus_matrix/pr_consensus' + ('' if utils.is_remote_server() else '.exe')

def get_union_method_name(community_detection_method, weighted):
    return 'u{sep}{weighted}{sep}{community_detection_method}'.format(
        weighted='w' if weighted else 'u',
        community_detection_method=community_detection_method,
        sep=settings.CONSENSUS_CLUSTERING_ID_SEPARATOR,
    )

def compute_consensus_partition(networks_list, method, **kwargs):
    values=method.split(settings.CONSENSUS_CLUSTERING_ID_SEPARATOR)
    if values[0] == 'u' and values[1] in ['w', 'u']:
        weighted = (values[1] == 'w')
        community_detection_method = values[2]
        g_union = Network.graph_union(networks_list, weighted=weighted, networks_weights=kwargs.get('networks_weights'))
        return static_clustering.compute_partition(
            network=g_union,
            method=community_detection_method,
            **kwargs
        )
    # Consensus Matrix - from http://www.nature.com/articles/srep00336
    elif values[0] == 'cons_matr':
        community_detection_method = values[1]
        alpha = 0.5 if len(values) < 3 else float(values[2])

        base_dir = kwargs.get('base_dir', settings.TMP_DIR + 'cons_matr/' + utils.id_generator(5) + '/')
        base_output_filename = kwargs.get('base_output_filename', utils.id_generator(5))
        if not os.path.exists(base_dir):
            os.makedirs(base_dir)

        logger.debug("Creating individual partitions for consensus matrix")
        pts_files = []
        p = None
        all_nodes = []

        for i,g in enumerate(networks_list):
            all_nodes += g.get_nx_graph().nodes()
            p = static_clustering.compute_partition(
                network=g,
                method=community_detection_method
            )
            f_name = base_dir + 'p_%s.txt' % i
            pts_files.append(f_name)
            with open(f_name, 'w') as outfile:
                for c, objs in p.get_clusters_objects().iteritems():
                    outfile.write("\t".join(objs) + "\n")

        # create input file contains names of files with individual partitions
        input_file = base_dir + 'pts'
        log_file = base_dir + 'pr_cons_log.txt'
        with open(input_file, 'w') as infile:
            infile.write("\n".join(pts_files))

        logger.info("Running pr_consensus executable")
        run_command = '{exe} \'{input}\' {alpha} > \'{log}\''.format(
            exe=CONS_MATR_EXEC,
            input=input_file,
            alpha=alpha,
            log=log_file,
        )
        logger.debug("Run command: %s" % run_command)
        subprocess.call(run_command, shell=True)

        # cluster the resulted consensus matrix to obtain consensus partition
        g_cons = Network.load(input_file + '_net.dat', read_method='edgelist', data=(('weight',float),))

        # it may happen that some nodes will be missing (as they are isolated) => add them
        all_nodes = set(all_nodes)
        if len(all_nodes) > g_cons.get_nodes_count():
            g_cons_nodes = set(g_cons.get_nodes())
            g_cons.get_nx_graph().add_nodes_from(all_nodes - g_cons_nodes)

        # in case all partitions contain only one cluster the resulted file will be empty
        if g_cons.get_edges_count() == 0:
            logger.warning("pr_consensus produced empty file - all partitions likely have one cluster")
            return p

        return static_clustering.compute_partition(
                network=g_cons,
                method=community_detection_method
            )
    elif values[0] == 'avg_mod':
        return MultistepMethod.compute_consensus_partition(
            networks_list=networks_list,
            **kwargs
        )

    raise Exception("Unknown consensus clustering method '%s'" % method)

def get_consensus_partition_method_description(method):
    values=method.split(settings.CONSENSUS_CLUSTERING_ID_SEPARATOR)
    if values[0] == 'u' and values[1] in ['w', 'u']:
        return "Compute {weighted} union of segment graphs and " \
               "then apply {community_detection_method} clustering method".format(
            weighted="weighted" if (values[1] == 'w') else "unweighted",
            community_detection_method=static_clustering.PARTITION_METHODS[values[2]]
        )
    elif values[0] == 'avg_mod':
        return "Average Modularity method based on Louvain Method"
    raise Exception("Unknown consensus clustering method '%s'" % method)

def get_consensus_partition_quality_metric_limits(metric):
    values = metric.split(settings.CONSENSUS_CLUSTERING_ID_SEPARATOR)
    if values[0] == 'avg':
        return static_clustering.get_partition_quality_metric_limits(values[1])
    raise Exception("Unknown consensus clustering scoring method '%s'" % metric)

def compute_consensus_partition_quality(networks_list, partition, metric):
    values = metric.split(settings.CONSENSUS_CLUSTERING_ID_SEPARATOR)
    if values[0] == 'avg':
        community_quality_metric = values[1]
        qualities = []
        for s in networks_list:
            qualities.append(
                static_clustering.compute_partition_quality(
                    network=s,
                    partition=partition,
                    metric=community_quality_metric,
                )
            )
        return np.average(qualities)
    raise Exception("Unknown consensus clustering scoring method '%s'" % metric)

def get_consensus_partition_metric_description(metric):
    values = metric.split(settings.CONSENSUS_CLUSTERING_ID_SEPARATOR)
    if values[0] == 'avg':
        return "avg of partition quality according to {community_quality_metric} metric across snapshots".format(
            community_quality_metric=static_clustering.PARTITION_QUALITY_METRICS[values[1]]
        )
    raise Exception("Unknown consensus clustering scoring method '%s'" % metric)