import cPickle
import draw_utils
import numpy as np
from ..method import Method
from ..dynamic_network import DynamicNetwork
from ..draw import com_ev_draw, method_draw, ranking_draw
from collections import OrderedDict
from ..methods import methods_utils
from ..community_evolution import CommunityEvolution
from .. import settings
from ..evaluation import eval_utils, com_ev_eval_across_networks, method_eval_pr_roc_across_networks
from .. import report_params
from ..utils import parse_user_args_string_into_values_list
from ..scoring_methods import scoring_methods_utils
from ..other.palettable.wesanderson import GrandBudapest3_6, GrandBudapest1_4
from ..methods.naive import NaiveMethod
from .. import static_clustering

def naive_ic_title_func(method_id):
    if '(aic)' in method_id:
        return 'AIC'
    elif '(bic)' in method_id:
        return 'BIC'

def naive_obj_title_func(method_id):
    if '(aic)' in method_id:
        return '$Q_P$ (AIC)'
    elif '(bic)' in method_id:
        return '$Q_B$ (BIC)'
    elif '(avg_sn)(mod)' in method_id:
        return '$Q_P$ (Modularity)'
    elif '(avg_sn)(cond_q)' in method_id:
        return '$Q_P$ (Conductance)'
    elif '(avg_sn)(norm_cut_q)' in method_id:
        return '$Q_P$ (Normalized Cut)'
    elif '(avg_sn)(avg_odf_q)' in method_id:
        return '$Q_P$ (Average-ODF)'
    return method_id

def naive_cons_sum_title_func(method_id):
    consensus_clustering_method_id, scoring_method_id, compute_type, model_selection = NaiveMethod._parse_method_id(
        method_id
    )
    return static_clustering.PARTITION_METHODS[consensus_clustering_method_id.replace('u^w^', '')]

def naive_cons_title_func(method_id):
    consensus_clustering_method_id, scoring_method_id, compute_type, model_selection = NaiveMethod._parse_method_id(
        method_id
    )
    if 'u^w^' in consensus_clustering_method_id:
        return "Sum graph (%s)" % static_clustering.PARTITION_METHODS[consensus_clustering_method_id.replace('u^w^', '')]
    elif 'cons_matr' in consensus_clustering_method_id:
        return "Consensus matrix (%s)" % static_clustering.PARTITION_METHODS[consensus_clustering_method_id.replace('cons_matr^', '')]
    elif 'avg_mod' in consensus_clustering_method_id:
        return "Average-Louvain"
    return method_id

def naive_search_title_func(method_id):
    consensus_clustering_method_id, scoring_method_id, compute_type, model_selection = NaiveMethod._parse_method_id(
        method_id
    )
    return NaiveMethod.COMPUTE_TYPE_TITLE_DICT[compute_type].capitalize()

def get_title_func(methods_ids_key):
    if methods_ids_key == 'multistep':
        return lambda x: x.replace('multistep-', '')
    elif methods_ids_key == 'ghrg':
        return lambda x: x.replace('ghrg-','')
    elif methods_ids_key == 'naive-ic':
        return naive_ic_title_func
    elif methods_ids_key == 'naive-obj':
        return naive_obj_title_func
    elif methods_ids_key == 'naive-cons-sum':
        return naive_cons_sum_title_func
    elif methods_ids_key == 'naive-cons':
        return naive_cons_title_func
    elif methods_ids_key == 'naive-search':
        return naive_search_title_func
    elif methods_ids_key.startswith('test'):
        return lambda x: x
    return method_id_to_title

def method_id_to_title(method_id):
    if 'graphscope' in method_id:
        return "GraphScope"
    elif 'multistep' in method_id:
        if '-3.0' in method_id:
            return "Multi-Step$^\\star$"
        return "Multi-Step"#$^\\ast$"
    elif 'ghrg' in method_id:
        return 'GHRG'
    elif 'segment_ground_truth' in method_id:
        return 'Ground truth'
    elif 'naive' in method_id:
        return "SCOUT"
    return methods_utils.get_method_title(method_id)

def draw_scoring_methods_ranking_bars(m_comparison_key, n_comparison_key, scoring_methods_ids, input_dir = None, **kwargs):
    """
    Using results from 'evalute_methods_score_rankings', show comparison of methods
    """
    if input_dir is None:
        input_dir = settings.EVAL_DIR + '_rankings/'
    output_dir = input_dir + '%s/%s/' % (m_comparison_key, n_comparison_key)
    for sc_m_id in scoring_methods_ids:
        for p_l in kwargs.get('precision_levels', [0.0]):
            # for m_key in methods_labels:
            #     data[m_key] = defaultdict(int)
            data = cPickle.load(
                open(input_dir + '%s/%s/%s_%s.cpickle' % (m_comparison_key,n_comparison_key, sc_m_id, p_l), 'r')
            )
                # for m_key,rankings in temp_data.iteritems():
                #     for rank,count in rankings.iteritems():
                #         data[m_key][rank] += count

            kwargs['figsize'] = kwargs.get('figsize') or (5, 5)
            ranking_draw.draw_ranking_bar(
                data=data,
                output_dir=output_dir,
                output_filename='%s_%s' % (sc_m_id, p_l),
                series_names_sorted=["graphscope", "multistep", "ghrg", "naive"],
                series_labels_dict={
                    "graphscope": "GraphScope",
                    "multistep": "Multi-Step" if ('(cl_res)(aupr)(segment_ground_truth)' in scoring_methods_ids) else "Multi-Step$^\\star$",
                    "ghrg": "GHRG",
                    "naive": "SCOUT",
                },
                always_include_na=True, # always include n/a label into the legend
                **kwargs
            )
    return output_dir

def draw_scoring_methods_n_changepoints_ranking_bars(m_comparison_key, n_comparison_key, scoring_methods_ids, input_dir = None, **kwargs):
    """
    Using results from 'evalute_methods_score_rankings_n_changepoints', show comparison of methods for the solutions with fixed numbers of changepoints
    """
    if input_dir is None:
        input_dir = settings.EVAL_DIR + '_rankings/'
    for sc_m_id in scoring_methods_ids:
        for p_l in [0]:
            data = {}
            # for m_key in methods_labels:
            #     data[m_key] = defaultdict(int)
            data = cPickle.load(
                # open(settings.NETWORKS_BASE_DIR + 'n_changepoints_%s_%s.cpickle' % (sc_m_id, p_l), 'r')
                open(input_dir + '%s/%s/n_changepoints_%s_%s.cpickle' % (m_comparison_key,n_comparison_key, sc_m_id, p_l), 'r')
            )
                # for m_key,rankings in temp_data.iteritems():
                #     for rank,count in rankings.iteritems():
                #         data[m_key][rank] += count

            t_data = OrderedDict()
            for n_changepoints, rank_dict in data.iteritems():
                n_segments = n_changepoints + 1
                t_data[str(n_segments)] = {
                    "naive": rank_dict['multistep'][2], # when naive is first
                    "tie": sum(rank_dict['naive'].values()) - rank_dict['naive'][2] - rank_dict['multistep'][2], # tie
                    "multistep": rank_dict['naive'][2] # when multistep is first
                }

            ranking_draw.draw_ranking_bar(
                data=t_data,
                output_dir=input_dir + '%s/%s/' % (m_comparison_key, n_comparison_key) ,
                output_filename='%s_%s' % (sc_m_id, p_l),
                values_types=["naive", "tie", "multistep"],
                x_label="$l$",#Number of segments",
                values_types_labels={
                    "naive": "SCOUT superior",
                    "tie": "Tie",
                    "multistep": "Multi-Step superior",
                },
                figsize=(12, 5),
                xticks_rotation=0,
                **kwargs

            )

def _get_colors(methods_ids_key, **kwargs):
    if 'best' in methods_ids_key or 'test' in methods_ids_key:
        if kwargs.get('format') == 'eps': # with added transparency, since eps does not support it
            colors = ['#7db7e1', '#fbb661', '#80ca86', '#f37976',] + settings.LINE_COLORS[4:]
        else:
            colors = settings.LINE_COLORS
    else: # comparing methods parameters
        colors = GrandBudapest3_6.hex_colors + GrandBudapest1_4.hex_colors
    return colors

def draw_all_running_times(networks_dict_key, methods_ids_key, **kwargs):
    networks_dict = OrderedDict()
    for n_k, n_ids_str in report_params.REPORT_NETWORKS_DICT_BY_KEY.get(networks_dict_key).iteritems():
        networks_dict[n_k] = parse_user_args_string_into_values_list(n_ids_str)

    methods_ids = []
    for temp_m in report_params.REPORT_METHODS_BY_KEY[methods_ids_key]:
        methods_ids += parse_user_args_string_into_values_list(temp_m)

    output_dir = kwargs.get('output_dir', settings.EVAL_DIR + '_adhoc/')
    output_dir += networks_dict_key + '/' + methods_ids_key + '/'

    title_func = get_title_func(methods_ids_key)

    data = OrderedDict()
    error_bars = OrderedDict()
    for n_key, n_ids in networks_dict.iteritems():
        for m_id in methods_ids:
            run_times = [Method.load_runtime(n_id, m_id) for n_id in n_ids]
            run_times = [t for t in run_times if t is not None]
            m_key = title_func(m_id)
            if m_key not in data:
                data[m_key] = OrderedDict()
                error_bars[m_key] = OrderedDict()
            if len(run_times)==0:
                data[m_key][n_key] = None
                error_bars[m_key][n_key] = 0
            else:
                data[m_key][n_key] = np.average(run_times)
                error_bars[m_key][n_key] = np.std(run_times)

    kwargs['y_lim'] = kwargs.get('y_lim', (0, 100000))
    kwargs['log_scale'] = kwargs.get('log_scale', True)
    kwargs['bar_colors'] = kwargs.get('colors', _get_colors(methods_ids_key,**kwargs))
    kwargs['legend_n_cols'] = kwargs.get('legend_n_cols', 2)
    kwargs['x_label'] = kwargs.get('x_label', "Snapshot size")
    kwargs['default_missing_value'] = kwargs['y_lim'][1]

    draw_utils.draw_bar(
        data=data,
        y_err=error_bars,
        output_dir=output_dir,
        output_filename='run_time',
        y_label="Running time (sec)", #  "Running time in sec" + (" (log scale)" if kwargs.get('log_scale') else ""),
        **kwargs
    )

    return output_dir


def draw_plots_per_measure(networks_dict_key, methods_ids_key, draw_com_evs_scores=True, draw_com_evs_sim=False, draw_pr_roc=False,
                           other_method_id='segment_ground_truth', **kwargs):
    networks_dict = OrderedDict()
    for n_k, n_ids_str in report_params.REPORT_NETWORKS_DICT_BY_KEY.get(networks_dict_key).iteritems():
        networks_dict[n_k] = parse_user_args_string_into_values_list(n_ids_str)
    methods_ids = []
    for temp_m in report_params.REPORT_METHODS_BY_KEY[methods_ids_key]:
        methods_ids += parse_user_args_string_into_values_list(temp_m)

    colors = kwargs.get('colors', _get_colors(methods_ids_key,**kwargs))

    output_dir = kwargs.get('output_dir', settings.EVAL_DIR + '_adhoc/')
    output_dir += networks_dict_key + '/' + methods_ids_key + '/'
    if 'legend_n_cols' not in kwargs:
        kwargs['legend_n_cols'] = 2 # 3 if (len(methods_ids) in [3,6]) else 2

    title_func = get_title_func(methods_ids_key)

    c_evs_dict = OrderedDict()
    for n_key, networks_ids in networks_dict.iteritems():
        c_evs_dict[n_key] = OrderedDict()
        for m_id in methods_ids:
            c_evs = []
            ms = []
            for n_id in networks_ids:
                m = Method.load_state(n_id, m_id)
                if m is None:
                    c_evs.append(None)
                    continue
                ms.append(m)
                c_evs.append(m.get_community_evolution())
            c_evs_dict[n_key][title_func(m_id)] = c_evs

    if draw_com_evs_sim:
        com_ev_eval_across_networks.evaluate_com_evs_against_other_across_networks_dict(
            networks_c_evs_dict=c_evs_dict,
            methods_ids=methods_ids,
            networks_dict=networks_dict,
            other_method_id=other_method_id,
            output_dir=output_dir,
            add_subdirs=False,
            com_evs_similarity_methods_ids=['(gl_p)(nmi)'],
            overwrite=True,
            draw_ranking_slopegraph=False,
            use_default_com_evs_sim_methods=False,
            bar_colors=colors,
            **kwargs
        )
    if draw_pr_roc:
        method_eval_pr_roc_across_networks.evaluate_methods_changepoint_classification_across_networks_dict(
            methods_ids=methods_ids,
            networks_dict=networks_dict,
            other_method_id=other_method_id,
            output_dir=output_dir,
            add_subdirs=False,
            overwrite=True,
            draw_ranking_slopegraph=False,
            method_title_func=title_func,
            bar_colors=colors,
            **kwargs
        )

    if draw_com_evs_scores:
        if other_method_id != 'gt_events':
            for n_key, networks_ids in networks_dict.iteritems():
                c_evs = []
                for n_id in networks_ids:
                    c_evs.append(CommunityEvolution.load(n_id, other_method_id))
                c_evs_dict[n_key][method_id_to_title(other_method_id)] = c_evs

        kwargs['scoring_methods_ids'] = kwargs.get(
            'scoring_methods_ids',
            ['(sgmq)(num)', '(avg_sn)(mod)',]# '(bic)']
        )

        com_ev_draw.draw_dict_networks_com_evs_scores_bars_avg(
            networks_c_evs_dict=c_evs_dict,
            output_dir=output_dir,
            add_subdirs=False,
            overwrite=True,
            use_default_scoring_methods=False,
            gt_title=method_id_to_title(other_method_id),
            bar_colors=colors,
            draw_pairwise_corr_matrices=True,
            **kwargs
        )
    return output_dir

def draw_plots_per_network_key(networks_dict_key, methods_ids_key, other_method_id='segment_ground_truth', draw_sim_scatter_plots=True, **kwargs):
    networks_dict = OrderedDict()
    for n_k, n_ids_str in report_params.REPORT_NETWORKS_DICT_BY_KEY.get(networks_dict_key).iteritems():
        networks_dict[n_k] = parse_user_args_string_into_values_list(n_ids_str)
    methods_ids = []
    for temp_m in report_params.REPORT_METHODS_BY_KEY[methods_ids_key]:
        methods_ids += parse_user_args_string_into_values_list(temp_m)

    dot_colors = kwargs.get('colors', _get_colors(methods_ids_key, **kwargs))
    output_dir = kwargs.get('output_dir', settings.EVAL_DIR + '_adhoc/')
    output_dir += networks_dict_key + '/' + methods_ids_key + '/'

    title_func = get_title_func(methods_ids_key)
    show_y_label_n_keys = kwargs.get('show_y_label_n_keys')

    include_network_keys = kwargs.get('include_network_keys')

    for n_key, networks_ids in networks_dict.iteritems():
        if include_network_keys is not None and n_key not in include_network_keys:
            continue
        c_evs_dict = OrderedDict()
        for m_id in methods_ids:
            c_evs = []
            for n_id in networks_ids:
                m = Method.load_state(n_id, m_id)
                c_evs.append(m.get_community_evolution() if m is not None else None)
            c_evs_dict[title_func(m_id)] = c_evs

        c_evs_gts_list = []
        for n_id in networks_ids:
            c_evs_gts_list.append(CommunityEvolution.load(n_id, other_method_id))

        if draw_sim_scatter_plots:
            show_y_label = True
            if show_y_label_n_keys is not None:
                show_y_label = n_key in show_y_label_n_keys
            com_ev_draw.draw_list_com_evs_similarities_to_other_scatter_avg(
                c_evs_dict=c_evs_dict,
                c_evs_gts_list=c_evs_gts_list,
                output_dir=output_dir,
                output_filename_prefix=n_key.replace(' ', '_').lower(),
                overwrite=True,
                show_y_label=show_y_label,
                dot_colors=dot_colors,
                **kwargs
            )

    # plots = []
    # for i,n_key in enumerate(networks_dict.keys()):
    #     fig = sg.fromfile(output_dir + 'scatter-com_evs_similarity/' + n_key + '_(avg_seg_t)(ami)-vs-(avg_sn_p)(ami).svg')
    #     plots.append(fig.getroot())
    #     plots[-1].moveto(i*480, 0)
    #
    # fig = sg.SVGFigure(width=str(480*5), height="400") # height 400 - no legend
    # fig.append(plots)
    # fig.save(output_dir + 'scatter-com_evs_similarity/combined.svg')

    return output_dir

def draw_methods_score_n_changepoints(networks_dict_key, methods_ids_key, **kwargs):
    networks_dict = OrderedDict()
    for n_k, n_ids_str in report_params.REPORT_NETWORKS_DICT_BY_KEY.get(networks_dict_key).iteritems():
        networks_dict[n_k] = parse_user_args_string_into_values_list(n_ids_str)
    methods_ids = []
    for temp_m in report_params.REPORT_METHODS_BY_KEY[methods_ids_key]:
        methods_ids += parse_user_args_string_into_values_list(temp_m)

    colors = kwargs.get('colors', _get_colors(methods_ids_key, **kwargs))

    output_dir = kwargs.get('output_dir', settings.EVAL_DIR + '_adhoc/')
    output_dir += networks_dict_key + '/' + methods_ids_key + '/'
    output_dir += 'timeseries-n_changepoints-score/'

    title_func = get_title_func(methods_ids_key)

    n_keys_show_legend = kwargs.get('n_keys_show_legend')
    show_x_label_n_keys = kwargs.get('show_x_label_n_keys')
    show_y_label_n_keys = kwargs.get('show_y_label_n_keys')

    extra_scoring_methods = []
    if kwargs.get('extra_scoring_methods'):
        for sc in kwargs.get('extra_scoring_methods'):
            extra_scoring_methods.append(scoring_methods_utils.create_scoring_method_by_id(sc))
        kwargs.pop('extra_scoring_methods')
    else:
        extra_scoring_methods = [scoring_methods_utils.create_scoring_method_by_id("(avg_sn)(mod)")]

    for n_key, n_ids_list in networks_dict.iteritems():
        if len(n_ids_list) != 1:
            raise Exception("Should be exactly one network")
        n_id = n_ids_list[0]
        temp_m_list = [Method.load_state(n_id, m_id) for m_id in methods_ids]
        line_colors = [] # to make sure colors are consistent
        m_list = []
        for i,m in enumerate(temp_m_list):
            if m is None:
                continue
            m_list.append(m)
            line_colors.append(colors[i % len(colors)])

        method_draw.draw_list_solution_scores_n_changepoints_timeseries(
            m_list,
            DynamicNetwork.load(n_id),
            output_images_dir=output_dir,
            extra_scoring_methods=extra_scoring_methods,
            output_filename_prefix=n_key.replace(' ', '_').lower(),
            format=kwargs.get('format', 'eps'),
            use_methods_scoring_methods=kwargs.get('use_methods_scoring_methods', False),
            method_title_func=title_func,
            overwrite=True,
            line_colors=line_colors,
            add_subdirs=False,
            show_legend=n_keys_show_legend is None or n_key in n_keys_show_legend,
            show_x_label=show_x_label_n_keys is None or n_key in show_x_label_n_keys,
            show_y_label=show_y_label_n_keys is None or n_key in show_y_label_n_keys,
            legend_bbox=None if kwargs.get('legend_bbox_n_keys') is None else kwargs.get('legend_bbox_n_keys').get(n_key),
            **kwargs
        )
    return output_dir

def draw_changepoints(network_id, methods_ids_key, **kwargs):
    methods_ids = []
    for temp_m in report_params.REPORT_METHODS_BY_KEY[methods_ids_key]:
        methods_ids += parse_user_args_string_into_values_list(temp_m)

    title_func = get_title_func(methods_ids_key)

    max_n_changepoints = kwargs.get('max_n_changepoints', 10)

    dn = DynamicNetwork.load(network_id)

    data = OrderedDict()
    for method_id in methods_ids:
        m = Method.load_state(network_id, method_id)
        if m is None:
            continue
        data[title_func(method_id)] = np.argsort(m.get_timepoints_scores())[-1:-max_n_changepoints-1:-1]

    x_labels_to_show = None
    if 'cow' in network_id:
        x_labels_to_show = [str(x) for x in range(1830, 2010, 10)] + ['1816', '2012']
    elif 'voteview' in network_id:
        x_labels_to_show = ['1789', '1799', '1809', '1819', '1829',
                            '1839', '1849', '1859', '1869', '1879',
                            '1889', '1899', '1909', '1919', '1929',
                            '1939', '1949', '1959', '1969', '1979',
                            '1989', '1999', '2009']
    # elif 'aging' in network_id:
    #     x_labels_to_show = ['20', '30', '50', '80']
    annotate_lines = None
    if 'aging' in network_id:
        ages = {}
        x_shift = lambda x: int(x) - 20
        for i,t in enumerate(dn.get_timepoints_labels()):
            ages[i] = x_shift(t)
        kwargs['x_coord'] = lambda x: ages[x]

        age_stages = [35, 50]
        if int(dn.get_timepoints_labels()[-1])>=82:
            age_stages.append(80)

        kwargs['annotate_lines'] = []
        for a in age_stages:
            kwargs['annotate_lines'].append({
                'x': x_shift(a),
                'x_label': a
            })

        kwargs['annotate_intervals'] = []
        if 80 not in age_stages:
            kwargs['annotate_intervals'].append(
                {
                    'x_start': x_shift(60),
                    'x_end': x_shift(80),
                    'x_start_label': 60,
                    'x_end_label': 80,
                }
            )

        for x in age_stages:
            kwargs['annotate_intervals'].append(
                {
                    'x_start': x_shift(x-2),
                    'x_end': x_shift(x+2),
                    'x_start_label': x-2,
                    'x_end_label': x+2,
                }
            )


        kwargs['show_y_ticks'] = False # only one method

    kwargs['output_dir'] = kwargs.get('output_dir', settings.EVAL_DIR + '_adhoc/')
    kwargs['output_filename'] = kwargs.get('output_filename', network_id.replace('/','_') + '-' + methods_ids_key + '-changepoints')
    kwargs['format'] = kwargs.get('format', 'eps')
    draw_utils.draw_changepoints(
        data=data,
        x_labels=dn.get_timepoints_labels(),
        x_labels_to_show=x_labels_to_show,
        **kwargs
    )

