import os
import numpy as np

from ..draw import ranking_draw
from .. import settings
from ..draw import draw_utils, corr_draw
from ..log import logger
from collections import OrderedDict
from ..classification_result import ClassificationResult

def draw_precision_recall(classification_results_dict, output_dir, output_filename='pr', **kwargs):
    logger.debug("Drawing Precision-Recall curve")
    if os.path.exists(output_dir + output_filename + '.' + settings.CHART_EXT):
        logger.debug("Precision-Recall curve already has been already drawn: %s"
                     % (output_dir + output_filename + '.' + settings.CHART_EXT)
        )
        if not kwargs.get('overwrite', True):
            logger.debug("No overwrite: aborting")
            return
    data = {}
    for label, cl_res in classification_results_dict.iteritems():
        data[label] = np.column_stack((cl_res.get_recalls(), cl_res.get_precisions()))
    draw_utils.draw_timeseries(
        data=data,
        output_dir=output_dir,
        output_filename=output_filename,
        x_label="Recall",
        y_label="Precision",
        x_lim=(0, 1),
        y_lim=(0, 1),
        show_legend=True,
        **kwargs
    )

def draw_roc(classification_results_dict, output_dir, output_filename='roc', **kwargs):
    logger.debug("Drawing ROC curve")
    if os.path.exists(output_dir + output_filename + '.' + settings.CHART_EXT):
        logger.debug("ROC curve already has been already drawn: %s"
                     % (output_dir + output_filename + '.' + settings.CHART_EXT)
        )
        if not kwargs.get('overwrite', True):
            logger.debug("No overwrite: aborting")
            return
    data = {}
    for label, cl_res in classification_results_dict.iteritems():
        data[label] = np.column_stack((cl_res.get_fprs(), cl_res.get_tprs()))

    draw_utils.draw_timeseries(
        data=data,
        output_dir=output_dir,
        output_filename=output_filename,
        x_label="False Positive Rate",
        y_label="True Positive Rate",
        x_lim=(0, 1),
        y_lim=(0, 1),
        show_legend=True,
        lines=[
            {  # annotate random performance
                'x_start': 0,
                'x_end': 1,
                'y_start': 0,
                'y_end': 1,
            }
        ],
        **kwargs
    )

def draw_bars(classification_results_dict, output_dir, output_filename='bars', **kwargs):
    logger.debug("Drawing bar plot (AUPR, AUROC, Max F1-score)")
    if os.path.exists(output_dir + output_filename + '.' + settings.CHART_EXT):
        logger.debug("Bar plot already has been already drawn: %s"
                     % (output_dir + output_filename + '.' + settings.CHART_EXT)
        )
        if not kwargs.get('overwrite', True):
            logger.debug("No overwrite: aborting")
            return
    data = OrderedDict()
    y_err = OrderedDict()
    show_error_bars = False

    corr_data = OrderedDict()
    corr_data["AUPR"] = []
    corr_data["AUROC"] = []
    corr_data["Max F1-score"] = []

    for label, cl_res_list in classification_results_dict.iteritems():
        if type(cl_res_list) is not list:
           cl_res_list = [cl_res_list]
        else:
            show_error_bars = True

        aupr_scores = [c.get_aupr() for c in cl_res_list]
        auroc_scores = [c.get_auroc() for c in cl_res_list]
        max_f1_scores = [c.get_max_f1score() for c in cl_res_list]

        data[label] = OrderedDict()
        data[label]["AUPR"] = np.average(aupr_scores)
        data[label]["AUROC"] = np.average(auroc_scores)
        data[label]["Max F1-score"] = np.average(max_f1_scores)

        if show_error_bars:
            y_err[label] = OrderedDict()
            y_err[label]["AUPR"] = np.std(aupr_scores)
            y_err[label]["AUROC"] = np.std(auroc_scores)
            y_err[label]["Max F1-score"] = np.std(max_f1_scores)

        corr_data["AUPR"] += aupr_scores
        corr_data["AUROC"] += auroc_scores
        corr_data["Max F1-score"] += max_f1_scores

    draw_utils.draw_bar(
        data=data,
        y_err=y_err,
        output_dir=output_dir,
        output_filename=output_filename,
        x_label="Metrics",
        y_label="Values",
        y_lim=(0, 1),
        show_legend=True,
        show_error_bars=show_error_bars,
        **kwargs
    )

    # draw pairwise correlation matrix for classification results
    if True: # or kwargs.get('draw_pairwise_corr_matrices', False) and len(corr_data.keys()) > 1:
        corr_draw.draw_pairwise_correlation_matrices(
            data=corr_data,
            corr_methods='all',
            output_dir=output_dir + 'pairwise_corr_matr/',
            output_filename_prefix='pr_roc',
            show_color_bar=True,
            save_data=True,
        )

def draw_bars_dict(classification_results_dict, output_dir, output_filename='bars', **kwargs):
    """
    Draw bars plots with groups
    :param classification_results_dict: group => series => classification_results list
    :param output_dir:
    :param output_filename:
    :param kwargs:
    :return:
    """
    show_error_bars = False
    data = OrderedDict()
    error_bars = OrderedDict()
    value_type_to_func = {
        "AUPR": ClassificationResult.get_aupr,
        "AUROC": ClassificationResult.get_auroc,
        "Max F1-score": ClassificationResult.get_max_f1score,
    }
    for value_type in ["AUPR", "AUROC", "Max F1-score"]:
        output_filename = value_type.lower().replace(" ", "_")
        if os.path.exists(output_dir + output_filename + '.' + settings.CHART_EXT):
            logger.debug("Bar plot already has been already drawn: %s"
                         % (output_dir + output_filename + '.' + settings.CHART_EXT)
            )
            if not kwargs.get('overwrite', True):
                logger.debug("No overwrite: aborting")
                continue
        for group, cl_res_dict in classification_results_dict.iteritems():
            for label, cl_res_list in cl_res_dict.iteritems():
                if type(cl_res_list) is not list:
                   cl_res_list = [cl_res_list]
                else:
                    show_error_bars = True
                if label not in data:
                    data[label] = OrderedDict()
                    error_bars[label] = OrderedDict()

                scores = [value_type_to_func[value_type](c) for c in cl_res_list if c is not None]
                scores = [s for s in scores if s is not None]

                if len(scores) == 0:
                    logger.warning("No valid scores for %s and %s" % (label, group))
                    data[label][group] = None
                    error_bars[label][group] = 0
                else:
                    data[label][group] = np.average(scores)
                    error_bars[label][group] = np.std(scores)

        draw_utils.draw_bar(
                data=data,
                y_err=error_bars,
                output_dir=output_dir,
                output_filename=output_filename,
                y_label=value_type,
                y_lim=(0, 1),
                show_error_bars=show_error_bars,
                **kwargs
            )

        # draw ranking
        if kwargs.get('draw_ranking_slopegraph', False) and len(data.keys()) > 1 and len(data.values()[0])>1:
            ranking_draw.draw_slopegraph(
                data=data,
                output_dir=output_dir + 'slopegraph-ranking/',
                output_filename=output_filename,
                as_ranking=True,
            )