import os
from collections import OrderedDict

from ..draw import ranking_draw
from .. import settings
from ..log import logger
from ..partition import Partition
from ..community_evolution import CommunityEvolution
from .. import method
import sankey
import draw_utils
import partition_draw
from .. import static_clustering
from ..com_evs_similarity_methods import com_evs_similarity_methods_utils
from ..scoring_methods import scoring_methods_utils
import corr_draw
import numpy as np

def draw_com_ev(com_ev, overwrite=True, **kwargs):
    draw_segments_community_evolution(com_ev)
    draw_segmentation(com_ev, show_ground_truth=kwargs.get('show_ground_truth', True), overwrite=overwrite)
    if kwargs.get('draw_partitions_qualities_timeseries', False):
        draw_list_partitions_qualities_timeseries(
            c_evs=[com_ev],
            output_dir=com_ev.get_images_dir(),
            show_ground_truth=kwargs.get('show_ground_truth', True),
            overwrite=overwrite
        )
    if kwargs.get('draw_partitions_qualities_heatmaps', False):
        draw_snapshots_segments_partitions_qualities_heatmaps(com_ev, overwrite=overwrite)
    if kwargs.get('draw_partitions_comparison_heatmaps', False):
        draw_self_partitions_comparison_heatmaps(com_ev, overwrite=overwrite)

def draw_segmentation(com_ev, output_dir=None, output_filename=None, others_to_compare=None, show_ground_truth=True,
                      overwrite=False):
    draw_segmentations(
        c_evs_list=[com_ev] + (others_to_compare or []),
        output_dir=output_dir or com_ev.get_images_dir(),
        output_filename=output_filename, show_ground_truth=show_ground_truth,
        overwrite=overwrite
    )

def draw_segments_community_evolution(com_ev, output_dir=None, output_filename=None, **kwargs):
    logger.debug("Drawing segments community evolution")
    sankey.generate_community_flow_webpage(
        partitions=CommunityEvolution.rename_clusters_as_ints_per_unique_adjacent_cluster(
            com_ev.get_segments_partitions()
        ),
        output_dir=output_dir or com_ev.get_images_dir(),
        output_filename=output_filename or 'sankey-partitions.html',
        changepoints_with_start_end=com_ev.get_changepoints_with_start_end(),
        timepoints_labels=None if com_ev.get_dynamic_network() is None else com_ev.get_dynamic_network().get_timepoints_labels(),
        objects_labels=None if com_ev.get_dynamic_network() is None else com_ev.get_dynamic_network().get_nodes_labels(),
        **kwargs
    )


def draw_segmentations(c_evs_list, output_dir, output_filename,
                       dynamic_network=None, show_ground_truth=True,
                       c_evs_titles=None, overwrite=False):
    if os.path.exists(output_dir + (output_filename or 'segmentation') + '.' + settings.CHART_EXT):
        logger.debug("Segmentation has been already drawn: %s"
                     % output_dir + (output_filename or 'segmentation') + '.' + settings.CHART_EXT)
        if not overwrite:
            logger.debug("No overwrite: aborting")
            return

    dynamic_network = dynamic_network or c_evs_list[0].get_dynamic_network()
    segments = OrderedDict()
    methods_ids = set()
    for i, cev in enumerate(c_evs_list):
        k = c_evs_titles[i] if c_evs_titles else cev.get_method_title()
        segments[k] = cev.get_segments_intervals()
        methods_ids.add(cev.get_method_id())
    if show_ground_truth and dynamic_network is not None:
        segment_ground_truth = dynamic_network.get_segment_ground_truth()
        if segment_ground_truth is not None and segment_ground_truth.get_method_id() not in methods_ids:
            segments[
                segment_ground_truth.get_method_title()
            ] = segment_ground_truth.get_segments_intervals()

    draw_utils.draw_segmentation(
        data=segments,
        x_label="Timepoints",
        output_dir=output_dir,
        output_filename=output_filename or 'segmentation',
    )

def draw_snapshots_segments_partitions_qualities_heatmaps(com_ev, **kwargs):
        output_dir = com_ev.get_images_dir() + 'heatmaps-partitions-qualities-per-t/'
        if os.path.exists(output_dir):
            logger.debug('Snapshots segments partitions qualities heatmaps has been already drawn: %s' % output_dir)
            if not kwargs.get('overwrite', True):
                logger.debug("No overwrite: aborting")
                return
        partition_draw.draw_partitions_qualities_heatmap(
            networks=com_ev.get_dynamic_network().get_snapshots(),
            partitions=com_ev.get_segments_partitions(),
            output_dir=output_dir,
            vertical_lines=[
                {
                    'x_list': com_ev.get_changepoints(),
                    'linewidth': 2,
                }
            ],
            x_labels=kwargs.get('x_labels'),
            y_labels=kwargs.get('y_labels', com_ev.get_segments_intervals_labels()),
            y_label=com_ev.get_method_title(),
        )

def draw_list_partitions_similarities_to_other_timeseries(c_evs_list, c_ev_gt, output_dir,
                                                     dyn_network=None,  overwrite=True):
    logger.debug("Drawing time-series of partition similarities to {c_ev_gt} per timepoint".format(
        c_ev_gt=c_ev_gt.get_method_title(),
    ))
    dn = dyn_network or c_evs_list[0].get_dynamic_network()
    output_dir += 'timeseries-snapshot-partitions_similarity/'

    vertical_lines = None
    if dn.get_segment_ground_truth() is not None:
        vertical_lines=[
            {
                'title': "ground truth",
                'x_list': [t - 0.5 for t in dn.get_segment_ground_truth().get_changepoints()],
                'y_text_relative': 0.2,
                'fontsize': 10,
            }
        ]
    for metric in settings.DEFAULT_PARTITION_SIMILARITY_METRICS:
        metric_title = Partition.SIMILARITY_METRICS[metric]
        if os.path.exists(output_dir + metric + '.' + settings.CHART_EXT):
            logger.debug('Timeseries of partition similarities has been already drawn: %s'
                         % output_dir + metric + '.' + settings.CHART_EXT)
            if not overwrite:
                logger.debug("No overwrite: aborting")
                continue
        data = OrderedDict()
        for i in xrange(len(dn.get_timepoints())):
            p_gt = c_ev_gt.get_snapshot_partition(i)
            if p_gt is None:
                continue
            for c_ev in c_evs_list:
                p_c = c_ev.get_snapshot_partition(i)
                if p_c is None:
                    continue
                if c_ev.get_method_title() not in data:
                    data[c_ev.get_method_title()] = []
                data[c_ev.get_method_title()].append(
                    (
                        i,
                        Partition.compute_partitions_similarity(
                            partition_true=p_gt,
                            partition_predicted=p_c,
                            metric=metric,
                            use_common_objects=True
                        )
                    )
                )
        if len(data) > 0:
            draw_utils.draw_timeseries(
                data=data,
                output_dir=output_dir,
                output_filename=metric,
                x_label="Snapshot",
                y_label=metric_title,
                y_lim=Partition.get_similarity_metric_limits(metric),
                # output_data_dir=images_dir + 'images_data/',
                vertical_lines=vertical_lines,
            )

def draw_list_partitions_qualities_timeseries(c_evs, output_dir, dynamic_network=None,
                                           show_ground_truth=True, overwrite=False):
    logger.debug("Drawing time-series of partition qualities")
    output_dir += 'timeseries-snapshot-partition_quality/'
    dynamic_network = dynamic_network or c_evs[0].get_dynamic_network()
    if os.path.exists(output_dir):
        logger.debug("Snapshots segments partitions quality plot comparison has been already drawn: %s" % output_dir)
        if not overwrite:
            logger.debug("No overwrite: aborting")
            return

    snapshots_partitions_dict = OrderedDict()
    for ce in c_evs + (dynamic_network.get_ground_truths_list() if show_ground_truth else []):
        if len(ce.get_segments_partitions()) == 0:
            logger.warning("Community evolution %s does not have segment partitions" % ce.get_method_id())
        snapshots_partitions_dict[ce.get_method_title()] = [
                ce.get_snapshot_partition(i) for i in xrange(len(dynamic_network.get_snapshots()))
        ]

    vertical_lines = None
    if dynamic_network.get_segment_ground_truth() is not None:
        vertical_lines=[
            {
                'title': "ground truth",
                'x_list': [t - 0.5 for t in dynamic_network.get_segment_ground_truth().get_changepoints()],
                'y_text_relative': 0.2,
                'fontsize': 10,
            }
        ]

    for metric in settings.DEFAULT_PARTITION_QUALITY_METRICS:
        metric_title = static_clustering.PARTITION_QUALITY_METRICS[metric]
        data = OrderedDict()
        for method_title, snapshots_partitions in snapshots_partitions_dict.iteritems():
            data[method_title] = []
            for i, network in enumerate(dynamic_network.get_snapshots()):
                if snapshots_partitions[i] is not None:
                    data[method_title].append(
                        (i, static_clustering.compute_partition_quality(network, snapshots_partitions[i], metric))
                    )
            if len(data[method_title]) == 0:
                del data[method_title]

        draw_utils.draw_timeseries(
            data=data,
            output_dir=output_dir,
            output_filename=metric,
            x_label="Snapshot",
            y_label=metric_title,
            y_lim=(0, 1.05) if metric != 'log_l' else (None, 0.0),
            # output_data_dir=output_dir + 'images_data/',
            vertical_lines=vertical_lines
        )

def draw_self_partitions_comparison_heatmaps(com_ev, overwrite=False):
    """
    Draw heatmaps-partitions comparison of self vs self
    :return:
    """
    logger.debug("Drawing heatmaps for '{method_id}' self partition comparison".format(
        method_id=com_ev.get_method_id(),
    ))

    partition_draw.draw_partitions_comparison_heatmaps(
        partitions_1=com_ev.get_segments_partitions(),
        partitions_2=com_ev.get_segments_partitions(),
        output_dir=com_ev.get_images_dir() + 'heatmaps-partitions/',
        x_labels=com_ev.get_segments_intervals_labels(),
        y_labels=com_ev.get_segments_intervals_labels(),
        x_label=com_ev.get_method_title(),
        y_label=com_ev.get_method_title(),
        overwrite=overwrite,
    )

def draw_list_com_evs_scores_bars(c_evs_list, output_dir, scoring_methods_ids=None,
                                  show_ground_truth=True, overwrite=True, **kwargs):
    c_evs_dict = OrderedDict()
    for c_ev in c_evs_list:
        c_evs_dict[c_ev.get_method_title()] = [c_ev]
    if show_ground_truth:
        for c_ev_gt in c_evs_list[0].get_dynamic_network().get_ground_truths_list():
            c_evs_dict[c_ev_gt.get_method_title()] = [c_ev_gt]
    draw_list_com_evs_scores_bars_avg(
        c_evs_dict=c_evs_dict,
        output_dir=output_dir,
        scoring_methods_ids=scoring_methods_ids,
        overwrite=overwrite,
        show_error_bars=False,
        **kwargs
    )

def draw_list_com_evs_scores_bars_avg(c_evs_dict, output_dir, scoring_methods_ids=None, overwrite=True, **kwargs):
    """
    Plot bars of community evolutions scores with error bars
    :param output_dir:
    :param c_evs_dict: Dictionary method_id => list of c_evs
    :param scoring_methods_ids:
    :param overwrite:
    :param kwargs:
    :return:
    """
    logger.debug("Evaluating community evolutions according to scoring methods via bar")

    output_dir += 'bar-score/'
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    # group scoring methods into different images
    # e.g., custom - contains scoring methods from community evolutions, default - default ones
    sc_m_dict = {}
    if scoring_methods_ids is not None and len(scoring_methods_ids) > 0:
        sc_m_dict['custom'] = scoring_methods_ids
    if kwargs.get('include_default_scoring_methods', True):
        sc_m_dict['default'] = settings.DEFAULT_SCORING_METHODS_IDS
    if len(sc_m_dict) == 0:
        logger.warning("No scoring methods for analysis")
        return

    output_filename_prefix = kwargs.get('output_filename_prefix')

    for sc_m_type, sc_m_ids_list in sc_m_dict.iteritems():
        if sc_m_ids_list is None or len(sc_m_ids_list) == 0:
            logger.warning("No scoring methods for %s analysis" % sc_m_type)
            continue

        if os.path.exists(output_dir + sc_m_type + '.' + settings.CHART_EXT):
            logger.debug(
                "Community evolution scores have been already computed: %s"
                % (output_dir + sc_m_type + '.' + settings.CHART_EXT)
            )
            if not overwrite:
                logger.debug("No overwrite: aborting")
                continue
        scoring_methods = [
            scoring_methods_utils.create_scoring_method_by_id(x)
            for x in sc_m_ids_list
        ]

        # for drawing pairwise correlation matrix
        data = OrderedDict()
        error_bars = OrderedDict()
        corr_data = OrderedDict()
        for m_title, c_evs_list in c_evs_dict.iteritems():
            data[m_title] = OrderedDict()
            error_bars[m_title] = OrderedDict()
            for sc_m in scoring_methods:
                scores = [sc_m.evaluate_community_evolution(c_ev) for c_ev in c_evs_list]
                scores = [s for s in scores if s is not None]
                if len(scores) > 0:
                    data[m_title][sc_m.get_id()] = np.average(scores)
                    error_bars[m_title][sc_m.get_id()] = np.std(scores)
                    if sc_m.get_id() not in corr_data:
                        corr_data[sc_m.get_id()] = []
                    corr_data[sc_m.get_id()] += scores
            if len(data[m_title]) == 0:
                del data[m_title]
                del error_bars[m_title]

        y_min = min(m.get_limits()[0] for m in scoring_methods if m.get_limits() is not None)
        y_max = max(m.get_limits()[1] for m in scoring_methods if m.get_limits() is not None)

        output_filename = sc_m_type
        if output_filename_prefix is not None:
            output_filename = '%s_%s' % (output_filename_prefix, output_filename)

        draw_utils.draw_bar(
            data=data,
            y_err=error_bars,
            output_dir=output_dir,
            output_filename=output_filename,
            x_label="Scoring methods",
            y_label="Scores",
            y_lim=(y_min, y_max),
            **kwargs
        )

        # draw pairwise correlation matrix for scoring methods
        if kwargs.get('draw_pairwise_corr_matrices', False) and len(corr_data.keys()) > 1:
            corr_draw.draw_pairwise_correlation_matrices(
                data=corr_data,
                corr_methods='all',
                output_dir=output_dir + 'pairwise_corr_matr/',
                output_filename_prefix=sc_m_type,
                show_color_bar=True,
                save_data=True,
            )


def draw_dict_networks_com_evs_scores_bars_avg(networks_c_evs_dict, output_dir, scoring_methods_ids=None, overwrite=True, **kwargs):
    """
    Plot bars of community evolutions scores with error bars
    :param output_dir:
    :param networks_c_evs_dict: Dictionary network_key => method_id => list of c_evs
    :param scoring_methods_ids:
    :param overwrite:
    :param kwargs:
    :return:
    """
    logger.debug("Evaluating community evolutions according to scoring methods via bar")

    if kwargs.get('add_subdirs', True):
        output_dir += 'bar-score/'
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)
    output_filename_prefix = kwargs.get('output_filename_prefix')

    sc_m_ids_list = []
    if kwargs.get('use_default_scoring_methods', True):
        sc_m_ids_list = settings.DEFAULT_SCORING_METHODS_IDS[:]
    if scoring_methods_ids is not None:
        sc_m_ids_list += scoring_methods_ids
        sc_m_ids_list = list(set(sc_m_ids_list))
    if len(sc_m_ids_list) == 0:
        logger.warning("No scoring methods for analysis")
        return
    gt_title = kwargs.get('gt_title')

    sc_limits = kwargs.get('sc_limits')

    for sc_m_id in sc_m_ids_list:
        threshold_values = [] # values to be shown not as bars, but as threshold value
        if os.path.exists(output_dir + sc_m_id + '.' + settings.CHART_EXT):
            logger.debug(
                "Community evolution scores have been already computed: %s"
                % (output_dir + sc_m_id + '.' + settings.CHART_EXT)
            )
            if not overwrite:
                logger.debug("No overwrite: aborting")
                continue
        scoring_method = scoring_methods_utils.create_scoring_method_by_id(sc_m_id)
        data = OrderedDict()
        error_bars = OrderedDict()
        for n_key, c_evs_dict in networks_c_evs_dict.iteritems():
            for m_title, c_evs_list in c_evs_dict.iteritems():
                scores = [scoring_method.evaluate_community_evolution(c_ev) if c_ev is not None else None for c_ev in c_evs_list]
                scores = [s for s in scores if s is not None]

                if m_title == gt_title:
                    threshold_values.append(np.average(scores))
                    continue

                # if len(scores) == 0:
                #     logger.warning("No valid scores for %s and network key %s" % (m_title, n_key))
                #     continue

                if m_title not in data:
                    data[m_title] = OrderedDict()
                    error_bars[m_title] = OrderedDict()
                data[m_title][n_key] = np.average(scores) if len(scores) > 0 else None
                error_bars[m_title][n_key] = np.std(scores) if len(scores) > 0 else None
        if len(data) == 0:
            logger.warning("No data for plotting")
            continue
        if sc_limits is not None and sc_m_id in sc_limits:
            y_min, y_max = sc_limits[sc_m_id]
        else:
            y_min = scoring_method.get_limits()[0] if scoring_method.get_limits() is not None else None
            y_max = scoring_method.get_limits()[1] if scoring_method.get_limits() is not None else None

        output_filename = sc_m_id
        if output_filename_prefix is not None:
            output_filename = '%s_%s' % (output_filename_prefix, output_filename)

        draw_utils.draw_bar(
            data=data,
            y_err=error_bars,
            output_dir=output_dir,
            output_filename=output_filename,
            y_lim=(y_min, y_max),
            y_label=scoring_methods_utils.get_scoring_method_title(sc_m_id) if kwargs.get('show_y_label', True) else "",
            threshold_values=None if len(threshold_values) == 0 else threshold_values,
            **kwargs
        )

        # draw ranking
        if kwargs.get('draw_ranking_slopegraph', False) and len(data.keys()) > 1 and len(data.values()[0])>1:
            ranking_draw.draw_slopegraph(
                data=data,
                output_dir=output_dir + 'slopegraph-ranking/',
                output_filename=sc_m_id,
                as_ranking=True,
            )

def draw_list_com_evs_similarities_to_other_bars(c_evs_list, c_ev_gt, output_dir,
                                                 com_evs_similarity_methods_ids=None,
                                                 dyn_network=None, overwrite=True, **kwargs):
    c_evs_dict = OrderedDict()
    for c_ev in c_evs_list:
        c_evs_dict[c_ev.get_method_title()] = [c_ev]

    draw_list_com_evs_similarities_to_other_bars_avg(
        c_evs_dict=c_evs_dict,
        c_evs_gts_list=[c_ev_gt],
        output_dir=output_dir,
        com_evs_similarity_methods_ids=com_evs_similarity_methods_ids,
        overwrite=overwrite,
        show_error_bars=False,
        **kwargs
    )

def draw_list_com_evs_similarities_to_other_bars_avg(c_evs_dict, output_dir,
                                                     c_evs_gts_list=None, c_evs_gts_dict=None,
                                                     com_evs_similarity_methods_ids=None,
                                                     **kwargs):
    if c_evs_gts_list is None and c_evs_gts_dict is None:
        raise Exception("Need other methods for comparison")
    # group scoring methods into different images
    # e.g., custom - contains scoring methods from community evolutions, default - default ones
    sim_m_dict = settings.DEFAULT_COM_EVS_SIMILARITY_METHODS_IDS_DICT if c_evs_gts_list is None or c_evs_gts_list[0].has_partitions()\
        else settings.DEFAULT_COM_EVS_SIMILARITY_AFFINITY_METHODS_IDS_DICT
    if com_evs_similarity_methods_ids is not None:
        sim_m_dict['custom'] = com_evs_similarity_methods_ids
    if len(sim_m_dict) == 0:
        logger.warning("No similarity methods for analysis")
        return

    output_dir += 'bar-com_evs_similarity/'
    for sim_m_type, sim_m_ids_list in sim_m_dict.iteritems():
        if sim_m_ids_list is None or len(sim_m_ids_list) == 0:
            logger.warning("No similarity methods for %s analysis" % sim_m_type)
            continue

        if os.path.exists(output_dir + sim_m_type + '.' + settings.CHART_EXT):
            logger.debug(
                "Community evolution similarities have been already computed: %s"
                % (output_dir + sim_m_type + '.' + settings.CHART_EXT)
            )
            if not kwargs.get('overwrite', True):
                logger.debug("No overwrite: aborting")
                continue
        com_evs_similarity_methods = [
            com_evs_similarity_methods_utils.create_com_evs_similarity_method_by_id(x)
            for x in sim_m_ids_list
        ]

        # for drawing pairwise correlation matrix
        data = OrderedDict()
        error_bars = OrderedDict()
        corr_data = OrderedDict()
        for m_title, c_evs_list in c_evs_dict.iteritems():
            data[m_title] = OrderedDict()
            error_bars[m_title] = OrderedDict()
            for sim_m in com_evs_similarity_methods:
                other_methods_list = c_evs_gts_list if c_evs_gts_dict is None else c_evs_gts_dict[m_title]
                scores = [sim_m.compute_similarity(c_evs_list[i], other_methods_list[i]) for i in xrange(len(c_evs_list))]
                scores = [s for s in scores if s is not None]
                if len(scores) > 0:
                    data[m_title][sim_m.get_id()] = np.average(scores)
                    error_bars[m_title][sim_m.get_id()] = np.std(scores)

                    if sim_m.get_id() not in corr_data:
                        corr_data[sim_m.get_id()] = []
                    corr_data[sim_m.get_id()] += scores
            if len(data[m_title]) == 0:
                del data[m_title]
                del error_bars[m_title]

        if len(data) > 0:
            y_min = min(m.get_limits()[0] for m in com_evs_similarity_methods if m.get_limits() is not None)
            y_max = max(m.get_limits()[1] for m in com_evs_similarity_methods if m.get_limits() is not None)
            draw_utils.draw_bar(
                data=data,
                y_err=error_bars,
                output_dir=output_dir,
                output_filename=sim_m_type,
                x_label="Metrics",
                y_label="Values",
                y_lim=(y_min, y_max),
                **kwargs
            )

            # draw pairwise correlation matrix for similarity methods
            if kwargs.get('draw_pairwise_corr_matrices', False) and len(corr_data.keys()) > 1:
                corr_draw.draw_pairwise_correlation_matrices(
                    data=corr_data,
                    corr_methods='all',
                    output_dir=output_dir + 'pairwise_corr_matr/',
                    output_filename_prefix=sim_m_type,
                    show_color_bar=True,
                    save_data=True,
                )

def draw_dict_networks_com_evs_similarities_to_other_bars_avg(networks_c_evs_dict, output_dir,
                                                     networks_c_evs_gts_list=None, c_evs_gts_dict=None,
                                                     com_evs_similarity_methods_ids=None,
                                                     **kwargs):
    """
    Compare com evs for all network keys
    :param networks_c_evs_dict:
    :param output_dir:
    :param networks_c_evs_gts_list:
    :param c_evs_gts_dict:
    :param com_evs_similarity_methods_ids:
    :param overwrite:
    :param kwargs:
    :return: Produce one figure per com ev similarity - for all networks keys and methods
    """

    if networks_c_evs_dict is None and networks_c_evs_gts_list is None:
        raise Exception("Need other methods for comparison")

    sim_m_list = []
    if kwargs.get('use_default_com_evs_sim_methods', True):
        sim_m_list = settings.DEFAULT_COM_EVS_SIMILARITY_METHODS_IDS_LIST[:]
    if com_evs_similarity_methods_ids is not None:
        sim_m_list += com_evs_similarity_methods_ids
        sim_m_list = list(set(sim_m_list))
    if len(sim_m_list) == 0:
        logger.warning("No similarity methods for analysis")
        return

    if kwargs.get('add_subdirs', True):
        output_dir += 'bar-com_evs_similarity/'

    for sim_m_id in sim_m_list:
        if os.path.exists(output_dir + sim_m_id + '.' + settings.CHART_EXT):
            logger.debug(
                "Community evolution similarities have been already computed: %s"
                % (output_dir + sim_m_id + '.' + settings.CHART_EXT)
            )
            if not kwargs.get('overwrite', True):
                logger.debug("No overwrite: aborting")
                continue
        com_evs_similarity_method = com_evs_similarity_methods_utils\
            .create_com_evs_similarity_method_by_id(sim_m_id)
        data = OrderedDict()
        error_bars = OrderedDict()
        for n_key, c_evs_dict in networks_c_evs_dict.iteritems():
            for m_title, c_evs_list in c_evs_dict.iteritems():
                other_methods_list = networks_c_evs_gts_list[n_key] if c_evs_gts_dict is None else c_evs_gts_dict[m_title][n_key]
                scores = [
                    com_evs_similarity_method.compute_similarity(c_evs_list[i], other_methods_list[i]) for i in xrange(len(c_evs_list))
                    if c_evs_list[i] is not None
                    ]
                scores = [s for s in scores if s is not None]

                if m_title not in data:
                    data[m_title] = OrderedDict()
                    error_bars[m_title] = OrderedDict()
                if len(scores) == 0:
                    logger.warning("No valid scores for %s and network key %s" % (m_title, n_key))
                    data[m_title][n_key] = None
                    error_bars[m_title][n_key] = 0
                data[m_title][n_key] = np.average(scores)
                error_bars[m_title][n_key] = np.std(scores)

        if len(data) == 0:
            logger.warning("No data for plotting")
            continue
        
        y_min = com_evs_similarity_method.get_limits()[0] if com_evs_similarity_method.get_limits() is not None else None
        y_max = com_evs_similarity_method.get_limits()[1] if com_evs_similarity_method.get_limits() is not None else None

        draw_utils.draw_bar(
            data=data,
            y_err=error_bars,
            output_dir=output_dir,
            output_filename=sim_m_id,
            y_label=com_evs_similarity_methods_utils.get_com_evs_similarity_title(sim_m_id),
            y_lim=(y_min, y_max),
            **kwargs
        )

        # draw ranking
        if kwargs.get('draw_ranking_slopegraph', False) and len(data.keys()) > 1 and len(data.values()[0])>1:
            ranking_draw.draw_slopegraph(
                data=data,
                output_dir=output_dir + 'slopegraph-ranking/',
                output_filename=sim_m_id,
                as_ranking=True,
            )

def draw_list_com_evs_similarities_to_other_scatter(c_evs_list, c_ev_gt, output_dir,
                                                    dyn_network=None, overwrite=True, **kwargs):
    c_evs_dict = OrderedDict()
    for c_ev in c_evs_list:
        c_evs_dict[c_ev.get_method_title()] = [c_ev]

    draw_list_com_evs_similarities_to_other_scatter_avg(
        c_evs_dict=c_evs_dict,
        c_evs_gts_list=[c_ev_gt],
        output_dir=output_dir,
        overwrite=overwrite,
        show_error_bars=False,
        **kwargs
    )

def draw_list_com_evs_similarities_to_other_scatter_avg(c_evs_dict, output_dir,
                                                        c_evs_gts_list=None, c_evs_gts_dict=None,
                                                        overwrite=True, **kwargs):
    if c_evs_gts_list is None and c_evs_gts_dict is None:
        raise Exception("Need other methods for comparison")
    logger.debug("Comparing community evolution similarities to other via scatter")
    output_dir += 'scatter-com_evs_similarity/'
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)
    use_avg = kwargs.get('use_avg', True)
    output_filename_prefix = kwargs.get('output_filename_prefix')
    com_evs_methods_ids_pairs = kwargs.get('com_evs_methods_ids_pairs', settings.DEFAULT_COM_EVS_SCATTER_PLOT_METHODS_IDS_PAIRS)

    for sim_m_id_1, sim_m_id_2 in com_evs_methods_ids_pairs:
        sim_m_1 = com_evs_similarity_methods_utils.create_com_evs_similarity_method_by_id(sim_m_id_1)
        sim_m_2 = com_evs_similarity_methods_utils.create_com_evs_similarity_method_by_id(sim_m_id_2)

        output_filename = '%s-vs-%s' % (sim_m_id_1, sim_m_id_2)
        if output_filename_prefix is not None:
            output_filename = '%s_%s' % (output_filename_prefix, output_filename)
        if os.path.exists(output_dir + output_filename + '.' + settings.CHART_EXT):
            logger.debug(
                "Community evolution similarities have been already computed: %s"
                % (output_dir + output_filename + '.' + settings.CHART_EXT)
            )
            if not overwrite:
                logger.debug("No overwrite: aborting")
                return

        data = OrderedDict()
        x_err = OrderedDict()
        y_err = OrderedDict()

        for m_title, c_evs_list in c_evs_dict.iteritems():
            other_methods_list = c_evs_gts_list if c_evs_gts_dict is None else c_evs_gts_dict[m_title]

            scores_1 = [sim_m_1.compute_similarity(c_evs_list[i], other_methods_list[i]) for i in xrange(len(c_evs_list)) if c_evs_list[i] is not None]
            scores_1 = [s for s in scores_1 if s is not None]
            scores_2 = [sim_m_2.compute_similarity(c_evs_list[i], other_methods_list[i]) for i in xrange(len(c_evs_list)) if c_evs_list[i] is not None]
            scores_2 = [s for s in scores_2 if s is not None]

            if len(scores_1) > 0 and len(scores_2) > 0:
                if not use_avg: # each score-pair is by itself
                    data[m_title] = []
                    for i in xrange(max(len(scores_1), len(scores_2))):
                        data[m_title].append((scores_1[i], scores_2[i]))
                else:
                    data[m_title] = [(np.average(scores_1), np.average(scores_2))]
                    x_err[m_title] = [(np.std(scores_1))]
                    y_err[m_title] = [(np.std(scores_2))]
            else:
                data[m_title] = []


        if len(data) > 0:
            x_lim = kwargs.get('x_lim') or sim_m_1.get_limits()
            x_range = None if x_lim[0] is None or x_lim[1] is None else x_lim[1] - x_lim[0]
            y_lim = kwargs.get('y_lim') or sim_m_2.get_limits()
            y_range = None if y_lim[0] is None or y_lim[1] is None else y_lim[1] - y_lim[0]
            draw_utils.draw_scatter(
                data=data,
                x_err=x_err if use_avg else None,
                y_err=y_err if use_avg else None,
                output_dir=output_dir,
                output_filename=output_filename,
                x_label=kwargs.get(
                        'x_label',
                        com_evs_similarity_methods_utils.get_com_evs_similarity_title(sim_m_1.get_id())
                    ) if kwargs.get('show_x_label', True) else None,
                y_label=kwargs.get(
                        'y_label',
                        com_evs_similarity_methods_utils.get_com_evs_similarity_title(sim_m_2.get_id())
                    ) if kwargs.get('show_y_label', True) else None,
                x_lim=x_lim if x_range is None else (x_lim[0] - x_range * 0.05, x_lim[1] + x_range * 0.05),
                y_lim=y_lim if y_range is None else (y_lim[0] - y_range * 0.05, y_lim[1] + y_range * 0.05),
                show_error_bars=kwargs.get('show_error_bars', True),
                format=kwargs.get('format', settings.CHART_EXT),
                show_legend=kwargs.get('show_legend', True),
                annotate=kwargs.get('annotate') if kwargs.get('show_y_label', True) else None,
                legend_size=kwargs.get('legend_size'),
                dot_colors=kwargs.get('dot_colors'),
                legend_n_cols=kwargs.get('legend_n_cols'),
                x_label_size=kwargs.get('x_label_size'),
                y_label_size=kwargs.get('y_label_size'),
                legend_bbox=kwargs.get('legend_bbox'),
                dot_size=kwargs.get('dot_size'),
                tick_size=kwargs.get('tick_size'),
            )