import os
from .. import settings
import draw_utils
import numpy as np
from scipy.stats import pearsonr, spearmanr
from ..log import logger

CORR_METHODS_TITLES = {
    'pearson': "Pearson",
    'spearman': "Spearman",
}

_CORR_METHODS_LIMITS = {
    'pearson': (-1.0, 1.0),
    'spearman': (-1.0, 1.0),
}
_CORR_METHODS_FUNCS = {
    'pearson': pearsonr,
    'spearman': spearmanr,
}

def draw_pairwise_correlation_matrices(data, corr_methods, output_dir,
                 output_filename_prefix, **kwargs):
    corr_methods = CORR_METHODS_TITLES.keys() if corr_methods == 'all' else corr_methods
    for corr_method in corr_methods:
        draw_pairwise_correlation_matrix(
            data=data,
            corr_method=corr_method,
            output_dir=output_dir,
            output_filename_prefix=output_filename_prefix,
            **kwargs
        )

def draw_pairwise_correlation_matrix(data, corr_method, output_dir,
                 output_filename_prefix, format = settings.CHART_EXT, show_color_bar=True, **kwargs):

    logger.debug("Computing pairwise correlations matrix for {keys} (for {dir}/{fnm}) using {corr} correlation".format(
        keys=data.keys(),
        dir=output_dir,
        fnm=output_filename_prefix,
        corr=corr_method
    ))

    n = len(data.keys())
    corr_matr = np.zeros((n, n))
    pval_matr = np.zeros((n, n))
    corr_func = _CORR_METHODS_FUNCS[corr_method]
    x_labels = []
    y_labels = []
    output_filename = output_filename_prefix + '-' + corr_method

    for i,d1 in enumerate(data.iteritems()):
        x_labels.append(d1[0])
        y_labels.append(d1[0])
        for j,d2 in enumerate(data.iteritems()):
            c_value = (_CORR_METHODS_LIMITS[corr_method][1], 0.0)
            # to avoid issues when all elements are the same (e.g., [1,1,1] and [1,1,1])
            if d1[1] != d2[1]:
                c_value = corr_func(d1[1], d2[1])
            corr_matr[i,j], pval_matr[i,j] = c_value

    if kwargs.get('save_data', True):
        output_data_dir = output_dir + '_data/'
        if not os.path.exists(output_data_dir):
            os.makedirs(output_data_dir)
        for d_type in ['corr', 'pval']:
            out_data_filename = output_data_dir + 'data-%s-%s.txt' % (output_filename, d_type)
            np.savetxt(
                fname=out_data_filename,
                X=corr_matr if d_type == 'corr' else pval_matr
            )

    draw_utils.draw_heatmap(
        data=corr_matr,
        output_dir=output_dir,
        output_filename=output_filename,
        x_labels=x_labels,
        y_labels=y_labels,
        format=format,
        show_color_bar=show_color_bar,
        color_bar_limits=_CORR_METHODS_LIMITS[corr_method],
        **kwargs
    )