import os

import matplotlib
matplotlib.use('Agg') # Must be before importing matplotlib.pyplot or pylab!
import matplotlib.pyplot as plt
from matplotlib.ticker import MaxNLocator, MultipleLocator
import matplotlib.patches as patches
import numpy as np
from .. import settings
from ..log import logger
import cPickle

def draw_timeseries(data, output_dir, output_filename, y_err=None,
                    x_label="", y_label="", x_lim=None, y_lim=None,
                    show_legend=True, output_data_dir=None, format = settings.CHART_EXT, **kwargs):
    line_colors = kwargs.get('line_colors', settings.LINE_COLORS)
    #needed to set ticks for axis x
    data_x,x = [],[]
    out_data_file = None
    if output_data_dir is not None:
        if not os.path.exists(output_data_dir):
            os.makedirs(output_data_dir)
        out_data_file = open(output_data_dir + "data_%s.txt" % output_filename, 'wb')
        out_data_file.write("timeseries\n")

    if kwargs.get('save_data', True):
        if not os.path.exists(output_dir+'_data/'):
            os.makedirs(output_dir+'_data/')
        with open(output_dir+'_data/' + output_filename + '.cpickle', 'wb') as pickle_file:
            cPickle.dump(data, pickle_file)

    for i_series, series in enumerate(data.keys()):
        series_data = data[series]

        #limit on max number of series in one plot
        if i_series >= len(settings.LINE_MARKERS):
            raise Exception("Too many series: %s" % len(data.keys()))

        #get data for axis x and axis y
        if isinstance(series_data, np.ndarray):
            data_x = series_data[:, 0]
            data_y = series_data[:, 1]
        else:
            data_x = [d[0] for d in series_data]
            data_y = [d[1] for d in series_data]

        if out_data_file is not None:
            out_data_file.write("series %s\n" % series)
            out_data_file.writelines("%s,%s\n" % (d[0], d[1]) for d in series_data)

        #axis x need to numeric - use integers
        #later set labels by using ticks for axis x
        #x = range(0, len(data_x))

        #draw a plot for a series
        #NOTE: eps does not support alpha
        plt.plot(data_x, data_y,
            settings.LINE_MARKERS[i_series] + '-',
            label=series,
            color=line_colors[i_series % len(line_colors)] if line_colors else None,
            ms=12 if len(data_x) < 2 else 5 if len(data_x) < 20 else 0,
            lw=0 if len(data_x) < 2 else 1 if len(data_x) < 20 else 4,
            mec=line_colors[i_series % len(line_colors)] if line_colors else None,
            alpha=0.8
        )

        #add error bars
        if y_err is not None:
            plt.errorbar(
                x,
                data_y,
                yerr = y_err.get(series,None),
                c = line_colors[i_series]  if line_colors else None
            )

    if out_data_file is not None:
        out_data_file.close()

    #set min and max values
    if x_lim is not None:
        plt.xlim(x_lim)
    if y_lim is not None:
        plt.ylim(y_lim)

    # Annotations
    vertical_lines = kwargs.get('vertical_lines')
    if vertical_lines is not None:
        for line in vertical_lines:
            for x_line in line['x_list']:
                if 'title' in line and len(line['x_list']) < 20:
                    plt.text(
                        x=x_line,
                        y=line['y_text_relative'] * (plt.ylim()[1] - plt.ylim()[0]) + plt.ylim()[0],
                        s=line['title'],
                        rotation='vertical',
                        fontsize=line.get('fontsize', 12),
                        alpha=line.get('alpha', 0.8),
                    )
                plt.axvline(
                    x=x_line,
                    color=line.get('color', 'k'),
                    linestyle=line.get('linestyle', '--'),
                    linewidth=line.get('linewidth', 1 if len(line['x_list']) < 20 else 0.5) ,
                    alpha=line.get('alpha', 0.8),
                )
    lines = kwargs.get('lines')
    if lines is not None and len(lines) < 20:
        for line in lines:
            plt.plot(
                [line.get('x_start'), line.get('x_end')],
                [line.get('y_start'), line.get('y_end')],
                color=line.get('color', 'k'),
                linestyle=line.get('linestyle', '--'),
                linewidth=line.get('linewidth', 1),
                alpha=line.get('alpha', 0.8),
            )
    dots = kwargs.get('dots')
    if dots is not None:
       for d in dots:
           for x,y in d['points_list']:
               plt.plot(
                   x, y,
                   color=d.get('color','black'),
                   marker=d.get('marker','o'),
                   alpha=d.get('alpha', None),
                   markersize=d.get('markersize', 10 if len(dots) < 20 else 2)
               )

    #set title
    #if params.get('title','') != '':ylim
    #    plt.title(params.get('title',''))

    #set labels for axis x and axis y
    plt.xlabel(x_label, fontsize=30 if len(x_label)<30 else 20)
    plt.ylabel(y_label, fontsize=30 if len(y_label)<30 else 20)

    lgd = None
    if show_legend:
        ncol = kwargs.get('legend_n_cols', 2) # int(len(data.keys()) / 8.0 + 1)
        lgd = plt.legend(
            loc='upper center',
            bbox_to_anchor=kwargs.get('legend_bbox') or (0.5, -0.14),
            prop={
                'size':24 if max(map(len,data.keys())) < 70 else 16
            },
            framealpha=0.5,
            ncol=ncol,
            numpoints=1, # show only one mark (instead of two)
        )

    plt.tick_params(
        axis='both',
        labelsize=22,
    )

    # set ticks for axis x
    plt.gca().xaxis.set_major_locator(
        MultipleLocator(
            base=1 if len(data_x) <= 10 else
                5 if len(data_x) <= 50 else
                10 if len(data_x) <= 100 else 20
        ) if len(data_x) <= 200 else MaxNLocator(nbins=min(len(data_x) + 1, 20))
    )
    if len(data_x) > 50:
        plt.gca().xaxis.set_minor_locator(
            MultipleLocator(
                base=5 if len(data_x) <= 100 else 10
            ) if len(data_x) < 200 else MaxNLocator(nbins=min(len(data_x) + 1, 40))
        )
    # draw a grid
    # plt.grid()

    #create output directory if it does not exists
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    #save plot to a file
    output_path = output_dir + output_filename + '.' + format
    logger.debug("Saving timeseries plot to %s" % output_path)
    plt.savefig(output_path, bbox_extra_artists=(lgd,) if lgd else None, bbox_inches='tight', format=format, dpi=settings.CHART_DPI)

    #clear plot for the next plot
    plt.close()

    return output_path

def draw_segmentation(data, output_dir, output_filename,
                    x_label="", y_label="",
                    show_legend=True,
                    output_data_dir=None, format = settings.CHART_EXT, **kwargs):
    """
    Draw line segmentations. Each segmentation is depicted as a sequence of horizontal lines
    :param data:
    :param output_dir:
    :param output_filename:
    :param x_label:
    :param y_label:
    :param show_legend:
    :param output_data_dir:
    :param format:
    :param kwargs:
    :return:
    """

    min_x = 0
    max_x = 0
    n_series = len(data.keys())
    for i_series, series in enumerate(data.keys()):
        x_list = []
        y_value = len(data.keys()) - i_series - 0.5
        c_value = settings.LINE_COLORS[i_series % len(settings.LINE_COLORS)] if settings.LINE_COLORS else None
        for s in data[series]:
            x_list += [s[0], s[1], None]
            min_x = min(min_x, s[0], s[1])
            max_x = max(max_x, s[0], s[1])

            # denote endpoints with markers
            if s[0] != s[1]:
                for x_value, marker in [(s[0], '<'), (s[1], '>')]:
                    plt.plot(
                        x_value, y_value,
                        color=c_value,
                        marker=marker,
                        markersize=5 if len(data[series]) < 10 else 1,
                    )
            else:
                plt.plot(
                    s[0], y_value,
                    color=c_value,
                    marker='D',
                    markersize=5 if len(data[series]) < 10 else 1,
                )

        y_list = [y_value, y_value, None] * len(data[series])

        plt.plot(
            x_list, y_list,
            label=series,
            color=c_value,
            linewidth=2 if n_series < 20 else 0.5,
        )

    #set labels for axis x and axis y
    plt.xlabel(x_label, fontsize=20 if len(x_label)<30 else 14)
    plt.ylabel(y_label, fontsize=20 if len(y_label)<30 else 14)

    lgd = None
    if show_legend and n_series < 20:
        ncol = kwargs.get('legend_n_cols', 2) # int(len(data.keys()) / 8.0 + 1)
        lgd = plt.legend(
            loc='upper center',
            bbox_to_anchor=(0.5, -0.2),
            prop={
                'size': 16 if max(map(len,data.keys())) < 50 else 10
            },
            framealpha=0.5,
            ncol=ncol,
        )

    plt.ylim((0, len(data.keys())))
    # remove y ticks
    plt.gca().yaxis.set_major_locator(plt.NullLocator())
    # set ticks for axis x
    plt.gca().xaxis.set_major_locator(MaxNLocator(integer=True, nbins=min(max_x-min_x+1, 20)))

    # draw a grid
    plt.grid()

    #create output directory if it does not exists
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    #save plot to a file
    output_path = output_dir + output_filename + '.' + format
    logger.debug("Saving segmentation plot to %s" % output_path)
    plt.savefig(output_path, bbox_extra_artists=(lgd,), bbox_inches='tight', format=format, dpi=settings.CHART_DPI)

    #clear plot for the next plot
    plt.close()

    return output_path


def draw_heatmap(data, output_dir, output_filename, x_labels=None, y_labels=None,
                 format = settings.CHART_EXT, show_color_bar=True, **kwargs):
    fig, ax = plt.subplots()
    colormap = plt.cm.Blues  # nlcmap(plt.cm.Blues, [0, 20, 35, 50])
    heatmap = ax.pcolor(data.T, cmap=colormap, edgecolors='gray', vmin=kwargs.get('vmin'), vmax=kwargs.get('vmax'))

    # turn off the frame
    ax.set_frame_on(False)

    # put the major ticks at the middle of each cell
    if kwargs.get('make_square_shape', True):
        max_xticks = max_yticks = max(data.shape) # for non-square heatmaps, to make sure they are not scaled to square
    else:
        max_xticks, max_yticks = data.shape
    ax.set_xticks(np.arange(max_xticks) + 0.5, minor=False)
    ax.set_yticks(np.arange(max_yticks) + 0.5, minor=False)

    # want a more natural, table-like display
    ax.invert_yaxis()
    ax.xaxis.tick_top()

    x_labels_size = None
    if x_labels is not None:
        x_labels_size = 12 if max(map(len, x_labels)) < 10 else 8
    ax.set_xticklabels(x_labels or [], minor=False, size=x_labels_size)

    y_labels_size = None
    if y_labels is not None:
        y_labels_size = 12 if max(map(len, y_labels)) < 10 else 8
    ax.set_yticklabels(y_labels or [], minor=False, size=y_labels_size)

    # rotate the x labels
    plt.xticks(rotation=90)
    ax.grid(False)
    # Turn off all the ticks
    ax = plt.gca()
    for t in ax.xaxis.get_major_ticks():
        t.tick1On = False
        t.tick2On = False
    for t in ax.yaxis.get_major_ticks():
        t.tick1On = False
        t.tick2On = False

    # Annotations
    vertical_lines = kwargs.get('vertical_lines')
    if vertical_lines:
        for line in vertical_lines:
            for x_line in line['x_list']:
                if 'title' in line and len(line['x_list']) < 20:
                    ax.text(
                        x=x_line,
                        y=line['y_text_relative'] * data.shape[1],
                        s=line['title'],
                        rotation='vertical',
                        fontsize=line.get('fontsize', 12),
                        alpha=line.get('alpha', 0.8),
                    )

                ax.axvline(
                    x=x_line,
                    ymin=max(0, 1 - (1.0 * (data.shape[1]))/(data.shape[0] - 0.5)),
                    ymax=1,
                    color=line.get('color', 'k'),
                    linestyle=line.get('linestyle','--'),
                    linewidth=line.get('linewidth', 1 if len(line['x_list']) < 20 else 0.5),
                    alpha=line.get('alpha', 0.8),
                )
    horizontal_lines = kwargs.get('horizontal_lines')
    if horizontal_lines:
        for line in horizontal_lines:
            for y_line in line['y_list']:
                if 'title' in line:
                    ax.text(
                        x=line['x_text_relative'] * data.shape[0],
                        y=y_line,
                        s=line['title'],
                        rotation='horizontal',
                        fontsize=line.get('fontsize', 12),
                        alpha=line.get('alpha', 0.8),
                    )
                ax.axhline(
                    y=y_line,
                    xmin=0,
                    xmax=min(1, (1.0 * data.shape[0])/data.shape[1]),
                    color=line.get('color', 'k'),
                    linestyle=line.get('linestyle','--'),
                    linewidth=line.get('linewidth'),
                    alpha=line.get('alpha', 0.8),
                )

    # if len(data)>50:
    #     plt.rcParams['font.size'] = 5
    # else:
    #     plt.rcParams['font.size'] = 8

    if kwargs.get('x_label') is not None:
        plt.xlabel(
            kwargs.get('x_label'),
            fontsize=20 if len(kwargs.get('x_label')) < 30 else
            12 if len(kwargs.get('x_label')) < 50 else
            10 if len(kwargs.get('x_label')) < 70 else 8
        )
        ax.xaxis.set_label_position('top')
    if kwargs.get('y_label') is not None:
        plt.ylabel(
            kwargs.get('y_label'),
            fontsize=20 if len(kwargs.get('y_label')) < 30 else
            12 if len(kwargs.get('y_label')) < 50 else
            10 if len(kwargs.get('y_label')) < 70 else 8
        )

    if show_color_bar:
        cbar = plt.colorbar(heatmap)
        if kwargs.get('color_bar_limits') is not None:
            lim = kwargs.get('color_bar_limits')
            cbar.set_clim(vmin=lim[0], vmax=lim[1])
            cbar.draw_all() # redraw colorbar with new limits
        if kwargs.get('color_bar_label') is not None:
            cbar.set_label(kwargs.get('color_bar_label'), fontsize=20 if len(kwargs.get('color_bar_label')) < 30 else 14)

    #create output directory if it does not exists
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    #save chart to a file
    output_path = output_dir + output_filename + '.' + format
    logger.debug("Saving heatmap plot to %s" % output_path)
    plt.tight_layout()
    plt.savefig(output_path, dpi=settings.CHART_DPI)

    #clear plot for the next chart
    plt.close()

    return output_path

def draw_bar(data, output_dir, output_filename, y_err=None,
                    x_label="", y_label="", y_lim=None,
                    show_legend=True, format = settings.CHART_EXT, **kwargs):
    bar_colors = kwargs.get('bar_colors', settings.LINE_COLORS)
    bar_width = kwargs.get('bar_width', 0.5)
    #needed to set ticks for axis x
    data_x,x = [],[]
    groups = None
    n_series = len(data.keys())

    if kwargs.get('figsize'):
        fig = plt.figure(figsize=kwargs.get('figsize'))

    if kwargs.get('save_data', True):
        if not os.path.exists(output_dir+'_data/'):
            os.makedirs(output_dir+'_data/')
        with open(output_dir+'_data/' + output_filename + '.cpickle', 'wb') as pickle_file:
            cPickle.dump(data, pickle_file)
        if y_err is not None:
            with open(output_dir + '_data/' + output_filename + '_yerr.cpickle', 'wb') as pickle_file:
                cPickle.dump(y_err, pickle_file)

    for i_series, series in enumerate(data.keys()):
        #limit on max number of series in one plot
        if i_series >= 15:
            raise Exception("Too many series: %s" % len(data.keys()))

        if groups is not None and groups != data[series].keys():
            raise Exception("Data groups should be the same across different series")
        groups = data[series].keys()

        values = data[series].values()
        yerr = None
        present_values = [i for i in xrange(len(groups)) if values[i] is not None]
        values_present = [values[i] for i in present_values]
        if (y_err is not None and y_err.has_key(series) and kwargs.get('show_error_bars', True)):
            yerr = [y_err.get(series).values()[i] for i in present_values]

        if len(present_values) > 0:
            plt.bar(
                [bar_width + i * (n_series + 1) * bar_width + i_series * bar_width  for i in present_values],
                values_present,
                width=bar_width,
                label=series,
                yerr=yerr,
                color=bar_colors[i_series % len(bar_colors)] if bar_colors else None,
                alpha=0.8,
                error_kw=dict(ecolor='black'),
                log=kwargs.get('log_scale', False),
            )

        missing_values = [i for i in xrange(len(groups)) if values[i] is None]
        if len(missing_values)>0:
            default_missing_value = kwargs.get('default_missing_value', 0)
            plt.bar(
                [bar_width + i * (n_series + 1) * bar_width + i_series * bar_width  for i in missing_values],
                [default_missing_value] * len(missing_values),
                width=bar_width,
                label=None if len(present_values) > 0 else series,
                hatch = '/' if default_missing_value>0 else None,
                color=bar_colors[i_series % len(bar_colors)] if bar_colors else None,
                alpha=0.8,
            )

    threshold_values = kwargs.get('threshold_values', None)
    if threshold_values is not None:
        for i,v in enumerate(threshold_values):
            plt.plot(
                [
                    (i + 1) * bar_width + i * n_series * bar_width - 0.25 * bar_width,
                    (i + 1) * bar_width + i * n_series * bar_width + (n_series + 0.25) * bar_width
                ],
                [v,v],
                color='black',
                linestyle='dotted',
                linewidth=1,
                alpha=1,
            )

    xticks_rotation = kwargs.get('xticks_rotation', 0 if sum(map(len, groups))<20 else 30)
    xticks_shift = 0# kwargs.get('xticks_shift') or (0 if xticks_rotation in [0, 90] else 0.9)
    plt.xticks(
        [bar_width + i * (n_series + 1) * bar_width + bar_width * n_series / 2.0  - xticks_shift * bar_width * n_series for i in xrange(len(groups))],
        groups,
        rotation=xticks_rotation,
    )
    plt.xlim(0, bar_width * n_series * len(groups) + bar_width * (len(groups) + 1))
    plt.tick_params(
        axis='x',          # changes apply to the x-axis
        which='both',      # both major and minor ticks are affected
        bottom='off',      # ticks along the bottom edge are off
        top='off',         # ticks along the top edge are off
        labelbottom='on', # labels along the bottom edge are on
        labelsize=24 if sum(map(len, groups))<70 else 16,
    )

    if not kwargs.get('show_x_ticks', True):
        plt.gca().xaxis.set_major_locator(plt.NullLocator())

    plt.tick_params(
        axis='y',
        labelsize=22,
    )

    if kwargs.get('y_ticks'):
        plt.yticks(kwargs.get('y_ticks'))

    #set min and max values
    if y_lim is not None:
        plt.ylim(y_lim)

    # annotate
    annotate_text = None
    if kwargs.get('annotate'):
        annotate_params = kwargs.get('annotate')
        annotate_text = plt.text(
            annotate_params.get('x'),
            annotate_params.get('y'),
            annotate_params.get('text'),
            ha="center",
            va="center",
            rotation='vertical',
            size=annotate_params.get('size'),
            weight=annotate_params.get('weight')
        )

    #set labels for axis x and axis y
    plt.xlabel(x_label, fontsize=30 if len(x_label)<30 else 20)
    plt.ylabel(y_label, fontsize=30 if len(y_label)<30 else 20)

    lgd = None
    if show_legend:
        ncol = kwargs.get('legend_n_cols', 2)# int(len(data.keys()) / 8.0 + 1)
        lgd = plt.legend(
            loc='upper center',
            bbox_to_anchor=kwargs.get('legend_bbox', (0.5, -0.39)),
           #  bbox_to_anchor=(0.5, -0.2 if xticks_rotation == 0 else -0.22),
            prop={
                'size':24 if max(map(len,data.keys())) < 50 else 16
            },
            framealpha=0.5,
            ncol=ncol,
        )


    extra_artists = []
    if lgd is not None:
        extra_artists.append(lgd)
    if annotate_text is not None:
        extra_artists.append(annotate_text)

    #create output directory if it does not exists
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    #save chart to a file
    output_path = output_dir + output_filename + '.' + format
    logger.debug("Saving bar plot to %s" % output_path)
    plt.savefig(output_path, bbox_extra_artists=extra_artists, bbox_inches='tight', dpi=settings.CHART_DPI)

    #clear plot for the next chart
    plt.close()

    return output_path


def draw_scatter(data, output_dir, output_filename, x_err=None, y_err=None,
                 x_label="", y_label="", x_lim=None, y_lim=None,
                 format = settings.CHART_EXT, show_legend=True, **kwargs):

    colors = kwargs.get('dot_colors', settings.LINE_COLORS)

    if kwargs.get('save_data', True):
        if not os.path.exists(output_dir + '_data/'):
            os.makedirs(output_dir + '_data/')
        with open(output_dir + '_data/' + output_filename + '_data.cpickle', 'wb') as pickle_file:
            cPickle.dump(data, pickle_file)
        if x_err is not None:
            with open(output_dir + '_data/' + output_filename + '_xerr.cpickle', 'wb') as pickle_file:
                cPickle.dump(x_err, pickle_file)
        if y_err is not None:
            with open(output_dir + '_data/' + output_filename + '_yerr.cpickle', 'wb') as pickle_file:
                cPickle.dump(y_err, pickle_file)

    for i_series, series in enumerate(data.keys()):
        xs = []
        ys = []
        for x,y in data[series]:
            xs.append(x)
            ys.append(y)
        plt.scatter(
            xs,
            ys,
            label=series,
            s=kwargs.get('dot_size') or 75,
            color=colors[i_series],
            marker=settings.LINE_MARKERS[i_series],
            alpha=kwargs.get('alpha', 1)
        )

        if kwargs.get('show_error_bars', True) and\
                ((x_err is not None and x_err.has_key(series)) or (y_err is not None and y_err.has_key(series))):
            plt.errorbar(
                xs,
                ys,
                xerr=x_err.get(series)
                    if (x_err is not None and x_err.has_key(series)) else None,
                yerr=y_err.get(series)
                    if (y_err is not None and y_err.has_key(series)) else None,
                ecolor=colors[i_series] if colors else 'black',
            )


    # annotate
    annotate_text = None
    if kwargs.get('annotate'):
        annotate_params = kwargs.get('annotate')
        annotate_text = plt.text(
            annotate_params.get('x'),
            annotate_params.get('y'),
            annotate_params.get('text'),
            ha="center",
            va="center",
            rotation='vertical',
            size=annotate_params.get('size'),
            weight=annotate_params.get('weight'),
        )

    #set title for the chart
    if kwargs.get('title'):
        plt.title(kwargs.get('title'))

    #set labels for axis x and axis y
    if x_label is not None:
        plt.xlabel(x_label, fontsize=kwargs.get('x_label_size') or (30 if len(x_label)<30 else 20))
    if y_label is not None:
        plt.ylabel(y_label, fontsize=kwargs.get('y_label_size') or (30 if len(y_label)<30 else 20))

    #set min and max values
    if x_lim is not None:
        plt.xlim(x_lim)
    if y_lim is not None:
        plt.ylim(y_lim)

    plt.tick_params(
        axis='both',
        labelsize=kwargs.get('tick_size') or 22,
    )

    lgd = None
    if show_legend:
        ncol = kwargs.get('legend_n_cols') or 2 # int(len(data.keys()) / 8.0 + 1)
        lgd = plt.legend(
            loc='upper center',
            bbox_to_anchor=kwargs.get('legend_bbox') or (0.5, -0.17),
            prop={
                'size':kwargs.get('legend_size') or (24 if max(map(len,data.keys())) < 50 else 16),
            },
            framealpha=0.5,
            scatterpoints=1,
            ncol=ncol,
        )

    #create output directory if it does not exists
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    extra_artists = []
    if lgd is not None:
        extra_artists.append(lgd)
    if annotate_text is not None:
        extra_artists.append(annotate_text)

    #save chart to a file
    output_path = output_dir + output_filename + '.' + format
    logger.debug("Saving scatter plot to %s" % output_path)
    plt.savefig(output_path, bbox_extra_artists=extra_artists, bbox_inches='tight', dpi=settings.CHART_DPI)

    #clear plot for the next chart
    plt.close()

    return output_path


def draw_changepoints(data, output_dir, output_filename, x_labels,
                    format = settings.CHART_EXT, **kwargs):
    y_labels = data.keys()
    y_min = 1
    y_max = len(data.keys())
    x_min = 0
    x_max = len(x_labels) - 1

    fig = plt.figure(figsize=(25, 2 if len(data.keys())==1 else 5))

    # remove axis
    ax = plt.subplot(111)
    ax.spines["top"].set_visible(False)
    ax.spines["bottom"].set_visible(False)
    ax.spines["right"].set_visible(False)
    ax.spines["left"].set_visible(False)
    ax.get_xaxis().tick_bottom()
    ax.get_yaxis().tick_left()

    if kwargs.get('annotate_lines'):
        for l in kwargs.get('annotate_lines'):
            # if x_tick not in x_labels:
            #     continue
            plt.axvline(
                x=l['x'],
                color="red",
                linestyle='-',
                linewidth=2,
            )
            ax.text(l['x'], y_max + 0.58, l.get('x_label', l['x']), color='red', size=20, ha='center', va='center')

    if kwargs.get('annotate_intervals'):
        for r in kwargs.get('annotate_intervals'):
            ax.add_patch(
                patches.Rectangle(
                    (r['x_start'], 0),  # (x,y)
                    r['x_end'] - r['x_start'],  # width
                    y_max + 0.5,  # height
                    #alpha=0.2,
                    facecolor="#EEEEEE",
                    edgecolor='none',
                    zorder=1,  # below scatter plot
                ),
            )
            ax.text(r['x_start'], y_max + 0.58, r.get('x_start_label', r['x_start']), color='gray', size=20, ha='center', va='center')
            ax.text(r['x_end'], y_max + 0.58, r.get('x_end_label', r['x_end']), color='gray', size=20, ha='center', va='center')

    # map snapshot index to x position
    x_coord = kwargs.get('x_coord', lambda x: x)

    max_n_changepoints = max(len(data[series]) for series in data)
    if max_n_changepoints > 3:
        color_range = (0, 0.8) # color intensity
    else:
        color_range = (0, 0.7)  # color intensity
    for i, series in enumerate(data.keys()):
        for j, changepoint in enumerate(data[series]):
            plt.scatter(
                [x_coord(changepoint)],
                [i + 1],
                label=series,
                s=kwargs.get('change_point_size', 100),
                edgecolors='none',
                color=str(color_range[0] + (color_range[1]-color_range[0])*(1.0*j) / (max_n_changepoints-1)),
                zorder=2, # on top intervals
            )

    plt.ylim(y_min - 0.5, y_max + 0.5)
    plt.xlim(x_min - 2 if kwargs.get('show_y_ticks', True) else x_min, x_max + 2)

    ax.yaxis.set_tick_params(pad=10)
    ax.xaxis.set_tick_params(pad=10)

    x_labels_to_show = kwargs.get('x_labels_to_show') or x_labels
    xticks_rotation = 0 if sum(map(len, x_labels_to_show)) < 20 else 45\
        if sum(map(len, x_labels_to_show)) < 40 else 90

    x_ticks = []
    x_ticks_labels = []
    for i in range(x_min, x_max + 1):
        if x_labels[i] in x_labels_to_show:
            x_ticks.append(i)
            x_ticks_labels.append(x_labels[i])


    if kwargs.get('show_y_ticks', True):
        plt.yticks(range(y_min, y_max + 1), y_labels, fontsize=24)
        #ax.set_yticklabels(y_labels, {'weight': 'bold'})
    else:
        plt.gca().yaxis.set_visible(False)
        #plt.gca().yaxis.set_major_formatter(plt.NullFormatter())
        plt.plot(
            [x_coord(x_ticks[0]), x_coord(x_ticks[-1])],
            [y_min,y_min],
            color="0.5",
            linestyle='-',
            linewidth=0.2,
        )

    plt.xticks(map(x_coord, x_ticks), x_ticks_labels, rotation=xticks_rotation, fontsize=24)
    for x_tick in x_ticks + [x_max]:
        plt.axvline(
            x=x_coord(x_tick),
            color="0.5",
            linestyle='--',
            linewidth=0.2,
        )

    for y in xrange(y_min, y_max + 1):
        plt.scatter(
            map(x_coord, range(x_max + 1)),
            [y] * (x_max + 1),
            s=kwargs.get('time_point_size', 2),
            color="black",
            edgecolors='none',
            zorder=3,  # on top
        )
    # Remove the tick marks; they are unnecessary with the tick lines we have just plotted.
    # plt.tick_params(axis="both", which="both", bottom="off", top="off",
    #                 labelbottom="on", left="off", right="off", labelleft="on")

    #create output directory if it does not exists
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    #save plot to a file
    output_path = output_dir + output_filename + '.' + format
    logger.debug("Saving change points plot to %s" % output_path)
    plt.savefig(output_path, bbox_inches='tight', format=format, dpi=settings.CHART_DPI)

    #clear plot for the next plot
    plt.close()

    return output_path


from matplotlib.colors import LinearSegmentedColormap
class nlcmap(LinearSegmentedColormap):
    """A nonlinear colormap"""

    name = 'nlcmap'

    def __init__(self, cmap, levels):
        self.cmap = cmap
        # @MRR: Need to add N for backend
        self.N = cmap.N
        self.monochrome = self.cmap.monochrome
        self.levels = np.asarray(levels, dtype='float64')
        self._x = self.levels / self.levels.max()
        self._y = np.linspace(0.0, 1.0, len(self.levels))

    #@MRR Need to add **kw for 'bytes'
    def __call__(self, xi, alpha=1.0, **kw):
        """docstring for fname"""
        # @MRR: Appears broken?
        # It appears something's wrong with the
        # dimensionality of a calculation intermediate
        #yi = stineman_interp(xi, self._x, self._y)
        yi = np.interp(xi, self._x, self._y)
        return self.cmap(yi, alpha)