import os
from collections import OrderedDict

import ranking_draw
from ..log import logger
from ..draw import draw_utils
from ..scoring_methods import scoring_methods_utils
from .. import settings
from ..draw import com_ev_draw
import numpy as np
from ..methods import methods_utils
from ..method import Method

def draw_list_timepoints_scores_timeseries(methods_list, dynamic_network, output_images_dir, **kwargs):
    """
    Draw timeseries of how changepoint-like timepoints are according to different methods
    :param methods_list:
    :param dynamic_network:
    :param output_images_dir:
    :return:
    """
    logger.debug("Comparing methods changepoint scores for timepoints")
    if os.path.exists(output_images_dir + 'timeseries-t-score.' + settings.CHART_EXT):
        logger.debug("Comparison of methods' changepoint scores for timepoints "
                     "has been already computed %s" % output_images_dir + 'timeseries-t-score.' + settings.CHART_EXT)
        if not kwargs.get('overwrite', False):
            logger.debug("No overwrite: aborting")
            return

    data = OrderedDict()
    for m in methods_list:
        if not hasattr(m,'get_timepoints_scores'):
            continue
        data[m.get_title()] = list(enumerate(m.get_timepoints_scores()))[1:]
    if len(data) == 0:
        logger.warning(
            "None of the methods %s supports 'get_timepoints_scores'" % (", ".join(m.get_id() for m in methods_list))
        )
        return
    vertical_lines = None
    dots = []
    # for each method, highlight changepoints from its actual solution
    for i_m, m in enumerate(methods_list):
        if not hasattr(m,'get_timepoints_scores'):
            continue
        dots += [
            {
                'points_list': [
                    (
                        t_c,
                        m.get_timepoints_scores()[t_c]
                    ) for t_c in m.get_community_evolution().get_changepoints()
                ],
                'color': settings.LINE_COLORS[i_m] if settings.LINE_COLORS else None,
                'marker': 'o',
                'markersize': 10,
                'alpha': 0.25,
            }
        ]
    if dynamic_network.get_segment_ground_truth() is not None:
        vertical_lines = [
             {
                'title': "ground truth",
                'x_list': dynamic_network.get_segment_ground_truth().get_changepoints(),
                'y_text_relative': 0.2,
                'fontsize': 10,
             }
        ]

    draw_utils.draw_timeseries(
        data=data,
        output_dir=output_images_dir,
        output_filename='timeseries-t-score',
        x_label="Timepoint",
        y_label="Score",
        y_lim=(0, 1),
        # output_data_dir=output_images_data_dir,
        vertical_lines=vertical_lines,
        dots=dots,
    )

def draw_segmentations_n_changepoints(method, output_images_dir=None, **kwargs):
    """
    Create a plot with method's segmentations for different fixed number of changepoints (from zero to number of snapshots)
    :param Method method:
    :param DynamicNetwork dynamic_network:
    :param output_images_dir:
    :return:
    """
    logger.debug("Comparing method's segmentations for fixed numbers of changepoints")
    if not hasattr(method,'get_solution_for_n_changepoints'):
        logger.warning(
            "Method %s does not support solutions for a fixed number of changepoints" % method.get_title()
        )
        return

    output_dir = output_images_dir or method.get_images_dir()
    if os.path.exists(output_dir + 'segmentation-n_changepoints.' + settings.CHART_EXT):
        logger.debug("Comparison of method's segmentations for fixed numbers of changepoints "
                     "has been already computed %s" % output_dir + 'segmentation-n_changepoints.' + settings.CHART_EXT)
        if not kwargs.get('overwrite', True):
            logger.debug("No overwrite: aborting")
            return

    dynamic_network = kwargs.get('dynamic_network') or method.get_dynamic_network()
    res = []
    n_timepoints = len(dynamic_network.get_timepoints())
    # to avoid too many series
    ks = np.arange(1, n_timepoints, max(1, n_timepoints / 10))

    for k in ks:
        res.append(method.get_solution_for_n_changepoints(k))
    com_ev_draw.draw_segmentations(
        c_evs_list=res,
        output_dir=output_dir,
        output_filename='segmentation-n_changepoints',
        show_ground_truth=True,
        c_evs_titles=["{k} changepoint{plural}".format(k=k, plural="s" if k!=1 else "") for k in ks],
        overwrite=kwargs.get('overwrite', True)
    )

def draw_list_solution_scores_n_changepoints_timeseries(methods_list, dynamic_network, output_images_dir, **kwargs):
    """
    Draw timeseries of methods' solution scores according to different scoring methods
    across different fixed numbers of changepoints (from zero to number of snapshots)
    I.e., for a fixed scoring method, compare how scores for each method vary as we change number of changepoints
    :param methods_list:
    :param dynamic_network:
    :param output_images_dir:
    :param extra_scoring_methods:
    :param use_methods_scoring_methods:
    :return:
    """
    logger.debug("Comparing methods' solutions for fixed numbers of changepoints")

    output_dir = output_images_dir
    if kwargs.get('add_subdirs'):
        output_dir += 'timeseries-n_changepoints-score/'

    if os.path.exists(output_dir):
        logger.debug("Comparison of methods' solutions for fixed numbers of changepoints "
                     "has been already computed %s" % output_dir)
        if not kwargs.get('overwrite', False):
            logger.debug("No overwrite: aborting")
            return

    x_axis_as_l = kwargs.get('x_axis_as_l', False) # show x axis as number of segments instead of number of change points

    method_title_func = kwargs.get('method_title_func', methods_utils.get_method_title)
    # changepoints_dot_count = kwargs.get('changepoints_dot_count', 15)
    changepoints_step = 1 # max(1, (len(dynamic_network.get_timepoints()) + 1) / changepoints_dot_count)

    # list of scoring methods for evaluating solutions
    scoring_methods = set(kwargs.get('extra_scoring_methods') or [])
    if kwargs.get('use_methods_scoring_methods', True):
        for m in methods_list:
            if not hasattr(m,'get_scoring_method'):
                continue
            scoring_methods.add(m.get_scoring_method())
        scoring_methods = sorted(list(scoring_methods), key=lambda x: x.get_id())
    if len(scoring_methods) == 0:
        logger.warning("No scoring methods available")
        logger.info("Using default scoring methods: %s" % settings.DEFAULT_SCORING_METHODS_IDS)
        scoring_methods = [
            scoring_methods_utils.create_scoring_method_by_id(x)
            for x in sorted(settings.DEFAULT_SCORING_METHODS_IDS)
        ]

    scd_methods_ids = kwargs.get('scd_methods_ids', []) # methods to be shown as SCD solutions - i.e., just as dots, instead of lines
    scd_methods = []
    colors = kwargs.get('line_colors', settings.LINE_COLORS)
    scd_colors = []

    # compare the best solutions for different numbers of changepoints
    c_evs = OrderedDict()
    for i,m in enumerate(methods_list):
        show_as_scd_method = m.get_id() in scd_methods_ids
        if not hasattr(m,'get_solution_for_n_changepoints'):
            logger.warning(
                "Method %s does not support solutions for different numbers of changepoints" % m.get_title()
            )
            show_as_scd_method = True
        res = []

        if show_as_scd_method:
            scd_methods.append(m)
            scd_colors.append(colors[i % len(colors)])
            res.append(m.get_community_evolution())
        else:
            for k in np.arange(0, len(dynamic_network.get_timepoints()), changepoints_step):
                res.append(m.get_solution_for_n_changepoints(k))
        c_evs[m] = res

    if len(c_evs) == 0:
        logger.warning("No suitable methods available")
        return

    vertical_lines = []
    if dynamic_network.get_segment_ground_truth() is not None:
        vertical_lines += [
             {
                'title': "ground truth",
                'x_list': [len(dynamic_network.get_segment_ground_truth().get_changepoints()) + (1 if x_axis_as_l else 0)],
                'y_text_relative': 0.2,
                'fontsize': 10,
            }
        ]

    if len(scd_methods) > 0:
        for i, scd_m in enumerate(scd_methods):
            vertical_lines += [
                {
                    # 'title': method_title_func(scd_m.get_id()),
                    'x_list': [len(scd_m.get_community_evolution().get_changepoints()) + (1 if x_axis_as_l else 0)],
                    'y_text_relative': 0.2,
                    'fontsize': 10,
                    'color': scd_colors[i],
                }
            ]


    for sm in scoring_methods:
        data = OrderedDict()
        for m, m_c_evs in c_evs.iteritems():
            res = []
            for i in xrange(len(m_c_evs)):
                res.append(
                    (
                        len(m_c_evs[i].get_changepoints()) + (1 if x_axis_as_l else 0),
                        sm.evaluate_community_evolution(m_c_evs[i])
                    )
                )
            data[method_title_func(m.get_id())] = res

        dots = []
        # for each method, highlight its number of changepoints in an actual solution
        if kwargs.get('highlight_sol_n_changepoints', False):
            for i_m, m in enumerate(c_evs.keys()):
                dots += [
                    {
                        'points_list': [
                            (
                                len(m.get_community_evolution().get_changepoints()) + (1 if x_axis_as_l else 0),
                                sm.evaluate_community_evolution(m.get_community_evolution())
                            )
                        ],
                        'color': colors[i_m],
                        'marker': 'o',
                        'markersize': 10,
                        'alpha': 0.25,
                    }
                ]

        if dynamic_network.get_segment_ground_truth() is not None:
            dots.append(
                {
                    'points_list': [
                        (
                            len(dynamic_network.get_segment_ground_truth().get_changepoints()) + (1 if x_axis_as_l else 0),
                            sm.evaluate_community_evolution(dynamic_network.get_segment_ground_truth())
                        )
                    ],
                    'color': 'limegreen',
                    'marker': '*',
                    'markersize': 10,
                    'alpha': 0.7,
                }
            )


        output_filename = sm.get_id()
        if kwargs.get('output_filename_prefix') is not None:
            output_filename = "%s_%s" % (kwargs.get('output_filename_prefix'), output_filename)

        draw_utils.draw_timeseries(
            data=data,
            output_dir=output_dir,
            output_filename=output_filename,
            x_label=("$l$" if x_axis_as_l else "Number of change points") if kwargs.get('show_x_label', True) else "",
            y_label=(scoring_methods_utils.get_scoring_method_title(sm.get_id())) if kwargs.get('show_y_label', True) else "",
            y_lim=sm.get_limits(),
            # output_data_dir=output_images_data_dir,
            vertical_lines=vertical_lines,
            dots=dots,
            **kwargs
        )

def draw_list_methods_run_time_bars(methods_list, output_dir, **kwargs):
    methods_dict = OrderedDict()
    for m in methods_list:
        methods_dict[m.get_title()] = [m]

    draw_list_methods_run_time_bars_avg(
        methods_dict=methods_dict,
        output_dir=output_dir,
        show_error_bars=False,
        **kwargs
    )

def draw_list_methods_run_time_bars_avg(methods_dict, output_dir, **kwargs):
    """
    Plot bars of methods running times with error bars
    :param output_dir:
    :param methods_dir: Dictionary method_id => list of methods
    :param overwrite:
    :param kwargs:
    :return:
    """
    logger.debug("Evaluating methods' running times via bar")

    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    data = OrderedDict()
    error_bars = OrderedDict()
    for m_title, ms_list in methods_dict.iteritems():
        run_times = [m.get_run_time() for m in ms_list]
        data[m_title] = {'': np.average(run_times) }
        error_bars[m_title] = {'': np.std(run_times) }

    if os.path.exists(output_dir + 'bar-run_time.' + settings.CHART_EXT):
        logger.debug("Comparison of methods' running times "
                     "has been already computed %s" % output_dir + 'bar-run_time.' + settings.CHART_EXT)
        if not kwargs.get('overwrite', False):
            logger.debug("No overwrite: aborting")
            return
    draw_utils.draw_bar(
        data=data,
        #y_err=error_bars,
        output_dir=output_dir,
        output_filename='bar-run_time',
        x_label="Networks",
        y_label="Running time, s",
        y_lim=(0, None),
        log_scale=True,
        **kwargs
    )

def draw_list_methods_networks_run_time_bars_avg(networks_methods_dict, output_dir, **kwargs):
    """
    Plot bars of methods running times with error bars
    (for plotting running times of multiple methods for multiple networks (each with possible many instances)
    :param output_dir:
    :param networks_methods_dict: Dictionary network_key =>  {method_id => list of methods}
    :param overwrite:
    :param kwargs:
    :return:
    """
    logger.debug("Evaluating methods' running times via bar")

    if os.path.exists(output_dir + 'bar-run_time.' + settings.CHART_EXT):
        logger.debug("Comparison of methods' running times "
                     "has been already computed %s" % output_dir + 'bar-run_time.' + settings.CHART_EXT)
        if not kwargs.get('overwrite', False):
            logger.debug("No overwrite: aborting")
            return

    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    data = OrderedDict()
    error_bars = OrderedDict()
    for n_key, methods_dict in networks_methods_dict.iteritems():
        for m_title, ms_list in methods_dict.iteritems():
            run_times = [m.get_run_time() for m in ms_list]
            if m_title not in data:
                data[m_title] = OrderedDict()
                error_bars[m_title] = OrderedDict()
            data[m_title][n_key] = np.average(run_times)
            error_bars[m_title][n_key] = np.std(run_times)

    draw_utils.draw_bar(
        data=data,
        y_err=error_bars,
        output_dir=output_dir,
        output_filename='bar-run_time',
        x_label=kwargs.get('x_label', "Networks"),
        y_label="Running time, s",
        y_lim=kwargs.get('y_lim', (0, None)),
        log_scale=True,
        **kwargs
    )

    # draw ranking
    if kwargs.get('draw_ranking_slopegraph', False) and len(data.keys()) > 1 and len(data.values()[0])>1:
        ranking_draw.draw_slopegraph(
            data=data,
            output_dir=output_dir + 'slopegraph-ranking/',
            output_filename='run_time',
            as_ranking=True,
        )