import os

import matplotlib
matplotlib.use('Agg') # Must be before importing matplotlib.pyplot or pylab!
import matplotlib.pyplot as plt

import networkx as nx
from .. import settings

def draw_network(network, output_dir, node_positions = None,
                 nodes_colors_dict=None, node_attr_value_color_dict = None,
                 edge_colors = 'gray', node_labels = None,
                 output_filename = None, format = settings.CHART_EXT):
    """
    Draw network
    :param network:
    :param output_dir:
    :param node_size:
    :param font_size:
    :param nodes_colors_dict: Dictionary node => cluster
    :param node_attr_value_color_dict:
    :param edge_colors:
    :param node_labels:
    :param output_filename:
    :return: Output file location
    """

    if node_positions is None:
        node_positions = nx.graphviz_layout(network.get_nx_graph(), prog='fdp', args='-Goverlap=false')

    #Draw parameters
    node_size=300
    font_size=10
    edge_width=1
    with_labels=True

    #If network is too large, change some parameters
    if network.get_nodes_count()>50:
        node_size=100
        font_size=7
        edge_width=0.5

    #node colors
    node_colors = 'gray'
    if nodes_colors_dict is not None:
        if node_attr_value_color_dict is None:  # transform attribute values to the values in the range [0,1]
            #find all unique attr values
            node_attr_value_color_dict = dict()
            for node_name in network.get_nx_graph().nodes():
                node_attr_value = nodes_colors_dict[node_name]
                if node_attr_value is not None:
                    node_attr_value_color_dict[node_attr_value]=0

            #assign a value between 0 and 1 for each unique attr value
            #(evenly distribute values in the range [0,1])
            i=0
            incr = 1.0/max(1,len(node_attr_value_color_dict.keys()))
            for node_attr_value in node_attr_value_color_dict:
                i += incr
                node_attr_value_color_dict[node_attr_value] = i

        #assign colors (values in [0,1]) to nodes according to their attribute values
        node_colors = []
        for node_name in network.get_nx_graph().nodes():
            node_attr_value = nodes_colors_dict[node_name]
            node_colors.append(node_attr_value_color_dict.get(node_attr_value,0))

    #draw a network
    nx.draw(
        network.get_nx_graph(),
        node_positions,
        node_color=node_colors,
        edge_color=edge_colors,
        width=edge_width,
        with_labels = with_labels and node_labels is None,
        node_size=node_size,
    )

    if with_labels and node_labels is not None :
        nx.draw_networkx_labels(
            network.get_nx_graph(),
            node_positions,
            node_labels,
            font_size=font_size
        )

    #if there is no specified output filename, use network name instead
    if output_filename is None:
        output_filename = network.get_name()

    #if output directory does not exists, create it
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    #save figure to the file
    image_file = output_dir + output_filename + '.' + format
    plt.savefig(image_file, format=format, dpi = settings.CHART_DPI)

    #clear canvas
    plt.clf()

    return node_positions,image_file
