from .. import settings
from ..log import logger
import os
import shutil
from ..draw import adhoc_draw

def convert_figure_name(name):
    new_name = name.replace('_(', '_').replace('(', '_').replace(')', '_').replace('_.', '.')
    if new_name.startswith('_'):
        new_name = new_name[1:]
    if new_name.endswith('_'):
        new_name = new_name[:-1]
    return new_name

def draw_methods_parameters_figures():
    """
    Figures for Section 3.1
    :return:
    """

    default_params = {
        'networks_dict_key': 'rnd16n50',
        'draw_com_evs_scores': False,
        'draw_com_evs_sim': True,
        'draw_pr_roc': False,
        'x_label': "Configuration",
        'format': 'eps',
        'scoring_methods_ids': ['(sgmq)(num)'],
        'log_scale': False,
        'sc_limits': {'(sgmq)(num)': (0, 17)},
    }
    configs = {
        'multistep': {
            'methods_ids_key': 'multistep',
            'legend_n_cols': 3,
            'draw_com_evs_scores': True,
        },
        'ghrg': {
            'methods_ids_key': 'ghrg',
            'legend_n_cols': 3,
        },
        'scout_ic': {
            'methods_ids_key': 'naive-ic',
            'draw_com_evs_scores': True,
        },
        'scout_obj': {
            'methods_ids_key': 'naive-obj',
            'legend_n_cols': 1,
        },
        'scout_cons_sum': {
            'methods_ids_key': 'naive-cons-sum',
            'draw_run_time': True,
            'legend_n_cols': 1,
        },
        'scout_cons': {
            'methods_ids_key': 'naive-cons',
            'draw_run_time': True,
            'legend_n_cols': 1,
        },
        'scout_search': {
            'methods_ids_key': 'naive-search',
            'draw_run_time': True,
            'run_time_log_scale': True,
            'run_time_y_lim': (0, 10**3),
            'legend_n_cols': 1,
        },
    }

    figures_dir = settings.EVAL_DIR + '_paper_figures/supplement/'
    if not os.path.exists(figures_dir):
        os.makedirs(figures_dir)

    for c_name,c_params in configs.iteritems():
        logger.info("Creating figures for %s" % c_name)
        params = default_params.copy()
        params.update(c_params)

        input_dir = adhoc_draw.draw_plots_per_measure(**params)

        if params.get('draw_run_time'):
            params['y_lim'] = params.get('run_time_y_lim', (0, 400))
            params['log_scale'] = params.get('run_time_log_scale', params.get('log_scale'))
            adhoc_draw.draw_all_running_times(**params)

        output_dir = figures_dir + c_name + '/'
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)

        logger.debug("Copying figures")
        if params.get('draw_com_evs_scores'):
            for sc in params.get('scoring_methods_ids'):
                shutil.copyfile(input_dir + sc + '.eps', output_dir  + convert_figure_name(sc) + '.eps')
        if params.get('draw_com_evs_sim'):
            shutil.copyfile(input_dir + '(gl_p)(nmi).eps', output_dir + convert_figure_name('(gl_p)(nmi).eps'))
        if params.get('draw_pr_roc'):
            shutil.copyfile(input_dir + 'aupr.eps', output_dir + convert_figure_name('aupr.eps'))
        if params.get('draw_run_time'):
            shutil.copyfile(input_dir + 'run_time.eps', output_dir + convert_figure_name('run_time.eps'))


def draw_synthetic_scatter_plots():
    input_dir = adhoc_draw.draw_plots_per_network_key(
        networks_dict_key='rnd16n100',
        methods_ids_key='best',
        format='eps',
        include_network_keys = ['4 segments'],
        legend_n_cols=2,
        legend_bbox=(0.5, -0.39), #(1.3, 1),
        # x_lim=(0.5, 1),
        # y_lim=(0.5, 1),
    )
    output_dir = settings.EVAL_DIR + '_paper_figures/supplement/synthetic_100/'
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)
    shutil.copyfile(
        input_dir + 'scatter-com_evs_similarity/4_segments_(avg_seg_t)(nmi)-vs-(avg_sn_p)(nmi).eps',
        output_dir + convert_figure_name('4_segments_(avg_seg_t)(nmi)-vs-(avg_sn_p)(nmi).eps')
    )

    for n_size in [50, 100, 500, 1000]:
        networks_dict_key = 'rnd16n%s' % n_size
        input_dir = adhoc_draw.draw_plots_per_network_key(
            networks_dict_key=networks_dict_key,
            methods_ids_key='best',
            format='eps',
            show_x_label=n_size in [1000],
            show_y_label_n_keys=['1 segment'],
            show_legend=n_size in [1000],
            legend_size=44,
            legend_n_cols=1,
            legend_bbox=(0.5, -0.27),
            x_label_size=48,
            y_label_size=48,
            tick_size=34,
            dot_size=150,
            annotate={
                'x': -0.4,
                'y': 0.5,
                'text': n_size, # "$%s$-node snapshots" % n_size,
                'size': 52,
                'weight': 'bold',
            }
        )

        logger.debug("Copying figures")

        output_dir = settings.EVAL_DIR + '_paper_figures/supplement/synthetic_%s/' % n_size
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)
        for i in [1, 4, 2, 8, 16]:
            shutil.copyfile(
                input_dir + 'scatter-com_evs_similarity/%s_segment%s_(avg_seg_t)(nmi)-vs-(avg_sn_p)(nmi).eps' % (i, "s" if i!=1 else ""),
                output_dir  + convert_figure_name('%s_segment%s_(avg_seg_t)(nmi)-vs-(avg_sn_p)(nmi).eps' % (i, "s" if i!=1 else ""))
            )

def draw_synthetic_bar_plots():
    input_dir = adhoc_draw.draw_all_running_times(
        networks_dict_key='all-synthetic',
        methods_ids_key='best-runtime',
        x_label="# of nodes per snapshot",
        y_lim=(0, 10**5),
        log_scale=True,
        format='eps',
        legend_n_cols=1,
        legend_bbox=(1.17, 1), # or (0.5, -0.16)
        figsize=(15, 5),
    )
    logger.debug("Copying figures")
    output_dir = settings.EVAL_DIR + '_paper_figures/paper/synthetic/'
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)
    shutil.copyfile(input_dir + 'run_time.eps', output_dir + 'run_time.eps')

    for n_size in [50, 100, 500, 1000]:
        networks_dict_key = 'rnd16n%s' % n_size

        adhoc_draw.draw_plots_per_measure(
            networks_dict_key=networks_dict_key,
            methods_ids_key='best-m',
            draw_com_evs_scores=False,
            draw_com_evs_sim=False,
            draw_pr_roc=True,
            x_label="Configuration" if n_size in [100, 1000] else "",
            format='eps',
            show_x_ticks=True, #n_size in [100, 1000],
            show_legend=n_size in [100, 1000],
            annotate={
                'x': -1.7,
                'y': 0.5,
                'text': n_size, # "$%s$-node snapshots" % n_size,
                'size': 40,
                'weight': 'bold',
            } if n_size != 100 else None,
        )

        input_dir = adhoc_draw.draw_plots_per_measure(
            networks_dict_key=networks_dict_key,
            methods_ids_key='best',
            draw_com_evs_scores=True,
            scoring_methods_ids=['(avg_sn)(mod)'],
            draw_com_evs_sim=True,
            draw_pr_roc=False,
            x_label="Configuration" if n_size in [100, 1000] else "",
            format='eps',
            show_x_ticks=True, #n_size in [100, 1000],
            show_legend=n_size in [100, 1000],
        )

        input_dir_m = adhoc_draw.draw_plots_per_measure(
            networks_dict_key=networks_dict_key,
            methods_ids_key='best-m',
            draw_com_evs_scores=True,
            scoring_methods_ids=['(sgmq)(num)'],
            sc_limits={'(sgmq)(num)': (0, 17)},
            draw_com_evs_sim=False,
            draw_pr_roc=False,
            x_label="Configuration" if n_size in [500, 1000] else "",
            show_y_label=n_size in [50, 500],
            format='eps',
            show_x_ticks=True, #n_size in [500, 1000],
            show_legend=n_size in [500, 1000]
        )

        # adhoc_draw.draw_all_running_times(
        #     networks_dict_key=networks_dict_key,
        #     methods_ids_key='best-runtime',
        #     x_label="Configuration",
        #     y_lim=(0, 10**5),
        #     log_scale=True,
        #     format='eps',
        # )


        logger.debug("Copying figures")
        in_main_paper = n_size in [100]
        output_dir = settings.EVAL_DIR + '_paper_figures/%s/synthetic_%s/' % ("paper" if in_main_paper else "supplement", n_size)
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)
        shutil.copyfile(input_dir_m + 'aupr.eps', output_dir + convert_figure_name('aupr.eps'))
        shutil.copyfile(input_dir + '(avg_sn)(mod).eps', output_dir  + convert_figure_name('(avg_sn)(mod).eps'))
        shutil.copyfile(input_dir + '(gl_p)(nmi).eps', output_dir  + convert_figure_name('(gl_p)(nmi).eps'))

        output_dir = settings.EVAL_DIR + '_paper_figures/supplement/synthetic_%s/' % (n_size)
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)
        shutil.copyfile(input_dir_m + '(sgmq)(num).eps', output_dir  + convert_figure_name('(sgmq)(num).eps'))

def draw_rankings_figures():

    configs = [
        {
            'scoring_methods_ids': [
                '(com_evs_sim)((gl_p)(nmi))(segment_ground_truth)'
            ],
            'show_y_label': False,
            'show_legend': True,
            'show_legend_large': True,
            'show_legend_small': False,
            'show_series_labels': True,
            'is_in_paper': True,
            'xticks_rotation': 15,
            'figsize': (10, 5),
        },
        {
            'scoring_methods_ids': [
                '(avg_sn)(mod)',
            ],
            'show_y_label': True,
            'show_legend': False,
        },
        {
            'scoring_methods_ids': [
                '(com_evs_sim)((avg_sn_p)(nmi))(segment_ground_truth)'
            ],
            'show_y_label': False,
            'show_legend': True,
        },
        {
            'scoring_methods_ids': [
                '(com_evs_sim)((avg_seg_t)(nmi))(segment_ground_truth)'
            ],
            'show_y_label': True,
            'show_legend': False,
        },
        {
            'exclude_extreme_configs': True,
            'scoring_methods_ids': [
                '(cl_res)(aupr)(segment_ground_truth)'
            ],
            'show_y_label': False,
            'show_legend': True,
        },
    ]

    m_comparison_key = 'ghrg-vs-graphscope-vs-multistep-vs-naive'

    output_dir_paper = settings.EVAL_DIR + '_paper_figures/paper/'
    output_dir_supplement = settings.EVAL_DIR + '_paper_figures/supplement/'

    for c in configs:
        for n_comparison_key in ['small-synthetic-configs', 'large-synthetic-configs']:
            if c.get('exclude_extreme_configs'):
                n_comparison_key += '-nonextreme'
            are_large_networks = n_comparison_key.startswith('large-')
            y_label = ("$500$- and $1000$-node \nsnapshots" if are_large_networks else "$50$- and $100$-node \nsnapshots")

            input_dir = adhoc_draw.draw_scoring_methods_ranking_bars(
                n_comparison_key=n_comparison_key,
                m_comparison_key=m_comparison_key,
                scoring_methods_ids=c['scoring_methods_ids'],
                input_dir=settings.EVAL_DIR + '_rankings_config_best_cpp/',
                precision_levels=[0],
                format='eps',
                show_legend=c.get('show_legend') and
                            ((c.get('show_legend_large', False) and are_large_networks)
                            or (c.get('show_legend_small', True) and not are_large_networks)),
                y_label=y_label if c['show_y_label'] else '',
                show_series_labels=c.get('show_series_labels', are_large_networks),
                y_ticks_raw_counts=False,
                xticks_rotation=c.get('xticks_rotation'),
                figsize=c.get('figsize'),
            )

            output_subdir = 'rankings_500_1000/' if are_large_networks else 'rankings_50_100/'
            output_dir = (output_dir_paper if c.get('is_in_paper', False) else output_dir_supplement) + output_subdir
            if not os.path.exists(output_dir):
               os.makedirs(output_dir)
            for sc in c['scoring_methods_ids']:
                shutil.copyfile(input_dir + sc + '_0.eps', output_dir + convert_figure_name(sc + '_0.eps'))


def draw_real_world_figures():
    networks_dict_key = 'all-realworld'
    methods_ids_key = 'best-m'
    legend_bbox = (0.5, -0.51)
    xticks_rotation = 50

    adhoc_draw.draw_plots_per_measure(
        networks_dict_key=networks_dict_key,
        methods_ids_key=methods_ids_key,
        draw_com_evs_scores=False,
        draw_com_evs_sim=False,
        draw_pr_roc=True,
        other_method_id="gt_events",
        x_label="Network",
        format='eps',
        legend_bbox=legend_bbox,
        xticks_rotation=xticks_rotation,
    )

    adhoc_draw.draw_plots_per_measure(
        networks_dict_key=networks_dict_key,
        methods_ids_key=methods_ids_key,
        draw_com_evs_scores=True,
        scoring_methods_ids=['(avg_sn)(mod)',],# '(sgmq)(num)'],
        sc_limits = {'(sgmq)(num)': (0, 50)},
        draw_com_evs_sim=False,
        draw_pr_roc=False,
        other_method_id="gt_events",
        x_label="Network",
        format='eps',
        legend_bbox=legend_bbox,
        xticks_rotation=xticks_rotation,
    )

    input_dir = adhoc_draw.draw_all_running_times(
        networks_dict_key=networks_dict_key,
        methods_ids_key=methods_ids_key,
        x_label="Network",
        y_lim=(0.2, 10**5), # start from 0.2 so that all values are visible
        legend_bbox=legend_bbox,
        xticks_rotation=xticks_rotation,
        log_scale=True,
        format='eps',
    )

    logger.debug("Copying figures")

    output_dir = settings.EVAL_DIR + '_paper_figures/paper/real_world/'
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)
    shutil.copyfile(input_dir + 'aupr.eps', output_dir  + convert_figure_name('aupr.eps'))
    shutil.copyfile(input_dir + '(avg_sn)(mod).eps', output_dir  + convert_figure_name('(avg_sn)(mod).eps'))


    output_dir = settings.EVAL_DIR + '_paper_figures/supplement/real_world/'
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)
    shutil.copyfile(input_dir + '(sgmq)(num).eps', output_dir + convert_figure_name('(sgmq)(num).eps'))
    shutil.copyfile(input_dir + 'run_time.eps', output_dir + convert_figure_name('run_time.eps'))

def draw_real_world_cscd_figures():
    input_dir = adhoc_draw.draw_methods_score_n_changepoints(
        networks_dict_key='all-realworld',
        methods_ids_key='best-n-changepoints',
        extra_scoring_methods=['(avg_sn)(mod)'],
        scd_methods_ids=['multistep-0.0'], # GraphScope and GHRG are added as SCD-only methods automatically, since they do not support CSCD
        x_axis_as_l=True,
        n_keys_show_legend=['Enron', 'Senate'],
        show_x_label_n_keys=['High School', 'Senate', 'Reality Mining', 'Enron'],
        show_y_label_n_keys=['High School', 'Senate', 'Hypertext', 'Enron'],
        legend_bbox_n_keys={'Senate': (0.5, -0.51)},
    )

    logger.debug("Copying figures")
    output_dir = settings.EVAL_DIR + '_paper_figures/paper/real_world/'
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)
    main_n_keys = ['senate']
    for n_key in main_n_keys:
        shutil.copyfile(input_dir + '%s_(avg_sn)(mod).eps' % n_key, output_dir + convert_figure_name('%s_(avg_sn)(mod).eps' % n_key))

    output_dir = settings.EVAL_DIR + '_paper_figures/supplement/real_world/'
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)
    other_n_keys = ['hypertext', 'amd_hope', 'high_school', 'reality_mining', 'enron']
    for n_key in other_n_keys:
        shutil.copyfile(input_dir + '%s_(avg_sn)(mod).eps' % n_key, output_dir + convert_figure_name('%s_(avg_sn)(mod).eps' % n_key))

def draw_real_world_changepoints_figures():
    output_dir = settings.EVAL_DIR + '_paper_figures/supplement/real_world/'
    adhoc_draw.draw_changepoints(
        network_id='voteview/1year_0.4',
        methods_ids_key='best-m',
        output_dir=output_dir,
        output_filename=convert_figure_name('senate_changepoints'),
    )

    output_dir = settings.EVAL_DIR + '_paper_figures/paper/real_world/'
    for n_id in ["aging/orig_80"]:  # , "aging/orig"]:
        adhoc_draw.draw_changepoints(
            network_id=n_id,
            methods_ids_key='best-m',
            # format='pdf',
            max_n_changepoints=3 if "_8" in n_id else 7,
            time_point_size=10,
            change_point_size=250,
            output_dir=output_dir,
            output_filename=convert_figure_name('aging_changepoints'),
        )

