import os
from ..log import logger
import draw_utils
import numpy as np
from ..partition import Partition
from ..static_clustering import compute_partition_quality
import networkx as nx
import network_draw
from .. import settings

def draw_partition(partition, output_dir, output_filename = None, ground_truth_clustering = None, object_names = None,
                        ground_truth_clustering_colors = None):
        """
        Plot clustering as a graph where clusters are cliques. Use ground truth for node coloring.
        """
        g = nx.Graph()

        #Cannot use object_names as node names - since in two captions are the same, different nodes will be treated as one
        for c, objects in partition.get_clusters_objects().iteritems():
            nodes = None
            object_captions = objects
            if ground_truth_clustering is not None:
                    nodes = [(x,{'ground_truth_cluster': ground_truth_clustering[x],'description': object_names[x]}) for x in objects]
            else:
                    nodes = objects

            g.add_nodes_from(nodes)
            g.add_edges_from([(x,y) for x in object_captions for y in object_captions])


        #save for Cytoscape
        # nx.write_edgelist(g,output_dir + output_filename + '.txt')
        # output_csv = csv.writer(open(output_dir + output_filename + '.csv', "wb"))
        # output_csv.writerow(['#','Description', 'Ground Truth'])
        # for node in g.nodes(data=True):
        #     output_csv.writerow([node[0],node[1]['description'], node[1]['ground_truth_cluster']])

        network_draw.draw_network(g, output_dir,
            #node_positions = nx.spring_layout(g),
            node_attr_value_color_dict = 'ground_truth_cluster',
            output_filename= 'clustering' if output_filename is None else output_filename,
            #node_attr_value_color_dict= ground_truth_clustering_colors,
            edge_colors= 'gray',
            #node_labels= dict(zip(range(0,g.order()), object_names))#dict node name => node label
        )

def draw_partitions_comparison_heatmaps(partitions_1, partitions_2,
                                   output_dir, **kwargs):
    logger.debug("Drawing heatmaps for partitions comparison ({count_1} partitions vs {count_2} partitions)".format(
        count_1 = len(partitions_1),
        count_2 = len(partitions_2)
    ))

    output_filename_prefix = kwargs.get('output_filename_prefix', '')

    for metric in settings.DEFAULT_PARTITION_SIMILARITY_METRICS:
        metric_title = Partition.SIMILARITY_METRICS[metric]
        output_filename = output_filename_prefix + ('-' if output_filename_prefix != '' else '') + metric
        if os.path.exists(output_dir + output_filename + '.' + settings.CHART_EXT):
            logger.debug('Partitions comparison heatmap has been already drawn: %s'
                         % output_dir + output_filename + '.' + settings.CHART_EXT)
            if not kwargs.get('overwrite', False):
                logger.debug("No overwrite: aborting")
                continue
        data = np.zeros((len(partitions_1), len(partitions_2)))
        for i, p_1 in enumerate(partitions_1):
            for j, p_2 in enumerate(partitions_2):
                data[i, j] = Partition.compute_partitions_similarity(
                    partition_true=p_2,
                    partition_predicted=p_1,
                    metric=metric
                )

        draw_utils.draw_heatmap(
            data=data,
            output_dir=output_dir,
            output_filename=output_filename,
            x_labels=kwargs.get('x_labels') or range(len(partitions_1)),
            y_labels=kwargs.get('y_labels') or range(len(partitions_2)),
            x_label=kwargs.get('x_label'),
            y_label=kwargs.get('y_label'),
            vmin=0,
            vmax=1,
            color_bar_label=metric_title,
            vertical_lines=kwargs.get('vertical_lines'),
            horizontal_lines=kwargs.get('horizontal_lines'),
        )


def draw_partitions_qualities_heatmap(networks, partitions,
                                   output_dir, **kwargs):
    """
    Heatmap of network partition qualities (i.e., each network vs each partition)
    :param networks:
    :param partitions:
    :param output_dir:
    :param output_filename_prefix:
    :param kwargs:
    :return:
    """
    logger.debug("Drawing heatmap for partitions quality ({count_1} networks vs {count_2} partitions)".format(
        count_1 = len(networks),
        count_2 = len(partitions)
    ))
    output_filename_prefix = kwargs.get('output_filename_prefix','')

    for metric, metric_title in settings.DEFAULT_PARTITION_QUALITY_METRICS:
        data = np.zeros((len(networks), len(partitions)))

        for i, network in enumerate(networks):
            for j, partition in enumerate(partitions):
                data[i, j] = compute_partition_quality(network, partition, metric)

        draw_utils.draw_heatmap(
            data=data,
            output_dir=output_dir,
            output_filename=output_filename_prefix + ('-' if output_filename_prefix != '' else '') + metric,
            x_labels=kwargs.get('x_labels') or map(str, range(len(networks))),
            y_labels=kwargs.get('y_labels') or map(str, range(len(networks))),
            x_label=kwargs.get('x_label', "Snapshots"),
            y_label=kwargs.get('y_label', "Partitions"),
            vmin=0 if metric != 'log_l' else None,
            vmax=1 if metric != 'log_l' else 0,
            color_bar_label=metric_title,
            vertical_lines=kwargs.get('vertical_lines'),
            horizontal_lines=kwargs.get('horizontal_lines'),
        )