import os
from collections import defaultdict
from matplotlib import pyplot as plt
from .. import settings
from ..log import logger
from operator import add
import numpy as np
from matplotlib.ticker import MaxNLocator, MultipleLocator

def draw_ranking_bar(data, output_dir, output_filename, y_err=None,
                    x_label="", y_label="", y_lim=None,
                    show_legend=True, output_data_dir=None, format = settings.CHART_EXT, **kwargs):

    bar_width = kwargs.get('bar_width', 1)
    #needed to set ticks for axis x
    data_x,x = [],[]
    groups = None
    series_names = kwargs.get('series_names_sorted', data.keys())
    series_labels_dict = kwargs.get('series_labels_dict', None)
    series_labels = series_names[:]
    if series_labels_dict is not None:
        for i,series in enumerate(series_names):
            series_labels[i] = series_labels_dict.get(series, series)

    if kwargs.get('figsize'):
        plt.figure(figsize=kwargs.get('figsize'))

    n_series = len(series_names)
    # series_names = kwargs.get('series_names')
    values_types = kwargs.get('values_types')
    bar_colors = kwargs.get('bar_colors')
    has_na = False
    if values_types is None:
        values_types = range(1, len(data.keys())+1) # for k methods => 1,2,...,k

        for series in series_names:
            if 'n/a' in data[series].keys():
                has_na = True
                break
        if kwargs.get('always_include_na') or has_na:
            values_types += ['n/a']
        values_types_labels = {}
        for v in values_types:
            l = ''
            if v == 'n/a':
                l = 'N/A'
            else:
                suf = 'th'
                if v == 1:
                    suf = 'st'
                elif v == 2:
                    suf = 'nd'
                elif v == 3:
                    suf ='rd'
                l = "$%s^{%s}$" % (v, suf)
            values_types_labels[v] = l
    else:
        values_types_labels = kwargs.get('values_types_labels', dict(zip(values_types, values_types)))
    if bar_colors is None:
        gray_colors = map(str, np.linspace(0.0, 1.0, len(values_types) +
               (0 if (kwargs.get('always_include_na') or has_na) else 1))) # so that white corresponds to n/a (and not to the last one)
        bar_colors = kwargs.get('bar_colors', gray_colors)
    n_values_types = len(values_types)


    # transform data (i.e., value types now become series)
    total_sum = sum(data[series_names[0]].values())
    for i in xrange(1,n_series):
        total_sum_test = sum(data[series_names[i]].values())
        if total_sum != total_sum_test:
            raise Exception('Number of test cases for each key must match!')
    t_data = {}
    for v_t in values_types:
        t_data[v_t] = {}
        for series in series_names:
            t_data[v_t][series] = data[series].get(v_t, 0) / (1.0 * total_sum)


    prev_series_values=[0] * n_series
    for i_value_type, value_type in enumerate(values_types):
        #limit on max number of series in one plot
        if i_value_type >= 15:
            raise Exception("Too many values types: %s" % len(data.keys()))

        curr_series_values = [t_data[value_type][series] for series in series_names]
        plt.bar(
            [i * bar_width for i in xrange(n_series)],
            curr_series_values,
            bottom=prev_series_values,
            width=bar_width,
            label=values_types_labels[value_type],
            color=bar_colors[i_value_type],
        )
        prev_series_values = map(add, prev_series_values, curr_series_values)

    xticks_rotation = kwargs.get('xticks_rotation') or 90 # 0 if sum(map(len, series_labels))<50 else 20
    plt.xticks(
        [i * bar_width + bar_width / 2.0 for i in xrange(n_series)],
        series_labels,
        rotation=xticks_rotation,
    )
    if kwargs.get('y_ticks_raw_counts'):
        plt.yticks(
            np.linspace(0, 1, total_sum + 1),
            [i for i in xrange(total_sum + 1)],
        )
    else:
        plt.yticks(
            [i*0.2 for i in xrange(6)],
            ['%s%%' % (i*20) for i in xrange(6)],
        )
        plt.gca().yaxis.set_minor_locator(
            MultipleLocator(base=0.1)
        )

    plt.tick_params(
        axis='y',          # changes apply to the x-axis
        which='major',      # both major and minor ticks are affected
        labelsize=24,
        pad=10,
    )
    plt.tick_params(
        axis='x',          # changes apply to the x-axis
        which='both',      # both major and minor ticks are affected
        bottom='off',      # ticks along the bottom edge are off
        top='off',         # ticks along the top edge are off
        labelbottom='on', # labels along the bottom edge are on
        labelsize=28 if sum(map(len, series_names))<30 else 14,
        pad=10,
    )
    if not kwargs.get('show_series_labels', True):
        plt.setp(plt.axes().get_xticklabels(), visible=False)

    #set min and max values
    if y_lim is not None:
        plt.ylim(y_lim)
    else:
        plt.ylim(0, 1)

    #set labels for axis x and axis y
    plt.xlabel(x_label, fontsize=40 if len(x_label)<30 else 32)
    plt.ylabel(y_label, fontsize=40 if len(y_label)<30 else 32)

    lgd = None
    bbox_extra_artists = None
    if show_legend:
        # ncol = n_values_types # int(n_values_types / 8.0 + 1)
        lgd = plt.legend(
            loc=2, # right
            bbox_to_anchor=(1.01, 1.0),
            #bbox_to_anchor=(0.5, -0.1 if xticks_rotation == 0 else -0.12),
            prop={
                'size':28,
            },
            framealpha=0.5,
            ncol=1,
        )
        bbox_extra_artists = (lgd,)

    #create output directory if it does not exists
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    #save chart to a file
    output_path = output_dir + output_filename + '.' + format
    logger.debug("Saving ranking bar plot to %s" % output_path)
    plt.savefig(output_path, bbox_extra_artists=bbox_extra_artists, bbox_inches='tight', dpi=settings.CHART_DPI)

    #clear plot for the next chart
    plt.close()

    return output_path


# Slopegraph - to visualize ranking over time
# see http://charliepark.org/slopegraphs/
def draw_slopegraph(data, output_dir, output_filename, as_ranking=True,
                    format = settings.CHART_EXT, **kwargs):
    line_colors = kwargs.get('line_colors', settings.LINE_COLORS)
    x_labels = []
    color_dict = {}
    for i,series in enumerate(data.keys()):
        color_dict[series] = line_colors[i % len(line_colors)]
        for x_tick in data[series].keys():
            if x_tick not in x_labels:
                x_labels.append(x_tick)

    data_values = {}
    y_min = y_max = None
    for i,series in enumerate(data.keys()):
        data_values[series] = [0] * len(x_labels)
        for i_x,x_tick in enumerate(x_labels):
            data_values[series][i_x] = data[series][x_tick]
        if not as_ranking:
            if y_min is not None:
                y_min = min(y_min, *data_values[series])
            else:
                y_min = min(data_values[series])

            if y_max is not None:
                y_max = max(y_max, *data_values[series])
            else:
                y_max = max(data_values[series])

    if as_ranking:
        for i in xrange(len(x_labels)):
            v = [-data_values[series][i] for series in data.keys()] # use negative values so that higher values are ranked higher
            # note:can use scipy.stats.rankdata with method 'min', but server uses an older version of scipy without method parameter
            v_ranks = [1 + sum(1 for y in v if y < x) for x in v]

            for i_series, series in enumerate(data.keys()):
                data_values[series][i] = v_ranks[i_series]
        y_min = 1
        y_max = len(data.keys())
    x_min = 0
    x_max = len(x_labels) - 1


    # remove axis
    ax = plt.subplot(111)
    ax.spines["top"].set_visible(False)
    ax.spines["bottom"].set_visible(False)
    ax.spines["right"].set_visible(False)
    ax.spines["left"].set_visible(False)
    ax.get_xaxis().tick_bottom()
    ax.get_yaxis().tick_left()

    x_axis_pad = 0.1
    y_axis_pad = 0.1
    plt.ylim(y_min - y_axis_pad, y_max + y_axis_pad)
    plt.xlim(x_min - x_axis_pad, x_max + x_axis_pad)

    if as_ranking:
        y_titles = []
        for i in xrange(y_min, y_max + 1):
            if i == 1:
                x = "st"
            elif i==2:
                x = "nd"
            elif i==3:
                x = "rd"
            else:
                x = "th"
            y_titles.append("$%s^{%s}$" % (i, x))
        plt.yticks(range(y_min, y_max + 1), y_titles)
        # make sure the higher ranking are above
        plt.gca().invert_yaxis()
    xticks_rotation = 0 if sum(map(len, x_labels))<10 else 20
    plt.xticks(
        range(len(x_labels)),
        x_labels,
        rotation=xticks_rotation,
    )

    # horizontal lines
    if as_ranking:
        plt.ylabel("Rank", fontsize=20)
        for y in xrange(y_min, y_max + 1):
            plt.plot(
                [x_min, x_max],
                [y, y],
                color='black',
                linestyle='--',
                linewidth=0.2,
                alpha=0.2,
            )

    # Remove the tick marks; they are unnecessary with the tick lines we just plotted.
    plt.tick_params(axis="both", which="both", bottom="off", top="off",
                    labelbottom="on", left="off", right="off", labelleft="on")

    # dict: value => series_list - where each series ends - determines the location of series' title
    # in case of tie there can be several series ending with the same title
    last_values = defaultdict(list)
    for i,series in enumerate(data.keys()):
        if kwargs.get('highlight_series', None) is not None and series in kwargs.get('highlight_series'):
            lw = 3
            alpha = 0.9
        else:
            lw = 2
            alpha= 0.8

        # Plot each line separately with its own color
        plt.plot(
            range(len(x_labels)),
            data_values[series],
            settings.LINE_MARKERS[0] + '-',
            ms=5, lw=lw, alpha=alpha,
            color=color_dict[series],
            markeredgecolor=color_dict[series],
        )

        last_values[data_values[series][-1]].append(series)


    # draw series's titles
    label_y_shift = 0
    label_x_shift = 0.05
    label_y_mult_shift = 0.02 # in case multiple series end with the same value
    for i, s_list in last_values.iteritems():
        n_s = len(s_list)
        for j, series in enumerate(s_list):
            plt.text(
                x_max + x_axis_pad + label_x_shift,
                i + label_y_shift + (i - (n_s - 1) / 2.0) * label_y_mult_shift,
                series,
                fontsize=12,
                color=color_dict[series]
            )


    #create output directory if it does not exists
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    #save plot to a file
    output_path = output_dir + output_filename + '.' + format
    logger.debug("Saving slopegraph plot to %s" % output_path)
    plt.savefig(output_path, bbox_inches='tight', format=format, dpi=settings.CHART_DPI)

    #clear plot for the next plot
    plt.close()

    return output_path