import os

from .. import utils, settings
from .. import settings
from ..log import logger
from collections import defaultdict

def generate_community_flow_webpage(partitions, output_dir, output_filename,
                                                changepoints_with_start_end=None, **kwargs):
    logger.debug("Generating Sankey community evolution webpage for %s partitions" % len(partitions))
    if len(partitions) < 2:
        logger.warning(
            "Cannot generate community evolutions for less than 2 partitions: %s available" % len(partitions)
        )
        return

    if not os.path.exists(output_dir):
        os.makedirs(output_dir)
    output_path = output_dir + output_filename

    width_timepoints = len(partitions)
    if changepoints_with_start_end is not None:
        width_timepoints = changepoints_with_start_end[-1] - changepoints_with_start_end[0] + 1

    # timepoints labels
    labels = []
    timepoints_labels = kwargs.get('timepoints_labels', None)
    if timepoints_labels is None:
        timepoints_labels = range(len(partitions))
    for i in timepoints_labels:
        labels.append("'%s'" % i)
    timepoints_labels_string = ", ".join(labels)

    min_cluster_size = kwargs.get('min_cluster_size', 0)

    labels = {}
    objects_names = kwargs.get('objects_labels', None)
    if objects_names is None:
        objects_names = {}
    # clusters labels
    for p in partitions:
        for c, objs in p.get_clusters_objects().iteritems():
            if len(objs) < min_cluster_size:
                continue
            labels[c] = "'" + ", ".join([str(objects_names.get(obj, obj)) for obj in objs]) + "'"
    clusters_objects_string = ", ".join(labels[c] for c in sorted(labels.keys())).decode().encode('utf-8')
    sankey_flow_data_csv_string, link_objs_dict = _get_community_evolution_flow_csv_string(
            partitions,
            changepoints_with_start_end=changepoints_with_start_end,
            **kwargs
        )

    link_descr_str = []
    for c_prev in link_objs_dict.keys():
        temp_str = []
        for c_curr in link_objs_dict[c_prev].keys():
            temp_str.append("'%s': '%s'" % (c_curr, ", ".join([str(objects_names.get(obj, obj)) for obj in link_objs_dict[c_prev][c_curr]])))
        link_descr_str.append(("'%s': {" % c_prev) +", ".join(temp_str) +  "}")
    link_descr_str = ", ".join(link_descr_str)

    template_values = {
        'node_width': 20,
        'node_padding': 20,
        'total_width': width_timepoints * 50 + (len(partitions) - 1) * 30,
        'total_height': max(p.get_clusters_count() for p in partitions) * 120,
        'sankey_flow_data_csv_string': sankey_flow_data_csv_string,
        'timepoints_labels': timepoints_labels_string,
        'links_labels': link_descr_str,
        'clusters_objects': clusters_objects_string,
        'use_unique_nodes_colors': 1 if kwargs.get('use_unique_nodes_colors', False) else 0,
        'include_non_flow_objects':  1 if kwargs.get('include_non_flow_objects', True) else 0,
    }
    template = file(settings.SANKEY_WEBPAGE_TEMPLATE, 'r').read()
    open(output_path, mode='wb').write(template % template_values)


def _get_community_evolution_flow_csv_string(partitions, changepoints_with_start_end=None, **kwargs):
    link_objs_dict = {} # objects that participate in each "link" (i.e., which moved from c_prev to c_curr)
    res = ['source,target,value']
    prev_P = partitions[0]
    curr_P = None

    min_cluster_size = kwargs.get('min_cluster_size', 0)

    widths = None
    if changepoints_with_start_end is not None:
        widths = []
        for i in xrange(1, len(changepoints_with_start_end)):
            widths.append(changepoints_with_start_end[i] - changepoints_with_start_end[i-1])
        # last time point is inclusive for the last partition
        widths[-1] += 1

    # to objects not shown from flow (i.e., they just appear/disappear in a cluster)
    c_prev_flow_objects = defaultdict(list) # objects from previous partition that appear in flow
    c_curr_flow_objects = defaultdict(list) # objects from current partition that appear in flow
    for i in xrange(1,len(partitions) + 1):

        # note: i iterates to len(partitions) inclusively so that process, if needed, the last partition for 'include_non_flow_objects'
        if i < len(partitions):
            curr_P = partitions[i]

            #for each pair of clusters, count common elements
            for c_prev, c_prev_objects in prev_P.get_clusters_objects().iteritems():
                if len(c_prev_objects) < min_cluster_size:
                    continue
                for c_curr, c_curr_objects in curr_P.get_clusters_objects().iteritems():
                    if len(c_curr_objects) < min_cluster_size:
                        continue
                    common_objects = utils.get_intersection(c_prev_objects, c_curr_objects)
                    link_weight = len(common_objects)
                    c_prev_flow_objects[c_prev] += common_objects
                    c_curr_flow_objects[c_curr] += common_objects
                    if link_weight>0:
                        if c_prev not in link_objs_dict:
                            link_objs_dict[c_prev] = {}
                        link_objs_dict[c_prev][c_curr] = common_objects

                        c_id_1 = '{c_1}_{start_time_1}_{width_1}'.format(
                            c_1=c_prev,
                            start_time_1=changepoints_with_start_end[i-1]
                                if changepoints_with_start_end is not None else i-1,
                            width_1=widths[i-1] if widths is not None else 1,
                        )
                        c_id_2 = '{c_2}_{start_time_2}_{width_2}'.format(
                            c_2=c_curr,
                            start_time_2=changepoints_with_start_end[i]
                                if changepoints_with_start_end is not None else i,
                            width_2=widths[i] if widths is not None else 1,
                        )

                        res.append(
                            '{c_id_1},{c_id_2},{weight}'.format(
                                c_id_1=c_id_1,
                                c_id_2=c_id_2,
                                weight=link_weight,
                            )
                        )

        # include objects that appear/disappear (i.e., do not participate in flow) - self-edges
        # note: objects are drawn for in/out coming edges (c_out_flow_objects contains all such objects for a cluster),
        # so if an object is neither in an in/out edge - add it as self loop
        if kwargs.get('include_non_flow_objects', True):
            for c_prev, c_prev_objects in prev_P.get_clusters_objects().iteritems():
                n_not_shown = len(c_prev_objects) - len(set(c_prev_flow_objects[c_prev]))
                if len(c_prev_objects) < min_cluster_size:
                    continue
                if n_not_shown > 0: # objects are not present in the previous partition
                    c_id = '{c}_{start_time}_{width}'.format(
                        c=c_prev,
                        start_time=changepoints_with_start_end[i-1]
                            if changepoints_with_start_end is not None else i-1,
                        width=widths[i-1] if widths is not None else 1,
                    )
                    res.append(
                        '{c_id},{c_id},{weight}'.format(
                            c_id=c_id,
                            weight=n_not_shown,
                        )
                    )
        c_prev_flow_objects = c_curr_flow_objects
        prev_P = curr_P
    return '\n'.join(res), link_objs_dict