import os
import settings
import utils
import shutil
from network import Network
import community_evolution
from log import logger

class DynamicNetwork(object):
    _CACHE = {}

    def __init__(self, network_id, snapshots, ground_truths=None, ground_truths_affinity=None,
                 description=None, timepoints_labels=None, nodes_labels=None):
        self._network_id = network_id
        self._snapshots = snapshots

        # # make sure all snapshots have the same node set
        # all_nodes = self.get_all_nodes()
        # for sn in self._snapshots:
        #     sn.get_nx_graph().add_nodes_from(all_nodes)
        # print [sn.get_nodes_count() for sn in self._snapshots]

        self._description = description
        #TODO: combine community evolution and affinity matrix evolution ground truths
        # ground truth community evolutions
        self._ground_truths = ground_truths or []
        for c_ev in self.get_ground_truths_list():
            c_ev.set_network_id(network_id)
        # ground truth affinity matrix evolutions
        self._ground_truths_affinity = ground_truths_affinity or []
        for c_ev in self.get_ground_truths_affinity_list():
            c_ev.set_network_id(network_id)

        self._timepoints_labels = timepoints_labels
        self._nodes_labels = nodes_labels

    def get_id(self):
        return self._network_id

    def get_timepoints(self):
        return  range(0, len(self._snapshots))

    def get_timepoints_labels(self):
        return self._timepoints_labels \
            if (hasattr(self,'_timepoints_labels') and self._timepoints_labels is not None) \
            else self.get_timepoints()

    def get_nodes_labels(self):
        return self._nodes_labels if hasattr(self,'_timepoints_labels') else None

    def get_snapshots(self):
        return self._snapshots

    def get_segment_ground_truth(self):
        # TODO: refactor this; make usage of this function more generalizable
        if hasattr(self, '_segment_ground_truth'): # backwards compatibility
            return self._segment_ground_truth
        for c_ev in self.get_ground_truths_list():
            if c_ev.get_method_id() == 'segment_ground_truth': # or (c_ev.get_segments_partitions() is not None and len(c_ev.get_changepoints()) > 0):
                return c_ev

    def get_ground_truth(self, method_id):
        for gt in (self.get_ground_truths_list() + self.get_ground_truths_affinity_list()):
            if gt.get_method_id() == method_id:
                return gt
        return None

    def get_ground_truths_list(self):
        return self._ground_truths

    def get_ground_truths_affinity_list(self):
        return self._ground_truths_affinity

    def get_dir(self):
        return DynamicNetwork.get_network_dir(self._network_id)

    def get_snapshots_dir(self):
        return DynamicNetwork.get_network_snapshots_dir(self._network_id)

    def get_ground_truth_dir(self):
        return DynamicNetwork.get_ground_truth_dir(self._network_id)

    def get_images_dir(self):
        return DynamicNetwork.get_network_images_dir(self._network_id)

    def get_reports_dir(self):
        return DynamicNetwork.get_network_reports_dir(self._network_id)

    def get_methods_comparison_dir(self):
        return DynamicNetwork.get_network_methods_comparison_dir(self._network_id)

    def get_description(self):
        return self._description
    def set_description(self, description):
        self._description = description

    def get_summary(self):
        segment_ground_truth_descr = None
        if self.get_segment_ground_truth() is not None:
            segment_ground_truth_descr = "{method_title}: " \
               "changepoints - [{changepoints}] for time interval [{t_start}-{t_end}]; " \
               "partitions sizes - [{communities_sizes}]".format(
                    method_title=self.get_segment_ground_truth().get_method_title(),
                    changepoints=", ".join(map(str, self.get_segment_ground_truth().get_changepoints()))
                        if self.get_segment_ground_truth() is not None else "unknown",
                    t_start=self.get_timepoints_labels()[0],
                    t_end=self.get_timepoints_labels()[-1],
                    communities_sizes=", ".join(
                    map(
                        str,
                        [
                            p.get_clusters_count() for p in self.get_segment_ground_truth().get_segments_partitions()
                        ]
                    )
                    ) if self.get_segment_ground_truth() is not None else "unknown",
               )

        return "Dynamic network '{network_id}' - " \
               "number of snapshots: {n_networks}; {segment_ground_truth_descr}. " \
               "".format(
            network_id=self._network_id,
            n_networks=len(self._snapshots),
            segment_ground_truth_descr=segment_ground_truth_descr or ''
        )

    @staticmethod
    def get_networks_list_all_nodes(networks_list):
        res = []
        for sn in networks_list:
            res += sn.get_nx_graph().nodes()
        return sorted(list(set(res)), cmp=utils.cmp_str_int)

    def get_all_nodes(self, t_start=None, t_end=None):
        t_start = t_start or self.get_timepoints()[0]
        t_end = t_end or self.get_timepoints()[-1]
        return DynamicNetwork.get_networks_list_all_nodes(
            networks_list=self.get_snapshots()[t_start:t_end+1]
        )

    def get_computed_methods_ids_list(self):
        return DynamicNetwork.get_network_computed_methods_ids_list(self.get_id())

    @staticmethod
    def get_network_computed_methods_ids_list(network_id):
        return [
            m for m in utils.get_subdir_list_from_dir(DynamicNetwork.get_network_methods_dir(network_id), full_path=False)
            if not m.startswith('_')
        ]

    def get_nodes(self):
        res = []
        for sn in self.get_snapshots():
            res += sn.get_nodes()
        return sorted(set(res))

    def save(self, **kwargs):
        logger.debug("Saving dynamic network {network_id}".format(network_id=self._network_id))
        if not os.path.exists(self.get_dir()):
            os.makedirs(self.get_dir())

        for i in xrange(len(self._snapshots)):
            self._snapshots[i].save(
                DynamicNetwork.get_network_snapshots_dir(self._network_id),
                str(self.get_timepoints_labels()[i]) if kwargs.get('use_timepoints_labels_snapshots_names') else str(i),
                'edgelist',
                save_data=kwargs.get('save_data', True)
            )
        for c_ev in self.get_ground_truths_list():
            c_ev.save(
                as_ground_truth=True,
            )

        if self.get_description() is not None:
            with open(self.get_dir() + 'description.txt', 'wb') as outfile:
                outfile.write(self.get_description())

    @staticmethod
    def load(network_id):
        """
        Load dynamic network by id
        :param network_id:
        :return: DynamicNetwork
        """
        if network_id in DynamicNetwork._CACHE:
            logger.debug("Obtaining dynamic network '%s' from cache" % network_id)
        else:
            logger.debug("Loading dynamic network '%s'" % network_id)
            # ground truth community evolutions
            ground_truths = []
            ground_truths_affinity = []
            for gt_file in utils.get_file_list_from_dir(
                DynamicNetwork.get_network_ground_truth_dir(network_id),
                include_dirs=True,
                full_path=True
            ):
                if gt_file.startswith('_'):
                    continue
                method_id = os.path.basename(gt_file).replace('.txt','')
                # temporary fix: ignore snapshot ground truth
                if method_id == 'snapshot_ground_truth':
                    continue
                c_ev = community_evolution.CommunityEvolution.load_from_path(
                    input_path=gt_file,
                    method_id=method_id,
                    network_id=network_id
                )
                if c_ev is not None:
                    if c_ev.has_partitions():
                        ground_truths.append(c_ev)
                    else:
                        ground_truths_affinity.append(c_ev)

            DynamicNetwork._CACHE[network_id] = DynamicNetwork(
                network_id=network_id,
                snapshots=DynamicNetwork.load_networks_list(network_id),
                ground_truths=ground_truths,
                ground_truths_affinity=ground_truths_affinity,
                description=DynamicNetwork.load_description(network_id),
                timepoints_labels=DynamicNetwork.load_timepoints_labels(network_id),
                nodes_labels=DynamicNetwork.load_nodes_labels(network_id),
            )
        return DynamicNetwork._CACHE[network_id]

    @staticmethod
    def delete(network_id):
        logger.debug("Deleting network '%s' and all its data " % network_id)
        shutil.rmtree(DynamicNetwork.get_network_dir(network_id))

    @staticmethod
    def parse_network_id(network_id):
        if '/' in network_id:
            v = network_id.split('/')
            network_input_dir = v[0] if len(v)>1 else ''
            network_name = v[1 if len(v)>1 else 0]
            network_instance = v[2] if len(v)>2 else ''
        else:
            network_input_dir = ''
            network_name = network_id
            network_instance = ''
        return network_input_dir, network_name, network_instance

    @staticmethod
    def get_snapshots_count(network_id):
        return len(
            DynamicNetwork.get_sorted_snapshot_files(
                DynamicNetwork.get_network_snapshots_dir(network_id)
            )
        )

    @staticmethod
    def get_network_dir(network_id):
        return settings.NETWORKS_BASE_DIR + network_id + '/'

    @staticmethod
    def get_network_snapshots_dir(network_id):
        return DynamicNetwork.get_network_dir(network_id) + 'snapshots/'

    @staticmethod
    def get_network_methods_dir(network_id):
        return DynamicNetwork.get_network_dir(network_id) + 'methods/'

    @staticmethod
    def get_network_ground_truth_dir(network_id):
        return DynamicNetwork.get_network_dir(network_id) + 'ground_truth/'

    @staticmethod
    def get_network_images_dir(network_id):
        return DynamicNetwork.get_network_dir(network_id) + 'images/'

    @staticmethod
    def get_network_reports_dir(network_id):
        return settings.REPORTS_DIR + 'networks/' + network_id + '/'

    @staticmethod
    def get_network_methods_comparison_dir(network_id):
        """
        Directory where methods comparisons are located (one-vs-one or more methods at a time)
        Within this directory, each type of comparison is located in separate directory (named as key).
        List of the methods involved in a particular comparison is located within methods.txt file inside
        a corresponding subdirectory
        :param network_id:
        :return:
        """
        return DynamicNetwork.get_network_dir(network_id) + 'methods_comparison/'

    @staticmethod
    def get_compared_methods_ids_by_key(network_id, comparison_methods_key):
        """
        Given comparison type, get list of method ids involved in it
        :param network_id:
        :param comparison_methods_key:
        :return:
        """
        with open(
            DynamicNetwork.get_network_methods_comparison_dir(network_id) + comparison_methods_key + '/methods.txt'
        ) as infile:
            return [m.replace('\n','') for m in infile.readlines()]

    @staticmethod
    def get_sorted_snapshot_files(snapshots_dir, full_path=False):
        return sorted(
            utils.get_file_list_from_dir(snapshots_dir, ext="txt", full_path=full_path),
            cmp=utils.cmp_str_int
        )
    @staticmethod
    def load_networks_list(network_id):
        snapshots_dir = DynamicNetwork.get_network_snapshots_dir(network_id)
        if not os.path.exists(snapshots_dir):
            raise Exception("Directory does not exist: '%s'" % snapshots_dir)
        snapshots_files = DynamicNetwork.get_sorted_snapshot_files(snapshots_dir, full_path=True)
        networks_list = []
        for f in snapshots_files:
            networks_list.append(Network.load(f, 'edgelist'))
        return networks_list

    @staticmethod
    def load_timepoints_labels(network_id):
        snapshots_dir = DynamicNetwork.get_network_snapshots_dir(network_id)
        if not os.path.exists(snapshots_dir):
            raise Exception("Directory does not exist: '%s'" % snapshots_dir)
        return [
            s.replace('.txt', '') for s in DynamicNetwork.get_sorted_snapshot_files(snapshots_dir, full_path=False)
        ]

    @staticmethod
    def load_description(network_id):
        descr_path = DynamicNetwork.get_network_dir(network_id) + 'description.txt'
        res = None
        if os.path.exists(descr_path):
            with open(descr_path, 'r') as infile:
                res = infile.read()
        return res

    @staticmethod
    def load_nodes_labels(network_id):
        nodes_labels_path = DynamicNetwork.get_network_dir(network_id) + 'nodes_labels.txt'
        res = {}
        if os.path.exists(nodes_labels_path):
            with open(nodes_labels_path, 'r') as infile:
                for line in infile.readlines():
                    values = line.strip().split(',')
                    res[values[0]] = values[1]
        return res

    def create_dynamic_network_per_changepoints(self, changepoints, new_network_id, **kwargs):
        """
        Create new dynamic network where each snapshot corresponds to a segment as defined by changepoints
        :param changepoints:
        :param new_network_id:
        :return:
        """
        i_prev = 0
        snapshots = []
        timepoints_labels = []

        prev_snapshots = []
        if kwargs.get('nodes_list') is None:
            prev_snapshots = self.get_snapshots()
        else:
            for sn in self.get_snapshots():
                nx_graph = sn.get_nx_graph().subgraph(kwargs.get('nodes_list'))
                prev_snapshots.append(Network(nx_graph))

        for i_curr in xrange(len(self.get_snapshots()) + 1):
            if i_curr in changepoints or i_curr == len(self.get_snapshots()):
                g_union = Network.graph_union(
                    prev_snapshots[i_prev:i_curr],
                    weighted=True
                )
                if kwargs.get('min_nodes_count') is None or g_union.get_nodes_count() >= kwargs.get('min_nodes_count'):
                    snapshots.append(g_union)
                    timepoints_labels.append(self.get_timepoints_labels()[i_prev])
                i_prev = i_curr

        return DynamicNetwork(
            network_id=new_network_id,
            snapshots=snapshots,
            timepoints_labels=timepoints_labels,
            nodes_labels=self.get_nodes_labels(),
        )