from collections import defaultdict
import numpy as np
import networkx as nx
from ..dynamic_network import DynamicNetwork
from ..method import Method
from ..community_evolution import CommunityEvolution
from ..classification_result import ClassificationResult
from ..log import logger
from ..utils import parse_user_args_string_into_values_list

def get_solutions_n_segments(networks_ids, methods_dict):
    """
    Get average number of segments in each method's solution
    :param networks_ids:
    :param methods_dict:
    :return:
    """
    segments_counts = defaultdict(list)
    for n_id in networks_ids:
        for m_key,m_ids in methods_dict.iteritems():
            for m_id in m_ids:
                m = Method.load_state(n_id, m_id)
                segments_counts[m_key].append(len(m.get_community_evolution().get_changepoints()) + 1)
    for m_key in methods_dict.keys():
        segments_counts[m_key] = np.average(segments_counts[m_key])
    return segments_counts


def print_changepoints_scores(methods_ids, c_ev_gt):
    dn = c_ev_gt.get_network()
    n_id = dn.get_id()
    gt_scores = [(i in c_ev_gt.get_changepoints()) if c_ev_gt is not None else False for i in dn.get_timepoints()]
    # gt_scores = [(y in ['1855',  '1880', '1900', '1915', '1920', '1940', '1945', '1950',  '1955', '1990']) for i,y in enumerate(dn.get_timepoints_labels())]
    for m_id in methods_ids:
        m = Method.load_state(
            network_id=n_id,
            method_id=m_id,
        )
        if m is None:
            logger.warning("Method does not exist: %s" % m_id)
            continue
        m._network_id = n_id
        c = ClassificationResult(
            ground_truth_labels=gt_scores[1:],
            predicted_scores=m.get_timepoints_scores()[1:],
        )
        for i in dn.get_timepoints()[1:]:
            logger.debug(
                "%s - method=%s, gt=%s" % (dn.get_timepoints_labels()[i], m.get_timepoints_scores()[i], gt_scores[i])
            )
        if c_ev_gt is not None:
            logger.info("Classification results summary for method '{m_id}' vs '{gt_id}' : {summary} ".format(
                m_id=m.get_id(),
                gt_id=c_ev_gt.get_method_id(),
                summary=c.get_summary(),
            ))
    return

def print_top_changepoints(methods_ids, dn, max_count = 10):
    for m_id in methods_ids:
        m = Method.load_state(
            network_id=dn.get_id(),
            method_id=m_id,
        )
        if m is None:
            print "Skipping %s" % m_id
            continue
        top_timepoints = np.argsort(m.get_timepoints_scores())[-1:-max_count-1:-1]
        print m_id
        for i,t in enumerate(top_timepoints):
            print(
                "%s" % (dn.get_timepoints_labels()[t])
            )

def print_network_size(dn):
    """
    For a dynamic network print its number of nodes/edges in general and per snapshot
    :param dn:
    :return:
    """
    ns = []
    es = []
    for sn in dn.get_snapshots():
        ns.append(sn.get_nodes_count())
        es.append(sn.get_edges_count())
    logger.info("Avg n: %s +- %s" % (np.mean(ns), np.std(ns)))
    logger.info("Avg e: %s +- %s" % (np.mean(es), np.std(es)))
    logger.info("Total e: %s" % np.sum(es))
    logger.info("Total n: %s" % len(dn.get_all_nodes()))

def print_check_synthetic_networks(n_ids):
    """
    Print check of synthetic networks and their ground truths  (e.g., the actual intra/inter cluster densities)
    :param n_ids:
    :return:
    """
    min_cluster_size = 1000
    max_clusters_count = 0
    actual_inter_edges = 0
    max_inter_edges = 0
    actual_intra_edges = 0
    max_intra_edges = 0
    for n_id in n_ids:

        sgt = CommunityEvolution.load(n_id, 'segment_ground_truth')
        dn = DynamicNetwork.load(n_id)
        for i_sn, sn in enumerate(dn.get_snapshots()):
            block_graph = nx.blockmodel(
                G=sn.get_nx_graph(),
                partitions=sgt.get_snapshot_partition(i_sn).get_clusters_objects().values(),
                multigraph=False,
            )
            blocks = block_graph.nodes(data=True)
            res = 0.0
            for i, i_attr in blocks:
                for j, j_attr in blocks:
                    if j > i:
                        break
                    if i == j:
                        max_intra_edges += i_attr['nnodes'] * (i_attr['nnodes'] - 1.0) / 2
                        actual_intra_edges += i_attr['nedges']
                    else:
                        max_inter_edges += i_attr['nnodes'] * j_attr['nnodes']
                        actual_inter_edges += block_graph[i][j].get('weight', 1.0) if block_graph.has_edge(i, j) else 0.0

        min_temp_cluster_size = 1000
        max_temp_clusters_count = 0
        for p in sgt.get_segments_partitions():
            min_temp_cluster_size = min(min_cluster_size,  min(p.get_clusters_sizes()))
            max_temp_clusters_count = max(max_temp_clusters_count, p.get_clusters_count())
        print "%s - %s" % (min_temp_cluster_size, max_temp_clusters_count)
        min_cluster_size = min(min_cluster_size, min_temp_cluster_size)
        max_clusters_count = max(max_clusters_count, max_temp_clusters_count)
    print "Min cluster size: %s" % min_cluster_size
    print "Max clusters count: %s" % max_clusters_count
    print "Intra: %s/%s = %s" % (actual_intra_edges, max_intra_edges, 1.0*actual_intra_edges/max_intra_edges)
    print "Inter: %s/%s = %s" % (actual_inter_edges, max_inter_edges, 1.0 * actual_inter_edges / max_inter_edges)
