import os
from ..log import logger
from ..draw import (
    sankey,
    com_ev_draw,
    partition_draw,
    aff_matr_draw,
)
import eval_utils

def evaluate_community_evolution(c_ev, dyn_network=None, overwrite=False, **kwargs):
    logger.debug("Evaluating community evolution for {method_id}".format(method_id=c_ev.get_method_id()))
    dyn_network = dyn_network or c_ev.get_dynamic_network()

    com_ev_draw.draw_com_ev(
        c_ev,
        show_ground_truth=kwargs.get('show_ground_truth', True),
        overwrite=overwrite,
        draw_partitions_qualities_timeseries=True,
        draw_partitions_qualities_heatmaps=False,
        draw_partitions_comparison_heatmaps=False,
    )
    for gt in dyn_network.get_ground_truths_list():
        if c_ev.get_method_id() != gt.get_method_id():
            compare_community_evolutions(
                c_evs_list=[c_ev, gt],
                dyn_network=dyn_network,
                overwrite=overwrite,
                draw_partitions_qualities_timeseries=False,
                draw_partition_comparison_heatmaps=False,
            )

    for gt in dyn_network.get_ground_truths_affinity_list():
        if c_ev.get_method_id() != gt.get_method_id():
            evaluate_com_evs_against_affinity(
                c_evs_list=[c_ev],
                aff_matr_ev=gt,
                dyn_network=dyn_network,
                overwrite=overwrite,
            )
def compare_community_evolutions(c_evs_list, dyn_network=None, output_dir=None,
                                 comparison_methods_key=None, overwrite=True, **kwargs):
    """
    :param c_evs_list:
    :param dyn_network:
    :param comparison_methods_key: Subdirectory where the results of comparison will be saved
    :param overwrite:
    :return:
    """
    methods_ids_list = [c.get_method_id() for c in c_evs_list]
    logger.debug("Drawing comparison of community evolutions {c_ev_list}".format(
        c_ev_list=", ".join(methods_ids_list),
    ))
    dn = dyn_network or c_evs_list[0].get_dynamic_network()
    if output_dir is None:
        output_dir = eval_utils.init_comparison_dir(dn.get_id(), methods_ids_list, comparison_methods_key)
    output_images_dir = output_dir + 'images/'
    if not os.path.exists(output_images_dir):
        os.makedirs(output_images_dir)

    logger.debug("Drawing per-timepoint partition comparison Sankey plots")
    sankey_dir = output_images_dir + 'sankey-partitions-per-t/'
    draw_sankey = True
    if os.path.exists(sankey_dir):
        logger.debug('Per-timepoint partition comparison Sankey plots have been already drawn: %s' % sankey_dir)
        if not overwrite:
            logger.debug("No overwrite: aborting")
            draw_sankey = False
    if draw_sankey:
        for i in xrange(len(dn.get_timepoints())):
            sankey.generate_community_flow_webpage(
                partitions=[c.get_snapshot_partition(i) for c in c_evs_list if c.get_snapshot_partition(i) is not None],
                output_dir=sankey_dir,
                output_filename='%s.html' % i,
                use_unique_nodes_colors=True,
                objects_labels=dyn_network.get_nodes_labels(),
            )

    logger.debug("Drawing segmentations comparison")
    com_ev_draw.draw_segmentations(c_evs_list=c_evs_list, output_dir=output_images_dir,
        output_filename='segmentation', show_ground_truth=True,
        overwrite=overwrite
    )

    if kwargs.get('draw_partitions_qualities_timeseries', False):
        com_ev_draw.draw_list_partitions_qualities_timeseries(
            c_evs=c_evs_list,
            dynamic_network=dn,
            output_dir=output_images_dir,
            show_ground_truth=True,
            overwrite=overwrite,
        )

    com_ev_draw.draw_list_com_evs_scores_bars(
        c_evs_list=c_evs_list,
        dynamic_network=dn,
        output_dir=output_images_dir,
        scoring_methods_ids=kwargs.get('scoring_methods_ids'),
        include_default_scoring_methods=True,
        show_ground_truth=True,
        draw_pairwise_corr_matrices=True,
        overwrite=overwrite
    )

    if len(methods_ids_list) == 2 and kwargs.get('one_vs_one_comparison', True):
        ce_1 = c_evs_list[0]
        ce_2 = c_evs_list[1]

        com_ev_draw.draw_list_partitions_similarities_to_other_timeseries(
            c_evs_list=[ce_1],
            c_ev_gt=ce_2,
            output_dir=output_images_dir,
            overwrite=overwrite,
        )

        if kwargs.get('draw_partition_comparison_heatmaps', False) and\
                ce_1.get_segments_partitions() is not None and ce_2.get_segments_partitions() is not None:
            logger.debug("Drawing heatmaps for segments partition comparison")
            partition_draw.draw_partitions_comparison_heatmaps(
                partitions_1=ce_1.get_segments_partitions(),
                partitions_2=ce_2.get_segments_partitions(),
                output_dir=output_images_dir + 'heatmaps-partitions/',
                x_labels=ce_1.get_segments_intervals_labels(),
                y_labels=ce_2.get_segments_intervals_labels(),
                x_label=ce_1.get_method_title(),
                y_label=ce_2.get_method_title(),
                overwrite=overwrite,
            )
            logger.debug("Drawing heatmaps for per-snapshot partition comparison")
            partition_draw.draw_partitions_comparison_heatmaps(
                partitions_1=[p[1] for p in ce_1.generate_snapshot_partition_pairs()],
                partitions_2=[p[1] for p in ce_2.generate_snapshot_partition_pairs()],
                output_dir=output_images_dir + 'heatmaps-partitions-per-t/',
                x_label=ce_1.get_method_title(),
                y_label=ce_2.get_method_title(),
                overwrite=overwrite,
            )

        com_ev_draw.draw_list_com_evs_similarities_to_other_bars(
            c_evs_list=[ce_1],
            c_ev_gt=ce_2,
            output_dir=output_images_dir,
            dyn_network=dn,
            overwrite=overwrite,
        )

        com_ev_draw.draw_list_com_evs_similarities_to_other_scatter(
            c_evs_list=[ce_1],
            c_ev_gt=ce_2,
            output_dir=output_images_dir,
            dyn_network=dn,
            overwrite=overwrite,
        )

    return output_images_dir

def evaluate_com_evs_against_other(c_evs_list, c_ev_gt, dyn_network=None, output_dir=None,
                                   comparison_methods_key=None, overwrite=True):
    """
    Evaluate community evolution against other (i.e., ground truth)
    :param c_evs_list:
    :param c_ev_gt:
    :param dyn_network:
    :param output_dir:
    :param comparison_methods_key:
    :param overwrite:
    :return:
    """
    methods_ids_list = [c.get_method_id() for c in c_evs_list]
    logger.debug("Evaluating community evolutions {c_ev_list} against {gt}".format(
        c_ev_list=", ".join(methods_ids_list),
        gt=c_ev_gt.get_method_id(),
    ))

    dn = dyn_network or c_evs_list[0].get_dynamic_network()
    if output_dir is None:
        output_dir = eval_utils.init_comparison_dir(dn.get_id(), methods_ids_list, comparison_methods_key)
    output_images_dir = output_dir + 'images/evaluation_against_other/' + c_ev_gt.get_method_id() + '/'

    for func in [
        com_ev_draw.draw_list_partitions_similarities_to_other_timeseries,
        com_ev_draw.draw_list_com_evs_similarities_to_other_bars,
        com_ev_draw.draw_list_com_evs_similarities_to_other_scatter,
    ]:
        func(
            c_evs_list=c_evs_list,
            c_ev_gt=c_ev_gt,
            output_dir=output_images_dir,
            dyn_network=dn,
            overwrite=overwrite,
        )


def evaluate_com_evs_against_affinity(c_evs_list, aff_matr_ev, dyn_network=None, output_dir=None,
                                 comparison_methods_key=None, overwrite=True, **kwargs):

    methods_ids_list = [c.get_method_id() for c in c_evs_list]
    logger.debug("Evaluating community evolutions {c_ev_list} vs affinity matrix evolution {a_m_ev}".format(
        c_ev_list=", ".join(methods_ids_list),
        a_m_ev=aff_matr_ev.get_method_id(),
    ))
    dn = dyn_network or c_evs_list[0].get_dynamic_network()
    output_images_dir = ""
    if output_dir is None:
        if len(methods_ids_list) > 1:
            output_dir = eval_utils.init_comparison_dir(dn.get_id(), methods_ids_list, comparison_methods_key)
            output_images_dir = output_dir + 'images/evaluation_against_other/' + aff_matr_ev.get_method_id() + '/'
        else:
            output_dir = eval_utils.init_comparison_dir(
                dn.get_id(),
                methods_ids_list + [aff_matr_ev.get_method_id()],
                comparison_methods_key
            )
            output_images_dir = output_dir + 'images/'

    logger.debug("Drawing segmentations comparison")
    com_ev_draw.draw_segmentations(c_evs_list=c_evs_list + [aff_matr_ev], output_dir=output_images_dir,
        output_filename='segmentation', show_ground_truth=True,
        overwrite=overwrite
    )

    for func in [
        aff_matr_draw.draw_list_partitions_internal_qualities_timeseries,
        com_ev_draw.draw_list_com_evs_similarities_to_other_bars,
    ]:
        func(
            c_evs_list=c_evs_list,
            c_ev_gt=aff_matr_ev,
            output_dir=output_images_dir,
            dyn_network=dn,
            overwrite=overwrite,
        )