from ..log import logger
from ..draw import method_draw
import eval_utils
import com_ev_eval
from collections import OrderedDict
from ..classification_result import ClassificationResult
from ..draw import classification_result_draw
from ..dynamic_network import DynamicNetwork
from ..method import Method

def evaluate_method(method, **kwargs):
    logger.debug("Evaluating method {method_id}".format(method_id=method.get_id()))

    com_ev_eval.evaluate_community_evolution(method.get_community_evolution(), overwrite=kwargs.get('overwrite', False))

    if kwargs.get('evaluate_changepoint_classification', True):
        for c_ev_gt in method.get_dynamic_network().get_ground_truths_list():
            if c_ev_gt.get_changepoints() is not None and len(c_ev_gt.get_changepoints()) > 0:
                evaluate_changepoint_classification(
                    methods_list=[method],
                    ground_truth_c_ev=c_ev_gt,
                    overwrite=kwargs.get('overwrite', False)
                )

    if kwargs.get('draw_segmentations_n_changepoints', True):
       method_draw.draw_segmentations_n_changepoints(method, overwrite=kwargs.get('overwrite', False))

    output_images_dir = method.get_output_dir() + 'images/'
    method_draw.draw_list_solution_scores_n_changepoints_timeseries(
        [method],
        method.get_dynamic_network(),
        output_images_dir=output_images_dir,
        overwrite=kwargs.get('overwrite', False)
    )
    method_draw.draw_list_timepoints_scores_timeseries(
        [method],
        method.get_dynamic_network(),
        output_images_dir=output_images_dir,
        overwrite=kwargs.get('overwrite', False)
    )

def compare_methods(methods_ids, network_id, comparison_methods_key=None, **kwargs):
    """
    For a set of methods, compare them against each other
    :param methods_ids:
    :param dyn_network:
    :param comparison_methods_key:
    :param kwargs:
    :return:
    """
    methods_list = [Method.load_state(network_id, m_id) for m_id in methods_ids]
    dyn_network = DynamicNetwork.load(network_id)
    logger.debug("Comparing methods {methods_list} for '{dynamic_network}'".format(
        methods_list=", ".join(m.get_title() for m in methods_list),
        dynamic_network=dyn_network.get_id(),
    ))

    comparison_methods_key = comparison_methods_key \
                             or eval_utils.generate_methods_comparison_key(methods_ids)
    output_dir = dyn_network.get_methods_comparison_dir() + comparison_methods_key + '/'
    output_images_dir = output_dir + 'images/'

    # compare final solutions
    # note: this will also init comparison dir
    com_ev_eval.compare_community_evolutions(
        c_evs_list=[m.get_community_evolution() for m in methods_list],
        dyn_network=dyn_network,
        comparison_methods_key=comparison_methods_key,
        overwrite=kwargs.get('overwrite', False),
        scoring_methods_ids=[m.get_scoring_method_id() for m in methods_list if hasattr(m, 'get_scoring_method_id')],
        one_vs_one_comparison=False,
        draw_partitions_qualities_timeseries=True,
        draw_partition_comparison_heatmaps=False,
    )

    # evaluate final solutions against ground truths
    for gt in dyn_network.get_ground_truths_list():
        com_ev_eval.evaluate_com_evs_against_other(
            c_evs_list=[m.get_community_evolution() for m in methods_list],
            c_ev_gt=gt,
            dyn_network=dyn_network,
            comparison_methods_key=comparison_methods_key,
            overwrite=kwargs.get('overwrite', False),
        )
    for gt in dyn_network.get_ground_truths_affinity_list():
        com_ev_eval.evaluate_com_evs_against_affinity(
            c_evs_list=[m.get_community_evolution() for m in methods_list],
            aff_matr_ev=gt,
            dyn_network=dyn_network,
            comparison_methods_key=comparison_methods_key,
            overwrite=kwargs.get('overwrite', False),
        )


    if dyn_network.get_segment_ground_truth() is not None:
        evaluate_changepoint_classification(
            methods_list=methods_list,
            dyn_network=dyn_network,
            output_images_dir=output_images_dir,
            comparison_methods_key=comparison_methods_key,
            ground_truth_c_ev=dyn_network.get_segment_ground_truth(),
            overwrite=kwargs.get('overwrite', False)
        )

    method_draw.draw_list_solution_scores_n_changepoints_timeseries(
        methods_list,
        dyn_network,
        output_images_dir=output_images_dir,
        **kwargs
    )

    method_draw.draw_list_timepoints_scores_timeseries(
        methods_list,
        dyn_network,
        output_images_dir=output_images_dir,
        **kwargs
    )

    method_draw.draw_list_methods_run_time_bars(
        methods_list,
        output_dir=output_images_dir,
        **kwargs
    )



def evaluate_changepoint_classification(methods_list, ground_truth_c_ev, dyn_network=None, output_images_dir=None,
                                        comparison_methods_key=None, **kwargs):
    """
    :param method: Method
    :param kwargs:
    :return:
    """
    logger.debug(
            "Computing classification performance results for method's changepoints identification for %s" % ground_truth_c_ev.get_method_id()
    )

    methods_list_filtered = []
    for m in methods_list:
        if hasattr(m,'get_timepoints_scores'):
            methods_list_filtered.append(m)
    if len(methods_list_filtered) == 0:
        logger.warning(
            "None of the methods %s supports changepoint-like scores"
            % (", ".join(m.get_id() for m in methods_list))
        )
        return
    dyn_network = dyn_network or methods_list[0].get_dynamic_network()
    gt_scores = [(i in ground_truth_c_ev.get_changepoints()) for i in dyn_network.get_timepoints()]

    if len(set(gt_scores[1:])) == 1:
        logger.warning("'{gt_id}' has only one timepoint label - aborting classification results analysis".format(
            gt_id=ground_truth_c_ev.get_method_id(),
        ))
        return


    if output_images_dir is None:
        if len(methods_list_filtered) == 1:
            output_images_dir = eval_utils.init_comparison_dir(
                dyn_network.get_id(),
                [methods_list_filtered[0].get_id(), ground_truth_c_ev.get_method_id()],
                comparison_methods_key
            ) + 'images/'
        else:
            output_images_dir = eval_utils.init_comparison_dir(
                dyn_network.get_id(),
                [m.get_id() for m in methods_list_filtered],
                comparison_methods_key
            ) + 'images/'
    if len(methods_list_filtered) > 1:
        output_images_dir += 'evaluation_against_other/' + ground_truth_c_ev.get_method_id() + '/'
    output_images_dir += 'pr_roc/changepoints/'

    classification_results_dict=OrderedDict()
    for m in methods_list_filtered:
        # note: first timepoint is never a changepoint and thus can be ignored
        c = ClassificationResult(
            ground_truth_labels=gt_scores[1:],
            predicted_scores=m.get_timepoints_scores()[1:],
        )
        # c.save(
        #     output_dir=m.get_output_dir() + 'pr_roc/changepoints/',
        #     output_file='vs-' + ground_truth_c_ev.get_method_id(),
        # )
        logger.info("Classification results summary for method '{m_id}' vs '{gt_id}' : {summary} ".format(
            m_id=m.get_id(),
            gt_id=ground_truth_c_ev.get_method_id(),
            summary=c.get_summary(),
        ))
        classification_results_dict[m.get_title()] = c

    for func in [
        classification_result_draw.draw_precision_recall,
        classification_result_draw.draw_roc,
        classification_result_draw.draw_bars
    ]:
        func(
            classification_results_dict=classification_results_dict,
            output_dir=output_images_dir,
            **kwargs
        )
