from ..log import logger
from ..draw import method_draw
from ..draw import com_ev_draw
import eval_utils
import com_ev_eval_across_networks
from collections import OrderedDict
from .. import settings
from ..methods import methods_utils
from ..dynamic_network import DynamicNetwork
from .. import utils
from ..method import Method
import method_eval_pr_roc_across_networks

# TODO: refactor method eval across networks and across networks dict
# networks dict - network key => network ids (e.g., "RN1 => [instances 1-5])
# Several types of evaluation:
# 1. Methods list across networks list => for each method, average its results across all networks
# e.g., M1, M2 for RN1, RN2, RN3 (rn - random network) => each plot has 2 series, each averaged across 3 networks
# 2. Method across networks dict (e.g., key: network name, values: random instances) => for a method,
# average its results across each network name
# e.g., M1 for RN1 {inst 1-5}, RN2 {inst 1-5}, RN3 {inst 1-5} => each plot has 3 series, each averaged across 5 networks
# 3. Methods list across networks dict => in this case, either create a separate image or we can just combine together images of type 1 for each key:
# i.e., run analysis of type 1 separately for each value list of the networks dict
# e.g., M1, M2, M3, M4, M5 and RN1 {inst 1-5}, RN2 {inst 1-5}, RN3 {inst 1-5}
# => one plot with 5 series (one per method), 3 groups (one per network key), each value averaged across 5 networks


# Evaluation of type 1
def evaluate_methods_across_networks(methods_ids, networks_ids, **kwargs):
    """
    For a set of methods, compare them against each other
    i.e. one figure per network key (each group in figure - measure, series - methods)
    """
    logger.debug("Comparing methods {methods_ids} across networks '{networks_ids}'".format(
        methods_ids=", ".join(methods_ids),
        networks_ids=", ".join(networks_ids),
    ))

    comparison_methods_key = kwargs.get('methods_key_name') or eval_utils.generate_methods_comparison_key(methods_ids)
    comparison_networks_key = eval_utils.get_netwoks_list_comparison_key(networks_ids, **kwargs)
    output_dir = settings.EVAL_DIR + comparison_networks_key + '/' + comparison_methods_key + '/'

    c_evs_dict = OrderedDict()
    methods_dict = OrderedDict()
    for m_id in methods_ids:
        c_evs = []
        ms = []
        for n_id in networks_ids:
            m = Method.load_state(n_id, m_id)
            ms.append(m)
            c_evs.append(m.get_community_evolution())
        c_evs_dict[methods_utils.get_method_title(m_id)] = c_evs
        methods_dict[methods_utils.get_method_title(m_id)] = ms


    output_images_dir = output_dir + 'images/'

    # evaluate against ground truths
    if kwargs.get('draw_against_other', True):
        networks_ground_truths_ids = []
        if kwargs.get('show_ground_truth', True):
            for n_id in networks_ids:
                dn = DynamicNetwork.load(n_id)
                networks_ground_truths_ids.append([])
                for c_ev_gt in (dn.get_ground_truths_list() + dn.get_ground_truths_affinity_list()):
                    networks_ground_truths_ids[-1].append(c_ev_gt.get_method_id())

        for gt_id in utils.get_intersection(*networks_ground_truths_ids):
            if kwargs.get('draw_com_evs_sim', True):
                com_ev_eval_across_networks.evaluate_com_evs_against_other_across_networks(
                    methods_ids=methods_ids,
                    networks_ids=networks_ids,
                    other_method_id=gt_id,
                    output_dir=output_images_dir,
                    c_evs_dict=c_evs_dict,
                    **kwargs
                )

            if kwargs.get('draw_changepoints_pr_roc', False):
                method_eval_pr_roc_across_networks.evaluate_methods_changepoint_classification_across_networks(
                    methods_ids=methods_ids,
                    networks_ids=networks_ids,
                    other_method_id=gt_id,
                    output_dir=output_images_dir,
                    **kwargs
                )

    if kwargs.get('show_ground_truth', True):
        for n_id in networks_ids:
            dn = DynamicNetwork.load(n_id)
            for c_ev_gt in (dn.get_ground_truths_list() + dn.get_ground_truths_affinity_list()):
                gt_title = c_ev_gt.get_method_title()
                if gt_title not in c_evs_dict:
                    c_evs_dict[gt_title] = []
                c_evs_dict[gt_title].append(c_ev_gt)

    if kwargs.get('draw_com_evs_scores', True):
            methods_list = [methods_utils.create_method_by_id(m_id) for m_id in methods_ids]
            com_ev_draw.draw_list_com_evs_scores_bars_avg(
                c_evs_dict=c_evs_dict,
                output_dir=output_images_dir,
                scoring_methods_ids=[m.get_scoring_method_id() for m in methods_list if hasattr(m, 'get_scoring_method_id')],
                **kwargs
            )

    if kwargs.get('draw_run_time', True):
        method_draw.draw_list_methods_run_time_bars_avg(
            methods_dict=methods_dict,
            output_dir=output_images_dir,
            **kwargs
        )


# Evaluation of type 2
def evaluate_method_across_networks_dict(method_id, networks_dict, **kwargs):
    """
    Evaluate methods across networks dict
    i.e. one figure per method (each group in figure - network key, series - measures)
    :param method_id:
    :param networks_dict:
    :param kwargs:
    :return:
    """
    logger.debug("Evaluate method {method_id} across networks keys '{networks_ids}'".format(
        method_id=method_id,
        networks_ids=", ".join(networks_dict),
    ))

    comparison_methods_key = eval_utils.generate_methods_comparison_key([method_id])
    comparison_networks_key = kwargs.get('networks_key_name') or eval_utils.generate_networks_comparison_key(networks_dict)
    output_dir = settings.EVAL_DIR + comparison_networks_key + '/' + comparison_methods_key + '/'

    method_obj = methods_utils.create_method_by_id(method_id)
    c_evs_dict = OrderedDict()
    methods_dict = OrderedDict()
    for n_key, networks_ids in networks_dict.iteritems():
        c_evs = []
        ms = []
        for n_id in networks_ids:
            m = Method.load_state(n_id, method_id)
            ms.append(m)
            c_evs.append(m.get_community_evolution())
        c_evs_dict[n_key] = c_evs
        methods_dict[n_key] = ms

    output_images_dir = output_dir + 'images/'

    # evaluate against ground truths
    if kwargs.get('draw_against_other', True):
        networks_ground_truths_ids = []
        if kwargs.get('show_ground_truth', True):
            for n_key, networks_ids in networks_dict.iteritems():
                for n_id in networks_ids:
                    dn = DynamicNetwork.load(n_id)
                    networks_ground_truths_ids.append([])
                    for c_ev_gt in (dn.get_ground_truths_list() + dn.get_ground_truths_affinity_list()):
                        networks_ground_truths_ids[-1].append(c_ev_gt.get_method_id())

        if len(networks_ground_truths_ids) > 0:
            for gt_id in utils.get_intersection(*networks_ground_truths_ids):
                if kwargs.get('draw_com_evs_sim', True):
                    com_ev_eval_across_networks.evaluate_com_ev_against_other_across_networks_dict(
                        method_id=method_id,
                        networks_dict=networks_dict,
                        other_method_id=gt_id,
                        output_dir=output_images_dir,
                        **kwargs
                    )

                if kwargs.get('draw_changepoints_pr_roc', False):
                    method_eval_pr_roc_across_networks.evaluate_method_changepoint_classification_across_networks_dict(
                        method_id=method_id,
                        networks_dict=networks_dict,
                        other_method_id=gt_id,
                        output_dir=output_images_dir,
                        **kwargs
                    )


    if kwargs.get('draw_com_evs_scores', True):
        com_ev_draw.draw_list_com_evs_scores_bars_avg(
            c_evs_dict=c_evs_dict,
            output_dir=output_images_dir,
            scoring_methods_ids=[method_obj.get_scoring_method_id()] if hasattr(method_obj, 'get_scoring_method_id') else None,
            draw_pairwise_corr_matrices=True,
            **kwargs
        )

    if kwargs.get('draw_run_time', True):
        method_draw.draw_list_methods_run_time_bars_avg(
            methods_dict=methods_dict,
            output_dir=output_images_dir,
            **kwargs
        )



# Evaluation of type 3
def evaluate_methods_across_networks_dict(methods_ids, networks_dict, **kwargs):
    """
    For a set of methods, compare them against each other
    i.e. one figure per measure (each group in figure - network key, series - methods)
    """
    logger.debug("Comparing methods {methods_ids} across networks dict '{networks_dict}'".format(
        methods_ids=", ".join(methods_ids),
        networks_dict=kwargs.get('networks_key_name') or ", ".join(networks_dict.keys()),
    ))

    comparison_methods_key = kwargs.get('methods_key_name') or eval_utils.generate_methods_comparison_key(methods_ids)
    comparison_networks_key = kwargs.get('networks_key_name') or eval_utils.generate_networks_comparison_key(networks_dict)
    output_dir = settings.EVAL_DIR + comparison_networks_key + '/' + comparison_methods_key + '/'
    output_images_dir = output_dir + 'images/'

    c_evs_dict = OrderedDict()
    networks_methods_dict = OrderedDict()
    for n_key, networks_ids in networks_dict.iteritems():
        networks_methods_dict[n_key] = OrderedDict()
        c_evs_dict[n_key] = OrderedDict()
        for m_id in methods_ids:
            c_evs = []
            ms = []
            for n_id in networks_ids:
                m = Method.load_state(n_id, m_id)
                ms.append(m)
                c_evs.append(m.get_community_evolution())
            c_evs_dict[n_key][methods_utils.get_method_title(m_id)] = c_evs
            networks_methods_dict[n_key][methods_utils.get_method_title(m_id)] = ms

    # evaluate against ground truths
    if kwargs.get('draw_against_other', True):
        networks_key_ground_truths_ids = OrderedDict()
        for n_key, networks_ids in networks_dict.iteritems():
            networks_ground_truths_ids = []
            for n_id in networks_ids:
                dn = DynamicNetwork.load(n_id)
                networks_ground_truths_ids.append([])
                for c_ev_gt in (dn.get_ground_truths_list() + dn.get_ground_truths_affinity_list()):
                    networks_ground_truths_ids[-1].append(c_ev_gt.get_method_id())
        networks_key_ground_truths_ids[n_key] = utils.get_intersection(*networks_ground_truths_ids)
        for gt_id in utils.get_intersection(*networks_key_ground_truths_ids.values()):
            if kwargs.get('draw_com_evs_sim', True):
                com_ev_eval_across_networks.evaluate_com_evs_against_other_across_networks_dict(
                    methods_ids=methods_ids,
                    networks_dict=networks_dict,
                    other_method_id=gt_id,
                    output_dir=output_images_dir,
                    networks_c_evs_dict=c_evs_dict,
                    **kwargs
                )

            if kwargs.get('draw_changepoints_pr_roc', False):
                method_eval_pr_roc_across_networks.evaluate_methods_changepoint_classification_across_networks_dict(
                    methods_ids=methods_ids,
                    networks_dict=networks_dict,
                    other_method_id=gt_id,
                    output_dir=output_images_dir,
                    **kwargs
                )

    if kwargs.get('show_ground_truth', True):
        for n_key, networks_ids in networks_dict.iteritems():
            for n_id in networks_ids:
                dn = DynamicNetwork.load(n_id)
                for c_ev_gt in (dn.get_ground_truths_list() + dn.get_ground_truths_affinity_list()):
                    gt_title = c_ev_gt.get_method_title()
                    if gt_title not in c_evs_dict[n_key]:
                        c_evs_dict[n_key][gt_title] = []
                    c_evs_dict[n_key][gt_title].append(c_ev_gt)

    if kwargs.get('draw_run_time', True):
        method_draw.draw_list_methods_networks_run_time_bars_avg(
            networks_methods_dict=networks_methods_dict,
            output_dir=output_images_dir,
            draw_ranking_slopegraph=True,
            **kwargs
        )

    if kwargs.get('draw_com_evs_scores', True):
            methods_list = [methods_utils.create_method_by_id(m_id) for m_id in methods_ids]
            com_ev_draw.draw_dict_networks_com_evs_scores_bars_avg(
                networks_c_evs_dict=c_evs_dict,
                output_dir=output_images_dir,
                scoring_methods_ids=[m.get_scoring_method_id() for m in methods_list if hasattr(m, 'get_scoring_method_id')],
                draw_ranking_slopegraph=True,
                **kwargs
            )


