from ..log import logger
from collections import OrderedDict
from ..classification_result import ClassificationResult
from ..draw import classification_result_draw
from ..methods import methods_utils
from ..community_evolution import CommunityEvolution
from ..method import Method

def evaluate_methods_changepoint_classification_across_networks(methods_ids, networks_ids, other_method_id,
                                                                output_dir, **kwargs):

    logger.debug("Evaluating changepoint classification of methods {methods_ids} across networks '{networks_ids}' "
                 "according to {gt}".format(
        methods_ids=", ".join(methods_ids),
        networks_ids=", ".join(networks_ids),
        gt=other_method_id,
    ))
    output_images_dir = output_dir + 'evaluation_against_other/' + other_method_id + '/pr_roc/changepoints/'

    methods_dict = OrderedDict()
    c_evs_gts_list = []
    for n_id in networks_ids:
        c = CommunityEvolution.load(n_id, other_method_id)
        if c is None:
            continue
        c_evs_gts_list.append(c)
        for m_id in methods_ids:
            m_title = methods_utils.get_method_title(m_id)
            if m_title not in methods_dict:
                methods_dict[m_title] = []
            m = Method.load_state(n_id, m_id)
            if not hasattr(m,'get_timepoints_scores'):
                continue
            methods_dict[m_title].append(m)

    if len(methods_dict) == 0:
        logger.warning(
            "None of the methods %s supports changepoint-like scores"
            % (", ".join(methods_ids))
        )
        return

    classification_results_dict=OrderedDict()
    for m_title, ms in methods_dict.iteritems():
        cs = []
        for i in xrange(len(ms)):
            gt_scores = [
                    (j in c_evs_gts_list[i].get_changepoints())
                    for j in c_evs_gts_list[i].get_dynamic_network().get_timepoints()
            ]
            if len(set(gt_scores[1:])) == 1:
                logger.warning(
                    "'{gt_id}' has only one timepoint label for network '{n_id}' - skipping".format(
                    gt_id=c_evs_gts_list[i].get_method_id(),
                    n_id=networks_ids[i],
                ))
                continue
            # note: first timepoint is never a changepoint and thus can be ignored
            c = ClassificationResult(
                ground_truth_labels=gt_scores[1:],
                predicted_scores=ms[i].get_timepoints_scores()[1:],
            )
            # c.save(
            #     output_dir=ms[i].get_output_dir() + 'pr_roc/changepoints/',
            #     output_file='vs-' + c_evs_gts_list[i].get_method_id(),
            # )
            cs.append(c)
        if len(cs) == 0:
            continue

        classification_results_dict[m_title] = cs

    if len(classification_results_dict) == 0:
        logger.warning("No methods for analysis - aborting classification results analysis")
        return

    for func in [
        classification_result_draw.draw_bars
    ]:
        func(
            classification_results_dict=classification_results_dict,
            output_dir=output_images_dir,
            **kwargs
        )

def evaluate_method_changepoint_classification_across_networks_dict(method_id, networks_dict, other_method_id,
                                                                    output_dir, **kwargs):

    logger.debug("Evaluating changepoint classification of method {method_id} across networks keys '{networks_dict}' "
                 "according to {gt}".format(
        method_id=method_id,
        networks_dict=kwargs.get('networks_key_name') or ", ".join(networks_dict.keys()),
        gt=other_method_id,
    ))
    output_images_dir = output_dir + 'evaluation_against_other/' + other_method_id + '/pr_roc/changepoints/'

    methods_dict = OrderedDict()
    c_evs_gts_dict = OrderedDict()
    for n_key, networks_ids in networks_dict.iteritems():
        ms_list = []
        c_evs_gts_list = []
        for n_id in networks_ids:
            c = CommunityEvolution.load(n_id, other_method_id)
            if c is None:
                continue
            if c.get_changepoints() is None:
                logger.warning("No changepoints for other method id %s and network id %s" % (other_method_id, n_id))
                continue
            m = Method.load_state(n_id, method_id)
            if not hasattr(m,'get_timepoints_scores'):
                continue
            c_evs_gts_list.append(c)
            ms_list.append(m)
        if len(ms_list) > 0:
            methods_dict[n_key] = ms_list
            c_evs_gts_dict[n_key] = c_evs_gts_list

    if len(methods_dict) == 0:
        logger.warning(
            "Method %s supports changepoint-like scores" % method_id
        )
        return

    classification_results_dict=OrderedDict()
    for n_key, ms in methods_dict.iteritems():
        cs = []
        for i in xrange(len(ms)):
            gt_scores = [
                    (j in c_evs_gts_dict[n_key][i].get_changepoints())
                    for j in c_evs_gts_dict[n_key][i].get_dynamic_network().get_timepoints()
            ]
            if len(set(gt_scores[1:])) == 1:
                logger.warning(
                    "'{gt_id}' has only one timepoint label for network '{n_id}' - skipping".format(
                    gt_id=c_evs_gts_dict[n_key][i].get_method_id(),
                    n_id=networks_dict[n_key][i],
                ))
                continue
            # note: first timepoint is never a changepoint and thus can be ignored
            c = ClassificationResult(
                ground_truth_labels=gt_scores[1:],
                predicted_scores=ms[i].get_timepoints_scores()[1:],
            )
            # c.save(
            #     output_dir=ms[i].get_output_dir() + 'pr_roc/changepoints/',
            #     output_file='vs-' + c_evs_gts_dict[n_key][i].get_method_id(),
            # )
            cs.append(c)
        if len(cs) == 0:
            continue

        classification_results_dict[n_key] = cs

    if len(classification_results_dict) == 0:
        logger.warning("No methods for analysis - aborting classification results analysis")
        return

    for func in [
        classification_result_draw.draw_bars
    ]:
        func(
            classification_results_dict=classification_results_dict,
            output_dir=output_images_dir,
            **kwargs
        )

def evaluate_methods_changepoint_classification_across_networks_dict(methods_ids, networks_dict, other_method_id,
                                                                output_dir, **kwargs):

    logger.debug("Evaluating changepoint classification of methods {methods_ids} across networks '{networks_dict}' "
                 "according to {gt}".format(
        methods_ids=", ".join(methods_ids),
        networks_dict=kwargs.get('networks_key_name') or ", ".join(networks_dict.keys()),
        gt=other_method_id,
    ))
    output_images_dir = output_dir
    if kwargs.get('add_subdirs', True):
        output_images_dir += 'evaluation_against_other/' + other_method_id + '/pr_roc/changepoints/'
    method_title_func = kwargs.get('method_title_func', methods_utils.get_method_title)

    networks_methods_dict = OrderedDict()
    c_evs_gts_list = OrderedDict()
    for n_key, networks_ids in networks_dict.iteritems():
        networks_methods_dict[n_key] = OrderedDict()
        c_evs_gts_list[n_key] = []
        for n_id in networks_ids:
            c = CommunityEvolution.load(n_id, other_method_id)
            if c is None:
                continue
            c_evs_gts_list[n_key].append(c)
            for m_id in methods_ids:
                m_title = method_title_func(m_id)
                if m_title not in networks_methods_dict[n_key]:
                    networks_methods_dict[n_key][m_title] = []
                m = Method.load_state(n_id, m_id)
                networks_methods_dict[n_key][m_title].append(m)

    if len(networks_methods_dict) == 0:
        logger.warning(
            "None of the methods %s supports changepoint-like scores"
            % (", ".join(methods_ids))
        )
        return

    classification_results_dict=OrderedDict()
    for n_key, methods_dict in networks_methods_dict.iteritems():
        classification_results_dict[n_key] = OrderedDict()
        for m_title, ms in methods_dict.iteritems():
            cs = []
            for i in xrange(len(ms)):
                gt_scores = [
                        (j in c_evs_gts_list[n_key][i].get_changepoints())
                        for j in c_evs_gts_list[n_key][i].get_dynamic_network().get_timepoints()
                ]
                if len(set(gt_scores[1:])) == 1:
                    logger.warning(
                        "'{gt_id}' has only one timepoint label for network '{n_id}' - skipping".format(
                        gt_id=c_evs_gts_list[n_key][i].get_method_id(),
                        n_id= networks_dict[n_key][i],
                    ))
                    continue
                if ms[i] is None or not hasattr(ms[i],'get_timepoints_scores'):
                    c = None
                else:
                    # note: first timepoint is never a changepoint and thus can be ignored
                    c = ClassificationResult(
                        ground_truth_labels=gt_scores[1:],
                        predicted_scores=ms[i].get_timepoints_scores()[1:],
                    )
                # c.save(
                #     output_dir=ms[i].get_output_dir() + 'pr_roc/changepoints/',
                #     output_file='vs-' + c_evs_gts_list[n_key][i].get_method_id(),
                # )
                cs.append(c)
            if len(cs) == 0:
                continue

            classification_results_dict[n_key][m_title] = cs
        if len(classification_results_dict[n_key]) == 0:
            del classification_results_dict[n_key]

    if len(classification_results_dict) == 0:
        logger.warning("No methods for analysis - aborting classification results analysis")
        return

    x_label = kwargs.get('x_label', "Networks")
    kwargs.pop('x_label')
    for func in [
        classification_result_draw.draw_bars_dict
    ]:
        func(
            classification_results_dict=classification_results_dict,
            output_dir=output_images_dir,
            x_label=x_label,
            **kwargs
        )