from ..log import logger
from ..dynamic_network import DynamicNetwork
from ..method import Method
from ..scoring_methods import scoring_methods_utils
import numpy as np
from collections import defaultdict
import cPickle
import os

def evalute_methods_score_rankings(methods_dict, networks_ids, scoring_method_ids, precision_levels=None, m_func=None, output_dir=None):
    """
    Compute rankings across all network instances
    """
    logger.debug("Comparing methods rankings according to %s scoring methods across %s networks"
                 % (", ".join(scoring_method_ids), len(networks_ids))
    )
    # for precision  level p, we assume to scores are equal if they are within fraction p of each other
    precision_levels = precision_levels or [0]
    scoring_methods = [scoring_methods_utils.create_scoring_method_by_id(sc_m_id) for sc_m_id in scoring_method_ids]

    m_keys = methods_dict.keys()
    res = {}
    for sc_m_id in scoring_method_ids:
        res[sc_m_id] = {}
        for p_l in precision_levels:
            res[sc_m_id][p_l] = {}
            for m_key in m_keys:
                res[sc_m_id][p_l][m_key] = defaultdict(int)

    for i_n,n_id in enumerate(networks_ids):
        logger.debug("{i_n}/{n_n}. Network '{n_id}'.".format(
                        i_n=i_n+1,
                        n_n=len(networks_ids),
                        n_id=n_id,
        ))
        dn = DynamicNetwork.load(n_id)
        m_res = {}
        for sc_m_id in scoring_method_ids:
            m_res[sc_m_id] = {}
        for m_key in m_keys:
            if m_func is not None:
                m_ids = m_func(m_key, dn)
            else:
                m_ids = methods_dict[m_key]
            if m_ids is None or len(m_ids)==0:
                for sc_m_id in scoring_method_ids:
                    m_res[sc_m_id][m_key] = np.NINF
            else:
                m_scores = {}
                for sc_m_id in scoring_method_ids:
                    m_scores[sc_m_id] = []

                for m_id in m_ids:
                    m = Method.load_state(n_id, m_id)
                    for sc_m in scoring_methods:
                        m_scores[sc_m.get_id()].append(
                            sc_m.evaluate_method(m,dn)
                        )
                for sc_m_id in scoring_method_ids:
                    m_res[sc_m_id][m_key] = np.max(m_scores[sc_m_id])

        for sc_m_id in scoring_method_ids:
            logger.debug("{sc_m_id} scores: {scores}".format(
                sc_m_id=sc_m_id,
                scores=", ".join(["%s - %s" % (m_res_pair[0], m_res_pair[1]) for m_res_pair in m_res[sc_m_id].iteritems()])
            ))

        for sc_m_id in scoring_method_ids:
            v = [m_res[sc_m_id][m_key] for m_key in m_keys]
            for p_l in precision_levels:
                v_ranks = [1 + sum(1 for y in v if y > x * (1 + p_l)) for x in v]
                for i_m,m_key in enumerate(m_keys):
                    pos = v_ranks[i_m] if m_res[sc_m_id][m_key] != np.NINF else 'n/a'
                    res[sc_m_id][p_l][m_key][pos] += 1

        logger.debug("Cumulative rankings: %s" % res)
    if output_dir is not None:
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)
        for sc_m_id in scoring_method_ids:
            for p_l in precision_levels:
                with open(output_dir + '%s_%s.cpickle' % (sc_m_id, p_l), 'wb') as out_f:
                    cPickle.dump(
                        res[sc_m_id][p_l],
                        out_f
                    )
                with open(output_dir + '%s_%s.txt' % (sc_m_id, p_l), 'wb') as out_f:
                    for m_key in m_keys:
                        out_f.write(
                            m_key + ", " + ", ".join(["%s - %s" % (m_res_pair[0], m_res_pair[1]) for m_res_pair in res[sc_m_id][p_l][m_key].iteritems()]) + "\n"
                        )

    for sc_m_id in scoring_method_ids:
        for p_l in precision_levels:
            logger.debug(
                "{sc_m_id}, precision level {p_l:.2f}%% ".format(
                    sc_m_id=sc_m_id,
                    p_l=p_l * 100,
                )
            )
            for m_key in m_keys:
                logger.debug(
                    m_key + ": " + ", ".join(["%s - %s" % (m_res_pair[0], m_res_pair[1]) for m_res_pair in res[sc_m_id][p_l][m_key].iteritems()])
                )
    return res

def evalute_methods_score_rankings_per_key(methods_dict, networks_dict,
                                           scoring_method_ids, precision_levels=None, m_func=None, output_dir=None
                                           ):
    """
    Compute rankings across all network keys (i.e., configurations)
    """
    logger.debug("Comparing methods rankings according to %s scoring methods across %s configurations"
                 % (", ".join(scoring_method_ids), len(networks_dict))
    )
    # for precision  level p, we assume to scores are equal if they are within fraction p of each other
    precision_levels = precision_levels or [0]
    scoring_methods = [scoring_methods_utils.create_scoring_method_by_id(sc_m_id) for sc_m_id in scoring_method_ids]

    m_keys = methods_dict.keys()
    res = {}
    for sc_m_id in scoring_method_ids:
        res[sc_m_id] = {}
        for p_l in precision_levels:
            res[sc_m_id][p_l] = {}
            for m_key in m_keys:
                res[sc_m_id][p_l][m_key] = defaultdict(int)

    i_nk = 1
    for n_key, n_ids in networks_dict.iteritems():
        logger.debug("{i_nk}/{n_nk}. Network key '{n_key}'.".format(
            i_nk=i_nk,
            n_nk=len(networks_dict),
            n_key=n_key,
        ))
        i_nk += 1

        m_scores_all = {}
        for sc_m_id in scoring_method_ids:
            m_scores_all[sc_m_id] = {}

        for n_id in n_ids:
            dn = DynamicNetwork.load(n_id)
            for m_key in m_keys:
                if m_func is not None:
                    m_ids = m_func(m_key, dn)
                else:
                    m_ids = methods_dict[m_key]
                if m_ids is None or len(m_ids) == 0:
                    for sc_m_id in scoring_method_ids:
                        m_scores_all[sc_m_id][m_key] = {}
                else:
                    for sc_m in scoring_methods:
                        sc_m_id = sc_m.get_id()
                        if m_key not in m_scores_all[sc_m_id]:
                            m_scores_all[sc_m_id][m_key] = defaultdict(list)
                        for m_id in m_ids:
                            m = Method.load_state(n_id, m_id)
                            m_scores_all[sc_m_id][m_key][m_id].append(
                                sc_m.evaluate_method(m, dn) if m is not None else None
                            )

        # for each method key, keep the best overall scores
        m_res = {} # for each m_key, the highest average score among all the corresponding m_ids
        m_best_ids = {}
        for sc_m_id in scoring_method_ids:
            m_res[sc_m_id] = {}
            m_best_ids[sc_m_id] = {}
            for m_key in m_keys:
                best_avg = None
                best_m_id = None
                for m_id, temp_values in m_scores_all[sc_m_id][m_key].iteritems():
                    if None in temp_values:
                        logger.warning('Method key %s - method id %s, found None, skipping' % (m_key, m_id))
                        continue
                    temp_avg = np.average(temp_values)
                    if best_avg is None or temp_avg > best_avg:
                        best_avg = temp_avg
                        best_m_id = m_id
                m_res[sc_m_id][m_key] = best_avg or -np.inf
                m_best_ids[sc_m_id][m_key] = best_m_id


        for sc_m_id in scoring_method_ids:
            logger.debug("{sc_m_id} scores: {scores}".format(
                sc_m_id=sc_m_id,
                scores=", ".join(["%s - %s" % (m_res_pair[0], m_res_pair[1]) for m_res_pair in m_res[sc_m_id].iteritems()])
            ))

        # compute rankings
        for sc_m_id in scoring_method_ids:
            v = [m_res[sc_m_id][m_key] for m_key in m_keys]
            for p_l in precision_levels:
                v_ranks = [1 + sum(1 for y in v if y > x * (1 + p_l)) for x in v]
                for i_m,m_key in enumerate(m_keys):
                    pos = v_ranks[i_m] if m_res[sc_m_id][m_key] != np.NINF else 'n/a'
                    res[sc_m_id][p_l][m_key][pos] += 1

        logger.debug("Cumulative rankings: %s" % res)

    if output_dir is not None:
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)
        for sc_m_id in scoring_method_ids:
            for p_l in precision_levels:
                with open(output_dir + '%s_%s.cpickle' % (sc_m_id, p_l), 'wb') as out_f:
                    cPickle.dump(
                        res[sc_m_id][p_l],
                        out_f
                    )
                with open(output_dir + '%s_%s.txt' % (sc_m_id, p_l), 'wb') as out_f:
                    for m_key in m_keys:
                        out_f.write(
                            m_key + ", " + ", ".join(["%s - %s" % (m_res_pair[0], m_res_pair[1]) for m_res_pair in res[sc_m_id][p_l][m_key].iteritems()]) + "\n"
                        )

    for sc_m_id in scoring_method_ids:
        for p_l in precision_levels:
            logger.debug(
                "{sc_m_id}, precision level {p_l:.2f}%% ".format(
                    sc_m_id=sc_m_id,
                    p_l=p_l * 100,
                )
            )
            for m_key in m_keys:
                logger.debug(
                    m_key + ": " + ", ".join(["%s - %s" % (m_res_pair[0], m_res_pair[1]) for m_res_pair in res[sc_m_id][p_l][m_key].iteritems()])
                )
    return res


def evalute_methods_score_rankings_n_changepoints(methods_dict, networks_ids, scoring_method_ids, n_timepoints, precision_levels=None, m_func=None, output_dir=None):
    logger.debug("Comparing methods rankings with solutions having different numbers of changepoints according to %s scoring methods across %s networks"
                 % (", ".join(scoring_method_ids), len(networks_ids))
    )
    # for precision  level p, we assume to scores are equal if they are within fraction p of each other
    precision_levels = precision_levels or [0]
    scoring_methods = [scoring_methods_utils.create_scoring_method_by_id(sc_m_id) for sc_m_id in scoring_method_ids]

    changepoints_values = range(n_timepoints)
    m_keys = methods_dict.keys()
    res = {}
    for sc_m_id in scoring_method_ids:
        res[sc_m_id] = {}
        for p_l in precision_levels:
            res[sc_m_id][p_l] = {}
            for n_changepoints in changepoints_values:
                res[sc_m_id][p_l][n_changepoints] = {}
                for m_key in m_keys:
                    res[sc_m_id][p_l][n_changepoints][m_key] = defaultdict(int)

    for i_n,n_id in enumerate(networks_ids):
        logger.debug("{i_n}/{n_n}. Network '{n_id}'.".format(
                        i_n=i_n+1,
                        n_n=len(networks_ids),
                        n_id=n_id,
        ))
        dn = DynamicNetwork.load(n_id)
        m_res = {}
        for sc_m_id in scoring_method_ids:
            m_res[sc_m_id] = {}
            for n_changepoints in changepoints_values:
                m_res[sc_m_id][n_changepoints] = {}
        for m_key in m_keys:
            if m_func is not None:
                m_ids = m_func(m_key, dn)
            else:
                m_ids = methods_dict[m_key]
            if m_ids is None or len(m_ids)==0:
                for sc_m_id in scoring_method_ids:
                    for n_changepoints in changepoints_values:
                        m_res[sc_m_id][n_changepoints][m_key] = np.NINF
            else:
                m_scores = {}
                for n_changepoints in changepoints_values:
                    m_scores[n_changepoints] = {}
                    for sc_m_id in scoring_method_ids:
                        m_scores[n_changepoints][sc_m_id] = []
                    for m_id in m_ids:
                        m = Method.load_state(n_id, m_id)
                        for sc_m in scoring_methods:
                                m_scores[n_changepoints][sc_m.get_id()].append(
                                    sc_m.evaluate_community_evolution(
                                        m.get_solution_for_n_changepoints(n_changepoints),
                                        dn
                                    )
                                )

                for n_changepoints in changepoints_values:
                    for sc_m_id in scoring_method_ids:
                        m_res[sc_m_id][n_changepoints][m_key] = np.max(m_scores[n_changepoints][sc_m_id])

        for sc_m_id in scoring_method_ids:
            for n_changepoints in changepoints_values:
                v = [m_res[sc_m_id][n_changepoints][m_key] for m_key in m_keys]
                for p_l in precision_levels:
                    v_ranks = [1 + sum(1 for y in v if y > x * (1 + p_l)) for x in v]
                    for i_m,m_key in enumerate(m_keys):
                        pos = v_ranks[i_m] if m_res[sc_m_id][n_changepoints][m_key] != np.NINF else 'n/a'
                        res[sc_m_id][p_l][n_changepoints][m_key][pos] += 1

        logger.debug("Cumulative rankings: %s" % res)
    if output_dir is not None:
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)
        for sc_m_id in scoring_method_ids:
            for p_l in precision_levels:
                with open(output_dir + 'n_changepoints_%s_%s.cpickle' % (sc_m_id, p_l), 'wb') as out_f:
                    cPickle.dump(
                        res[sc_m_id][p_l],
                        out_f
                    )

    return res