from ..log import logger
from ..dynamic_network import DynamicNetwork
from ..method import Method
from ..scoring_methods import scoring_methods_utils
from scipy.stats import ttest_rel, wilcoxon
import numpy as np
from collections import defaultdict, OrderedDict
import cPickle
import os

def evalute_methods_score_p_vals(methods_dict, networks_dict, scoring_method_ids,m_func=None, output_dir=None):
    logger.debug("Comparing methods rankings according to %s scoring methods across %s network keys"
                 % (", ".join(scoring_method_ids), len(networks_dict))
    )
    scoring_methods = [scoring_methods_utils.create_scoring_method_by_id(sc_m_id) for sc_m_id in scoring_method_ids]


    m_keys = methods_dict.keys()
    if len(m_keys) != 2:
        raise Exception("Need exactly 2 method keys!")

    m_key_a = m_keys[0]
    m_key_b = m_keys[1]
    logger.info("Comparing %s and %s" % (m_key_a, m_key_b))

    res = {}
    for sc_m_id in scoring_method_ids:
        res[sc_m_id] = OrderedDict()

    i_nk = 1
    for n_key, n_ids in networks_dict.iteritems():
        logger.debug("{i_nk}/{n_nk}. Network key '{n_key}'.".format(
                        i_nk=i_nk,
                        n_nk=len(networks_dict),
                        n_key=n_key,
        ))
        i_nk += 1
        m_res = {}
        for sc_m_id in scoring_method_ids:
            m_res[sc_m_id] = {}

        m_scores_all = {}
        for sc_m_id in scoring_method_ids:
            m_scores_all[sc_m_id] = {}

        for n_id in n_ids:
            dn = DynamicNetwork.load(n_id)
            for m_key in m_keys:
                if m_func is not None:
                    m_ids = m_func(m_key, dn)
                else:
                    m_ids = methods_dict[m_key]
                if m_ids is None or len(m_ids)==0:
                    for sc_m_id in scoring_method_ids:
                        m_scores_all[sc_m_id][m_key] = None
                else:
                    for sc_m in scoring_methods:
                        sc_m_id = sc_m.get_id()
                        if m_key not in m_scores_all[sc_m_id]:
                            m_scores_all[sc_m_id][m_key] = defaultdict(list)
                        for m_id in m_ids:
                            m = Method.load_state(n_id, m_id)
                            m_scores_all[sc_m_id][m_key][m_id].append(
                                sc_m.evaluate_method(m,dn) if m is not None else None
                            )

        # for each method key, keep the best overall scores
        m_scores = {}
        m_best_ids = {}
        for sc_m_id in scoring_method_ids:
            m_scores[sc_m_id] = {}
            m_best_ids[sc_m_id] = {}
            for m_key in m_keys:
                best_avg = None
                best_values = None
                best_m_id = None
                for m_id,temp_values in m_scores_all[sc_m_id][m_key].iteritems():
                    if None in temp_values:
                        logger.warning('Method key %s - method id %s, found None, skipping' % (m_key, m_id))
                        continue
                    temp_avg = np.average(temp_values)
                    if best_avg is None or temp_avg > best_avg:
                        best_avg = temp_avg
                        best_values = temp_values
                        best_m_id = m_id
                m_scores[sc_m_id][m_key] = best_values
                m_best_ids[sc_m_id][m_key] = best_m_id

        for sc_m_id in scoring_method_ids:
            logger.debug(sc_m_id)
            t_stat, t_p_val = ttest_rel(m_scores[sc_m_id][m_key_a], m_scores[sc_m_id][m_key_b])
            logger.info("Averages: %s - %s, %s - %s" % (m_key_a, np.average(m_scores[sc_m_id][m_key_a]), m_key_b, np.average(m_scores[sc_m_id][m_key_b])))
            logger.info("Paired t-test: statistic - %s, p-value - %s" % (t_stat, t_p_val))

            w_stat, w_p_val = wilcoxon(m_scores[sc_m_id][m_key_a], m_scores[sc_m_id][m_key_b])
            logger.info("Averages: %s - %s, %s - %s" % (
            m_key_a, np.average(m_scores[sc_m_id][m_key_a]), m_key_b, np.average(m_scores[sc_m_id][m_key_b])))
            logger.info("Wilcoxon signed-rank test: statistic - %s, p-value - %s" % (w_stat, w_p_val))

            res[sc_m_id][n_key] = {
                't-statistic': t_stat,
                't-test p-value': t_p_val,
                'wsrt-statistic': w_stat,
                'wsrt-test p-value': w_p_val,
                'avg score-%s' % m_key_a: np.average(m_scores[sc_m_id][m_key_a]),
                'best id-%s' % m_key_a: m_best_ids[sc_m_id][m_key_a],
                'avg score-%s' % m_key_b: np.average(m_scores[sc_m_id][m_key_b]),
                'best id-%s' % m_key_b: m_best_ids[sc_m_id][m_key_b],
            }

    if output_dir is not None:
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)
        for sc_m_id in scoring_method_ids:
            logger.debug(sc_m_id)
            with open(output_dir + '%s.cpickle' % sc_m_id, 'wb') as out_f:
                cPickle.dump(res[sc_m_id],out_f)
            with open(output_dir + '%s.csv' % sc_m_id, 'wb') as out_f:
                out_f.write("Configuration,t-test,t-test p-value, WSRT test, WSRT p-value,avg score-%s,best id-%s,avg score-%s,best id-%s\n" % (m_key_a, m_key_a, m_key_b, m_key_b))
                for n_key in res[sc_m_id]:
                    out_f.write("%s,%s,%s,%s,%s,%s,%s,%s,%s\n" % (
                        n_key,
                        res[sc_m_id][n_key]['t-statistic'],
                        res[sc_m_id][n_key]['t-test p-value'],
                        res[sc_m_id][n_key]['wsrt-statistic'],
                        res[sc_m_id][n_key]['wsrt-test p-value'],
                        res[sc_m_id][n_key]['avg score-%s' % m_key_a],
                        res[sc_m_id][n_key]['best id-%s' % m_key_a],
                        res[sc_m_id][n_key]['avg score-%s' % m_key_b],
                        res[sc_m_id][n_key]['best id-%s' % m_key_b],
                    ))
                    logger.debug("%s: %s" % (n_key, res[sc_m_id][n_key]))
    return res
