import math, os, subprocess, sys
import settings
import utils
import hashlib
from log import logger

class LatexBuilder(object):
    """
    A simple class for automatically generating  and compiling LaTeX reports in python
    """

    SUBFIGURE_RELATIVE_WIDTH = 0.94

    def __init__(self, report_name, title,
                 author=settings.REPORT_AUTHOR,
                 latex_template_file=settings.LATEX_TEMPLATE_FILE,
                 makefile_template_file=settings.MAKEFILE_TEMPLATE_FILE,
                 graphicspath = settings.NETWORKS_BASE_DIR):
        """

        :param author: Author of the report
        :param title: Title of the report
        :param latex_template_file: Latex template file to use
        :param makefile_template_file: makefile template file to use
        :param graphicspath: Path to figures
        :return:
        """
        self._report_name = report_name
        self._author = author
        self._title = title

        self._latex_template_file = latex_template_file
        self._makefile_template_file = makefile_template_file
        self._graphicspath = graphicspath

        self.pages = []

    def get_name(self):
        return self._report_name

    def add_page(self, text):
        """
        Add page to the report
        :param text:
        :return:
        """
        self.pages.append(text)

    def add_text(self, text):
        self.add_page("\\begin{lstlisting}[breaklines=true]\n" + text + "\n\end{lstlisting}")

    def add_section(self, section_name, level = 0, clearPage = False):
        """
        Add new section to the report
        :param section_name: Name of the section
        :param level: Level of the section
        :param clearPage: Whether to start section from new page
        :return:
        """
        self.add_page(
            "%s\%ssection{%s}" % (
                '\clearpage' if clearPage else '', 'sub' * level, self._escape_special_chars(section_name)
            )
        )

    def add_group_figure(self, title, images_list=None, columns_count=None,
                         images_captions_list=None, rows_images_list=None, rows_captions_list=None,
                         columns_captions_list=None,
                         image_label_suffix='', caption_suffix='',
                         image_angle=0, max_images_per_figure=None, **kwargs):
        """
        Add group figure
        :param title: Title of the main figure
        :param images_list: List of subfigures
        :param columns_count: Number of columns in the figure
        :param images_captions_list: List of captions for subfigures
        :param rows_images_list:
        :param rows_captions_list: List of captions for rows
        :param image_label_suffix: Suffix for image labels
        :param caption_suffix: Suffix for figure caption
        :param image_angle: Rotate images
        :param max_images_per_figure: If number of subfigures exceeds this number, split the figure into several parts
        :return:
        """
        if len(images_list) == 0:
            logger.warning("No images provided for %s image" % title)
            return

        if max_images_per_figure is None and columns_count is not None and kwargs.get('max_rows_per_figure'):
            max_images_per_figure = columns_count * kwargs.get('max_rows_per_figure')

        if max_images_per_figure is None or len(images_list)<max_images_per_figure:
            self.add_page(
                self._get_group_figure_latex_string(
                    title, images_list, columns_count, images_captions_list, rows_images_list, rows_captions_list,
                    columns_captions_list, image_label_suffix, caption_suffix, image_angle
                )
            )
        else: #  split into several figures
            remaining_figures = len(images_list)
            i_start = 0
            num_of_figures = int(math.ceil(remaining_figures / float(max_images_per_figure)))
            n_fig_images = min(remaining_figures,max_images_per_figure)
            row_i_start = 0
            while remaining_figures>0:
                self.add_page(
                    self._get_group_figure_latex_string(
                        title=title
                            + (' - Part %s/%s' % (int(i_start / max_images_per_figure)+1,num_of_figures)
                            if num_of_figures > 1 else ''),
                        images_list=images_list[i_start:(i_start + n_fig_images)],
                        columns_count=columns_count,
                        images_captions_list=images_captions_list[i_start:(i_start + n_fig_images)]
                        if images_captions_list is not None else None,
                        rows_images_list=rows_images_list[row_i_start:] if rows_images_list is not None else None,
                        rows_captions_list=rows_captions_list[row_i_start:] if rows_captions_list is not None else None,
                        columns_captions_list=columns_captions_list,
                        image_label_suffix=image_label_suffix,
                        caption_suffix=caption_suffix,
                        image_angle=image_angle,
                    )
                )

                i_start += n_fig_images
                remaining_figures -= n_fig_images
                row_i_start += n_fig_images / columns_count


    def save(self, output_file, create_makefile = False, makefile_pdf_quality = 'prepress'):
        """
        Save text file
        :param output_file:
        :param create_makefile:
        :param makefile_pdf_quality:
        :return:
        """

        if not output_file.find('.tex') == len(output_file) - 4:
            output_file += '.tex'

        if not os.path.exists(os.path.dirname(output_file)):
            os.makedirs(os.path.dirname(output_file))

        if isinstance(self._graphicspath, basestring):
            self._graphicspath = [self._graphicspath]

        template_values = {
            'author': self._author,
            'title': LatexBuilder._escape_special_chars(self._title),
            'graphicspath': ''.join(['{%s}' % path for  path in self._graphicspath]),
            'body': '\n\n'.join(self.pages)
        }

        template = file(self._latex_template_file, 'r').read()

        open(output_file, mode='wb').write(template % template_values)


        if create_makefile:
            template_values =  {
                'tex_filename': os.path.splitext(os.path.basename(output_file))[0],
                'pdf_quality': makefile_pdf_quality
            }

            template_makefile = file(self._makefile_template_file, 'r').read()

            open(os.path.dirname(output_file) + '/makefile', mode='w').write(
                template_makefile % template_values
            )

    @staticmethod
    def _escape_special_chars(s):
        return s.replace('_','{\_}').replace('#','\#').replace('^','\\textasciicircum{}')

    def _get_group_figure_latex_string(self, title, images_list=None, columns_count=None,
                                       images_captions_list=None, rows_images_list=None,
                                       rows_captions_list=None, columns_captions_list=None, image_label_suffix='',
                                       caption_suffix='', image_angle=0):
        """
        Get latex string for a figure containing sub figures
        :param images_list: List of all images to be included
        :param title: Title of the figure
        :param columns_count: Number of columns
        :param images_captions_list:
        :param rows_captions_list: List of all images to be included groupped by row
        :param image_label_suffix:
        :return:
        """

        title = LatexBuilder._escape_special_chars(title)
        if images_captions_list is not None:
            images_captions_list = map(LatexBuilder._escape_special_chars, images_captions_list)
        if rows_captions_list is not None:
            rows_captions_list = map(LatexBuilder._escape_special_chars, rows_captions_list)
        if columns_captions_list is not None:
            columns_captions_list = map(LatexBuilder._escape_special_chars, columns_captions_list)

        latex_string = '''
\\begin{figure}[H]
\centering\n'''

        #Either all images in one list or images grouped by row
        if images_list is None and rows_images_list is None:
            return ''

        if images_captions_list is not None and len(images_list) != len(images_captions_list):
            raise Exception(
                "There should be the same number of plots and captions: %s != %s",
                len(images_list),
                len(images_captions_list)
            )

        #If grouped by row, determine column count (max row length)
        if columns_count is None:
            if images_list is None:
                columns_count = 0
                for image_list_for_the_row in rows_images_list:
                    if len(image_list_for_the_row)>columns_count:
                        columns_count = len(image_list_for_the_row)
            else:
                columns_count = int(math.ceil(math.sqrt(len(images_list))))

        columns_count = min(columns_count, len(images_list))

        # if column_count > 9:
        #     image_width = str(0.75/(column_count))[:5]
        # elif column_count > 4:
        #     image_width = str(0.85/(column_count))[:5]
        # else:
        #     image_width = str(0.65/(column_count))[:5]

        image_width = str(LatexBuilder.SUBFIGURE_RELATIVE_WIDTH/(columns_count))[:5]
        latex_string += '\\begin{tabular}{%s}\n' % ('c' * columns_count)
        caption = '\caption{\\textbf{%s%s} '\
                  % (title, "." if (images_captions_list or rows_captions_list or columns_captions_list) else "")

        #If all images in one list, group them by row based on column count
        if rows_images_list is None:
            rows_images_list = []
            temp_list = []
            i=0
            for image in images_list:
                i+=1
                temp_list.append(image)
                if i%columns_count == 0:
                    rows_images_list.append(temp_list)
                    temp_list = []

            if len(temp_list)>0:
                rows_images_list.append(temp_list)

        if columns_captions_list is not None:
            for i_c in xrange(columns_count):
                caption += " Column {i_c} - {caption},".format(i_c=i_c + 1, caption=columns_captions_list[i_c])
            caption = caption[:-1] + '.'

        first_in_the_row_image_name = None #used for references in row captions
        i=0 #used for row captions
        k=0 #used for image captions
        for i_r, row in enumerate(rows_images_list):
            j=0 #used to determine the first image in the row
            for image_file in row:
                head, tail = os.path.split(image_file)
                fileName, fileExtension = os.path.splitext(tail)
                image_name = '%s_%s_%s_%s'\
                             % (fileName, k, j, utils.hash_string(head, size=5)) # image name is used for labels

                if j==0: #first image in the row
                    first_in_the_row_image_name = image_name
                j+=1

                latex_string += '\subfloat[]{\label{fig:%s}\includegraphics[width=%s\linewidth, angle=%s]{%s}} &\n' % (
                    image_name + image_label_suffix,
                    image_width,
                    image_angle,
                    image_file
                )

                #if needed, add image caption
                if images_captions_list is not None:
                    caption += ' \protect\subref{fig:%s} -  ' % (image_name + image_label_suffix)  + images_captions_list[k] + ','
                k+=1

            latex_string = latex_string[:-3] + ' \\'+'\\' + '\n'

            #if needed, add row caption
            if rows_captions_list is not None:
                caption += ' Row %s (\protect\subref{fig:%s}-\protect\subref{fig:%s}) -  '\
                           % (i_r + 1, first_in_the_row_image_name + image_label_suffix, image_name + image_label_suffix)\
                           + rows_captions_list[i] + ','
            i+=1

        caption = caption[:-1] + '.' + caption_suffix + '}'

        latex_string += '\end{tabular}%s\end{figure}\n' % caption

        return latex_string


    def _get_latex_figure_string(self, image_name, caption):
        """
        Get figure containing one image
        """
        return '''
\\begin{figure}[H]
\centering
\includegraphics[width=0.9\\textwidth]{''' + image_name + '''}
\caption{''' +  caption + '''}
\end{figure}
'''


    def get_compiled_dir(self):
        return self._compiled_dir

    def compile(self, output_directory, pdflatex = False):
        """
        Compile document
        :param output_directory: Output directory
        :param pdflatex: Use pdflatex
        :return:
        """
        if not os.path.exists(output_directory):
            os.makedirs(output_directory)

        output_report_dir = output_directory + self._report_name + '/'
        output_filename_prefix = output_report_dir + self._report_name
        self.save(output_filename_prefix + '.tex',
                  create_makefile=True,
                  makefile_pdf_quality='screen')

        logger.debug("Compiling report %s in directory %s" % (self._report_name, output_report_dir))
        if utils.is_remote_server():
            run_command = ['make','pdflatex' if pdflatex else 'pdf']
        else:
            run_command = [
                'pdflatex' if pdflatex else 'latex',
                '-synctex=1', '-interaction=nonstopmode',
                self._report_name + '.tex'
            ]
        # when not using make, run twice to create table of contents
        repeat_count = 1
        if not utils.is_remote_server():
            logger.debug("Compiling document twice to create table of contents")
            repeat_count = 2

        for _ in xrange(repeat_count):
            p = subprocess.Popen(
                run_command,
                stdout=subprocess.PIPE,
                cwd = output_directory + self._report_name
            )
            for line in iter(p.stdout.readline, ''):
                sys.stdout.write(line)
            p.stdout.close()

        self._compiled_dir = output_report_dir