import logging
import multiprocessing
import os
import sys
import time
import timeit
import traceback

import settings
import utils
from other.colorlog import ColoredFormatter


class MsgCounterHandler(logging.Handler):
    level2count = None

    def __init__(self, *args, **kwargs):
        super(MsgCounterHandler, self).__init__(*args, **kwargs)
        self.level2count = {}

    def emit(self, record):
        l = record.levelname
        if (l not in self.level2count):
            self.level2count[l] = 0
        self.level2count[l] += 1

class CustomFileHandler(logging.FileHandler):
    finish_status = 'finished'

    def __init__(self, filename, mode='a', encoding=None, delay=0):
        filename += '-temp.log'
        super(CustomFileHandler, self).__init__(filename, mode, encoding, delay)

    def close(self):
        super(CustomFileHandler, self).close()
        new_baseFilename = self.baseFilename.replace('temp.log', self.finish_status.lower() + '.log')
        os.rename(
            self.baseFilename,
            new_baseFilename
        )
        self.baseFilename = new_baseFilename
        self.mode = 'a'

def setup_logger():
    """Return a logger with a default ColoredFormatter."""
    logger = multiprocessing.get_logger()
    first_load = not hasattr(logger, '_configured')
    if first_load:
        logger._configured = True

        # count number of log records
        logger.addHandler(MsgCounterHandler())

        # log to console
        formatter = ColoredFormatter(
            "%(log_color)s%(levelname)-5.5s [%(processName)-12.12s] [%(asctime)s] %(reset)s %(message)s",
            datefmt="%a, %d %b %Y %H:%M:%S",
            reset=True,
            log_colors={
                'DEBUG':    'cyan',
                'INFO':     'green',
                'WARNING':  'yellow',
                'ERROR':    'red',
                'CRITICAL': 'red',
            }
        )
        handler = logging.StreamHandler()
        handler.setFormatter(formatter)
        logger.addHandler(handler)
        logger.setLevel(logging.DEBUG)

        # also save logs to file
        # note that processes still share the same file
        if utils.is_remote_server():
            logFormatter = logging.Formatter("%(asctime)s [%(processName)-12.12s] [%(levelname)-5.5s]  %(message)s")
            fileHandler = CustomFileHandler(
                "{0}/{1}-{2}-{3}".format(
                    settings.LOGS_DIR,
                    time.strftime('%Y-%m-%d-%H-%M-%S'),
                    os.path.basename(sys.argv[0]).replace('.py',''),
                    multiprocessing.current_process().name.lower(),
                ),
                mode='wb'
            )
            fileHandler.setFormatter(logFormatter)
            fileHandler.setLevel(logging.DEBUG)
            logger.addHandler(fileHandler)

            logger.info("Running command: %s" % ' '.join(sys.argv))

    return logger

def log_func_progress(func):
    def func_wrapper(*args, **kwargs):
        start = timeit.default_timer()
        success = False

        try:
           func(*args, **kwargs)
           success = True
        except Exception as ex:
            logger.error(traceback.format_exc())

        logger.info(
            "Number of log records: %s" % ", ".join(
                "%s - %s" % (k,v) for k,v in logger.handlers[0].level2count.iteritems()
            )
        )

        message = "SUCCESS" if success else "ERROR"
        end = timeit.default_timer()
        run_time =  utils.sec_to_time(end - start)
        logger.info("{message} ({run_time})".format(message=message, run_time=run_time))
        # rename log file according to the result
        if len(logger.handlers) > 2:
            logger.handlers[2].finish_status = message

    return func_wrapper

def close_filehandler():
   if len(logger.handlers) > 2:
       logger.handlers[2].close()

logger = setup_logger()