import os
import settings
import cPickle
from log import logger

class Method(object):
    def __init__(self, method_id):
        self._method_id = method_id
        self._clean_run = True
        self._network_id = None
        self._dynamic_network = None

    def get_id(self):
        raise NotImplementedError()
    @staticmethod
    def get_id_prefix():
        raise NotImplementedError()

    @staticmethod
    def is_ground_truth_id(network_id, method_id, sub_type=None):
        """
        Check whether method id corresponds to ground truth
        :param method_id:
        :return:
        """
        return os.path.exists(
            settings.NETWORKS_BASE_DIR + network_id + '/ground_truth/%s' % ('' if sub_type is None else ('_%s/' % sub_type)) + method_id + '/'
        )

    def run(self, dynamic_networks, **kwargs):
        raise NotImplementedError()

    @staticmethod
    def get_method_title(method_id):
        raise NotImplementedError()

    def get_title(self):
        return self.__class__.get_method_title(self.get_id())

    def get_community_evolution(self):
        """
        Get final solution
        :return: CommunityEvolution
        """
        raise NotImplementedError()


    def get_dynamic_network(self):
        if self._dynamic_network is None:
            from dynamic_network import DynamicNetwork
            self._dynamic_network = DynamicNetwork.load(self._network_id)

        return self._dynamic_network

    def set_dynamic_network(self, dyn_network):
        self._network_id = dyn_network.get_id()
        self._dynamic_network = dyn_network

    def get_network_id(self):
        return self._network_id

    def get_output_dir(self, network_id=None):
        return Method.get_method_output_dir(network_id or self.get_network_id(), self.get_id())

    def get_images_dir(self, network_id=None):
        return Method.get_method_images_dir(network_id or self.get_network_id(), self.get_id())

    @staticmethod
    def get_method_output_dir(network_id, method_id=None):
        # TODO: have a better way to separate computed methods and ground truths
        method_subdir_name = 'methods'
        if Method.is_ground_truth_id(network_id, method_id):
            method_subdir_name = 'ground_truth'
        elif Method.is_ground_truth_id(network_id, method_id, sub_type='affinity'):
            method_subdir_name = 'ground_truth/_affinity/'

        return settings.NETWORKS_BASE_DIR + network_id + '/' + method_subdir_name + '/'\
               + (method_id + '/' if method_id is not None else '')

    @staticmethod
    def get_method_images_dir(network_id, method_id):
        return Method.get_method_output_dir(network_id, method_id) + 'images/'

    @classmethod
    def is_corresponding_method_id(cls, method_id):
        return method_id.startswith(cls.get_id_prefix())

    def get_run_time(self):
        return NotImplementedError()

    def is_clean_run(self):
        """
        Whether this was a clean run or some files were reused from previous runs
        :return:
        """
        return self._clean_run

    def save_run_stats(self):
        if self.get_dynamic_network() is None or self.get_run_time() is None:
            raise Exception("No run data available")

        logger.debug("Saving run statistics")
        self.get_community_evolution().save()
        self.save_runtime(self.get_network_id())
        self.save_state()

    def save_runtime(self, network_id, run_time=None):
        run_time = run_time or self.get_run_time()
        output_path = self.get_output_dir() + 'time.txt'
        logger.debug("Saving running time to %s" % output_path)

        if not os.path.exists(os.path.dirname(output_path)):
            os.makedirs(os.path.dirname(output_path))
        else:
            t_old = Method.load_runtime(network_id, self.get_id())
            if t_old is not None and t_old < run_time - 0.001:
               logger.warning("Previous recorded running time %s < %s" % (t_old, run_time))
               return
        with open(output_path, "wb") as outfile:
            outfile.write(
                "%s\nclean_run=%s" % (run_time, self.is_clean_run())
            )

    def save_state(self):
        #TODO: to save memory, avoid saving unnecessary fields (e.g., do not save dynamic network) - save only network_id
        # temporary fix - to make it proper, can use __getstate__ and __setstate__
        self._dynamic_network = None
        self._community_evolution._dynamic_network = None
        output_path = self.get_output_dir() + 'state.cpickle'
        logger.debug("Saving method state to %s" % output_path)
        if not os.path.exists(os.path.dirname(output_path)):
            os.makedirs(os.path.dirname(output_path))
        with open(output_path, 'wb') as pickle_file:
            cPickle.dump(self, pickle_file)

    @staticmethod
    def load_state(network_id, method_id):
        """
        Load method with state by id
        :param network_id:
        :param method_id:
        :return: Method
        """
        pickle_path = Method.get_method_output_dir(network_id, method_id) + 'state.cpickle'
        if not os.path.exists(pickle_path):
            logger.warning("No saved state in %s" % pickle_path)
            return None
        with open(pickle_path, 'r') as pickle_file:
            return cPickle.load(pickle_file)

    @staticmethod
    def load_runtime(network_id, method_id):
        input_path = Method.get_method_output_dir(network_id, method_id) + 'time.txt'
        if not os.path.exists(input_path):
            return None
        with open(input_path, 'r') as in_file:
            lines = in_file.readlines()
            t = float(lines[0])
            # is_clean_run  = not "clean_run=False" in lines[1]
        # if not is_clean_run:
        #     logger.warning("No clean run in %s" % input_path)
        return t