import sys
import timeit
from .. import settings
from .. import utils
from ..log import logger
from ..method import Method
from .. import community_evolution
from ..draw import draw_utils
from ..scoring_methods import scoring_methods_utils
from .. import consensus_clustering

# Given G_1, G_2, ..., G_k => produce a common partition
class NaiveMethod(Method):

    COMPUTE_TYPE_TITLE_DICT = {
        'exh': "exhaustive",
        'bu': "bottom-up",
        'td': "top-down",
    }

    def __init__(self, method_id):
        self._method_id = method_id

        self._consensus_clustering_method_id, scoring_method_id, self._compute_type, self._model_selection = NaiveMethod._parse_method_id(
             self._method_id
        )
        self._scoring_method = scoring_methods_utils.create_scoring_method_by_id(
            scoring_method_id=scoring_method_id
        )

        Method.__init__(self, method_id)

    @staticmethod
    def _parse_method_id(method_id):
        values = method_id.split(settings.METHOD_ID_SEPARATOR)
        if values[0] != NaiveMethod.get_id_prefix():
            raise Exception("Cannot create NaiveMethod instance from method_id '%s'" % method_id)
        compute_type = 'exh' if len(values)<4 else values[3]
        model_selection = 'auto' if len(values)<5 else values[4]
        return values[1],values[2],compute_type, model_selection

    @staticmethod
    def get_method_title(method_id):
        consensus_clustering_method_id, scoring_method_id, compute_type, model_selection = NaiveMethod._parse_method_id(
             method_id
        )

        return "Naive, {compute_type}, cons clust - {consensus_clustering_method_id}, score={scoring_method_id}, model selection={model_selection}".format(
            compute_type=NaiveMethod.COMPUTE_TYPE_TITLE_DICT[compute_type],
            consensus_clustering_method_id=consensus_clustering_method_id,
            scoring_method_id=scoring_method_id,
            model_selection=model_selection,
        )

    @staticmethod
    def get_method_description(method_id):
        consensus_clustering_method_id, scoring_method_id, compute_type, model_selection = NaiveMethod._parse_method_id(
             method_id
        )

        return "This method searches through various time segmentations (via {compute_type} approach), " \
               "clustering each segment via a consensus clustering method '{consensus_clustering_method_id}' " \
               "({consensus_cluster_method_title}) " \
               "so that to maximize the '{scoring_method_id}' scoring function " \
               "where score={scoring_method_title} " \
               "(model selection - {model_selection}).".format(
            compute_type=NaiveMethod.COMPUTE_TYPE_TITLE_DICT[compute_type],
            consensus_clustering_method_id=consensus_clustering_method_id,
            consensus_cluster_method_title=consensus_clustering.get_consensus_partition_method_description(
                consensus_clustering_method_id
            ),
            scoring_method_id=scoring_method_id,
            scoring_method_title=scoring_methods_utils.get_scoring_method_title(scoring_method_id),
            model_selection=model_selection,
        )

    def get_id(self):
        return self._method_id

    @staticmethod
    def get_id_prefix():
        return 'naive'

    def get_scoring_method(self):
        return self._scoring_method

    def get_scoring_method_id(self):
        return self.get_scoring_method().get_id()

    def get_consensus_clustering_method_id(self):
        return self._consensus_clustering_method_id

    def get_solution_for_n_changepoints(self, n_changepoints):
        dp_record = self._dp[
            '{t_start}-{t_end}-{n_changepoints}'.format(
                t_start=0,
                # backward compatibility - earlier version of the methods used t_end inclusive
                t_end=len(self.get_dynamic_network().get_snapshots())-(0 if hasattr(self, '_compute_type') else 1),
                n_changepoints=n_changepoints,
            )
        ]
        return community_evolution.CommunityEvolution(
            segments_partitions=community_evolution.CommunityEvolution
                .rename_clusters_as_ints_per_unique_adjacent_cluster(
                dp_record[1],
            ),
            changepoints=dp_record[2],
            score=dp_record[0],
            method_id=self.get_id(),
            dynamic_network=self.get_dynamic_network(),
        )

    def get_timepoints_scores(self):
        """
        Get list of scores describing how changepoint-like each timepoint is
        :return:
        """
        # first check - backward compatibility
        if not hasattr(self, "_timepoints_scores") or self._timepoints_scores is None:
            n_timepoints = len(self.get_dynamic_network().get_timepoints())
            res = [0] * n_timepoints
            for i in xrange(1, len(res)):
                s = self.get_solution_for_n_changepoints(i)
                # score_curr = self.get_scoring_method().evaluate_community_evolution(s)

                # in case time point appears as a changepoint and then disappears later on
                # TODO: check why this happens
                # for t in xrange(len(res)):
                #     if res[t] > 0 and t not in s.get_changepoints():
                #         res[t] = 0

                for t in s.get_changepoints():
                    if res[t] == 0:
                        res[t] =  (n_timepoints - i)  / (1.0 * n_timepoints)
            self._timepoints_scores = res
        return self._timepoints_scores

    def get_community_evolution(self):
        if self.get_scoring_method_id() in ['(aic)', '(bic)']:
            return self._community_evolution
        if not hasattr(self,'_com_ev_opt'):
            max_score = None
            max_i = None

            if self._model_selection == 'auto':
                self._model_selection = '(bic)'
            model_selection_method_name = self._model_selection # either BIC or AIC

            ic_method = scoring_methods_utils.create_scoring_method_by_id(model_selection_method_name)
            for i in xrange(len(self.get_dynamic_network().get_snapshots())):
                sc = ic_method.evaluate_community_evolution(
                    community_evolution=self.get_solution_for_n_changepoints(i),
                )
                if max_score is None or sc > max_score:
                    max_i = i
                    max_score = sc
            self._com_ev_opt = self.get_solution_for_n_changepoints(max_i)

        return self._com_ev_opt

    def get_run_time(self):
        return self._run_time

    def run(self, dynamic_network, **kwargs):
        self._run_time = None
        self.set_dynamic_network(dynamic_network)
        logger.info("Running '{method_name}' method for '{network_id}'".format(
                method_name=self.get_title(),
                network_id=dynamic_network.get_id(),
            )
        )

        # for memoization
        # t_start-t_end-n_changepoints
        self._dp = {}

        self._n_calls_to_cons_clust = 0
        self._timepoints_scores = None

        compute_func_dict = {
            'exh': self._compute_exh,
            'bu': self._compute_bu,
            'td': self._compute_td,
        }

        start = timeit.default_timer()
        score, segments_partitions, changepoints = compute_func_dict[self._compute_type](
            self._dynamic_network.get_snapshots(),
            t_start=0,
            n_changepoints=None,
        )
        end = timeit.default_timer()
        self._run_time = end - start
        logger.info("Finished in %s" % utils.sec_to_time(self._run_time))
        logger.debug("Number of calls to consensus clustering: %s" % self._n_calls_to_cons_clust)

        self._community_evolution = community_evolution.CommunityEvolution(
            segments_partitions=community_evolution.CommunityEvolution
                .rename_clusters_as_ints_per_unique_adjacent_cluster(
                segments_partitions,
            ),
            changepoints=changepoints,
            score=score,
            method_id=self.get_id(),
            dynamic_network=dynamic_network,
        )

        self._result_score = self.get_scoring_method().evaluate_community_evolution(
                    self._community_evolution
        )
        logger.info(self._community_evolution.get_summary())
        logger.info(
            "Result score according to '{scoring_method}': {score}".format(
                scoring_method=self.get_scoring_method().get_id(),
                score=self._result_score
            )
        )

        self._segment_ground_truth_score = None
        if dynamic_network.get_segment_ground_truth() is not None:
            self._segment_ground_truth_score = self.get_scoring_method().evaluate_community_evolution(
                dynamic_network.get_segment_ground_truth()
            )
            logger.info(
                "{ground_truth_title} score according to '{scoring_method}': {score}".format(
                    ground_truth_title=dynamic_network.get_segment_ground_truth().get_method_title(),
                    scoring_method=self.get_scoring_method().get_id(),
                    score=self._segment_ground_truth_score
                )
            )

        return self._community_evolution

    def _compute_bu(self, networks_list, t_start=0, t_end=None, n_changepoints=None):
        """
        Bottom-up version of compute - do not perform exhaustive search, just do bottom-up approach
        (Note: it still uses memoization for consensus clustering)
        :param networks_list:
        :param t_start:
        :param n_changepoints:
        :return:
        """
        t_end = t_end or t_start + len(networks_list)

        # for stabilized louvain, use previous segment's partition as initialization
        is_stabilized_louvain = self.get_consensus_clustering_method_id().endswith('stab_louvain')
        if is_stabilized_louvain:
            prev_partition = None
            for i in xrange(t_start, t_end):
                temp_score, temp_partitions, temp_changepoints = self._compute_exh(
                    networks_list,
                    t_start=i,
                    t_end=i+1,
                    n_changepoints=0,
                    init_partition=prev_partition,
                )
                prev_partition = temp_partitions[0]
        best_score, curr_partitions, curr_changepoints = self._compute_exh(
            networks_list,
            t_start=t_start,
            t_end=t_end,
            n_changepoints=t_end-t_start-1
        )
        dp_key = '{t_start}-{t_end}-any'.format(
            t_start=t_start,
            t_end=t_end,
        )
        self._dp[dp_key] = best_score, curr_partitions, curr_changepoints

        for temp_n_merges in xrange(1, t_end-t_start):
            best_merge_score = None
            best_merge_partitions = None
            best_merge_changepoints = None
            for i_segment in xrange(len(curr_changepoints)):
                t_curr_segment_start = curr_changepoints[i_segment - 1] if i_segment > 0 else t_start
                t_next_segment_end = curr_changepoints[i_segment + 1] if i_segment + 1 < len(curr_changepoints) else t_end

                temp_merge_score, temp_merge_partitions, temp_merge_changepoints = self._compute_exh(
                    networks_list=networks_list,
                    t_start=t_curr_segment_start,
                    t_end=t_next_segment_end,
                    n_changepoints=0,
                    use_memoization=((not is_stabilized_louvain) or i_segment==0),
                    init_partition=curr_partitions[i_segment-1] if (is_stabilized_louvain and i_segment>0) else None,
                )

                temp_partitions = curr_partitions[:i_segment] + temp_merge_partitions + curr_partitions[i_segment + 2:]
                temp_changepoints = curr_changepoints[:i_segment] + curr_changepoints[i_segment + 1:]
                temp_score = self.get_scoring_method().evaluate(
                    networks_list=networks_list,
                    segments_partitions=temp_partitions,
                    changepoints=temp_changepoints,
                )

                if best_merge_score is None or temp_score > best_merge_score:
                    best_merge_score, best_merge_partitions, best_merge_changepoints = temp_score, temp_partitions, temp_changepoints

            curr_score, curr_partitions, curr_changepoints = best_merge_score, best_merge_partitions, best_merge_changepoints
            dp_key = '{t_start}-{t_end}-{changepoints_count}'.format(
                t_start=t_start,
                t_end=t_end,
                changepoints_count=t_end-t_start-1-temp_n_merges
            )
            self._dp[dp_key] = curr_score, curr_partitions, curr_changepoints
            if best_score is None or curr_score > best_score:
                dp_key = '{t_start}-{t_end}-any'.format(
                    t_start=t_start,
                    t_end=t_end,
                )
                self._dp[dp_key] = curr_score, curr_partitions, curr_changepoints
                best_score = curr_score

        dp_key = '{t_start}-{t_end}-{changepoints_count}'.format(
            t_start=t_start,
            t_end=t_end,
            changepoints_count=n_changepoints if n_changepoints is not None else "any"
        )
        return self._dp[dp_key]


    def _compute_td(self, networks_list, t_start=0, t_end=None, n_changepoints=None):
        """
        Top-down version of compute - do not perform exhaustive search, just do top-down approach
        (Note: it still uses memoization for consensus clustering)
        :param networks_list:
        :param t_start:
        :param n_changepoints:
        :return:
        """
        best_score = None
        best_split_partitions = None
        best_split_changepoints = []
        curr_partitions = []
        curr_changepoints = []
        t_end = t_end or t_start + len(networks_list)
        # for stabilized louvain, use previous segment's partition as initialization
        is_stabilized_louvain = self.get_consensus_clustering_method_id().endswith('stab_louvain')
        for temp_n_changepoints in xrange(t_end - t_start):
            i_segment = -1
            t_segment_start = 0
            best_split_score = None
            # try to split each current segment into two parts
            for t_segment_end in curr_changepoints + [t_end]:
                i_segment += 1
                if t_segment_end - t_segment_start < 2: # segment has only one snapshot
                    t_segment_start = t_segment_end
                    continue
                if is_stabilized_louvain and temp_n_changepoints > 0:
                    temp_split_score = None
                    temp_split_partitions = None
                    temp_split_changepoints = None
                    for t_split in xrange(t_segment_start+1, t_segment_end):
                        score_1, partitions_1, changepoints_1 = self._compute_exh(
                            networks_list, t_segment_start, t_split, n_changepoints=0,
                            use_memoization=(i_segment==0),
                            init_partition=curr_partitions[i_segment-1] if i_segment>0 else None,
                        )
                        score_2, partitions_2, changepoints_2 = self._compute_exh(
                            networks_list, t_split, t_segment_end, n_changepoints=0,
                            use_memoization=False,
                            init_partition=partitions_1[0],
                        )

                        temp_stab_partitions = (partitions_1 + partitions_2)
                        temp_stab_changepoints = [t_split]
                        temp_stab_score = self.get_scoring_method().evaluate(
                            networks_list=networks_list[t_segment_start:t_segment_end],
                            segments_partitions=temp_stab_partitions,
                            changepoints=[t_split-t_segment_start],
                        )
                        if temp_split_score is None or temp_stab_score > temp_split_score:
                            temp_split_score = temp_stab_score
                            temp_split_partitions = temp_stab_partitions
                            temp_split_changepoints = temp_stab_changepoints
                else:
                    temp_split_score, temp_split_partitions, temp_split_changepoints = self._compute_exh(
                        networks_list=networks_list,
                        t_start=t_segment_start,
                        t_end=t_segment_end,
                        n_changepoints=0 if temp_n_changepoints==0 else 1,
                    )
                t_segment_start = t_segment_end

                temp_partitions = curr_partitions[:i_segment] + temp_split_partitions + curr_partitions[i_segment + 1:]
                temp_changepoints = curr_changepoints[:i_segment] + temp_split_changepoints + curr_changepoints[i_segment:]
                temp_score = self.get_scoring_method().evaluate(
                    networks_list=networks_list,
                    segments_partitions=temp_partitions,
                    changepoints=temp_changepoints,
                )

                if best_split_score is None or temp_score > best_split_score:
                    best_split_score, best_split_partitions, best_split_changepoints = temp_score, temp_partitions, temp_changepoints

            curr_score, curr_partitions, curr_changepoints = best_split_score, best_split_partitions, best_split_changepoints
            dp_key = '{t_start}-{t_end}-{changepoints_count}'.format(
                t_start=t_start,
                t_end=t_end,
                changepoints_count=temp_n_changepoints
            )
            self._dp[dp_key] = curr_score, curr_partitions, curr_changepoints
            if best_score is None or curr_score > best_score:
                dp_key = '{t_start}-{t_end}-any'.format(
                    t_start=t_start,
                    t_end=t_end,
                    changepoints_count=temp_n_changepoints
                )
                self._dp[dp_key] = curr_score, curr_partitions, curr_changepoints
                best_score = curr_score

        dp_key = '{t_start}-{t_end}-{changepoints_count}'.format(
            t_start=t_start,
            t_end=t_end,
            changepoints_count=n_changepoints if n_changepoints is not None else "any"
        )
        return self._dp[dp_key]


    def _compute_exh(self, networks_list, t_start=0, t_end=None, n_changepoints=None, **kwargs):
        """
        Exhaustive version of compute - not a regular top-down, but instead an exhaustive search using DP which takes n*(n+1)/2 runs of consensus clustering;
        :param networks_list:
        :param t_start:
        :param t_end: exclusive end
        :param n_changepoints:
        :return:
        """
        best_score = None
        best_partitions = None
        best_changepoints = None

        t_end = t_end or len(networks_list)

        dp_key = '{t_start}-{t_end}-{changepoints_count}'.format(
            t_start=t_start,
            t_end=t_end,
            changepoints_count=n_changepoints if n_changepoints is not None else "any"
        )
        if dp_key in self._dp and kwargs.get('use_memoization', True):
            return self._dp[dp_key]

        sys.stdout.write("\rFinding the best split of {n_networks} network{plural} with "
                         "{changepoints_count} changepoints".format(
                n_networks=t_end - t_start,
                plural="s" if t_end - t_start!=1 else "",
                changepoints_count=n_changepoints if n_changepoints is not None else "any",
            )
        )
        sys.stdout.flush()

        if n_changepoints is None:
            # try all possible numbers of checkpoints
            for n_changepoints in xrange(t_end - t_start):
                temp_score, temp_partitions, temp_changepoints = self._compute_exh(
                        networks_list,
                        t_start,
                        t_end,
                        n_changepoints=n_changepoints
                )
                if best_score is None or temp_score > best_score:
                    best_score = temp_score
                    best_partitions = temp_partitions
                    best_changepoints = temp_changepoints
        else:
            if n_changepoints == 0:
                self._n_calls_to_cons_clust += 1

                partition = consensus_clustering.compute_consensus_partition(
                    networks_list=networks_list[t_start:t_end],
                    method=self.get_consensus_clustering_method_id(),
                    base_dir=self.get_output_dir() + 'aux/%s_%s/' % (t_start, t_end), # for auxiliary files
                    base_output_filename='%s_%s' % (t_start, t_end),
                    overwrite=True,
                    **kwargs
                )

                best_partitions = [partition]
                best_changepoints = []
                best_score = self.get_scoring_method().evaluate(
                    networks_list=networks_list[t_start:t_end],
                    segments_partitions=best_partitions,
                    changepoints=best_changepoints,
                )
            else:
                # insert one changepoint at t_split and optimally split the rest with num_changepoints-1 changepoints
                for t_split in xrange(t_start + 1, t_end):

                    # avoid unnecessary recursive calls
                    if n_changepoints == t_end - t_start - 1 and t_split > t_start + 1:
                        break

                    for n_changepoints_1 in xrange(n_changepoints):
                        n_changepoints_2 = n_changepoints - 1 - n_changepoints_1
                        if n_changepoints_1 >= t_split - t_start or n_changepoints_2 >= t_end - t_split:
                            continue
                        score_1, partitions_1, changepoints_1 = self._compute_exh(
                            networks_list, t_start, t_split, n_changepoints=n_changepoints_1
                        )
                        score_2, partitions_2, changepoints_2 = self._compute_exh(
                            networks_list, t_split, t_end, n_changepoints=n_changepoints_2
                        )

                        temp_partitions = (partitions_1 + partitions_2)
                        temp_changepoints = (changepoints_1 + [t_split] + changepoints_2)
                        temp_score = self.get_scoring_method().evaluate(
                            networks_list=networks_list[t_start:t_end],
                            segments_partitions=temp_partitions,
                            changepoints=[t-t_start for t in temp_changepoints],
                        )

                        if best_score is None or temp_score > best_score:
                            best_score, best_partitions, best_changepoints = temp_score, temp_partitions, temp_changepoints

                        # logger.debug("Best result for [{t_start}, {t_end}] with split at {t_split} "
                        #           "and total of {n_changepoints} changepoint{plural} has score {score} and partition " \
                        #           "sizes [{community_sizes}]\n".format(
                        #         t_start=t_start,
                        #         t_end=t_end - 1,
                        #         t_split=t_split,
                        #         n_changepoints=n_changepoints,
                        #         plural = "s" if n_changepoints != 1 else "",
                        #         score=temp_score,
                        #         community_sizes = ", ".join(map(str,[len(set(p.values())) for p in (partitions_1 + partitions_2)]))
                        #     )
                        # )

        # if best_score != None:
        #     segments = []
        #     segments_starts = [t_start] + best_changepoints + [t_end]
        #     for i in xrange(1, len(segments_starts)):
        #         segments.append('{start}-{end}'.format(start=segments_starts[i-1], end=segments_starts[i]-1))
        #     logger.info("Optimal split for [{t_start}, {t_end}] with {changepoints_count} changepoint{plural} is [{t_segments}] with score {score}".format(
        #             t_start=t_start,
        #             t_end=t_end - 1,
        #             changepoints_count=n_changepoints or "any",
        #             plural="s" if n_changepoints is None or n_changepoints > 1 else "",
        #             t_segments=", ".join(segments),
        #             score = best_score
        #         )
        #     )

        # to remove progress text
        sys.stdout.write("\r")
        sys.stdout.flush()

        self._dp[dp_key] = best_score, best_partitions, best_changepoints
        return self._dp[dp_key]