import sys
import timeit
from .. import settings
from .. import utils
from ..log import logger
from ..method import Method
from .. import community_evolution
from ..scoring_methods import scoring_methods_utils
import os
import subprocess
import networkx as nx
from multiprocessing import Queue
from ..parallel import CustomProcess, run_process_pool

class NaiveCppMethod(Method):
    Naive_EXEC = settings.BASE_DIR + 'dynamic_communities_cpp/naive/main_openmp' + ('' if utils.is_remote_server() else '.exe')

    def __init__(self, method_id):
        self._method_id = method_id

        self._static_clustering_method_id, _, self._scoring_method, self._model_selection = NaiveCppMethod._parse_method_id(
             self._method_id
        )

        if self._scoring_method in ['aic', 'bic']:
            logger.debug("As scoring method is '%s', using the same model selection" % self._scoring_method)
            self._model_selection = self._scoring_method

        self._n_changepoints_sol_cache = {}

        Method.__init__(self, method_id)

    @staticmethod
    def _parse_method_id(method_id):
        values = method_id.split(settings.METHOD_ID_SEPARATOR)
        if values[0] != NaiveCppMethod.get_id_prefix():
            raise Exception("Cannot create NaiveCppMethod instance from method_id '%s'" % method_id)
        is_old_version = 1 if "-sn-" in method_id else 0
        average_type = values[2] if (is_old_version and len(values)>2) else "sn" # note: for backward compatibility, currently not used
        scoring_method = values[2 + is_old_version]
        model_selection = values[3 + is_old_version] if len(values) > (3 +is_old_version) else scoring_method
        return values[1], average_type, scoring_method, model_selection

    @staticmethod
    def get_method_title(method_id):
        static_clustering_method_id, _, scoring_method, model_selection = NaiveCppMethod._parse_method_id(method_id)

        return "Naive-CPP, cons clust - sum {clustering_method_id}, scoring method - {scoring_method}, model selection={model_selection}".format(
            clustering_method_id=static_clustering_method_id,
            scoring_method=scoring_method,
            model_selection=model_selection,
        )

    @staticmethod
    def get_method_description(method_id):
        static_clustering_method_id, _, scoring_method, model_selection = NaiveCppMethod._parse_method_id(
             method_id
        )

        return "This method searches through various time segmentations (via bottom-up approach), " \
               "clustering each segment via a consensus clustering method 'sum {clustering_method_id}' " \
               "so that to maximize the '{scoring_method}' scoring function " \
               "(model selection - {model_selection}).".format(
            clustering_method_id=static_clustering_method_id,
            scoring_method=scoring_method,
            model_selection=model_selection,
        )

    def get_id(self):
        return self._method_id

    @staticmethod
    def get_id_prefix():
        return 'cppnaive'

    def get_static_clustering_method_id(self):
        return self._static_clustering_method_id

    def get_solution_for_n_changepoints(self, n_changepoints):
        if not n_changepoints in self._n_changepoints_sol_cache:
            self._n_changepoints_sol_cache[n_changepoints] = community_evolution.CommunityEvolution.load_from_path(
                self.get_output_dir() + 'results/%s.txt' % n_changepoints,
                method_id=self.get_id(),
                network_id=self.get_network_id(),
                convert_objects_to_ints=True,
                convert_objects_names_dict=self._integer_names_nodes,
            )
        return self._n_changepoints_sol_cache[n_changepoints]

    def get_timepoints_scores(self):
        """
        Get list of scores describing how changepoint-like each timepoint is
        :return:
        """
        # first check - backward compatibility
        if not hasattr(self, "_timepoints_scores") or self._timepoints_scores is None:
            n_timepoints = len(self.get_dynamic_network().get_timepoints())
            res = [0] * n_timepoints
            for i in xrange(1, len(res)):
                s = self.get_solution_for_n_changepoints(i)
                for t in s.get_changepoints():
                    if res[t] == 0:
                        res[t] =  (n_timepoints - i)  / (1.0 * n_timepoints)
            self._timepoints_scores = res
        return self._timepoints_scores

    def get_community_evolution(self):
        if not hasattr(self,'_community_evolution'):
            self._community_evolution = community_evolution.CommunityEvolution.load_from_path(
                self.get_output_dir() + 'results/_%s.txt' % self._model_selection,
                method_id=self.get_id(),
                network_id=self.get_network_id(),
                convert_objects_to_ints=True,
                convert_objects_names_dict=self._integer_names_nodes,
            )
            # bic_processes = []
            # bic_sc_m  = scoring_methods_utils.create_scoring_method_by_id("(bic)")
            # n_snapshots = len(self.get_dynamic_network().get_snapshots())
            # q = Queue(n_snapshots)
            # def calculate_best_bic(community_evolution, dynamic_network, i):
            #     sc = bic_sc_m.evaluate_community_evolution(
            #         community_evolution,
            #         dynamic_network
            #     )
            #     q.put((i, sc))
            #
            # for i in xrange(n_snapshots):
            #     p = CustomProcess(
            #         name="EvalProc-%s" % i,
            #         target=calculate_best_bic,
            #         kwargs={
            #             'community_evolution': self.get_solution_for_n_changepoints(i),
            #             'dynamic_network': self.get_dynamic_network(),
            #             'i': i
            #         },
            #     )
            #     bic_processes.append(p)
            # run_process_pool(bic_processes)
            # q_top = q.get()
            # max_i = q_top[0]
            # max_score = q_top[1]
            # for i in xrange(1, n_snapshots):
            #     q_top = q.get()
            #     if q_top[1] > max_score:
            #         max_i = q_top[0]
            #         max_score = q_top[1]
            # self._community_evolution = self.get_solution_for_n_changepoints(max_i)
        return self._community_evolution
        # return self.get_solution_for_n_changepoints(5) #len(self.get_dynamic_network().get_segment_ground_truth().get_changepoints()))


    def get_run_time(self):
        return self._run_time

    def _prepare_input(self, **kwargs):
        output_dir = self.get_output_dir() + 'input/'

        logger.info("Preparing input")

        # cppnaive needs all nodes to be integers in [0, |V|-1]
        # so, convert ['a', 'b', 'c'] to [0, 1, 2]
        all_nodes = self.get_dynamic_network().get_all_nodes()
        # 0 => 'a', 1 => 'b', 2 => 'c'
        self._integer_names_nodes = all_nodes
        # 'a' => 0, 'b' => 1, 'b' => x
        self._nodes_integer_names = dict(zip(all_nodes, range(len(all_nodes))))

        if os.path.exists(output_dir) and not kwargs.get('overwrite', False):
            logger.info('Preparing input: %s already exists' % output_dir)
            return output_dir
        elif not os.path.exists(output_dir):
            os.makedirs(output_dir)

        with open(self.get_output_dir() + 'integer_names_nodes.txt', 'wb') as outfile:
            for i,node in enumerate(self._integer_names_nodes):
                outfile.write('%s,%s\n' % (i, node))

        for i,sn in enumerate(self.get_dynamic_network().get_snapshots()):
            nx.write_edgelist(sn.get_nx_graph(), output_dir + '%s.txt' % i, data=False)
            with open(output_dir + '%s.txt' % i, 'wb') as outfile:
                for e in sn.get_nx_graph().edges_iter(data=True):
                    outfile.write('%s %s\n' % (self._nodes_integer_names[e[0]], self._nodes_integer_names[e[1]]))
                # add self-loops to make sure all nodes are present
                for v in self._nodes_integer_names.values():
                    outfile.write('%s %s\n' % (v, v))
        return output_dir


    def run(self, dynamic_network, **kwargs):
        self._run_time = None
        self.set_dynamic_network(dynamic_network)
        logger.info("Running '{method_name}' method for '{network_id}'".format(
                method_name=self.get_title(),
                network_id=dynamic_network.get_id(),
            )
        )
        self._timepoints_scores = None

        input_dir = self._prepare_input(**kwargs)
        input_files = sorted(
            utils.get_file_list_from_dir(input_dir, ext='txt', full_path=True),
            cmp=utils.cmp_str_int
        )

        output_dir = self.get_output_dir() + 'results/'
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)

        run_command = '{exe} -c {clust} -s {scoring_method} "{input_files}" "{output_dir}"'.format(
            exe=NaiveCppMethod.Naive_EXEC,
            clust=self.get_static_clustering_method_id(),
            scoring_method=self._scoring_method,
            input_files=",".join(input_files),
            output_dir=output_dir,
        )
        num_tries = 10
        if os.path.exists(self.get_output_dir() + 'results/_%s.txt' % self._model_selection):
            os.remove(self.get_output_dir() + 'results/_%s.txt' % self._model_selection)
        while True: # to avoid issues during batch run
            logger.debug("Run command: %s" % run_command)

            start = timeit.default_timer()
            subprocess.call(run_command, shell=True)
            end = timeit.default_timer()

            if not os.path.exists(self.get_output_dir() + 'results/_%s.txt' % self._model_selection) and num_tries>0:
                logger.warning("Run had errors, trying %s more times" % num_tries)
                num_tries -=1
                continue

            self._run_time = end - start
            logger.info("Finished in %s" % utils.sec_to_time(self._run_time))
            break

        scoring_method = scoring_methods_utils.create_scoring_method_by_id(
            '(avg_sn)(mod)' if self._scoring_method == 'mod' else '(%s)' % self._scoring_method
        )
        self._result_score = scoring_method.evaluate_community_evolution(
            community_evolution=self.get_community_evolution(),
            dynamic_network=self.get_dynamic_network(),
        )
        logger.info(self.get_community_evolution().get_summary())
        logger.info(
            "Result score according to '{scoring_method}': {score}".format(
                scoring_method=scoring_method.get_id(),
                score=self._result_score
            )
        )

        self._segment_ground_truth_score = None
        if dynamic_network.get_segment_ground_truth() is not None:
            self._segment_ground_truth_score = scoring_method.evaluate_community_evolution(
                dynamic_network.get_segment_ground_truth()
            )
            logger.info(
                "{ground_truth_title} score according to '{scoring_method}': {score}".format(
                    ground_truth_title=dynamic_network.get_segment_ground_truth().get_method_title(),
                    scoring_method=scoring_method.get_id(),
                    score=self._segment_ground_truth_score
                )
            )

            scm = scoring_methods_utils.create_scoring_method_by_id('(com_evs_sim)((gl_p)(ami))(segment_ground_truth)')
            logger.info("Similarity to segment ground truth according to (gl_p)(ami): %s" %
                scm.evaluate_community_evolution(
                    self.get_community_evolution(),
                    self.get_dynamic_network(),
                )
            )
