import os
import timeit
from ... import settings
from ... import utils
from ...log import logger
from ...method import Method
from ...partition import Partition
from ... import community_evolution
from ..other_methods.ghrg_utils import changepointDetection
import numpy as np
import subprocess
from collections import defaultdict
from ... import consensus_clustering

class GHRG(Method):
    BIN_DIR = settings.BASE_DIR + 'dynamic_communities_cpp/other_methods/ghrg/'
    fitHRG_EXEC = BIN_DIR + 'fitHRG_GPL_Bayes/fitHRG' + ('' if utils.is_remote_server() else '.exe')
    consensusHRG_EXEC = BIN_DIR + 'consensusHRG_GPL_Bayes/consensusHRG'+ ('' if utils.is_remote_server() else '.exe')
    ALPHA = 0.95

    def __init__(self, method_id):
        Method.__init__(self, method_id)
        self._window_size = int(GHRG._parse_method_id(method_id)[0])

    @staticmethod
    def _parse_method_id(method_id):
        values = method_id.split(settings.METHOD_ID_SEPARATOR)
        if values[0] != GHRG.get_id_prefix():
            raise Exception("Cannot create HRG instance from method_id '%s'" % method_id)
        return values[1:]

    @staticmethod
    def get_method_title(method_id):
        return "GHRG, w=%s" % GHRG._parse_method_id(method_id)[0]

    @staticmethod
    def get_method_description(method_id):
        return "This is GHRG method with window size %s" % GHRG._parse_method_id(method_id)[0]

    def get_window_size(self):
        return self._window_size

    def get_id(self):
        return self._method_id

    @staticmethod
    def get_id_prefix():
        return 'ghrg'

    def get_timepoints_scores(self):
        """
        Get list of scores describing how changepoint-like each timepoint is
        :return:
        """
        return self._t_scores

    def get_community_evolution(self):
        return self._community_evolution

    def get_run_time(self):
        return self._run_time

    @staticmethod
    def _prepare_input(dynamic_network, output_dir, **kwargs):

        if os.path.exists(output_dir) and not kwargs.get('overwrite', False):
            logger.info('Preparing input: %s already exists' % output_dir)
            return
        elif not os.path.exists(output_dir):
            os.makedirs(output_dir)

        all_nodes = dynamic_network.get_all_nodes()
        with open(output_dir + 'names.lut', 'wb') as f:
            f.write('virtual real\n')
            for i,u in enumerate(all_nodes):
                f.write('%s %s\n' % (i, u))

        for i_sn, sn in enumerate(dynamic_network.get_snapshots()):
            output_file = output_dir + '%s.pairs' % i_sn
            with open(output_file, 'wb') as out_f:
                for u,v in sn.get_edges():
                    out_f.write('%s %s\n' % (u, v))

    def _get_ghrg_dir(self):
        return self.get_output_dir() + 'results/'

    def _get_cp_detector(self):
        return self._cp_detector

    def _run_execs(self, t_start, t_end, is_segment_partition=False, **kwargs):
        ghrg_dir = self._get_ghrg_dir()
        cp_detector = self._get_cp_detector()

        paramStr="-n %s -m %i" % (cp_detector.namesFile, t_end - t_start + 1)

        run_label = 'seg' if is_segment_partition else 'joint'
        # for runs in the end where window size is smaller
        if not is_segment_partition and t_end - t_start + 1 != cp_detector.window:
            run_label = 'joint-w%s' % (t_end - t_start + 1)


        for gi in xrange(t_start, t_end + 1):
            paramStr+=" %s" % (ghrg_dir + '%s.pairs' % gi)

        paramStr+=(" > /dev/null" if utils.is_remote_server() else " > NUL")

        logger.info("Running fitHRG")
        best_dendro_file = ghrg_dir + '%s_%s_best-dendro.hrg' % (t_end, run_label)
        if not os.path.exists(best_dendro_file) or kwargs.get('overwrite', False):
            run_command = "%s -t %s %s" % (GHRG.fitHRG_EXEC, run_label, paramStr)
            logger.debug("Run command: %s" % run_command)
            start = timeit.default_timer()
            subprocess.call(run_command, shell=True)
            end = timeit.default_timer()
            self._run_time = end - start
            logger.debug("Finished in %s" % utils.sec_to_time(self._run_time))
        elif os.path.exists(best_dendro_file):
            logger.debug("fitHRG: using existing file %s" % best_dendro_file)
            self._clean_run = False

        logger.info("Running consensusHRG")
        cp_detector.consensusFilename=ghrg_dir + "%s_%s_s1-consensus.tree"\
                                                % (t_end, run_label)
        if not os.path.exists(cp_detector.consensusFilename) or kwargs.get('overwrite', False):
            run_command = "%s %s -f %s" % (
                GHRG.consensusHRG_EXEC,
                paramStr,
                best_dendro_file
            )
            logger.debug("Run command: %s" % run_command)
            start = timeit.default_timer()
            subprocess.call(run_command, shell=True)
            end = timeit.default_timer()
            self._run_time = end - start
            logger.debug("Finished in %s" % utils.sec_to_time(self._run_time))
        elif os.path.exists(cp_detector.consensusFilename):
            logger.debug("consensusHRG: using existing file %s" % cp_detector.consensusFilename)
            self._clean_run = False

    def _compute_consensus_clustering(self, t_start, t_end, **kwargs):
        logger.info("Computing consensus clustering for [%s,%s]" % (t_start, t_end))
        ghrg_dir = self._get_ghrg_dir()
        cp_detector = self._get_cp_detector()

        # compute consensus clustering
        self._run_execs(
           t_start=t_start,
           t_end=t_end,
           is_segment_partition=True,
            **kwargs
        )

        cons_tree = cp_detector.createDendroTreeFromConsensus(
            '%s%s_seg_s1-consensus.tree' % (ghrg_dir, t_end)
        )
        edgelist=[]
        networks_files = [ghrg_dir + '%s.pairs' % i for i in xrange(t_start, t_end + 1)]
        for networkFile in networks_files:
            with open(os.path.join(ghrg_dir, networkFile)) as f:
                edgelist.extend([edge.split() for edge in f.readlines()])
        depth_dict = defaultdict(list)
        q_best = None
        p_best = None
        for u in cons_tree.internal_nodes():
            depth_dict[u.level()].append(u)
        # try to cut the tree at different levels to get a flat partition and pick the best one
        logger.debug("Trying to cut tree to produce a partition")
        for i in xrange(1, max(depth_dict.keys()) + 1):
            cl_objects = {}
            used_nodes = []
            for j, u in enumerate(depth_dict[i]):
                nodes_list = [k.taxon.label for k in u.leaf_nodes()]
                cl_objects[j] = nodes_list
                used_nodes += nodes_list
            # add the rest of the nodes as separate clusters
            used_nodes = set(used_nodes)
            j = max(cl_objects.keys()) + 1
            for k in cons_tree.taxon_namespace:
                if k.label not in used_nodes:
                    cl_objects[j] = [k.label]
                    j += 1
            p = Partition.from_clusters_objects(cl_objects)
            q = consensus_clustering.compute_consensus_partition_quality(
                networks_list=self.get_dynamic_network().get_snapshots()[t_start:t_end + 1],
                partition=p,
                metric='avg^mod',
            )
            logger.debug(
                "Cutting at level %s: partition with %s cluster(s) and quality %s"
                % (i, p.get_clusters_count(), q)
            )
            if q_best is None or q > q_best:
                p_best = p
                q_best = q
        logger.debug(
            "Choosing partition with %s cluster(s) and quality %s as segment partition"
            % (p_best.get_clusters_count(), q_best)
        )
        return p_best

    def run(self, dynamic_network, **kwargs):
        self._run_time = None
        self.set_dynamic_network(dynamic_network)
        logger.info("Running '{method_name}' method for '{network_id}'".format(
                method_name=self.get_title(),
                network_id=dynamic_network.get_id(),
            )
        )

        logger.info("Preparing input")
        ghrg_dir = self._get_ghrg_dir()

        GHRG._prepare_input(dynamic_network, ghrg_dir, **kwargs)

        if not os.path.exists(ghrg_dir):
            os.makedirs(ghrg_dir)

        start = timeit.default_timer()

        ### -----------------------------------------------------------------------------------
        # Note: this is the adapted method detectAnomaliesInSequence from changepointDetection
        network_file_sequence =['%s.pairs' % i for i in xrange(len(self.get_dynamic_network().get_snapshots()))]
        self._cp_detector = changepointDetection.anomalyDetection(nsamples=1000 if utils.is_remote_server() else 50)
        self._cp_detector.namesFile = ghrg_dir + 'names.lut'
        self._cp_detector.window = self.get_window_size()
        self._cp_detector.importNames(self._cp_detector.namesFile)
        self._cp_detector.alpha = GHRG.ALPHA
        if not os.path.exists(ghrg_dir + 'results.txt') or kwargs.get('overwrite', True):
            # remove previous results file, as the new lines are appended
            if os.path.exists(ghrg_dir + 'results.txt'):
                logger.debug("Deleting previous results file: %s" % (ghrg_dir + 'results.txt'))
                os.remove(ghrg_dir + 'results.txt')
            t = 0
            while t < len(network_file_sequence) - 1:
                startindx = t
                endindx = min(t + self._cp_detector.window, len(network_file_sequence))

                logger.info("Processing window [%s,%s]" % (startindx, endindx - 1))

                self._run_execs(
                   t_start=startindx,
                   t_end=endindx - 1,
                   is_segment_partition=False,
                    **kwargs
                )

                self._cp_detector.currentNormalNetworks=[network_file_sequence[gi] for gi in xrange(startindx,endindx)]
                self._cp_detector.createNullModel(ghrg_dir, nProcessors=settings.MAX_PROCESSES_COUNT)
                self._cp_detector.testNetwork(ghrg_dir, self._cp_detector.currentNormalNetworks)

                changeDetected = self._cp_detector.calculateP_values(ghrg_dir)
                if changeDetected:
                    logger.info(
                        "Change detected in window [%s,%s] - timepoint %s with p-value %s"
                        % (
                            startindx,
                            endindx - 1,
                            t + np.argmax(self._cp_detector.pvals) + 1, np.max(self._cp_detector.pvals)
                        )
                    )
                    t+=np.argmax(self._cp_detector.pvals)
                else:
                    logger.info("No change detected in [%s,%s]" % (startindx, endindx - 1))
                    # no change in the last segment - no need to process smaller windows with the last snapshot
                    if endindx == len(network_file_sequence):
                        break
                t += 1
        elif os.path.exists(ghrg_dir + 'results.txt'):
            logger.info("Using existing results file: %sresults.txt" % ghrg_dir)
        ### -----------------------------------------------------------------------------------

        logger.debug("Processing results file")
        changepoints = [0]
        segments_partitions = []
        t_scores = [0.0] * len(self.get_dynamic_network().get_snapshots())
        with open(ghrg_dir + 'results.txt', 'r') as infile:
            for l in infile.readlines():
                values = l.replace('\n', '').split()
                t_start = int(values[0])
                t_end = int(values[1])
                pvals = map(float, values[2:])

                logger.info("Processing [%s,%s]" % (t_start, t_end))

                for i,p in enumerate(pvals):
                    t_scores[t_start + i + 1] = p

                i_pval_max = np.argmax(pvals)
                change_detected = pvals[i_pval_max] > GHRG.ALPHA
                if change_detected:
                    changepoints.append(t_start + i_pval_max + 1)
                    logger.info("Detected changepoint at t=%s" % changepoints[-1])
                    segments_partitions.append(
                        self._compute_consensus_clustering(
                            t_start=changepoints[-2],
                            t_end=changepoints[-1] - 1,
                            **kwargs
                        )
                    )
        # add the last segment
        segments_partitions.append(
            self._compute_consensus_clustering(
                t_start=changepoints[-1],
                t_end=len(t_scores) - 1,
                **kwargs
            )
        )
        del changepoints[0]
        self._t_scores = t_scores

        self._community_evolution = community_evolution.CommunityEvolution(
            segments_partitions=community_evolution.CommunityEvolution
                .rename_clusters_as_ints_per_unique_adjacent_cluster(
                    segments_partitions,
            ),
            changepoints=changepoints,
            method_id=self.get_id(),
            dynamic_network=self.get_dynamic_network(),
        )
        logger.info(self._community_evolution.get_summary())

        end = timeit.default_timer()
        self._run_time = end - start
        logger.info("Finished in %s" % utils.sec_to_time(self._run_time))