import os
import timeit
from ... import settings
from ... import utils
from ...log import logger
from ...method import Method
from ...partition import Partition
from ... import community_evolution
import subprocess

class GraphScope(Method):
    N_ITERATIONS = 1
    GraphScope_EXEC = 'java -cp ' \
                      '{target}social-graph-cluster-0.0.1-SNAPSHOT.jar{delimiter}' \
                      '{target}dependency/opencsv-2.3.jar{delimiter}' \
                      '{target}dependency/ejml-0.25.jar ' \
                      'sgc.SocialGraphCluster'.format(
        target=settings.BASE_DIR + 'dynamic_communities_java/other_methods/graphscope/target/',
        delimiter=';' if not utils.is_remote_server() else ':',
    )

    @staticmethod
    def get_method_title(method_id):
        return "GraphScope"

    @staticmethod
    def get_method_description(method_id):
        return "This is GraphScope method."

    def get_id(self):
        return self._method_id

    @staticmethod
    def get_id_prefix():
        return 'graphscope'

    def get_timepoints_scores(self):
        """
        Get list of scores describing how changepoint-like each timepoint is
        :return:
        """
        s_min = min(self._t_scores)
        s_max = max(self._t_scores)
        s_span = s_max - s_min
        return [(x - s_min) / s_span for x in self._t_scores]

    def get_community_evolution(self):
        return self._community_evolution

    def get_run_time(self):
        return self._run_time

    @staticmethod
    def _prepare_snapshots(dynamic_network, output_dir, **kwargs):

        if os.path.exists(output_dir) and not kwargs.get('overwrite', False):
            logger.info('Preparing input: %s already exists' % output_dir)
            return
        elif not os.path.exists(output_dir):
            os.makedirs(output_dir)

        all_nodes = dynamic_network.get_all_nodes()
        n_nodes = len(all_nodes)

        for i_sn, sn in enumerate(dynamic_network.get_snapshots()):
            output_file = output_dir + '%s.txt' % i_sn
            with open(output_file, 'wb') as out_f:
                out_f.write('%s %s\n' % (n_nodes, n_nodes))
                for u in all_nodes:
                    for v in all_nodes:
                        out_f.write('%s ' % (
                            1 if sn.get_nx_graph().has_edge(u, v) or u == v else 0
                        ))
                    out_f.write('\n')

    def run(self, dynamic_network, **kwargs):
        self._run_time = None
        self.set_dynamic_network(dynamic_network)
        logger.info("Running '{method_name}' method for '{network_id}'".format(
                method_name=self.get_title(),
                network_id=dynamic_network.get_id(),
            )
        )

        logger.info("Preparing input")
        GraphScope._prepare_snapshots(dynamic_network, output_dir=self.get_output_dir() + 'input/', **kwargs)

        if not os.path.exists(self.get_output_dir() + 'results/'):
            os.makedirs(self.get_output_dir() + 'results/')

        start = timeit.default_timer()

        stats_file = self.get_output_dir() +  'stats.txt'
        if not os.path.exists(stats_file) or kwargs.get('overwrite', True):
            logger.info("Running GraphScope executable")
            run_command = "{exe} {input}".format(
                exe=GraphScope.GraphScope_EXEC,
                input=self.get_output_dir(),
            )
            logger.debug("Run command: %s" % run_command)
            subprocess.call(run_command, shell=True)
        else:
            logger.info("Executable has been already run - using previous results")
            self._clean_run = False

        logger.debug("Processing output files")
        changepoints = []
        t_scores = []
        prev_seg_t = 0
        with open(stats_file, 'r') as infile:
            for l in infile.readlines()[1:]:
                values = l.replace('\n', '').split(',')
                curr_t = int(values[0])
                seg_t = int(values[1])
                if seg_t > prev_seg_t:
                    changepoints.append(curr_t)
                    prev_seg_t = seg_t
                t_scores.append(float(values[4]))
        self._t_scores = t_scores

        segments_partitions = []
        all_nodes = dynamic_network.get_all_nodes()
        for i in changepoints + [len(t_scores)]:
            # use the last partition from the segment as segment partition
            p_file = self.get_output_dir() + 'results/%s.rows.plist' % (i - 1)
            objects_clusters = {}
            with open(p_file, 'r') as outfile:
                for j, l in enumerate(outfile.readlines()):
                    objects_clusters[all_nodes[j]] = l.replace('\n', '')
            segments_partitions.append(
                Partition(objects_clusters)
            )

        self._community_evolution = community_evolution.CommunityEvolution(
            segments_partitions=community_evolution.CommunityEvolution
                .rename_clusters_as_ints_per_unique_adjacent_cluster(
                    segments_partitions,
            ),
            changepoints=changepoints,
            method_id=self.get_id(),
            dynamic_network=self.get_dynamic_network(),
        )
        logger.info(self._community_evolution.get_summary())

        end = timeit.default_timer()
        self._run_time = end - start
        logger.info("Finished in %s" % utils.sec_to_time(self._run_time))