import os
import timeit
import subprocess
from ... import settings
from ... import utils
from ...log import logger
from ...method import Method
from ...partition import Partition
from ...dynamic_network import DynamicNetwork
from ... import community_evolution

class MultistepMethod(Method):
    Multistep_comDyn_EXEC = settings.BASE_DIR + 'dynamic_communities_cpp/other_methods/multistep/com_dyn' + ('' if utils.is_remote_server() else '.exe')
    Multistep_hierarch_EXEC = settings.BASE_DIR + 'dynamic_communities_cpp/other_methods/multistep/hierarch' + ('' if utils.is_remote_server() else '.exe')

    def __init__(self, method_id):
        Method.__init__(self, method_id)
        self._sim_threshold = float(MultistepMethod._parse_method_id(method_id)[0])

    @staticmethod
    def get_method_title(method_id):
        # return MultistepMethod._parse_method_id(method_id)[0]
        return "Multi-Step, s=%s" % float(MultistepMethod._parse_method_id(method_id)[0])

    @staticmethod
    def _parse_method_id(method_id):
        values = method_id.split(settings.METHOD_ID_SEPARATOR)
        if values[0] != MultistepMethod.get_id_prefix():
            raise Exception("Cannot create MultistepMethod instance from method_id '%s'" % method_id)
        return values[1:]

    @staticmethod
    def get_method_description(method_id):
        return "This is Multi-Step method with similarity threshold %s" % float(MultistepMethod._parse_method_id(method_id)[0])

    def get_id(self):
        return self._method_id

    @staticmethod
    def get_id_prefix():
        return 'multistep'

    def get_solution_for_n_changepoints(self, n_changepoints, **kwargs):
        if self._sol_n_changepoints[n_changepoints] is None:
            n_snapshots = len(self.get_dynamic_network().get_snapshots())
            logger.info("Computing solution with %s changepoints" % n_changepoints)
            changepoints = sorted(set(range(1, n_snapshots)) - set(self._t_merge_timepoints_merge_order[:n_snapshots - n_changepoints - 1]))
            changepoints_with_start_end = [0] + changepoints + [n_snapshots]
            segments_intervals = []
            for i in xrange(1, len(changepoints_with_start_end)):
                segments_intervals.append((changepoints_with_start_end[i-1], changepoints_with_start_end[i] - 1))
            segments_partitions = []
            for s in segments_intervals:
                segments_partitions.append(
                    self._compute_consensus_partition(
                        t_start=s[0],
                        t_end=s[1],
                        **kwargs
                    )
                )

            self._sol_n_changepoints[n_changepoints] = {
                'changepoints': changepoints,
                'segments_partitions': segments_partitions,
            }
        return community_evolution.CommunityEvolution(
            segments_partitions=community_evolution.CommunityEvolution
                .rename_clusters_as_ints_per_unique_adjacent_cluster(
                self._sol_n_changepoints[n_changepoints]['segments_partitions'],
            ),
            changepoints=self._sol_n_changepoints[n_changepoints]['changepoints'],
            method_id=self.get_id(),
            dynamic_network=self.get_dynamic_network(),
        )
    def get_timepoints_scores(self):
        """
        Get list of scores describing how changepoint-like each timepoint is
        :return:
        """
        if not hasattr(self, "_timepoints_scores") or self._timepoints_scores is None:
            n_timepoints = len(self._t_similarity_scores)
            res = [0.0] * n_timepoints
            for i,t in enumerate(reversed(self._t_merge_timepoints_merge_order)):
                res[t] = (n_timepoints - i)  / (1.0 * n_timepoints)
            self._timepoints_scores = res
        return self._timepoints_scores

    def get_community_evolution(self):
        # note: sim threshold 3 means Multi-Step^star
        if self._sim_threshold == 3.0 and self.get_dynamic_network().get_segment_ground_truth() is not None:
            return self.get_solution_for_n_changepoints(len(self.get_dynamic_network().get_segment_ground_truth().get_changepoints()))
        return self._community_evolution

    def get_run_time(self):
        return self._run_time

    def get_shared_dir(self):
        """
        Get directory shared between different instances of the method for a given dynamic network
        :return:
        """
        return self.get_dynamic_network().get_dir() + 'methods/_shared/multistep/'

    @staticmethod
    def _prepare_input(networks_list, **kwargs):
        base_dir = kwargs.get('base_dir', settings.TMP_DIR + 'avg_mod/' + utils.id_generator(5) + '/')
        base_output_filename = kwargs.get('base_output_filename', utils.id_generator(5))
        output_file = base_dir + 'input/' + base_output_filename + '.txt'

        logger.info("Preparing input")

        if os.path.exists(output_file) and not kwargs.get('overwrite', False):
            logger.info('Preparing input: %s already exists' % output_file)
            return output_file
        elif not os.path.exists(os.path.dirname(output_file)):
            os.makedirs(os.path.dirname(output_file))

        with open(output_file,'wb') as f:
            #number of snapshots
            f.write('%s\n' % len(networks_list))

            #node list
            f.writelines(['%s\n' % n1 for n1 in DynamicNetwork.get_networks_list_all_nodes(networks_list)])

            #snapshots edge lists
            for sn in networks_list:
                f.writelines(['-2 1\n'] + ['%s %s\n' % (e[0], e[1]) for e in sn.get_nx_graph().edges_iter()])
        return output_file

    @staticmethod
    def compute_consensus_partition(networks_list, **kwargs):
        base_dir = kwargs.get('base_dir', settings.TMP_DIR + '_avg_mod/' + utils.id_generator(10) + '/')
        base_output_filename = kwargs.get('base_output_filename','%s_%s' % (len(networks_list), utils.id_generator(8)))
        com_dyn_output_file = base_dir + 'com_dyn/%s.txt' % base_output_filename

        if not os.path.exists(com_dyn_output_file) or kwargs.get('overwrite', False):
            if not os.path.exists(os.path.dirname(com_dyn_output_file)):
                os.makedirs(os.path.dirname(com_dyn_output_file))
            if 'base_dir' in kwargs:
                kwargs.pop('base_dir')
            if 'base_output_filename' in kwargs:
                kwargs.pop('base_output_filename')

            input_file = MultistepMethod._prepare_input(
                networks_list,
                output_dir=base_dir + 'input/',
                base_dir=base_dir,
                base_output_filename=base_output_filename,
                **kwargs
            )
            logger.info("Running com_dyn executable")
            run_command = '{exe} "{input}" > "{output}"'.format(
                exe=MultistepMethod.Multistep_comDyn_EXEC,
                input=input_file,
                output=com_dyn_output_file,
            )
            logger.debug("Run command: %s" % run_command)
            subprocess.call(run_command, shell=True)
        else:
            logger.info("Executable has been already run - using previous results")
            if kwargs.get('is_clean_run'):
                kwargs['is_clean_run'][0] = False

        return Partition.load(com_dyn_output_file)

    def _compute_consensus_partition(self, t_start, t_end, **kwargs):
        logger.info("Computing consensus clustering for [%s,%s]" % (t_start, t_end))
        # mutable object to check whether there was a clean run (i.e., without file reuse)
        is_clean_run = [True]
        res = MultistepMethod.compute_consensus_partition(
            networks_list=self.get_dynamic_network().get_snapshots()[t_start:t_end+1],
            base_dir=self.get_shared_dir(),
            base_output_filename='%s_%s' % (t_start, t_end),
            is_clean_run=is_clean_run,
            **kwargs
        )
        self._clean_run = self._clean_run and is_clean_run[0]
        return res

    def run(self, dynamic_network, **kwargs):
        self._run_time = None
        self.set_dynamic_network(dynamic_network)
        logger.info("Running '{method_name}' method for '{network_id}'".format(
                method_name=self.get_title(),
                network_id=dynamic_network.get_id(),
            )
        )

        logger.info("Preparing input")
        input_file = MultistepMethod._prepare_input(
            dynamic_network.get_snapshots(),
            base_dir=self.get_shared_dir(),
            base_output_filename='%s_%s' % (dynamic_network.get_timepoints()[0], dynamic_network.get_timepoints()[-1]),
            **kwargs
        )

        if not os.path.exists(self.get_output_dir()):
            os.makedirs(self.get_output_dir())

        start = timeit.default_timer()

        hierarch_output_file = self.get_shared_dir() + 'hierarch_output.txt'
        if not os.path.exists(hierarch_output_file) or kwargs.get('overwrite', True):
            logger.info("Running hierarch executable")
            # Note: use min sim threshold, to make sure get scores for all timepoints
            run_command = "{exe} {input} -sim_threshold '-2.0' --connex > {output}".format(
                exe=MultistepMethod.Multistep_hierarch_EXEC,
                input=input_file,
                output=hierarch_output_file,
            )
            logger.debug("Run command: %s" % run_command)
            subprocess.call(run_command, shell=True)
        else:
            logger.info("Executable has been already run - using previous results")
            self._clean_run = False

        logger.info("Converting output tree hierarchy to newick format")
        leave_encoding = {}
        tree_newick = ''
        n_snapshots = len(self.get_dynamic_network().get_snapshots())
        t_similarity_scores = [-2.0] * n_snapshots
        t_sim_scores_merge_order = []
        t_merge_timepoints_merge_order = []
        with open(hierarch_output_file,'r') as f:
            for line in f.readlines():
                line = line.strip().replace('\n','')

                # 'similarity' line goes before 'merge' line and contains the corresponding similarity value
                if line.startswith('similarity'):
                    similarity = float(line.replace('similarity=',''))
                    continue

                leaves = line.split(' ')
                if len(leaves)!=2:
                    raise Exception("Incorrect output line: %s" % line)

                l1_points = sorted(map(int,leaves[0][:-1].split(',')))
                l2_points = sorted(map(int,leaves[1][:-1].split(',')))

                # compute timepoint "changepoint" score
                curr_point = l2_points[0] if  l1_points[-1] < l2_points[0] else l1_points[0]
                t_similarity_scores[curr_point] = similarity
                t_sim_scores_merge_order.append(similarity)
                t_merge_timepoints_merge_order.append(curr_point)

                l1_key = ':'.join(map(str,l1_points))
                l1 = leave_encoding.get(l1_key,leaves[0][:-1])

                l2_key = ':'.join(map(str,l2_points))
                l2 = leave_encoding.get(l2_key,leaves[1][:-1])
                tree_newick = '(%s:%s,%s:%s)' % (l1,similarity,l2,similarity)

                line_key = ':'.join(sorted(line.replace(' ','')[:-1].split(',')))
                leave_encoding[line_key] = tree_newick

        with open(self.get_output_dir() + 't_sim.txt','wb') as f_similarities:
            for timepoint,score in enumerate(t_similarity_scores):
                f_similarities.write('%s,%s\n' % (timepoint, score))
        self._t_similarity_scores = t_similarity_scores
        self._t_sim_scores_merge_order = t_sim_scores_merge_order
        self._t_merge_timepoints_merge_order = t_merge_timepoints_merge_order
        self._tree_newick = tree_newick

        # computing solutions for different numbers of changepoints
        self._sol_n_changepoints = [None] * n_snapshots
        if kwargs.get('precompute_solutions_n_changepoints', True):
            for i in xrange(len(self._t_sim_scores_merge_order) + 1): # from n_snapshots-1 to 0 changepoints
                self.get_solution_for_n_changepoints(
                    n_changepoints=i,
                    **kwargs
                )

        n_sol_changepoints = n_snapshots - 1
        # once merge similarity is smaller or equal to threshold, merging is stopped => can get number of changepoints
        # as we have _t_sim_scores_merge_order, we can also get merging timepoints
        for i in xrange(n_snapshots-1):
            if t_sim_scores_merge_order[i] <= self._sim_threshold:
                break
            n_sol_changepoints -= 1
        # n_sol_changepoints = sum(1 for i in range(n_snapshots) if t_similarity_scores[i] <= self._sim_threshold) - 1
        self._community_evolution = self.get_solution_for_n_changepoints(
            n_changepoints=n_sol_changepoints,
            **kwargs
        )
        logger.info(self._community_evolution.get_summary())

        end = timeit.default_timer()
        self._run_time = end - start
        logger.info("Finished in %s" % utils.sec_to_time(self._run_time))
