import os
import networkx as nx
import numpy as np
import igraph
from log import logger

class Network():

    NETWORK_PROPERTIES = {
        'nodes_count': "Number of nodes",
        'edges_count': "Number of edges",
        'avg_degree': "Average node degree",
        'min_degree': "Minimum node degree",
        'max_degree': "Maximum node degree",
        'avg_clust_coeff': "Average clustering coefficient",
        'density': "Density",
        'triangles_count': "Number of triangles",
        'diameter': "Diameter",
        'connected_components_count': "Number of connected components",
        'radius': "Radius",
    }

    def __init__(self, nx_graph):
        self._nx_graph = nx_graph
        self._cache_properties = dict()

    def get_nx_graph(self):
        return self._nx_graph

    def is_directed(self):
        return isinstance(self.get_nx_graph(), nx.DiGraph)

    def get_nodes(self):
        return self.get_nx_graph().nodes()

    def get_edges(self):
        return self.get_nx_graph().edges()

    def get_nodes_count(self):
        """
        Get the number of nodes in the network.
        """
        return self.get_nx_graph().order()

    def get_edges_count(self):
        """
        Get the number of edges in the network.
        """
        return self.get_nx_graph().size()

    def get_avg_degree(self):
       if self.get_nodes_count()==0 or self.get_edges_count()==0:
        return 0

       return (2*float(self.get_nx_graph().size()))/self.get_nx_graph().order()

    def get_max_degree(self):
        if not self._cache_properties.has_key('max_degree'):
            degree_sequence=nx.degree(self.get_nx_graph()).values()
            if len(degree_sequence)>0:
                self._cache_properties['max_degree'] = max(degree_sequence)
            else:
                self._cache_properties['max_degree'] = 0

        return self._cache_properties['max_degree']

    def get_min_degree(self):
        if not self._cache_properties.has_key('min_degree'):
            degree_sequence=nx.degree(self.get_nx_graph()).values()
            if len(degree_sequence)>0:
                self._cache_properties['min_degree'] = min(degree_sequence)
            else:
                self._cache_properties['min_degree'] = 0

        return self._cache_properties['min_degree']


    def get_connected_components(self):
         return list(nx.connected_component_subgraphs(self.get_nx_graph()))

    def get_connected_components_count(self):
        return nx.number_connected_components(self.get_nx_graph())

    def get_largest_connected_component(self):
        if not self._cache_properties.has_key('largest_connected_component'):
            connected_components = self.get_connected_components()
            self._cache_properties['largest_connected_component'] = connected_components[0]

        return self._cache_properties['largest_connected_component']

    def get_avg_clust_coeff(self):
        if self.get_nodes_count()==0 or self.get_edges_count()==0:
            return 0

        if not self._cache_properties.has_key('avg_clust_coeff'):
            self._cache_properties['avg_clust_coeff'] = nx.average_clustering(self.get_nx_graph())

        return self._cache_properties['avg_clust_coeff']

    def get_diameter(self):
        if self.get_nodes_count()==0 or self.get_edges_count()==0:
            return 0

        if not self._cache_properties.has_key('diameter'):
            self._cache_properties['diameter'] = nx.diameter(self.get_largest_connected_component())

        return self._cache_properties['diameter']

    def get_radius(self):
        return nx.radius(self.get_largest_connected_component())

    def get_density(self):
        return nx.density(self.get_nx_graph())

    def get_triangles_count(self):
        return sum(nx.triangles(self.get_nx_graph()).values())

    def get_network_property(self, property_name):
        if property_name not in self.NETWORK_PROPERTIES:
            raise Exception("Unknown network property %s" % property_name)
        property_method = getattr(self, 'get_' + property_name)
        return property_method()


    def get_description(self):
        if not self._cache_properties.has_key('description'):
            descr =  "\nNodes: %d, edges: %d" % (self.get_nodes_count(),self.get_edges_count())
            descr += "\nDegree - Avg : %d, Max: %d, Min: %d" % (self.get_avg_degree(),self.get_max_degree(),self.get_min_degree())
            descr += "\nAverage clustering coefficient: %.2f" % self.get_avg_clust_coeff()
            descr += "\nConnected components: %d " % len(self.get_connected_components())
            descr += "\nLargest component - Size: %.1f%%" % (100.0 * self.get_largest_connected_component().order()/self.get_nodes_count())

            self._cache_properties['description'] = descr

        return self._cache_properties['description']

    def as_igraph(self):
        return Network.nx_graph_as_igraph(self.get_nx_graph())

    @staticmethod
    def nx_graph_as_igraph(nx_graph):
        iG = igraph.Graph()
        iG.add_vertices(nx_graph.nodes())
        for e in nx_graph.edges(data=True):
            if "weight" in e[2]:
                iG.add_edge(e[0], e[1], weight = float(e[2]["weight"]))
            else:
                iG.add_edge(e[0], e[1])
        return iG

    @staticmethod
    def graph_union(networks_list, weighted=False, networks_weights=None):
        G = nx.Graph()
        if len(networks_list)<1:
            return None
        elif len(networks_list)==1:
            return networks_list[0]
        for i, network in enumerate(networks_list):
            n_weight = 1.0 if networks_weights is None else networks_weights[i]
            for edge in network.get_nx_graph().edges_iter(data='weight', default=1.0):
                if weighted:
                    prev_weight = G[edge[0]][edge[1]]["weight"] if G.has_edge(edge[0], edge[1]) else 0.0
                    G.add_edge(edge[0], edge[1], weight=prev_weight + n_weight * edge[2])
                else:
                    G.add_edge(edge[0], edge[1])

        return Network(G)

    @staticmethod
    def get_common_edges(network_list):
        """
        Get list of common edges for all networks in the list
        :param network_list: List of the networks to check
        :return: List of the common edges
        """
        if len(network_list)<1:
            return None
        elif len(network_list)==1:
            return network_list[0].edges()

        common_edge_list = []
        #traverse each edge of the first network and check if it is common for all networks
        for edge in network_list[0].edges():
            is_common_edge = True
            for network in network_list[1:]:
                if not network.has_edge(edge[0],edge[1]):
                    is_common_edge = False
                    break

            if is_common_edge:
                common_edge_list.append(edge)

        return common_edge_list

    @staticmethod
    def get_union_edges(network_list):
        """
        Get union of edges for all networks in the list
        :param network_list: List of the networks to check
        :return: Union edges
        """
        if len(network_list)<1:
            return None
        elif len(network_list)==1:
            return network_list[0].edges()

        union_edge_list = []
        for network in network_list:
            union_edge_list += network.edges()


        return list(set(union_edge_list))

    @staticmethod
    def load(input_path, read_method="edgelist", **kwargs):
        if read_method == "edgelist":
            return Network(nx.read_edgelist(input_path, **kwargs))
        raise Exception("Unknown read method")

    def save(self, output_dir, output_filename, write_method = 'edgelist', save_data=True):
        logger.debug(
            "Saving network to {output_path} using {write_method} method".format(
                output_path=output_dir + output_filename,
                write_method=write_method,
            )
        )
        if not write_method in [
            'edgelist','weighted_edgelist','adjlist','multiline_adjlist',
            'gexf','gml','gpickle','graphml','yaml','pajek','shp', 'adjmatrix'
            ]:
            raise Exception

        #create output directory if it does not exists
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)


        #create write function
        if write_method != 'adjmatrix':
            write_method_func = getattr(nx,'write_' + write_method)
            if not save_data and write_method == 'edgelist':
                write_method_func(
                    self.get_nx_graph(),
                    output_dir + str(output_filename) + '.txt',
                    data=False
                )
            else:
                write_method_func(
                    self.get_nx_graph(),
                    output_dir + str(output_filename) + '.txt',
                )
                if save_data and write_method != 'edgelist':
                    logger.warning("Write method %s does not support writing edge data" % write_method)
        else:
            np.savetxt(output_dir + str(output_filename) + '.txt',
                nx.adj_matrix(self.get_nx_graph(),
                nodelist=self.get_nx_graph().nodes(data=False)),
                fmt='%f'
            )