import random
import itertools
import os
import sklearn.metrics as metrics
import sklearn.metrics.cluster.supervised as supervised
from log import logger
import utils
import numpy as np

class Partition(object):
    #clustering distance metrics - variation of information and information distance
    SIMILARITY_METRICS = {
        'nmi': "Normalized Mutual Information",
        'ami': "Adjusted for chance Mutual Information",
        'ari': "Adjusted for chance Rand index",
        'com': "Completeness",
        'hom': "Homogeneity",
        'vm': "V-measure",
    }

    def __init__(self, objects_clusters):
        if isinstance(objects_clusters, list):
            objects_clusters = dict(zip(range(0,len(objects_clusters)), objects_clusters))
        self._objects_clusters = objects_clusters
        self._clear_cache()

    @staticmethod
    def from_clusters_objects(clusters_objects):
        objects_clusters = dict()
        for cluster, objects in clusters_objects.iteritems():
            for o in objects:
                objects_clusters[o] = cluster
        return Partition(objects_clusters)

    @staticmethod
    def from_list_of_sets(list_of_sets):
        objects_clusters = dict()
        for i_cl, cl in enumerate(list_of_sets):
            for obj in cl:
                objects_clusters[obj] = i_cl
        return Partition(objects_clusters)

    def _clear_cache(self):
        self._clusters_objects = None
        self._objects_count = None
        self._clusters_count = None

    def get_objects_clusters(self):
        return self._objects_clusters

    def get_object_cluster(self, obj):
        return self.get_objects_clusters().get(obj)

    def get_clusters_objects(self):
        if self._clusters_objects is None:
            self._clusters_objects = {}
            for object, cluster in self.get_objects_clusters().iteritems():
                if not self._clusters_objects.has_key(cluster):
                    self._clusters_objects[cluster]=set()
                self._clusters_objects[cluster].add(object)
        return self._clusters_objects

    def get_cluster_objects(self, c):
        return self.get_clusters_objects().get(c)

    def get_objects_count(self):
        if self._objects_count is None:
            self._objects_count =  len(self.get_objects_clusters().keys())
        return self._objects_count

    def get_clusters_count(self):
        if self._clusters_count is None:
            self._clusters_count = len(set(self.get_objects_clusters().values()))
        return self._clusters_count

    def get_clusters_names(self):
        return sorted(self.get_clusters_objects().keys())

    def get_clusters_sizes(self):
        return [len(objs) for c,objs in self.get_clusters_objects().iteritems()]

    def get_objects_names(self):
        return sorted(self.get_objects_clusters().keys())

    def get_subpartition(self, objects_list):
        """
        Get subpartition covering only specified objects
        :param objects_list:
        :return:
        """
        new_objects_clusters = {}
        for o in objects_list:
            new_objects_clusters[o] = self.get_object_cluster(o)
        return Partition(new_objects_clusters)

    def rename_cluster(self, c_old, c_new):
        if c_new in self.get_clusters_names():
            raise Exception("Cluster '%s' already exists" % c_new)

        for o in self.get_objects_clusters():
            if c_old == self.get_object_cluster(o):
                self.get_objects_clusters()[o] = c_new
        self._clear_cache()

    @staticmethod
    def get_common_objects(p_1, p_2):
        """
        Get common objects for two partitions
        :param p_1:
        :param p_2:
        :return:
        """
        return utils.get_intersection(p_1.get_objects_names(), p_2.get_objects_names())

    def to_string(self, record_separator="\n"):
        res = "%s\n" % self.get_objects_count()
        for el,cl in self.get_objects_clusters().iteritems():
            res += "%s,%s%s" % (el, cl, record_separator)
        return res


    def to_array_form(self):
        """
        Get partition in array form: partition_array[i] - cluster of ith object, objects_names[i] - name of ith object
        :return:
        """
        objects_names = sorted(self.get_objects_clusters().keys())
        partition_array = []
        for o in objects_names:
            partition_array.append(self.get_object_cluster(o))
        return partition_array, objects_names

    def to_list_of_sets(self):
        """
        Get partition in list of lists form [set(cl_1_obj_1, cl_1_obj_2,...),set(cl_2_obj_1,...),...]
        :return:
        """
        res = []
        for cl_objs in self.get_clusters_objects().values():
            res.append(cl_objs)
        return res

    @staticmethod
    def load(input_path, **kwargs):
        with open(input_path, 'r') as input_f:
            lines = input_f.readlines()
            return Partition.from_lines(
                lines=lines,
                i_start=0,
                value_separator=kwargs.get('value_separator', ' '),
                n_records=len(lines),
                **kwargs
            )

    @staticmethod
    def from_string(input_string, **kwargs):
        lines = input_string.split(kwargs.get('record_separator'))
        return Partition.from_lines(lines, 0, **kwargs)

    @staticmethod
    def from_lines(lines, i_start=0, **kwargs):
        data = dict()
        n = kwargs.get('n_records')
        if n is None:
            n = int(lines[i_start])

        convert_objects_to_ints = kwargs.get('convert_objects_to_ints', False)
        convert_clusters_to_ints = kwargs.get('convert_clusters_to_ints', False)
        convert_objects_names_dict = kwargs.get('convert_objects_names_dict')
        for i in xrange(i_start,i_start + n):
            values = lines[i if kwargs.get('n_records') else i + 1].strip().split(kwargs.get('value_separator', ','))
            object_id = values[0] if not convert_objects_to_ints else int(values[0])
            if convert_objects_names_dict is not None:
                object_id = convert_objects_names_dict[object_id]
            cluster_id = values[1] if not convert_clusters_to_ints else int(values[1])
            data[object_id] = cluster_id

        return Partition(data)



    def save(self, output_dir, output_filename):
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)
        with open(output_dir + output_filename + '.txt','wb') as f:
            f.write(self.to_string())

    @staticmethod
    def get_similarity_metric_limits(metric=None):
        return (0, 1)

    @staticmethod
    def _constrained_sum_sample_pos(n, sum_value):
        """
        Return a randomly chosen list of n positive integers summing to sum_value.
        Each such list is equally likely to occur.
        """
        dividers = sorted(random.sample(xrange(1, sum_value), n - 1))
        return [a - b for a, b in zip(dividers + [sum_value], [0] + dividers)]

    @staticmethod
    def _split_list(lst, min_size, split_size):
        it = iter(lst)

        size = len(lst)
        for i in range(split_size - 1,0,-1):
            s = random.randint(min_size, size -  min_size * i)
            yield list(itertools.islice(it,0,s))
            size -= s
        yield list(it)

    @staticmethod
    def generate_random_partition(n, cluster_count=None, cluster_sizes=None):
        """
        Generate random partition of n elements
        :param n: Number of elements
        :param cluster_count: Number of clusters
        :param cluster_sizes: Distribution of cluster sizes
        """
        data = []
        if cluster_sizes is None:
            if cluster_count is None:
                cluster_count = random.randint(1,n)

            cluster_sizes = []  # Partition._constrained_sum_sample_pos(n, cluster_count)
            for i in xrange(cluster_count-1):
                cluster_sizes.append(n/cluster_count)
            cluster_sizes.append(n - sum(cluster_sizes))

        current_cluster = 0
        for i in cluster_sizes:
            data += [current_cluster] * i
            current_cluster+=1

        random.shuffle(data)
        return Partition(data)


    @staticmethod
    def compute_partitions_similarity(partition_true, partition_predicted, metric='nmi', use_common_objects=False):
        """
         Calculate similarity between two partitions
        :param partition_true: true labels
        :param partition_predicted: predicted labels
        :param metric:
        :param use_common_objects: Check whether objects names of partitions are the same;
        if not, use their intersection
        :return:
        """
        if metric not in Partition.SIMILARITY_METRICS:
            raise Exception("Unknown metric {metric}, should be one of {metrics}".format(
                    metric=metric,
                    metrics=Partition.SIMILARITY_METRICS.keys()
                )
            )

        if partition_true.get_objects_names() != partition_predicted.get_objects_names():
            common_objects = Partition.get_common_objects(partition_true, partition_predicted)
            logger.warning("Partitions do not cover the same object set "
                           "({size_1} vs {size_2}) - using intersection ({size_inter} objects)".format(
                size_1 = partition_true.get_objects_count(),
                size_2 = partition_predicted.get_objects_count(),
                size_inter = len(common_objects)
            ))
            partition_true = partition_true.get_subpartition(common_objects)
            partition_predicted = partition_predicted.get_subpartition(common_objects)

        p1_array, p1_objects_names = partition_true.to_array_form()
        p2_array, p2_objects_names = partition_predicted.to_array_form()
        res = None
        if metric == 'mi':
            res = metrics.mutual_info_score(p1_array, p2_array)
        elif metric == 'nmi':
            res = metrics.normalized_mutual_info_score(p1_array, p2_array)
        elif metric == 'ami':
            res = metrics.adjusted_mutual_info_score(p1_array, p2_array)
        elif metric == 'ari':
            res = metrics.adjusted_rand_score(p1_array, p2_array)
        elif metric == 'com':
            res = metrics.completeness_score(p1_array, p2_array)
        elif metric == 'hom':
            res = metrics.homogeneity_score(p1_array, p2_array)
        elif metric == 'vm':
            res = metrics.v_measure_score(p1_array, p2_array)

        return res
