import preprocess_utils
import datetime
import bisect
from .. import settings
from ..community_evolution import CommunityEvolution
from ..log import logger
from ..partition import Partition
from collections import defaultdict
import csv
import os
import numpy as np
from ..affinity_matrix import AffinityMatrix
from .. import utils
import networkx as nx
from .. import static_clustering
from ..network import Network

input_raw_dir = settings.NETWORKS_RAW_DATA_DIR + 'amd_hope/'

def create_co_location_network(min_norm_time_in_area, delta_t, threshold, persons_set):
    # get co-location info
    co_location_path = input_raw_dir + 'co_location_%s.txt' % (str(min_norm_time_in_area).replace('.', ''))
    if not os.path.exists(co_location_path):
        co_loc = defaultdict(list)
        with open(input_raw_dir + 'snapshot_summary.csv', 'r') as infile:
            header = {}
            reader = csv.reader(infile)
            for i_row, values in enumerate(reader):
                if i_row == 0:
                    header = dict(zip(values, range(len(values))))
                    continue
                time_period = values[header['TIME_PERIOD']]
                tag_id = values[header['TAG_ID']]
                area_id = values[header['AREA_ID']]
                normalized_time_in_area = float(values[header['NORMALIZED_TIME_IN_AREA']])

                if tag_id not in persons_set:
                    continue

                if normalized_time_in_area < min_norm_time_in_area:
                    continue

                # only date, no time
                if not ' ' in time_period:
                    continue

                co_loc['%s-%s' % (time_period, area_id)].append(tag_id)

        # create dynamic network file
        with open(co_location_path, 'wb') as outfile:
            for time_loc_id, tags in co_loc.iteritems():
                time_period, loc = time_loc_id.split('-')
                if len(tags) < 2:
                    continue
                for i in xrange(len(tags)):
                    for j in xrange(i + 1, len(tags)):
                        outfile.write(
                            '%s,%s,%s\n'
                            % (time_period, tags[i], tags[j])
                        )

    network_sub_name = '%s_%s_%s/' % (min_norm_time_in_area, delta_t, threshold)
    network_id = 'amd_hope/coloc/' + network_sub_name
    network_dir = settings.NETWORKS_BASE_DIR + network_id + '/'
    timepoints_labels = preprocess_utils.split_into_snapshots(
        input_temporal_network_file=co_location_path,
        output_dir_snapshots=network_dir + 'snapshots/',
        delta_t=delta_t,
        threshold=threshold,
        delimiter=',',
        datetime_format='%m/%d/%Y %I:%M:%S %p',
        min_t='7/18/2008 2:00:00 PM',
        as_directed_graph=False,
        as_weighted_graph=True,
        min_nodes_snapshot=5,
        min_edges_snapshot=5,
        save_snapshots_names_ordered=True,
    )

    return network_id, network_dir, timepoints_labels


def create_co_attendance_network(persons_talks, talks_attendees, talks_time, min_person_talks,
                                 max_talk_persons, delta_t, threshold,  persons_set, max_person_talks=-1,
                                 min_talk_persons=0):
    logger.info("Average talk attendance: %s" % np.average(map(len, talks_attendees.values())))
    logger.info("Average person talks: %s" % np.average(map(len, persons_talks.values())))
    persons_to_ignore = set()
    new_persons_talks = {}
    for person_id, talks in persons_talks.iteritems():
        if len(persons_talks[person_id]) < min_person_talks or \
                (max_person_talks > 0 and len(persons_talks[person_id]) > max_person_talks):
            logger.warning("Person %s attended either too few or too many talks: %s" % (person_id, len(talks)))
            persons_to_ignore.add(person_id)
        else:
            new_persons_talks[person_id] = talks[:]
    n_persons_active = len(persons_talks) - len(persons_to_ignore)
    logger.info("%s persons attended a feasible amount of talks" % n_persons_active)
    persons_talks = new_persons_talks

    talks_to_ignore = set()
    for talk_id, talk_time in talks_attendees.iteritems():
        attendees = talks_attendees[talk_id]
        if len(attendees) < min_talk_persons or (max_talk_persons > 0 and len(attendees) > max_talk_persons):
            logger.warning("Either too few people or too many persons attended talk %s: %s" % (talk_id, len(attendees)))
            talks_to_ignore.add(talk_id)
    logger.info("%s talks have a feasible attendance" % (len(talks_attendees) - len(talks_to_ignore)))

    # create dynamic network file
    co_attendance_path = input_raw_dir + '/attend_%s_%s_%s.txt' % (min_person_talks, min_talk_persons, max_talk_persons)
    if not os.path.exists(co_attendance_path):
        with open(co_attendance_path, 'wb') as outfile:
            for talk_id, talk_time in talks_time.iteritems():
                if talk_id in talks_to_ignore:
                    continue
                attendees = sorted(p for p in talks_attendees[talk_id])
                filtered_attendees = [p for p in attendees if p not in persons_to_ignore and p in persons_set]
                if len(filtered_attendees) < len(attendees):
                    logger.debug("Talk %s: %s persons with unfeasible attendance removed" % (
                    talk_id, len(attendees) - len(filtered_attendees)))
                for i in xrange(len(filtered_attendees)):
                    for j in xrange(i + 1, len(filtered_attendees)):
                        outfile.write(
                            '%s,%s,%s\n'
                            % (talk_time.strftime('%Y-%m-%d %H:%M:%S'), filtered_attendees[i], filtered_attendees[j])
                        )

    network_sub_name = '%s_%s_%s_%s_%s/' % (min_person_talks, min_talk_persons, max_talk_persons, delta_t, threshold)
    network_id = 'amd_hope/attend/' + network_sub_name
    network_dir = settings.NETWORKS_BASE_DIR + network_id + '/'
    timepoints_labels = preprocess_utils.split_into_snapshots(
        input_temporal_network_file=co_attendance_path,
        output_dir_snapshots=network_dir + 'snapshots/',
        delta_t=delta_t,
        threshold=threshold,
        delimiter=',',
        datetime_format='%Y-%m-%d %H:%M:%S',
        min_t='2008-07-18 14:00:00',
        as_directed_graph=False,
        as_weighted_graph=True,
        min_nodes_snapshot=5,
        min_edges_snapshot=5,
        save_snapshots_names_ordered=True,
    )

    return network_id, network_dir, timepoints_labels


def create_dynamic_network(network_type='coloc', delta_t='1week', threshold=1, **kwargs):
    # get talk attendance info
    talks_attendees = defaultdict(list)
    persons_talks = defaultdict(list)
    talks_time = {}
    times_talks = defaultdict(list)
    tracks_talks = defaultdict(list)
    talks_track = {}
    persons_tracks = {} # for each person, how many times s/he attended a given track
    with open(input_raw_dir + 'talk_presence.csv', 'r') as infile:
        header = {}
        reader = csv.reader(infile)
        for i_row, values in enumerate(reader):
            if i_row == 0:
                header = dict(zip(values, range(len(values))))
                continue
            talk_id = values[header['TALK_ID']]
            person_id = values[header['PERSON_ID']]
            track = values[header['TRACK']]

            talks_attendees[talk_id].append(person_id)
            persons_talks[person_id].append(talk_id)
            tracks_talks[track].append(talk_id)
            talks_track[talk_id] = track

            if person_id not in persons_tracks:
                persons_tracks[person_id] = {}
            if track not in persons_tracks[person_id]:
                persons_tracks[person_id][track] = 0
            persons_tracks[person_id][track] += 1

            if talk_id not in talks_time:
                # manual fix
                dt_string = values[header['TALK_TIME']]\
                    if values[header['TALK_TIME']] != '7/19/2008 11:59:00 PM'\
                    else '7/20/2008 12:00:00 AM'
                talk_time = datetime.datetime.strptime(dt_string, '%m/%d/%Y %I:%M:%S %p')
                talks_time[talk_id] = talk_time
                times_talks[talk_time].append(talk_id)
    times_of_talks = times_talks.keys()

    # interests - ground truth, need to get list of users with avilable info
    interests_names = {}
    with open(input_raw_dir + 'interests_list.csv', 'r') as infile:
        header = {}
        reader = csv.reader(infile)
        for i_row, values in enumerate(reader):
            if i_row == 0:
                header = dict(zip(values, range(len(values))))
                continue
            interest_id = values[header['INTEREST_ID']]
            interest_name = values[header['INTEREST_NAME']]
            interests_names[interest_id] = interest_name.strip().replace(' ', '_').lower()

    interests_persons = defaultdict(list)
    persons_interests = defaultdict(list)
    node_labels = {}
    with open(input_raw_dir + 'interests.csv', 'r') as infile:
        header = {}
        reader = csv.reader(infile)
        for i_row, values in enumerate(reader):
            if i_row == 0:
                header = dict(zip(values, range(len(values))))
                continue
            person_id = values[header['ID']]
            interest_id = values[header['INTEREST']]
            if person_id in node_labels:
                node_labels[person_id] = "-".join([node_labels[person_id], interests_names[interest_id]])
            else:
                node_labels[person_id] = interests_names[interest_id]
            persons_interests[person_id].append(interest_id)
            interests_persons[interest_id].append(person_id)
    logger.info("Have interests info for %s persons" % len(persons_interests))
    persons_list = sorted(persons_interests.keys())
    persons_set = set(persons_list)

    if network_type == 'attend':
        network_id, network_dir, timepoints_labels = create_co_attendance_network(
            persons_talks=persons_talks,
            talks_attendees=talks_attendees,
            talks_time=talks_time,
            min_person_talks=kwargs.get('min_person_talks'),
            min_talk_persons=kwargs.get('min_talk_persons'),
            max_talk_persons=kwargs.get('max_talk_persons'),
            delta_t=delta_t,
            threshold=threshold,
            persons_set=persons_set,
        )
    else:
        network_id, network_dir, timepoints_labels = create_co_location_network(
            min_norm_time_in_area=kwargs.get('min_norm_time_in_area'),
            delta_t=delta_t,
            threshold=threshold,
            persons_set=persons_set,
        )



    timepoints_datetimes = preprocess_utils.convert_timepoint_labels_to_dt(timepoints_labels, delta_t)

    talk_partitions = []
    for i in xrange(len(timepoints_datetimes)):
        talk_partitions.append(Partition({}))
    for talk_time,talk_ids in times_talks.iteritems():
        sn = bisect.bisect_left(timepoints_datetimes, talk_time)
        if 0 <= sn < len(timepoints_datetimes):
            if talk_time < timepoints_datetimes[sn]:
                sn -= 1
                if sn < 0:
                    continue
            objects_clusters = {}
            for i_talk, talk_id in enumerate(talk_ids):
                for person_id in talks_attendees[talk_id]:
                    if person_id not in persons_set:
                        continue
                    objects_clusters[person_id] = i_talk
            talk_partitions[sn] = Partition(objects_clusters)
    c_ev = CommunityEvolution(
        segments_partitions=talk_partitions,
        changepoints=range(1,len(timepoints_labels)),
        method_id='gt_talks',
        network_id=network_id,
        start_timepoint=0,
        end_timepoint=len(timepoints_labels)-1,
    )
    c_ev.save(as_ground_truth=True)

    # partition ground truths
    gt_partitions = {}

    # interests
    # affinities = []
    # for i in xrange(len(persons_list)):
    #     for j in xrange(i+1, len(persons_list)):
    #         affinities.append(
    #             (
    #                 persons_list[i],
    #                 persons_list[j],
    #                 len(utils.get_intersection(persons_interests[persons_list[i]], persons_interests[persons_list[j]]))
    #             )
    #         )
    # gt_partitions['interest'] = AffinityMatrix.from_affinities_list(affinities)
    #
    # # interests graph
    # g_interests = nx.MultiGraph()
    # for interest, persons in interests_persons.iteritems():
    #     for i in xrange(len(persons)):
    #         for j in xrange(i+1, len(persons)):
    #             g_interests.add_edge(persons[i], persons[j])
    #
    # # edge (i,j) - number of shared interests between i and j
    # g = preprocess_utils.convert_multidigraph_to_graph(g_interests, as_weighted_graph=True)
    # gt_partitions['interest_graph'] = static_clustering.compute_partition(
    #     network=Network(g),
    #     method='louvain',
    # )

    # countries and genders
    counties_names = {}
    with open(input_raw_dir + 'countries.csv', 'r') as infile:
        header = {}
        reader = csv.reader(infile)
        for i_row, values in enumerate(reader):
            if i_row == 0:
                header = dict(zip(values, range(len(values))))
                continue
            country_id = values[header['ID']]
            country_name = values[header['COUNTRY']]
            counties_names[country_id] = country_name.strip().replace(' ','_').lower()
    genders_names = {"1": "M", "2": "F"}
    persons_country = {}
    persons_gender = {}
    with open(input_raw_dir + 'person.csv', 'r') as infile:
        header = {}
        reader = csv.reader(infile)
        for i_row, values in enumerate(reader):
            if i_row == 0:
                header = dict(zip(values, range(len(values))))
                continue
            if person_id not in persons_set:
                continue
            person_id = values[header['ID']]
            country_id = values[header['COUNTRY']]
            gender_id = values[header['SEX']]
            if country_id != '' and country_id != '0':
                persons_country[person_id] = country_id
                if person_id in node_labels:
                    node_labels[person_id] = "--".join([node_labels[person_id], counties_names[country_id]])
                else:
                    node_labels[person_id] = counties_names[country_id]
            if gender_id != '':
                persons_gender[person_id] = gender_id
                if person_id in node_labels:
                    node_labels[person_id] = "--".join([node_labels[person_id],genders_names[gender_id]])
                else:
                    node_labels[person_id] = genders_names[gender_id]
    gt_partitions['country'] = Partition(persons_country)
    gt_partitions['gender'] = Partition(persons_gender)

    for gt_title, gt_partition in gt_partitions.iteritems():
        c_ev = CommunityEvolution(
            segments_partitions=[gt_partition],
            changepoints=None,
            method_id='gt_%s' % gt_title,
            network_id=network_id,
            start_timepoint=0,
            end_timepoint=len(timepoints_labels)-1,
        )
        c_ev.save(as_ground_truth=True)

    # load changepoints - events-only ground truth
    c_ev = CommunityEvolution(
        segments_partitions=None,
        changepoints=preprocess_utils.load_changepoints(
            settings.NETWORKS_RAW_DATA_DIR + 'amd_hope/events.txt', timepoints_datetimes,
            datetime_format='%m/%d/%Y %I:%M %p',
        ),
        method_id='gt_events',
        network_id=network_id,
        start_timepoint=0,
        end_timepoint=len(timepoints_labels)-1,
    )
    c_ev.save(as_ground_truth=True)


    # save node labels
    preprocess_utils.save_node_labels(network_dir, node_labels)

    # save description
    with open(settings.NETWORKS_BASE_DIR + network_id + '/description.txt', 'wb') as outfile:
        outfile.write(
            "This is AMD_HOPE talk-co-attendance network. The raw data was obtained from here https://networkdata.ics.uci.edu/data.php?id=110. "
            "Each snapshot contains data of {delta_t} and was constructed with threshold {threshold}".format(
                delta_t=delta_t,
                threshold=threshold
            )
        )
