# create networks for Bilateral Trade
import os
from collections import defaultdict
import pandas as pd
import numpy as np
from .. import settings
from ..community_evolution import CommunityEvolution
from ..partition import Partition
from ..affinity_matrix import AffinityMatrix
import preprocess_utils
import bisect
import datetime

NETWORK_RAW_DATA_DIR = settings.NETWORKS_RAW_DATA_DIR + "correlates_of_war/"

NETWORK_TYPE_RAW_DATA_DIR = {
    'bt': NETWORK_RAW_DATA_DIR + "bilateral_trade/",
    'fa': NETWORK_RAW_DATA_DIR + "formal_alliances/",
    'mid': NETWORK_RAW_DATA_DIR + "militarized_interstate_disputes/",
}

NETWORK_TYPE_TITLE = {
    'bt': "Bilateral Trade",
    'fa': "Formal Alliances",
    'mid': "Militarized Interstate Dispputes",
}

NETWORK_SUBTYPE_TITLE = {
    'fa': {
        'all': "All alliances",
        'def': "Only defense pacts",
    }
}

NETWORK_TYPE_RANGE = {
    'bt': (1870, 2009),
    'fa': (1810, 2012),
    'mid': (1810, 2010),
}

def get_bilateral_trade_edges(min_year=1870, max_year=2009):
    snapshots_edges = defaultdict(set)

    df = pd.read_csv(os.path.join(NETWORK_TYPE_RAW_DATA_DIR['bt'], "dyadic_trade_3.0.csv"))
    for index, record in df.iterrows():
        year = int(record["year"])
        if pd.isnull(record["importer1"]) or pd.isnull(record["importer2"]):
            continue
        country_code_1 = record["ccode1"]
        country_code_2 = record["ccode2"]
        # country_1 = record["importer1"].strip().replace(" ", "_").replace("-", "_").replace(".", "").lower()
        # country_2 = record["importer2"].strip().replace(" ", "_").replace("-", "_").replace(".", "").lower()
        flow_1 = float(record["flow1"])
        flow_2 = float(record["flow2"])
        if flow_1 >= 0:
            snapshots_edges[year].add((country_code_1, country_code_2, flow_1))
        if flow_2 >= 0:
            snapshots_edges[year].add((country_code_2, country_code_1, flow_2))

    return snapshots_edges


def formal_alliances(network_sub_type='all', min_year = 1810, max_year = 2012):
    snapshots_edges = defaultdict(set)

    df = pd.read_csv(os.path.join(NETWORK_TYPE_RAW_DATA_DIR['fa'], "alliance_v4.1_by_dyad.csv"))
    for index, record in df.iterrows():
        is_defense = record["defense"]
        if network_sub_type == 'def' and not is_defense:
            continue
        start_year = int(record["dyad_st_year"])
        end_year = int(record["dyad_end_year"]) if not np.isnan(record["dyad_end_year"]) else 2012

        start_year = max(start_year, min_year) if min_year is not None else start_year
        end_year = min(end_year, max_year) if max_year is not None else end_year

        years = range(start_year, end_year + 1)
        country_code_1 = record["ccode1"]
        country_code_2 = record["ccode2"]
        # state_name_1 = record["state_name1"].strip().replace(" ", "_").replace("-", "_").replace(".", "").lower()
        # state_name_2 = record["state_name2"].strip().replace(" ", "_").replace("-", "_").replace(".", "").lower()
        for y in years:
            snapshots_edges[y].add((country_code_1, country_code_2))

    return snapshots_edges


def militarized_interstate_disputes(min_year = 1810, max_year = 2010, positive_relationship=True):
    snapshots_edges = defaultdict(set)

    midb_df = pd.read_csv(os.path.join(NETWORK_TYPE_RAW_DATA_DIR['mid'], "mid_level/MIDB_4.01.csv"))
    grouped_by_dispute = midb_df.groupby("DispNum3")

    def create_inter_group_edges(side_a, side_b):
        for index_a, p_a in side_a.iterrows():
            start_year_a = p_a["StYear"]
            end_year_a = p_a["EndYear"]
            years_a = range(start_year_a, end_year_a + 1)
            country_code_a = p_a["code"]
            # country_iso3_a = p_a["StAbb"]
            for index_b, p_b in side_b.iterrows():
                country_code_b = p_b["code"]
                # country_iso3_b = p_b["StAbb"]
                if country_code_a == country_code_b:
                    continue
                start_year_b = p_b["StYear"]
                end_year_b = p_b["EndYear"]
                years_b = range(start_year_b, end_year_b + 1)
                # when p_a and p_b where both participating in the dispute
                for y in set(years_a).intersection(years_b):
                    snapshots_edges[y].add((country_code_a, country_code_b))

    for name, group in grouped_by_dispute:
        side_a = group[group["SideA"]==1]
        side_b = group[group["SideA"]==0]

        if positive_relationship:
            create_inter_group_edges(side_a, side_a)
            create_inter_group_edges(side_b, side_b)
        else:
            create_inter_group_edges(side_a, side_b)

    return snapshots_edges

def create_militarized_interstate_disputes_affinities(timepoints_datetimes,
                                                      same_side=True,
                                                      min_year = 1810, max_year = 2010,
                                                      min_hostility_level=-10):
    snapshots_affinities = []
    snapshots_disputes = []
    for i in xrange(len(timepoints_datetimes)):
        snapshots_affinities.append([])
        snapshots_disputes.append(defaultdict(int))

    midb_df = pd.read_csv(os.path.join(NETWORK_TYPE_RAW_DATA_DIR['mid'], "mid_level/MIDB_4.01.csv"))
    grouped_by_dispute = midb_df.groupby("DispNum3")

    def create_disputes_counts(side_a, side_b):
        for index_a, p_a in side_a.iterrows():
            country_code_a = p_a["ccode"]
            # country_iso3_a = p_a["StAbb"]
            start_year_a = p_a["StYear"]
            end_year_a = p_a["EndYear"]
            years_a = range(start_year_a, end_year_a + 1)
            host_level_a = p_a["HostLev"]
            for index_b, p_b in side_b.iterrows():
                country_code_b = p_b["ccode"]
                # country_iso3_b = p_b["StAbb"]
                if country_code_a == country_code_b:
                    continue
                start_year_b = p_b["StYear"]
                end_year_b = p_b["EndYear"]
                years_b = range(start_year_b, end_year_b + 1)
                host_level_b = p_b["HostLev"]
                if host_level_a < min_hostility_level and host_level_b < min_hostility_level:
                    continue
                # when p_a and p_b where both participating in the dispute
                for y in set(years_a).intersection(years_b):
                    sn = bisect.bisect_left(timepoints_datetimes, datetime.datetime.strptime(str(y), "%Y"))
                    if len(timepoints_datetimes) > sn and y < timepoints_datetimes[sn].year:
                        sn -= 1
                    if 0 < sn < len(timepoints_datetimes):
                        id_str = '%s-%s' % (min(country_code_a, country_code_b), max(country_code_a, country_code_b))
                        snapshots_disputes[sn][id_str] += 1

    for name, group in grouped_by_dispute:
        side_a = group[group["SideA"]==1]
        side_b = group[group["SideA"]==0]
        if same_side:
            create_disputes_counts(side_a, side_a)
            create_disputes_counts(side_b, side_b)
        else:
            create_disputes_counts(side_a, side_b)

    for i in xrange(len(timepoints_datetimes)):
        for id_str, count in snapshots_disputes[i].iteritems():
            country_code_a, country_code_b = id_str.split('-')
            snapshots_affinities[i].append((country_code_a, country_code_b, count))

    return snapshots_affinities

def get_country_code_dicts():
    country_code_to_name = {}
    country_code_to_iso3 = {}
    df = pd.read_csv(os.path.join(NETWORK_RAW_DATA_DIR, "COW country codes.csv"))
    for index, record in df.iterrows():
        code = record["CCode"]
        name = record["StateNme"].strip().lower()\
            .replace(" ", "_").replace("'", "").replace("&", "and").replace(".", "")
        iso3 = record["StateAbb"].lower()
        country_code_to_name[code] = name
        country_code_to_iso3[code] = iso3
    return country_code_to_name,country_code_to_iso3


def save_snapshots_edges(snapshots_edges, output_path):
    with open(output_path, 'wb') as outfile:
        for sn, edges in snapshots_edges.iteritems():
            for edge in edges:
                outfile.write(
                    "%s\t%s\n" % (sn, "\t".join(map(str, edge)))
                )


def create_dynamic_network(network_type, network_sub_type=None, delta_t='1year', threshold=1, **kwargs):
    # snapshots_edges - dict year => edge set
    if network_type == 'bt':
        snapshots_edges = get_bilateral_trade_edges()
    elif network_type == 'fa':
        snapshots_edges =  formal_alliances(network_sub_type, **kwargs)
    elif network_type == 'mid':
        snapshots_edges =  militarized_interstate_disputes()
    else:
        raise NotImplementedError()

    temporal_network_path = NETWORK_TYPE_RAW_DATA_DIR[network_type] + network_type + ('' if network_sub_type is None else '_%s' % network_sub_type) + '.txt'
    save_snapshots_edges(
        snapshots_edges,
        temporal_network_path
    )

    network_id = 'cow/' + network_type + '/'
    if network_sub_type is not None:
        network_id += network_sub_type + '_'
    network_id += '%s_%s' % (delta_t, threshold)
    if kwargs.get('min_year'):
        network_id += '_%s' % kwargs.get('min_year')
    if kwargs.get('max_year'):
        network_id += '_%s' % kwargs.get('max_year')
    network_dir = settings.NETWORKS_BASE_DIR + network_id + '/'

    timepoints_labels = preprocess_utils.split_into_snapshots(
        input_temporal_network_file=temporal_network_path,
        output_dir_snapshots=network_dir + 'snapshots/',
        delta_t=delta_t,
        threshold=threshold,
        delimiter='\t',
        datetime_format='%Y',
        min_t=str(min(snapshots_edges.keys())),
        get_weights_from_input_file=True,
        as_directed_graph=False,
        as_weighted_graph=True,
        min_nodes_snapshot=5,
        min_edges_snapshot=5,
        save_snapshots_names_ordered=True,
    )
    timepoints_datetimes = preprocess_utils.convert_timepoint_labels_to_dt(timepoints_labels, delta_t)

    country_code_to_name,country_code_to_iso3 = get_country_code_dicts()
    iso3_to_country_code = {}
    for code,iso3 in country_code_to_iso3.iteritems():
        iso3_to_country_code[iso3] = code

    # partition ground truth
    regions_objects_clusters = {}
    with open(NETWORK_RAW_DATA_DIR + 'regions.txt', 'r') as infile:
        curr_region = ''
        for line in infile.readlines():
            if len(line.strip()) == 0:
                continue
            values = line.lower().strip().split(" ")
            if values[0] == '-': # new region start
                curr_region = values[-1].replace("(", "").replace(")", "")
                continue
            iso3 = values[-1].replace("[", "").replace("]", "")
            regions_objects_clusters[iso3_to_country_code[iso3]] = curr_region

    c_ev = CommunityEvolution(
        segments_partitions=[Partition(regions_objects_clusters)],
        changepoints=None,
        method_id='gt_region',
        network_id=network_id,
        start_timepoint=0,
        end_timepoint=len(timepoints_labels)-1,
    )
    c_ev.save(as_ground_truth=True)

    # disputes ground truth
    for same_side in [True, False]:
        snapshots_affinities = create_militarized_interstate_disputes_affinities(
            timepoints_datetimes, same_side=same_side, min_hostility_level=3,
            min_year=kwargs.get('min_year'),
            max_year=kwargs.get('max_year'),
        )
        c_ev = CommunityEvolution(
            segments_partitions=[AffinityMatrix.from_affinities_list(snapshots_affinities[i], is_distance_matrix=not same_side) for i in xrange(len(snapshots_affinities))],
            changepoints=range(1,len(timepoints_labels)),
            method_id='gt_dispute%s' % ("_same" if same_side else ""),
            network_id=network_id,
            start_timepoint=0,
            end_timepoint=len(timepoints_labels)-1,
        )
        c_ev.save(as_ground_truth=True)

    # load changepoints - events-only ground truth
    # if os.path.exists(NETWORK_TYPE_RAW_DATA_DIR[network_type] + 'events.txt'):
    #     c_ev = CommunityEvolution(
    #         segments_partitions=None,
    #         changepoints=preprocess_utils.load_changepoints(
    #             NETWORK_TYPE_RAW_DATA_DIR[network_type] + 'events.txt', timepoints_datetimes,
    #             datetime_format='%Y',
    #         ),
    #         method_id='gt_events',
    #         network_id=network_id,
    #         start_timepoint=0,
    #         end_timepoint=len(timepoints_labels)-1,
    #     )
    #     c_ev.save(as_ground_truth=True)

    # save node labels
    preprocess_utils.save_node_labels(network_dir, country_code_to_name)

    # save description
    with open(settings.NETWORKS_BASE_DIR + network_id + '/description.txt', 'wb') as outfile:
        outfile.write(
            "This is Correlates of War '{title}'{subtype} network. The raw data was obtained from here http://www.correlatesofwar.org/. "
            "Each snapshot contains data of {delta_t} and was constructed with threshold {threshold}".format(
                title=NETWORK_TYPE_TITLE[network_type],
                subtype=("" if network_sub_type is None else " (%s)" % NETWORK_SUBTYPE_TITLE[network_type][network_sub_type]),
                delta_t=delta_t,
                threshold=threshold
            )
        )