import preprocess_utils
import datetime
from .. import settings
import bisect
from ..community_evolution import CommunityEvolution
from ..log import logger
from ..partition import Partition
from collections import defaultdict

def create_dynamic_network(delta_t='1week', threshold=1):
    network_sub_name = '%s_%s' % (delta_t, threshold)
    network_id = 'enron/' + network_sub_name
    network_dir = settings.NETWORKS_BASE_DIR + network_id + '/'

    timepoints_labels = preprocess_utils.split_into_snapshots(
        input_temporal_network_file=settings.NETWORKS_RAW_DATA_DIR + '/enron/enron.txt',
        output_dir_snapshots=network_dir + 'snapshots/',
        delta_t=delta_t,
        threshold=threshold,
        delimiter=' ',
        min_t=946684800, # from 2000-01-01
        # max_t=1017619200, # to 2002-04-01
        as_directed_graph=False,
        as_weighted_graph=True,
        min_nodes_snapshot=5,
        min_edges_snapshot=5,
        save_snapshots_names_ordered=True,
    )
    timepoints_datetimes = preprocess_utils.convert_timepoint_labels_to_dt(timepoints_labels, delta_t)

    # load changepoints - events-only ground truth
    c_ev = CommunityEvolution(
        segments_partitions=None,
        changepoints=preprocess_utils.load_changepoints(
            settings.NETWORKS_RAW_DATA_DIR + 'enron/events.txt', timepoints_datetimes
        ),
        method_id='gt_events',
        network_id=network_id,
        start_timepoint=0,
        end_timepoint=len(timepoints_labels)-1,
    )
    c_ev.save(as_ground_truth=True)


    # load partitions - partition-only ground truth
    node_labels = {}
    with open(settings.NETWORKS_RAW_DATA_DIR + 'enron/employees.txt', 'r') as infile:
        n_id = -1
        dep_partition_cl_objs = defaultdict(list)
        pos_partition_cl_objs = defaultdict(list)
        for line in infile.readlines():
            if line.startswith('#'):
                continue
            n_id += 1
            values = line.strip().split('\t')
            values = [v.strip().lower().replace(" ", "_") for v in values]

            node_labels[n_id] = "-".join(values)

            # partition by positions
            if len(values) > 2:
                pos = values[2]
                if pos == "":
                    pos = "n/a"
                elif pos == "managing_director__legal_department":
                    pos = "managing_director"
                if pos != "n/a" :
                    pos_partition_cl_objs[pos].append(n_id)

            # partition by departments
            if len(values) > 3:
                dep = values[3]
                if dep == "" and len(values) > 4:
                    dep = values[4]
                if dep == "":
                    dep = "n/a"
                if "risk" in dep:
                    dep = "risk"
                elif "legal" in dep:
                    dep = "legal"
                elif "logistics" in dep:
                    dep = "logistics"
                elif "enery" in dep:
                    dep = "energy"
                elif "government" in dep:
                    dep = "government"
                elif "global" in dep:
                    dep = "global"
                elif "wholesale" in dep:
                    dep = "wholesale"
                elif "analyst" in dep:
                    dep = "analyst"
                elif "specialist" in dep:
                    dep = "n/a"
                if dep != "n/a":
                    dep_partition_cl_objs[dep].append(n_id)

        for gt_id, p_cl_objs in {'dep': dep_partition_cl_objs, 'pos': pos_partition_cl_objs}.iteritems():
            logger.debug("Partition ground truth %s" % gt_id)
            filtered_p_cl_objs = {}
            for cl,objs in p_cl_objs.iteritems():
                if len(objs) < 2:
                    logger.warning("Cluster %s is too small: %s" % (cl, objs))
                    continue
                filtered_p_cl_objs[cl] = objs
                logger.debug("%s: %s" % (cl, objs))
            c_ev = CommunityEvolution(
                segments_partitions=[Partition.from_clusters_objects(p_cl_objs)],
                changepoints=None,
                method_id='gt_%s' % gt_id,
                network_id=network_id,
                start_timepoint=0,
                end_timepoint=len(timepoints_labels)-1,
            )
            c_ev.save(as_ground_truth=True)

    # save node labels
    preprocess_utils.save_node_labels(network_dir, node_labels)

    # save description
    with open(network_dir + 'description.txt', 'wb') as outfile:
        outfile.write(
            "This is Enron network. The raw data was obtained from here http://cis.jhu.edu/~parky/Enron/enron.html. "
            "Each snapshot contains data of {delta_t} and was constructed with threshold {threshold}".format(
                delta_t=delta_t,
                threshold=threshold
            )
        )

