import os
import re
import getopt
import sys
import bisect
import networkx as nx
import numpy as np
from .. import settings
from ..log import logger
import datetime
from collections import defaultdict
from dateutil.relativedelta import relativedelta
from dateutil.tz import tzlocal

def diff_month(d1, d2):
    return (d1.year - d2.year)*12 + d1.month - d2.month


def set_to_year_start(dt):
    return set_to_month_start(dt).replace(month=1)

def set_to_month_start(dt):
    return set_to_day_start(dt).replace(day=1)

def set_to_week_start(dt):
    return set_to_day_start(dt - datetime.timedelta(days=dt.weekday()))

def set_to_day_start(dt):
    midnight = datetime.time(0)
    return datetime.datetime.combine(dt.date(), midnight)

def set_to_hour_start(dt):
    hour_start = datetime.time(hour=dt.hour)
    return datetime.datetime.combine(dt.date(), hour_start)

def set_to_minute_start(dt):
    minute_start = datetime.time(hour=dt.hour, minute=dt.minute)
    return datetime.datetime.combine(dt.date(), minute_start)

def reset_dt_trailing(dt, delta_t_unit_type):
    if delta_t_unit_type=='year':
        return set_to_year_start(dt)
    elif delta_t_unit_type=='month':
        return set_to_month_start(dt)
    elif delta_t_unit_type == 'week':
        return set_to_week_start(dt)
    elif delta_t_unit_type == 'day':
        return set_to_day_start(dt)
    elif delta_t_unit_type == 'hour':
        return set_to_hour_start(dt)
    elif delta_t_unit_type == 'min':
        return set_to_minute_start(dt)
    else:
        raise NotImplementedError()

def get_timedelta_hours(td):
    return td.days * 24 + td.seconds // 3600

def get_timedelta_minutes(td):
    return td.days * 24 * 60 + td.seconds // 60

TIME_UNIT_SNAPSHOT_NAME_FORMAT = {
    'year': '%Y',
    'month': '%Y-%m',
    'week': '%Y-%m-%d',
    'day': '%Y-%m-%d',
    'hour': '%Y-%m-%d-%H',
    'min': '%Y-%m-%d-%H-%M',
}

def get_time_unit_type_from_delta_t(delta_t):
    delta_t_units = int(re.findall(r'\d+', delta_t)[0])
    if 'year' in delta_t:
        delta_t_unit_type = 'year'
    elif 'month' in delta_t:
        delta_t_unit_type = 'month'
    elif 'week' in delta_t:
        delta_t_unit_type = 'week'
    elif 'day' in delta_t:
        delta_t_unit_type = 'day'
    elif 'hour' in delta_t:
        delta_t_unit_type = 'hour'
    elif 'min' in delta_t:
        delta_t_unit_type = 'min'
    else:
        raise NotImplementedError() 
    return delta_t_units, delta_t_unit_type

def convert_timepoint_labels_to_dt(timepoints_labels, delta_t):
    timepoints_datetimes = []
    for t in timepoints_labels:
        _,time_unit_type = get_time_unit_type_from_delta_t(delta_t)
        timepoints_datetimes.append(
            datetime.datetime.strptime(
                t,
                TIME_UNIT_SNAPSHOT_NAME_FORMAT.get(time_unit_type)
            )
        )
    return timepoints_datetimes

def convert_multidigraph_to_graph(g_multi, threshold=1, as_directed_graph=False, as_weighted_graph=True):
    if not as_weighted_graph:
        return nx.Graph(g_multi) if not as_directed_graph else nx.DiGraph(g_multi)
    g = nx.Graph()  if not as_directed_graph else nx.DiGraph()
    for u,v,data in g_multi.edges_iter(data=True):
        w = data.get('weight', 1)
        if g.has_edge(u,v):
            g[u][v]['weight'] += w
        else:
            g.add_edge(u, v, weight=w)

    if threshold>1:
        tempNodeLabels = g.nodes()
        adjMatr = nx.to_numpy_matrix(g)
        newAdjMatr = np.maximum(adjMatr  - (threshold - 1) * np.ones(adjMatr.shape), np.zeros(adjMatr.shape))
        g = nx.relabel_nodes(nx.from_numpy_matrix(newAdjMatr), dict(zip(range(0, g.order()), tempNodeLabels)))
    return g

def save_node_labels(output_dir, node_labels, output_file_name = 'nodes_labels.txt'):
    with open(output_dir + output_file_name, 'wb') as outfile:
        for node,label in node_labels.iteritems():
            outfile.write('%s,%s\n' % (node,label))


def load_changepoints(input_file, timepoints_datetimes, datetime_format=None):
    changepoints = []
    with open(input_file, 'r') as infile:
        for l in infile.readlines():
            if l.startswith('#'):
                continue
            values = l.strip().split('#')
            ev_time_str = values[0]
            descr = '#'.join(values[1:])
            if datetime_format is None:
                if ev_time_str.count('-') == 1: # per-months events
                    ev_time = datetime.datetime.strptime(values[0], '%m-%Y')
                else:
                    ev_time = datetime.datetime.strptime(values[0], '%d-%m-%Y')
            else:
                ev_time = datetime.datetime.strptime(values[0], datetime_format)
            cp = bisect.bisect_left(timepoints_datetimes, ev_time)  # note: it will be one time point after the actual event
            if 0 < cp < len(timepoints_datetimes):
                if (ev_time-timepoints_datetimes[cp-1]) < (timepoints_datetimes[cp] - ev_time):
                    cp -=1
                changepoints.append(cp)
                logger.debug("%s - %s" % (cp, descr))
    return sorted(set(changepoints))


def save_snapshots_names_ordered(output_dir, snapshots_names):
    with open(output_dir + '/snapshots_names_ordered.txt', 'wb') as outfile:
        for i, sn in enumerate(snapshots_names):
            outfile.write("%s,%s\n" % (i, sn))

def split_into_snapshots(input_temporal_network_file, output_dir_snapshots, delta_t, threshold = 1,
                         delimiter=' ', **kwargs):

    min_t = kwargs.get('min_t', 0)
    max_t = kwargs.get('max_t', None)
    datetime_format = kwargs.get('datetime_format', None)
    min_edges_snapshot = kwargs.get('min_edges_snapshot', None)
    min_nodes_snapshot = kwargs.get('min_nodes_snapshot', None)
    t_convert_func = kwargs.get('t_convert_func', None)
    get_weights_from_input_file = kwargs.get('get_weights_from_input_file', False)

    delta_t_units, delta_t_unit_type = get_time_unit_type_from_delta_t(delta_t)

    #if output directory does not exists, create it
    if not os.path.exists(output_dir_snapshots):
        os.makedirs(output_dir_snapshots)

    snapshots = defaultdict(nx.MultiDiGraph)

    if t_convert_func is not None:
        min_t = t_convert_func(min_t)
    else:
        min_t = datetime.datetime.fromtimestamp(min_t) if datetime_format is None else datetime.datetime.strptime(min_t, datetime_format)

    min_snapshot = reset_dt_trailing(min_t, delta_t_unit_type)

    if max_t is not None:
        max_t = datetime.datetime.fromtimestamp(max_t) if datetime_format is None else datetime.datetime.strptime(max_t, datetime_format)
    with open(input_temporal_network_file, "r") as f:
      for line in f:
        values = line.strip().split(delimiter)
        t = values[0]
        u = values[1]
        v = values[2]
        w = float(values[3]) if (get_weights_from_input_file and len(values)>3) else None

        if t_convert_func is not None:
            t = t_convert_func(t)
        else:
            t = datetime.datetime.fromtimestamp(int(t)) if datetime_format is None else datetime.datetime.strptime(t, datetime_format)
        if (min_t is not None and t < min_t) or (max_t is not None and t > max_t):
            continue

        i_snapshot = -1
        if delta_t_unit_type=='year':
            i_snapshot = (t.year - min_snapshot.year) / delta_t_units
        elif delta_t_unit_type=='month':
            i_snapshot = diff_month(set_to_month_start(t), min_snapshot) / delta_t_units
        elif delta_t_unit_type == 'week':
            i_snapshot = (set_to_week_start(t) - min_snapshot).days / (7*delta_t_units)
        elif delta_t_unit_type == 'day':
            i_snapshot = (set_to_day_start(t) - min_snapshot).days / delta_t_units
        elif delta_t_unit_type == 'hour':
            i_snapshot = get_timedelta_hours(set_to_hour_start(t) - min_snapshot) / delta_t_units
        elif delta_t_unit_type == 'min':
            i_snapshot = get_timedelta_minutes(set_to_minute_start(t) - min_snapshot) / delta_t_units

        if w is not None:
            snapshots[i_snapshot].add_edge(u,v, weight=w)
        else:
            snapshots[i_snapshot].add_edge(u,v)

    snapshots_names = []
    for i_snapshot in snapshots.keys():
        if delta_t_unit_type=='year': # note: to handle dates before 1900, just take year as int
            snapshot_name = str((min_snapshot + relativedelta(years=delta_t_units * i_snapshot)).year)
        elif delta_t_unit_type=='month':
            snapshot_name = min_snapshot + relativedelta(months=delta_t_units * i_snapshot)
        elif delta_t_unit_type == 'week':
            snapshot_name = min_snapshot + relativedelta(weeks=delta_t_units * i_snapshot)
        elif delta_t_unit_type == 'day':
            snapshot_name = min_snapshot + relativedelta(days=delta_t_units * i_snapshot)
        elif delta_t_unit_type == 'hour':
            snapshot_name = min_snapshot + relativedelta(hours=delta_t_units * i_snapshot)
        elif delta_t_unit_type == 'min':
            snapshot_name = min_snapshot + relativedelta(minutes=delta_t_units * i_snapshot)
        else:
            snapshot_name = str(i_snapshot)

        if isinstance(snapshot_name, datetime.datetime):
            if kwargs.get('tz'):
                if snapshot_name.tzinfo == None:
                    snapshot_name = snapshot_name.replace(tzinfo=tzlocal())
                snapshot_name = snapshot_name.astimezone(kwargs.get('tz'))
            snapshot_name = snapshot_name.strftime(TIME_UNIT_SNAPSHOT_NAME_FORMAT[delta_t_unit_type])

        g = convert_multidigraph_to_graph(
            snapshots[i_snapshot],
            threshold=threshold,
            as_directed_graph=kwargs.get('as_directed_graph', False),
            as_weighted_graph=kwargs.get('as_weighted_graph', True),
        )
        logger.debug("Saving snapshot {i} - {name}: {nodes} nodes and {edges} edges".format(
            i=i_snapshot,
            name=snapshot_name,
            nodes=g.order(),
            edges=g.size()
        ))
        if min_edges_snapshot is not None and g.size() < min_edges_snapshot:
            logger.warning("Snapshot {i} does not have enough edges".format(i=i_snapshot))
            continue
        if min_nodes_snapshot is not None and g.order() < min_nodes_snapshot:
            logger.warning("Snapshot {i} does not have enough nodes".format(i=i_snapshot))
            continue

        nx.write_edgelist(
            g,
            output_dir_snapshots + snapshot_name + '.txt',
            data=True,
        )
        snapshots_names.append(snapshot_name)
    logger.debug("Finished saving snapshots: %s snapshots" % len(snapshots_names))
    if kwargs.get('save_snapshots_names_ordered', True):
        save_snapshots_names_ordered(
            output_dir=kwargs.get('dir_snapshots_names_ordered', os.path.dirname(os.path.dirname(output_dir_snapshots))),
            snapshots_names=snapshots_names
        )
    return snapshots_names

if __name__ == '__main__':
    inputSubDir = ''
    delimiter = ' '
    directed = False
    mintT0 = -1
    accumulating = False

    opts, args = getopt.getopt(sys.argv[1:], "s:d:m:ath", ["subdir", "help"])
    for o, a in opts:
        if o in ('-s', '--subdir'):
            inputSubDir = a  + '/'
        elif o == '-d':
            delimiter = a.strip()
            if delimiter == 'tab':
                delimiter = '\t'
        elif o == '-m':
            mintT0 = int(a)
        elif o =='--directed':
            directed = True
        elif o == '-a':
            accumulating = True
        if o in ('-h','--help'):
            print """Usage:
    %s network_name delta_t threshold
    -s,--subdir      Input/output subdirectory
    -d               Delimiter (use 'tab' for '\\t')
    -m               Starting time to consider
    -a               Accumulate edges across snapshots with time
    -h,--help        Print this"""  % sys.argv[0]
            sys.exit(0)

    networkName = args[0]
    deltaT = args[1]
    threshold = int(args[2])

    outputDir = settings.NETWORKS_BASE_DIR + inputSubDir + networkName + '/snapshots/snapshots_%s_%s%s/' % (deltaT, threshold, '_acc' if accumulating else '')

    split_into_snapshots(
            settings.NETWORKS_BASE_DIR + inputSubDir + networkName + '/' + networkName + '.txt',
            outputDir,
            deltaT,
            threshold,
            delimiter,
            min_t= mintT0,
    )