# code is based on https://github.com/j2kun/reality-mining
import scipy.io
from datetime import datetime, timedelta
import time
import sys, os
import itertools
import numpy
from collections import deque

from ..community_evolution import CommunityEvolution
from ..log import logger
from ..partition import Partition
from .. import settings
import preprocess_utils
from ..affinity_matrix import AffinityMatrix

RAW_DATA_DIR = settings.NETWORKS_RAW_DATA_DIR + 'reality_mining/'
#start_date = 1094601600 #First day of classes in 04-05 academic year
#end_date = 1133740800 #Last day of classes in 04-05 academic year

def hasNumeric(obj, field):
   try:
      obj[field][0][0]
      return True
   except:
      return False


def getNumeric(obj, field):
   return obj[field][0][0]

def getString(obj, field):
   return obj[field][0][0][0]

def hasString(obj, field):
   try:
      obj[field][0][0][0]
      return True
   except:
      return False

def hasArray(obj, field):
   try:
      obj[field][0]
      return True
   except:
      return False

def hasAttr(obj, field):
   try:
      obj[field]
      return True
   except:
      return False

def getArray(obj, field):
   return obj[field][0]


def validSubjects(allSubjects):
   return [s for s in allSubjects if hasNumeric(s,'mac') and hasNumeric(s,'my_hashedNumber')]


# idDicts: subjects -> {int: subject}, {float: (int, subject)}, {int: (int, subject)}
# First hash is contiguousId: subjectObject
# second hash is macAddress: contiguousId, subjectObject
# third hash is hashedNumber: contiguousId, subjectObject
# because the id dictionaries reference the subject object, we can replace
# the array of subject objects with these dictionaries.
def idDicts(subjects):
   return (dict((i, s) for (i,s) in enumerate(subjects)),
      dict((getNumeric(s,'mac'), (i, s)) for (i,s) in enumerate(subjects)),
      dict((getNumeric(s, 'my_hashedNumber'), (i, s)) for (i,s) in enumerate(subjects)))


def allCommEvents(idDictionary):
   events = []
   for subjectId, subject in idDictionary.items():
      if hasArray(subject, 'comm'):
         events.extend([(subjectId, event) for event in getArray(subject, 'comm')])

   logger.debug("%d total comm events" % len(events))
   return events


# extract those call events which are voice calls and only between
# two members of the study.
def callsWithinStudy(commEvents, hashNumDict):
   calls = [(subjectId, e) for (subjectId, e) in commEvents if getArray(e, 'description') == "Voice call"
                                             and getNumeric(e, 'hashNum') in hashNumDict]
   logger.debug("%d total calls within study" % len(calls))
   return calls


def convertDatetime(dt):
   return datetime.fromordinal(int(dt)) + timedelta(days=dt%1) - timedelta(days=366) - timedelta(hours=5)


def processCallEvents(callEvents, hashNumDict):
   processedCallEvents = []

   for subjectId, event in callEvents:
      direction = getArray(event, 'direction')
      duration = 0 if direction == 'Missed' else getNumeric(event, 'duration')
      date = convertDatetime(getNumeric(event, 'date'))
      hashNum = getNumeric(event, 'hashNum')
      otherPartyId = hashNumDict[hashNum][0]

      eventAsDict = {'subjectId': subjectId,
                      'direction': direction,
                      'duration': duration,
                      'otherPartyId': otherPartyId,
                      'date': date}
      processedCallEvents.append(eventAsDict)

   logger.debug("%d call event dictionaries" % len(processedCallEvents))
   return processedCallEvents


def inRange(dateRange, timevalue):
   start, end = dateRange
   unixTime = int(time.mktime(timevalue.timetuple()))
   return start <= unixTime <= end


def filterByDate(dateRange, events):
   filteredCalls = [e for e in events if inRange(dateRange, e['date'])]
   logger.debug("%d calls after filtering by date" % len(filteredCalls))
   return filteredCalls


def writeCallEvents(callEventDicts, filename):
   with open(filename, 'w') as outfile:
      outfile.write('subjectId\totherPartyId\tduration\tdirection\tdate\n')
      for d in callEventDicts:
         values = [d['subjectId'], d['otherPartyId'], d['duration'], d['direction'], d['date']]
         line = '\t'.join(("%s" % (v,)) for v in values)
         outfile.write('%s\n' % line)


def createPhoneCallDataset(idDictionaries):
   startDate = 1072915200  # use beginning of 2004, instead of start of classes 1095984000
   endDate = 1135987200 # use end of 2005, instead of end of classes 1105142400

   # this data contains the subject records as well
   idDict, macDict, hashNumDict = idDictionaries

   logger.debug("Extracting intra-study calls")
   calls = callsWithinStudy(allCommEvents(idDict), hashNumDict)

   logger.debug("Converting call events to a reasonable format")
   convertedCallEvents = processCallEvents(calls, hashNumDict)

   logger.debug("Filtering calls within the given date range")
   callsToWrite = filterByDate((startDate, endDate), convertedCallEvents)

   logger.debug("Writing the calls to reality-mining-calls.txt")
   writeCallEvents(callsToWrite, RAW_DATA_DIR + 'reality-mining-calls.txt')


# survey values are either numeric or numpy.nan, so we need special
# functions to account for means/maxes involving nan.
def mean(x, y):
   if numpy.isnan(x):
      return mean(0, y)
   if numpy.isnan(y):
      return mean(x, 0)

   return float(x + y) / 2


def myMax(x, y):
   if numpy.isnan(x):
      return myMax(0, y)
   if numpy.isnan(y):
      return myMax(x, 0)

   return max(x,y)


# added to extract research group affiliations
def createAffiliation(subjects, idDictionaries, affiliation_type = 'class'):
   type_to_field_dict = {
      'class': 'my_affil',
      'neighborhood': 'my_neighborhood',
      'research_group': 'my_group',
   }

   type_is_array_dict = {
      'class': True,
      'neighborhood': False,
      'research_group': False,
   }

   idDict, macDict, hashNumDict = idDictionaries

   logger.debug("Creating '%s' affiliation dataset" % affiliation_type)
   affiliations = []
   for i,s in idDict.iteritems():
      if not hasString(s, type_to_field_dict[affiliation_type]):
         logger.warning("No '%s' info for id %s" % (affiliation_type, i))
         continue
      affiliations.append(
         (i, getString(s,type_to_field_dict[affiliation_type]))
      )

   if len(affiliations) == 0:
      logger.warning("No affiliation data for '%s'" % affiliation_type)

   logger.debug("Writing the data")
   with open(RAW_DATA_DIR + 'reality-mining-%s.txt' % affiliation_type, 'w') as outfile:
      outfile.write('id\t%s\t\n' % affiliation_type)
      for values in affiliations:
         line = '\t'.join(map(str, values))
         outfile.write('%s\n' % line)


# For simplicity: take the avg of estimates, and the max of the friendship reporting
def getSurveyResponse(network, id1, id2):
   friends = myMax(network['friends'][id1][id2], network['friends'][id2][id1])
   inLabProximity = mean(network['lab'][id1][id2], network['lab'][id2][id1])
   outLabProximity = mean(network['outlab'][id1][id2], network['outlab'][id2][id1])

   return (id1, id2, friends, inLabProximity, outLabProximity)


def writeSurveyEvents(surveyRecords, filename):
   with open(filename, 'w') as outfile:
      outfile.write('id1\tid2\tclose-friends?\tinlab-proximity\toutlab-proximity\n')
      for values in surveyRecords:
         line = '\t'.join(("%s" % (v,)) for v in values)
         outfile.write('%s\n' % line)


def createFriendshipDataset(networkObj, idDictionaries):
   idDict, macDict, hashNumDict = idDictionaries

   networkIdDict = dict((i, (hashNum, hashNumDict[hashNum][0]))
         for i,hashNum in enumerate(getArray(networkObj, 'sub_sort'))
         if hashNum in hashNumDict) # this guarantees the subject is valid

   convertId = lambda i: networkIdDict[i][1]

   logger.debug("Creating network survey dataset (friendship/proximity/close friends)")
   networkSurvey = [getSurveyResponse(networkObj,i,j)
         for i,j in itertools.combinations(networkIdDict.keys(), 2)]

   logger.debug("Converting ids")
   convertedNetworkSurvey = [(convertId(x[0]), convertId(x[1]), x[2], x[3], x[4])
         for x in networkSurvey if x[2] != 0 or x[3] != 0 or x[4] != 0]

   logger.debug("Writing the survey data to reality-mining-survey.txt")
   writeSurveyEvents(convertedNetworkSurvey, RAW_DATA_DIR + 'reality-mining-survey.txt')


# turn each (date, tower) pair into a (dateInterval, tower) pair
# so we can compute the amount of time spent within one tower range,
# or the overlap of two people in the same tower rage.
def makeCellTowerIntervals(subject):
   events = subject['locs']
   dt = convertDatetime
   return [((dt(events[i][0]), dt(events[i+1][0])), events[i][1])
           for i in range(len(events) - 1) if events[i][1] > 0] # condition ensures there was signal


def dateIntervalOverlap(dtint1, dtint2):
   start1, end1 = dtint1
   start2, end2 = dtint2

   if start1 <= start2 <= end1:
      return (start2, min(end1, end2))
   elif start2 <= start1 <= end2:
      return (start1, min(end1, end2))
   else:
      return None


def listProximityEvents(intervals1, intervals2):
   if len(intervals1) == 0 or len(intervals2) == 0:
      logger.debug("Found an empty interval list?")
      return []

   D1, D2 = deque(intervals1), deque(intervals2)
   events = deque()

   logger.debug('Processing new pairs of intervals')
   dateInterval1, towerId1 = D1.popleft()
   dateInterval2, towerId2 = D2.popleft()
   while len(D1) > 0 and len(D2) > 0:
      if dateInterval2[0] >= dateInterval1[1]:
         dateInterval1, towerId1 = D1.popleft()
      elif dateInterval1[0] >= dateInterval2[1]:
         dateInterval2, towerId2 = D2.popleft()
      else:
         if towerId1 == towerId2:
            theOverlap = dateIntervalOverlap(dateInterval1, dateInterval2)
            if (theOverlap[1] - theOverlap[0]).total_seconds() > 1:
               events.append((theOverlap, towerId1))
               #logger.debug('Found a match! %s, %s at tower %s' % (theOverlap[0], theOverlap[1], towerId1))

         if dateInterval1[0] < dateInterval2[0]:
            dateInterval1, towerId1 = D1.popleft()
         else:
            dateInterval2, towerId2 = D2.popleft()

   return events


def writeProximityEvents(proxEventsDict, filename):
   with open(filename, 'w') as outfile:
      outfile.write('id1\tid2\tcellTower\tstart\tend\n')
      for k in proxEventsDict:
         id1, id2 = k
         for event in proxEventsDict[k]:
            values = [id1, id2, event[1], event[0][0], event[0][1]]
            line = '\t'.join(("%s" % (v,)) for v in values)
            outfile.write('%s\n' % line)


def createCellTowerDataset(idDictionaries):
   idDict, macDict, hashNumDict = idDictionaries

   logger.debug("Making cell tower intervals")
   cellTowerIntervals = dict((i, makeCellTowerIntervals(idDict[i])) for i in idDict)

   logger.debug("Computing cell tower proximity events")
   proximityEvents = dict(((i, j), listProximityEvents(cellTowerIntervals[i], cellTowerIntervals[j]))
         for i,j in itertools.combinations(cellTowerIntervals.keys(), 2))

   logger.debug("Writing proximity events to reality-mining-proximity.txt")
   writeProximityEvents(proximityEvents, RAW_DATA_DIR + 'reality-mining-proximity.txt')


def create_dynamic_network(delta_t='1hour', threshold=1):
   extracted_data_types = [
      'survey',
      'calls',
      # 'proximity',
      'class',
      'neighborhood',
      'research_group',
   ]
   processed_data = {}
   for data_type in extracted_data_types:
      processed_data[data_type] = os.path.exists(RAW_DATA_DIR + 'reality-mining-%s.txt' % data_type)

   # need to extract data from .mat file first
   if not all(processed_data.values()):
      logger.debug("Loading in matlab data - this takes a while and about 2gb memory")
      matlab_obj = scipy.io.loadmat(RAW_DATA_DIR + 'realitymining.mat')
      logger.debug("Done loading MATLAB data")

      logger.debug("Extracting valid subjects and creating id dictionaries")
      subjects = validSubjects(matlab_obj['s'][0])
      idDictionaries = idDicts(subjects)

      if not processed_data.get('survey', True):
         createFriendshipDataset(matlab_obj['network'][0][0], idDictionaries)
      if not processed_data.get('calls', True):
         createPhoneCallDataset(idDictionaries)
      if not processed_data.get('proximity', True):
         createCellTowerDataset(idDictionaries)

      for affiliation_type in ['class', 'neighborhood', 'research_group']:
         if not processed_data[affiliation_type]:
            createAffiliation(subjects, idDictionaries, affiliation_type)
   else:
      logger.info("No need to process MATLAB data - necessary fields are already extracted")

   # convert calls data to dynamic network format
   with open(RAW_DATA_DIR + 'reality-mining-calls.txt', 'r') as infile:
      with open(RAW_DATA_DIR + 'calls.txt','wb') as outfile:
         for i_l, l in enumerate(infile.readlines()):
            if i_l == 0:
               continue
            values = l.strip().split('\t')
            id_1 = values[0]
            id_2 = values[1]
            time = values[4].split('.')[0]
            outfile.write('%s\t%s\t%s\n' % (time, id_1, id_2))

   network_sub_name = '%s_%s' % (delta_t, threshold)
   network_id = 'reality_mining/calls/' + network_sub_name
   network_dir = settings.NETWORKS_BASE_DIR + network_id + '/'
   timepoints_labels = preprocess_utils.split_into_snapshots(
      input_temporal_network_file=settings.NETWORKS_RAW_DATA_DIR + '/reality_mining/calls.txt',
      output_dir_snapshots=network_dir + 'snapshots/',
      delta_t=delta_t,
      threshold=threshold,
      delimiter='\t',
      datetime_format='%Y-%m-%d %H:%M:%S',
      min_t='2004-08-30 00:00:00',
      max_t='2005-05-09 00:00:00',
      as_directed_graph=False,
      as_weighted_graph=True,
      min_nodes_snapshot=5,
      min_edges_snapshot=5,
      save_snapshots_names_ordered=True,
   )
   timepoints_datetimes = preprocess_utils.convert_timepoint_labels_to_dt(timepoints_labels, delta_t)

   # ground truth events
   c_ev = CommunityEvolution(
      segments_partitions=None,
      changepoints=preprocess_utils.load_changepoints(
         settings.NETWORKS_RAW_DATA_DIR + 'reality_mining/events.txt', timepoints_datetimes
      ),
      method_id='gt_events',
      network_id=network_id,
      start_timepoint=0,
      end_timepoint=len(timepoints_labels)-1,
   )
   c_ev.save(as_ground_truth=True)


   # ground truth partitions
   gt_partitions = {}
   node_labels = {}
   for gt_title in ['class', 'neighborhood', 'research_group']:
      obj_clusters = {}
      with open(RAW_DATA_DIR + 'reality-mining-%s.txt' % gt_title, 'r') as infile:
         for i_l, l in enumerate(infile.readlines()):
            if i_l == 0:
               continue
            values = l.strip().lower().split('\t')
            id_user = values[0]
            group = values[1].strip().replace(' ','_')
            obj_clusters[id_user] = group
            if id_user not in node_labels:
               node_labels[id_user] = group
            else:
               node_labels[id_user] = "-".join([node_labels[id_user], group])
      if len(obj_clusters) == 0:
         logger.warning("No ground truth partition data for '%s'" % gt_title)
         continue
      gt_partitions[gt_title] = Partition(obj_clusters)

   with open(RAW_DATA_DIR + 'reality-mining-survey.txt', 'r') as infile:
      affinities = {
         'friends': [],
         'in_lab': [],
         'out_lab': [],
      }
      for i_l, l in enumerate(infile.readlines()):
         if i_l == 0:
            continue
         values = l.strip().split('\t')
         id_1 = values[0]
         id_2 = values[1]
         friends = float(values[2])
         affinities['friends'].append((id_1, id_2, friends))

         in_lab = float(values[3])
         affinities['in_lab'].append((id_1, id_2, in_lab))

         out_lab = float(values[4])
         affinities['out_lab'].append((id_1, id_2, out_lab))
      for gt_title, affinities_list in affinities.iteritems():
         gt_partitions[gt_title] = AffinityMatrix.from_affinities_list(affinities_list)

   for gt_title, gt_partition in gt_partitions.iteritems():
      c_ev = CommunityEvolution(
         segments_partitions=[gt_partition],
         changepoints=None,
         method_id='gt_%s' % gt_title,
         network_id=network_id,
         start_timepoint=0,
         end_timepoint=len(timepoints_labels)-1,
      )
      c_ev.save(as_ground_truth=True)


   # save node labels
   preprocess_utils.save_node_labels(network_dir, node_labels)

   # save description
   with open(settings.NETWORKS_BASE_DIR + network_id + '/description.txt', 'wb') as outfile:
      outfile.write(
         "This is Reality Mining network. The raw data was obtained from here http://realitycommons.media.mit.edu/realitymining.html. "
         "Each snapshot contains data of {delta_t} and was constructed with threshold {threshold}".format(
             delta_t=delta_t,
             threshold=threshold
      )
   )