"""
 Evaluate a set of methods with respect to how they similar to another community evolution
 (e.g., per-snapshot partitions similarity, Precision-Recall for changepoint classification,
 similarities to another community evolution according to a given community evolution similarity method)
"""
import os
from collections import OrderedDict
from ..partition import Partition
from ..log import logger
from ..latex_builder import LatexBuilder
from ..dynamic_network import DynamicNetwork
from ..reports import reports_utils
from ..evaluation import eval_utils
from .. import utils
from .. import settings
from collections import defaultdict
from ..methods import methods_utils
from ..affinity_matrix import AffinityMatrix

def create_report_evaluation_against_other(methods_ids, networks_ids, **kwargs):
    other_methods_ids = set()
    for i_n,n_id in enumerate(networks_ids):
        images_dir = DynamicNetwork.get_network_methods_comparison_dir(n_id)\
                     + eval_utils.generate_methods_comparison_key(methods_ids) + '/images/'
        sub_dirs = utils.get_subdir_list_from_dir(images_dir + 'evaluation_against_other/')
        for sub_dir in sub_dirs:
            other_method_id = os.path.basename(sub_dir)
            other_methods_ids.add(other_method_id)
    if len(other_methods_ids) == 0:
        logger.warning("No evaluation against others")
        return

    for other_method_id in other_methods_ids:
        report = LatexBuilder(
            report_name='against-%s' % other_method_id,
            title="Evaluation against {other_method_id} for {methods} methods"
                  "in {networks} networks".format(
                other_method_id=methods_utils.get_method_title(other_method_id),
                methods=kwargs.get('methods_key_name') or ", ".join(methods_ids),
                networks=kwargs.get('networks_key_name') or ", ".join(networks_ids),
            ),
        )

        add_section_evaluation_against_other(
            report,
            networks_ids,
            {'all': methods_ids},
            other_method_id=other_method_id,
            clearPage=True,
            include_per_network=len(networks_ids) == 1 or kwargs.get('include_per_network', True),
            include_across_networks=len(networks_ids) > 1 and kwargs.get('include_across_networks', True),
        )

        report.compile(
            output_directory=settings.REPORTS_DIR + 'methods/' + kwargs.get('report_subdir_name', '')
                             + '/' + reports_utils.get_methods_networks_key(methods_ids, networks_ids, **kwargs) + '/',
            pdflatex=True
        )
        if kwargs.get('copy_to_tmp_dir', True):
            reports_utils.copy_report_to_tmp_dir(
                report,
                tmp_subdir='methods/' + kwargs.get('report_subdir_name', '') + '/',
                tmp_report_name=report.get_name() + '_'
                                + reports_utils.get_methods_networks_key(methods_ids, networks_ids, **kwargs),
            )


def add_section_evaluation_against_other(report, networks_ids, methods_ids_dict, other_method_id,
                                         section_level=None, section_name=None, **kwargs):
    logger.debug("Adding evaluation of '{methods_ids_dict}' methods against '{other_method_id}' "
                 "for '{networks_ids}' networks".format(
        methods_ids_dict=methods_ids_dict,
        other_method_id=other_method_id,
        networks_ids=", ".join(networks_ids),
    ))

    if section_level is not None:
        if section_name is not None:
            report.add_section(
                section_name=section_name,
                level=section_level,
                clearPage=kwargs.get('clearPage', False),
            )
        else:
            section_level -= 1
    else:
        section_level = -1

    if kwargs.get('include_per_network', True):
        add_section_timeseries_partitions_similarities(
            report,
            methods_ids_dict,
            networks_ids,
            other_method_id=other_method_id,
            classification_type='changepoints',
            section_level=section_level + 1,
            clearPage=False,
        )

    if 'clearPage' in kwargs:
        kwargs.pop('clearPage')
    add_section_pr_roc_results(
        report,
        methods_ids_dict,
        networks_ids,
        other_method_id=other_method_id,
        classification_type='changepoints',
        section_level=section_level + 1,
        clearPage=False,
        **kwargs
    )

    add_section_com_evs_similarity(
        report,
        methods_ids_dict,
        networks_ids,
        other_method_id=other_method_id,
        section_level=section_level + 1,
        **kwargs
    )


def add_section_timeseries_partitions_similarities(report, methods_ids_dict, networks_ids,
                               other_method_id, classification_type, key_label=None, section_level=None, **kwargs):

    if section_level is not None:
        report.add_section(
            section_name="Timeseries of partitions similarities",
            level=section_level,
            clearPage=kwargs.get('clearPage', False),
        )
    else:
        section_level = -1
    metrics = Partition.SIMILARITY_METRICS.keys() + AffinityMatrix.PARTITION_SIMILARITY_METRICS.keys()
    for metric in metrics:
        metric_title = Partition.SIMILARITY_METRICS.get(metric, AffinityMatrix.PARTITION_SIMILARITY_METRICS.get(metric))
        images_list = []
        rows_captions_list = []
        for i_n,n_id in enumerate(networks_ids):
            for k in sorted(methods_ids_dict.keys()):
                m_ids = methods_ids_dict[k]
                images_dir = DynamicNetwork.get_network_methods_comparison_dir(n_id)
                if len(m_ids) > 1:
                    images_dir += eval_utils.generate_methods_comparison_key(m_ids) + '/images/'\
                                  + 'evaluation_against_other/' + other_method_id + '/'
                else:
                    images_dir += eval_utils.generate_methods_comparison_key(m_ids + [other_method_id]) + '/images/'
                image_name = images_dir + 'timeseries-snapshot-partitions_similarity/{metric}'.format(
                    dir=images_dir,
                    metric=metric,
                )
                if os.path.exists(image_name + '.' + settings.CHART_EXT):
                    images_list.append(image_name)
            rows_captions_list.append("{n_id} network".format(n_id=n_id))

        if len(images_list) == 0:
            continue

        report.add_section(
            section_name=metric_title,
            level=section_level + 1,
            clearPage=False,
        )

        if len(methods_ids_dict.keys()) == 1:
            columns_captions_list = None
            columns_count = 3
            images_captions_list = rows_captions_list
            rows_captions_list = None
            rows_descr = ""
        else:
            columns_captions_list = sorted(methods_ids_dict.keys())
            columns_count = len(methods_ids_dict.keys())
            images_captions_list = None
            rows_descr = ", rows correspond to networks"

        report.add_group_figure(
            title="Timeseries of partition similarities over time (x-axis - time, y-axis - {metric})"
              "{rows_descr}{sep}{cols_descr}"
            .format(
                metric=metric_title,
                rows_descr=rows_descr,
                sep="" if rows_descr == "" else ", ",
                cols_descr="columns - to %s" % key_label.replace('_', ' ') if key_label is not None else "",
            ),
            images_list=images_list,
            images_captions_list=images_captions_list,
            rows_captions_list=rows_captions_list,
            columns_captions_list=columns_captions_list,
            columns_count=columns_count,
            max_rows_per_figure=4,
        )



def add_section_pr_roc_results(report, methods_ids_dict, networks_ids,
                               other_method_id, classification_type, key_label=None, section_level=None, **kwargs):
    if section_level is not None:
        report.add_section(
            section_name="Classification results for %s" % classification_type,
            level=section_level,
            clearPage=kwargs.get('clearPage', False),
        )
    else:
        section_level = -1
    funcs_list = []
    if kwargs.get('include_per_network', True):
        funcs_list.append(add_section_pr_roc_results_per_network)
    if kwargs.get('include_across_networks', True):
        funcs_list.append(add_section_pr_roc_results_across_networks)
    if len(funcs_list) == 0:
        logger.warning("No analysis (per network / across networks to include)")
        return
    for func in funcs_list:
        func(
            report,
            methods_ids_dict,
            networks_ids,
            other_method_id=other_method_id,
            classification_type=classification_type,
            key_label=None,
            section_level=section_level + 1,
            clearPage=False,
        )


def add_section_pr_roc_results_per_network(report, methods_ids_dict, networks_ids,
                               other_method_id, classification_type, key_label=None, section_level=None, **kwargs):
    if section_level is not None:
        report.add_section(
            section_name="Per network",
            level=section_level,
            clearPage=kwargs.get('clearPage', False),
        )
    else:
        section_level = -1

    chart_types = OrderedDict()
    chart_types['pr'] = "Precision-Recall"
    chart_types['roc'] = "Receiver Operating Characteristic"
    chart_types['bars'] = "Bars"
    for chart_type, chart_type_title in chart_types.iteritems():
        report.add_section(
            section_name=chart_type_title,
            level=section_level + 1,
            clearPage=False,
        )
        images_list = []
        rows_captions_list = []
        for i_n,n_id in enumerate(networks_ids):
            networks_has_images = False
            for k in sorted(methods_ids_dict.keys()):
                m_ids = methods_ids_dict[k]
                images_dir = DynamicNetwork.get_network_methods_comparison_dir(n_id)
                if len(m_ids) > 1:
                    images_dir += eval_utils.generate_methods_comparison_key(m_ids) + '/images/'\
                                  + 'evaluation_against_other/' + other_method_id + '/'
                else:
                    images_dir += eval_utils.generate_methods_comparison_key(m_ids + [other_method_id]) + '/images/'

                images_dir += 'pr_roc/' + classification_type + '/'
                if not os.path.exists(images_dir):
                    continue
                images_list.append(images_dir + chart_type)
                networks_has_images = True
            if networks_has_images:
                rows_captions_list.append("{n_id} network".format(n_id=n_id))

        if len(images_list) == 0:
            logger.warning("No images of type '%s'" % chart_type)
            return

        if len(methods_ids_dict.keys()) == 1:
            columns_captions_list = None
            columns_count = 3
            images_captions_list = rows_captions_list
            rows_captions_list = None
        else:
            columns_captions_list = sorted(methods_ids_dict.keys())
            columns_count = len(methods_ids_dict.keys())
            images_captions_list = None

        report.add_group_figure(
            title=chart_type_title,
            images_list=images_list,
            images_captions_list=images_captions_list,
            rows_captions_list=rows_captions_list,
            columns_captions_list=columns_captions_list,
            columns_count=columns_count,
            max_rows_per_figure=4,
        )

def add_section_pr_roc_results_across_networks(report, methods_ids_dict, networks_ids,
                               other_method_id, classification_type, key_label=None, section_level=None, **kwargs):
    if section_level is not None:
        report.add_section(
            section_name="Across networks",
            level=section_level,
            clearPage=kwargs.get('clearPage', False),
        )
    else:
        section_level = -1

    chart_types = OrderedDict()
    chart_types['bars'] = "Bars"
    for chart_type, chart_type_title in chart_types.iteritems():
        report.add_section(
            section_name=chart_type_title,
            level=section_level + 1,
            clearPage=False,
        )
        images_list = []

        for k in sorted(methods_ids_dict.keys()):
            m_ids = methods_ids_dict[k]
            images_dir = settings.EVAL_DIR + eval_utils.generate_networks_comparison_key(networks_ids) + '/'\
                     + eval_utils.generate_methods_comparison_key(m_ids) + '/images/'\
                      + 'evaluation_against_other/' + other_method_id + '/'

            images_dir += 'pr_roc/' + classification_type + '/'
            if not os.path.exists(images_dir):
                continue
            images_list.append(images_dir + chart_type)

        if len(images_list) == 0:
            logger.warning("No images of type '%s'" % chart_type)
            return

        if len(methods_ids_dict.keys()) == 1:
            columns_captions_list = None
            columns_count = 3
        else:
            columns_captions_list = sorted(methods_ids_dict.keys())
            columns_count = len(methods_ids_dict.keys())

        report.add_group_figure(
            title=chart_type_title,
            images_list=images_list,
            columns_captions_list=columns_captions_list,
            columns_count=columns_count,
            max_rows_per_figure=4,
        )


def add_section_com_evs_similarity(report, methods_ids_dict, networks_ids, other_method_id,
                                    section_level=None, **kwargs):
    if section_level is not None:
        report.add_section(
            section_name="Community evolutions similarities",
            level=section_level,
            clearPage=True,
        )
    else:
        section_level = -1

    if kwargs.get('include_per_network', True):
        report.add_section(
            section_name="Per network",
            level=section_level + 1,
            clearPage=True,
        )
        for func in [
            add_section_com_evs_similarity_scatter_per_network,
            add_section_com_evs_similarity_bar_per_network
        ]:
            func(
                report,
                methods_ids_dict,
                networks_ids,
                other_method_id=other_method_id,
                key_label=None,
                section_level=section_level + 2,
                clearPage=False,
            )
    if kwargs.get('include_across_networks', True):
        report.add_section(
            section_name="Across networks",
            level=section_level + 1,
            clearPage=True,
        )
        for func in [
            add_section_com_evs_similarity_scatter_across_networks,
            add_section_com_evs_similarity_bar_across_networks
        ]:
            func(
                report,
                methods_ids_dict,
                networks_ids,
                other_method_id=other_method_id,
                key_label=None,
                section_level=section_level + 2,
                clearPage=False,
            )


def add_section_com_evs_similarity_scatter_per_network(report, methods_ids_dict, networks_ids,
                                                       other_method_id, key_label=None, section_level=None, **kwargs):
    if section_level is not None:
        report.add_section(
            section_name="Scatter",
            level=section_level,
            clearPage=kwargs.get('clearPage', False),
        )

     # dict - similarity methods pair => list of comparisons
    sim_m_pairs_type_dict = defaultdict(set)
    for i_n, n_id in enumerate(networks_ids):
        for k, m_ids in methods_ids_dict.iteritems():
            images_dir = DynamicNetwork.get_network_methods_comparison_dir(n_id)
            if len(m_ids) > 1:
                images_dir += eval_utils.generate_methods_comparison_key(m_ids) + '/images/'\
                              + 'evaluation_against_other/' + other_method_id + '/'
            else:
                images_dir += eval_utils.generate_methods_comparison_key(m_ids + [other_method_id]) + '/images/'
            images_dir += 'scatter-com_evs_similarity/'
            for im in utils.get_file_list_from_dir(images_dir, full_path=False):
                sim_m_pair_id = im.replace('.' + settings.CHART_EXT, '')
                # to avoid comparing according to too many sim pairs
                sim_m_pairs_type_dict[sim_m_pair_id].add(k)

    if len(sim_m_pairs_type_dict)==0:
        logger.warning("No similarity methods pairs for scatter plots")
        return


    for sim_m_pair in sorted(sim_m_pairs_type_dict.keys()):
        ks_set = sim_m_pairs_type_dict[sim_m_pair]
        images_list = []
        rows_captions_list = []
        for i_n,n_id in enumerate(networks_ids):
            networks_has_images = False
            for k in sorted(methods_ids_dict.keys()):
                if k not in ks_set:
                    continue
                m_ids = methods_ids_dict[k]
                images_dir = DynamicNetwork.get_network_methods_comparison_dir(n_id)
                if len(m_ids) > 1:
                    images_dir += eval_utils.generate_methods_comparison_key(m_ids) + '/images/'\
                                  + 'evaluation_against_other/' + other_method_id + '/'
                else:
                    images_dir += eval_utils.generate_methods_comparison_key(m_ids + [other_method_id]) + '/images/'
                images_dir += 'scatter-com_evs_similarity/'
                images_list.append(images_dir + sim_m_pair)
                networks_has_images = True
            if networks_has_images:
                rows_captions_list.append("{n_id} network".format(n_id=n_id))

        if len(methods_ids_dict.keys()) == 1:
            columns_captions_list = None
            columns_count = 3
        else:
            columns_captions_list = sorted(ks_set)
            columns_count = len(ks_set)

        report.add_group_figure(
            title="Similarity scatter plot scores with avg/std per network",
            images_list=images_list,
            columns_captions_list=columns_captions_list,
            columns_count=columns_count,
            max_rows_per_figure=4,
        )

def add_section_com_evs_similarity_bar_per_network(report, methods_ids_dict, networks_ids,
                                               other_method_id, key_label=None, section_level=None, **kwargs):
    if section_level is not None:
        report.add_section(
            section_name="Bar",
            level=section_level,
            clearPage=kwargs.get('clearPage', False),
        )


    # dict - similarity methods list type => list of comparisons
    # (types can be e.g., default - using default fixed set of similarity methodss)
    sim_methods_type_dict = defaultdict(set)
    for i_n, n_id in enumerate(networks_ids):
        for k, m_ids in methods_ids_dict.iteritems():
            images_dir = DynamicNetwork.get_network_methods_comparison_dir(n_id)\
                         + eval_utils.generate_methods_comparison_key(m_ids) + '/images/evaluation_against_other/'\
                         + other_method_id +'/bar-com_evs_similarity/'
            for im in utils.get_file_list_from_dir(images_dir, full_path=False):
                sc_m_type_id = im.replace('.' + settings.CHART_EXT, '')
                # to avoid comparing according to too many scoring methods
                sim_methods_type_dict[sc_m_type_id].add(k)

    if len(sim_methods_type_dict)==0:
        logger.warning("No scoring methods for solutions score analysis")
        return


    for sim_m_type in sorted(sim_methods_type_dict.keys()):
        ks_set = sim_methods_type_dict[sim_m_type]
        images_list = []
        rows_captions_list = []
        for i_n,n_id in enumerate(networks_ids):
            networks_has_images = False
            for k in sorted(methods_ids_dict.keys()):
                if k not in ks_set:
                    continue
                m_ids = methods_ids_dict[k]
                images_dir = DynamicNetwork.get_network_methods_comparison_dir(n_id)\
                             + eval_utils.generate_methods_comparison_key(m_ids) + '/images/evaluation_against_other/'\
                         + other_method_id +'/bar-com_evs_similarity/'
                images_list.append(images_dir + sim_m_type)
                networks_has_images = True
            if networks_has_images:
                rows_captions_list.append("{n_id} network".format(n_id=n_id))

        if len(methods_ids_dict.keys()) == 1:
            columns_captions_list = None
            columns_count = 3
            images_captions_list = rows_captions_list
            rows_captions_list = None
            rows_descr = ""
        else:
            columns_captions_list = sorted(ks_set)
            columns_count = len(ks_set)
            images_captions_list = None
            rows_descr = ", rows correspond to networks"

        report.add_group_figure(
            title="Bar plots of similarity scores of the solutions per network according to {sim_m_type} similarity methods"
                  "{rows_descr}{sep}{cols_descr}"
                .format(
                sim_m_type=sim_m_type,
                rows_descr=rows_descr,
                sep="" if rows_descr == "" else ", ",
                cols_descr="columns - to %s" % key_label.replace('_', ' ') if key_label is not None else "",
            ),
            images_list=images_list,
            images_captions_list=images_captions_list,
            rows_captions_list=rows_captions_list,
            columns_captions_list=columns_captions_list,
            columns_count=columns_count,
            max_rows_per_figure=4,
        )

# TODO: refactor - extract per/across networks methods in a separate function (e.g., for bar and scatter)
def add_section_com_evs_similarity_scatter_across_networks(report, methods_ids_dict, networks_ids,
                                                           other_method_id, key_label=None, section_level=None, **kwargs):
    if section_level is not None:
        report.add_section(
            section_name="Scatter",
            level=section_level,
            clearPage=kwargs.get('clearPage', False),
        )

    # dict - similarity methods pair => list of comparisons
    sim_m_pairs_type_dict = defaultdict(set)
    for k, m_ids in methods_ids_dict.iteritems():
        images_dir = settings.EVAL_DIR + eval_utils.generate_networks_comparison_key(networks_ids) + '/'\
                     + eval_utils.generate_methods_comparison_key(m_ids) + '/images/'\
                      + 'evaluation_against_other/' + other_method_id + '/scatter-com_evs_similarity/'
        for im in utils.get_file_list_from_dir(images_dir, full_path=False):
            sim_m_pair_id = im.replace('.' + settings.CHART_EXT, '')
            # to avoid comparing according to too many sim pairs
            sim_m_pairs_type_dict[sim_m_pair_id].add(k)

    if len(sim_m_pairs_type_dict)==0:
        logger.warning("No similarity methods pairs for scatter plots")
        return

    use_one_figure = False
    images_list = []
    images_captions_list = []
    if len(methods_ids_dict) == 1:
        use_one_figure = True

    for sim_m_pair in sorted(sim_m_pairs_type_dict.keys()):
        ks_set = sim_m_pairs_type_dict[sim_m_pair]
        if not use_one_figure:
            images_list = []
        for k in sorted(methods_ids_dict.keys()):
            if k not in ks_set:
                continue
            m_ids = methods_ids_dict[k]
            images_dir = settings.EVAL_DIR + eval_utils.generate_networks_comparison_key(networks_ids) + '/'\
                 + eval_utils.generate_methods_comparison_key(m_ids) + '/images/'\
                  + 'evaluation_against_other/' + other_method_id + '/scatter-com_evs_similarity/'
            images_list.append(images_dir + sim_m_pair)
            images_captions_list.append(sim_m_pair.replace('-vs-',' vs '))

        if not use_one_figure:
            report.add_group_figure(
                title="Similarity scatter plot scores with avg/std across networks",
                images_list=images_list,
                images_captions_list=images_captions_list,
                columns_captions_list=sorted(ks_set),
                columns_count=len(ks_set),
                max_rows_per_figure=4,
            )
    if use_one_figure:
        report.add_group_figure(
            title="Similarity scatter plot scores with avg/std across networks",
            images_list=images_list,
            images_captions_list=images_captions_list,
            columns_captions_list=None,
            columns_count=3,
            max_rows_per_figure=4,
        )
def add_section_com_evs_similarity_bar_across_networks(report, methods_ids_dict, networks_ids,
                                   other_method_id, key_label=None, section_level=None, **kwargs):
    if section_level is not None:
        report.add_section(
            section_name="Bar",
            level=section_level,
            clearPage=kwargs.get('clearPage', False),
        )

    # dict - similarity methods list type => list of comparisons
    # (types can be e.g., default - using default fixed set of similarity methods)
    sim_methods_type_dict = defaultdict(set)
    for k, m_ids in methods_ids_dict.iteritems():
        images_dir = settings.EVAL_DIR + eval_utils.generate_networks_comparison_key(networks_ids) + '/'\
                     + eval_utils.generate_methods_comparison_key(m_ids) + '/images/evaluation_against_other/'\
                     + other_method_id + '/bar-com_evs_similarity/'
        for im in utils.get_file_list_from_dir(images_dir, full_path=False):
            sc_m_type_id = im.replace('.' + settings.CHART_EXT, '')
            # to avoid comparing according to too many scoring methods
            sim_methods_type_dict[sc_m_type_id].add(k)

    if len(sim_methods_type_dict)==0:
        logger.warning("No similarity methods for analysis")
        return

    use_one_figure = False
    images_list = []
    images_captions_list = []
    if len(methods_ids_dict) == 1:
        use_one_figure = True

    for sim_m_type in sorted(sim_methods_type_dict.keys()):
        ks_set = sim_methods_type_dict[sim_m_type]
        if not use_one_figure:
            images_list = []

        for k in sorted(methods_ids_dict.keys()):
            if k not in ks_set:
                continue
            m_ids = methods_ids_dict[k]
            images_dir = settings.EVAL_DIR + eval_utils.generate_networks_comparison_key(networks_ids) + '/'\
                         + eval_utils.generate_methods_comparison_key(m_ids) + '/images/evaluation_against_other/'\
                     + other_method_id + '/bar-com_evs_similarity/'
            images_list.append(images_dir + sim_m_type)
            images_captions_list.append(sim_m_type)

        if not use_one_figure:
            report.add_group_figure(
                title="Bar plots of similarity scores of the solutions across networks with avg/std "
                      "according to {sim_m_type} similarity methods "
                    .format(
                    sim_m_type=sim_m_type,
                ),
                images_list=images_list,
                images_captions_list=images_captions_list,
                columns_captions_list=sorted(ks_set),
                columns_count=len(ks_set),
                max_rows_per_figure=4,
            )
    if use_one_figure:
        report.add_group_figure(
            title="Bar plots of similarity scores of the solutions across networks with avg/std",
            images_list=images_list,
            images_captions_list=images_captions_list,
            columns_captions_list=None,
            columns_count=3,
            max_rows_per_figure=4,
        )