"""
 Evaluate a set of methods with respect to external criteria
 (e.g., time segmentation or score according to a given scoring method)
"""
import os
from ..log import logger
from ..latex_builder import LatexBuilder
from ..community_evolution import CommunityEvolution
from ..dynamic_network import DynamicNetwork
from ..reports import reports_utils
from ..evaluation import eval_utils
from .. import utils
from .. import settings
from .. import static_clustering
from collections import defaultdict
from ..scoring_methods import scoring_methods_utils

def create_report_evaluation_external_criteria(methods_ids, networks_ids, **kwargs):
    report = LatexBuilder(
        report_name='basic_evaluation',
        title="Basic analysis for {methods} methods in {networks} networks".format(
            methods=kwargs.get('methods_key_name') or ", ".join(methods_ids),
            networks=kwargs.get('networks_key_name') or ", ".join(networks_ids),
        ),
    )

    add_section_methods_comparison(
        report,
        networks_ids,
        {'all': methods_ids},
        clearPage=True,
        include_per_network=len(networks_ids) == 1 or kwargs.get('include_per_network', True),
        include_across_networks=len(networks_ids) > 1 and kwargs.get('include_across_networks', True),
    )

    report.compile(
        output_directory=settings.REPORTS_DIR + 'methods/' + kwargs.get('report_subdir_name', '')
                         + '/' + reports_utils.get_methods_networks_key(methods_ids, networks_ids, **kwargs) + '/',
        pdflatex=True
    )
    if kwargs.get('copy_to_tmp_dir', True):
        reports_utils.copy_report_to_tmp_dir(
            report,
            tmp_subdir='methods/' + kwargs.get('report_subdir_name', '') + '/',
            tmp_report_name=report.get_name() + '_'
                            + reports_utils.get_methods_networks_key(methods_ids, networks_ids, **kwargs),
        )



def add_section_methods_comparison(report, networks_ids, methods_ids_dict, section_level=None, section_name=None, **kwargs):
    logger.debug("Adding method comparison '{methods_ids_dict}' methods for '{networks_ids}' networks".format(
        methods_ids_dict=methods_ids_dict,
        networks_ids=", ".join(networks_ids),
    ))

    if section_level is not None:
        if section_name is not None:
            report.add_section(
                section_name=section_name,
                level=section_level,
                clearPage=kwargs.get('clearPage', False),
            )
        else:
            section_level -= 1
    else:
        section_level = -1

    for func in [
            add_section_description,
            add_section_time_segmentations,
            add_section_partition_qualities_timeseries,
            add_section_score_num_changepoints,
            add_section_scores_bar,
            add_section_run_times,
        ]:
            func(
                report,
                methods_ids_dict,
                networks_ids,
                section_level=section_level + 1,
                **kwargs
            )

def add_section_description(report, methods_ids_dict, networks_ids, section_level=None, **kwargs):
    if section_level is not None:
        report.add_section(
            "Descriptions",
            level=section_level,
            clearPage=kwargs.get('clearPage', False),
        )
    else:
        section_level = -1

    for dn_id in networks_ids:
        if len(networks_ids) > 1:
            report.add_section(
                dn_id,
                level=section_level + 1
            )
        for k, methods_ids in methods_ids_dict.iteritems():
            c_evs_list = [CommunityEvolution.load(dn_id, m) for m in methods_ids]
            for c_ev in c_evs_list:
                report.add_section(
                    c_ev.get_method_id(),
                    level=section_level + (2 if len(networks_ids) > 1 else 1)
                )
                if c_ev.get_description() is not None:
                    report.add_text(c_ev.get_description())
                else:
                    report.add_text(c_ev.get_summary())

def add_section_time_segmentations(report, methods_ids_dict, networks_ids, key_label=None, section_level=None,
                                   **kwargs):
    if section_level is not None:
        report.add_section(
            section_name="Time interval segmentation",
            level=section_level,
            clearPage=kwargs.get('clearPage', False),
        )
    images_list = []
    rows_captions_list = []
    for i_n,n_id in enumerate(networks_ids):
        for k in sorted(methods_ids_dict.keys()):
            m_ids = methods_ids_dict[k]
            images_dir = DynamicNetwork.get_network_methods_comparison_dir(n_id)\
                         + eval_utils.generate_methods_comparison_key(m_ids) + '/images/'
            images_list.append(images_dir + 'segmentation')
        rows_captions_list.append("{n_id} network".format(n_id=n_id))

    if len(methods_ids_dict.keys()) == 1:
        columns_captions_list = None
        columns_count = 3
        images_captions_list = rows_captions_list
        rows_captions_list = None
        rows_descr = ""
    else:
        columns_captions_list = sorted(methods_ids_dict.keys())
        columns_count = len(methods_ids_dict.keys())
        images_captions_list = None
        rows_descr = ", rows correspond to networks"

    report.add_group_figure(
        title="Time interval segmentation (x-axis - timepoints, y-axis - methods)"
              "{rows_descr}{sep}{cols_descr}"
            .format(
            rows_descr=rows_descr,
            sep="" if rows_descr == "" else ", ",
            cols_descr="columns - to %s" % key_label.replace('_', ' ') if key_label is not None else "",
        ),
        images_list=images_list,
        images_captions_list=images_captions_list,
        rows_captions_list=rows_captions_list,
        columns_captions_list=columns_captions_list,
        columns_count=columns_count,
        max_rows_per_figure=4,
    )


def add_section_scores_bar(report, methods_ids_dict, networks_ids,
                           section_level=None, **kwargs):
    if section_level is not None:
        report.add_section(
            section_name="Scores of the solutions",
            level=section_level,
            clearPage=kwargs.get('clearPage', True),
        )
    else:
        section_level = -1

    funcs_list = []
    if kwargs.get('include_per_network', True):
        funcs_list.append(add_section_scores_bar_per_network)
    if kwargs.get('include_aross_networks', True):
        funcs_list.append(add_section_scores_bar_across_networks)
    if len(funcs_list) == 0:
        logger.warning("No analysis (per network / across networks to include)")
        return
    for func in funcs_list:
        func(
            report,
            methods_ids_dict,
            networks_ids,
            key_label=None,
            section_level=section_level + 1,
            clearPage=False,
        )

def add_section_scores_bar_per_network(report, methods_ids_dict, networks_ids, key_label=None, section_level=None,
                                       **kwargs):
    if section_level is not None:
        report.add_section(
            section_name="Per network",
            level=section_level,
            clearPage=kwargs.get('clearPage', False),
        )

    # dict - score methods list type => list of comparisons
    # (types can be e.g., default - using default fixed set of scoring methods, or custom - selected based on methods)
    scoring_methods_type_dict = defaultdict(set)
    for i_n, n_id in enumerate(networks_ids):
        for k, m_ids in methods_ids_dict.iteritems():
            images_dir = DynamicNetwork.get_network_methods_comparison_dir(n_id)\
                         + eval_utils.generate_methods_comparison_key(m_ids) + '/images/bar-score/'
            for im in utils.get_file_list_from_dir(images_dir, full_path=False):
                sc_m_type_id = im.replace('.' + settings.CHART_EXT, '')
                # to avoid comparing according to too many scoring methods
                scoring_methods_type_dict[sc_m_type_id].add(k)

    if len(scoring_methods_type_dict)==0:
        logger.warning("No scoring methods for solutions score analysis")
        return


    for sc_m_type in sorted(scoring_methods_type_dict.keys()):
        ks_set = scoring_methods_type_dict[sc_m_type]
        images_list = []
        rows_captions_list = []
        for i_n,n_id in enumerate(networks_ids):
            networks_has_images = False
            for k in sorted(methods_ids_dict.keys()):
                if k not in ks_set:
                    continue
                m_ids = methods_ids_dict[k]
                images_dir = DynamicNetwork.get_network_methods_comparison_dir(n_id)\
                             + eval_utils.generate_methods_comparison_key(m_ids) + '/images/bar-score/'
                images_list.append(images_dir + sc_m_type)
                networks_has_images = True
            if networks_has_images:
                rows_captions_list.append("{n_id} network".format(n_id=n_id))

        if len(methods_ids_dict.keys()) == 1:
            columns_captions_list = None
            columns_count = 3
            images_captions_list = rows_captions_list
            rows_captions_list = None
            rows_descr = ""
        else:
            columns_captions_list = sorted(ks_set)
            columns_count = len(ks_set)
            images_captions_list = None
            rows_descr = ", rows correspond to networks"

        report.add_group_figure(
            title="Scores of the solutions according to {sc_m_type} scoring methods"
                  "{rows_descr}{sep}{cols_descr}"
                .format(
                sc_m_type=sc_m_type,
                rows_descr=rows_descr,
                sep="" if rows_descr == "" else ", ",
                cols_descr="columns - to %s" % key_label.replace('_', ' ') if key_label is not None else "",
            ),
            images_list=images_list,
            images_captions_list=images_captions_list,
            rows_captions_list=rows_captions_list,
            columns_captions_list=columns_captions_list,
            columns_count=columns_count,
            max_rows_per_figure=4,
        )

def add_section_scores_bar_across_networks(report, methods_ids_dict, networks_ids, key_label=None, section_level=None,
                                            **kwargs):
    if section_level is not None:
        report.add_section(
            section_name="Across networks",
            level=section_level,
            clearPage=kwargs.get('clearPage', False),
        )

    # dict - score methods list type => list of comparisons
    # (types can be e.g., default - using default fixed set of scoring methods, or custom - selected based on methods)
    scoring_methods_type_dict = defaultdict(set)
    for k, m_ids in methods_ids_dict.iteritems():
        images_dir = settings.EVAL_DIR + eval_utils.generate_networks_comparison_key(networks_ids) + '/'\
                     + eval_utils.generate_methods_comparison_key(m_ids) + '/images/bar-score/'
        for im in utils.get_file_list_from_dir(images_dir, full_path=False):
            sc_m_type_id = im.replace('.' + settings.CHART_EXT, '')
            # to avoid comparing according to too many scoring methods
            scoring_methods_type_dict[sc_m_type_id].add(k)

    if len(scoring_methods_type_dict)==0:
        logger.warning("No scoring methods for solutions score analysis")
        return


    for sc_m_type in sorted(scoring_methods_type_dict.keys()):
        ks_set = scoring_methods_type_dict[sc_m_type]
        images_list = []

        for k in sorted(methods_ids_dict.keys()):
            if k not in ks_set:
                continue
            m_ids = methods_ids_dict[k]
            images_dir = settings.EVAL_DIR + eval_utils.generate_networks_comparison_key(networks_ids) + '/'\
                         + eval_utils.generate_methods_comparison_key(m_ids) + '/images/bar-score/'
            images_list.append(images_dir + sc_m_type)

        if len(methods_ids_dict.keys()) == 1:
            columns_captions_list = None
            columns_count = 3
        else:
            columns_captions_list = sorted(ks_set)
            columns_count = len(ks_set)

        report.add_group_figure(
            title="Scores of the solutions according to {sc_m_type} scoring methods "
                  "with avg/std"
                .format(
                sc_m_type=sc_m_type,
            ),
            images_list=images_list,
            columns_captions_list=columns_captions_list,
            columns_count=columns_count,
            max_rows_per_figure=4,
        )


def add_section_run_times(report, methods_ids_dict, networks_ids,
                           section_level=None, **kwargs):
    if section_level is not None:
        report.add_section(
            section_name="Running times",
            level=section_level,
            clearPage=kwargs.get('clearPage', True),
        )
    else:
        section_level = -1

    funcs_list = []
    if kwargs.get('include_per_network', True):
        funcs_list.append(add_section_run_times_per_network)
    if kwargs.get('include_aross_networks', True):
        funcs_list.append(add_section_run_times_across_networks)
    if len(funcs_list) == 0:
        logger.warning("No analysis (per network / across networks to include)")
        return
    for func in funcs_list:
        func(
            report,
            methods_ids_dict,
            networks_ids,
            key_label=None,
            section_level=section_level + 1,
            clearPage=False,
        )

def add_section_run_times_per_network(report, methods_ids_dict, networks_ids, key_label=None, section_level=None,
                                            **kwargs):
    if section_level is not None:
        report.add_section(
            section_name="Per network",
            level=section_level,
            clearPage=kwargs.get('clearPage', False),
        )

    images_list = []
    rows_captions_list = []
    for i_n,n_id in enumerate(networks_ids):
        networks_has_images = False
        for k in sorted(methods_ids_dict.keys()):
            m_ids = methods_ids_dict[k]
            images_dir = DynamicNetwork.get_network_methods_comparison_dir(n_id)\
                         + eval_utils.generate_methods_comparison_key(m_ids) + '/images/'
            images_list.append(images_dir + 'bar-run_time')
            networks_has_images = True
        if networks_has_images:
            rows_captions_list.append("{n_id} network".format(n_id=n_id))

    if len(methods_ids_dict.keys()) == 1:
        columns_captions_list = None
        columns_count = 3
        images_captions_list = rows_captions_list
        rows_captions_list = None
        rows_descr = ""
    else:
        columns_captions_list = sorted(methods_ids_dict.keys())
        columns_count = len(methods_ids_dict.keys())
        images_captions_list = None
        rows_descr = ", rows correspond to networks"

    report.add_group_figure(
        title="Running times of the methods"
              "{rows_descr}{sep}{cols_descr}"
            .format(
            rows_descr=rows_descr,
            sep="" if rows_descr == "" else ", ",
            cols_descr="columns - to %s" % key_label.replace('_', ' ') if key_label is not None else "",
        ),
        images_list=images_list,
        images_captions_list=images_captions_list,
        rows_captions_list=rows_captions_list,
        columns_captions_list=columns_captions_list,
        columns_count=columns_count,
        max_rows_per_figure=4,
    )

def add_section_run_times_across_networks(report, methods_ids_dict, networks_ids, key_label=None, section_level=None,
                                            **kwargs):
    if len(networks_ids) < 2:
        return
        return
    if section_level is not None:
        report.add_section(
            section_name="Across networks",
            level=section_level,
            clearPage=kwargs.get('clearPage', False),
        )

    images_list = []

    for k in sorted(methods_ids_dict.keys()):
        m_ids = methods_ids_dict[k]
        images_dir = settings.EVAL_DIR + eval_utils.generate_networks_comparison_key(networks_ids) + '/'\
                     + eval_utils.generate_methods_comparison_key(m_ids) + '/images/'
        images_list.append(images_dir + 'bar-run_time')

    if len(methods_ids_dict.keys()) == 1:
        columns_captions_list = None
        columns_count = 3
    else:
        columns_captions_list = sorted(methods_ids_dict.keys())
        columns_count = len(methods_ids_dict.keys())

    report.add_group_figure(
        title="Running times of the methods with avg/std",
        images_list=images_list,
        columns_captions_list=columns_captions_list,
        columns_count=columns_count,
        max_rows_per_figure=4,
    )


def add_section_score_num_changepoints(report, methods_ids_dict, networks_ids, key_label=None,
                                       section_level=None, **kwargs):
    # dict - score method => list of comparisons
    scoring_methods_dict = defaultdict(set)
    for i_n, n_id in enumerate(networks_ids):
        for k, m_ids in methods_ids_dict.iteritems():
            images_dir = DynamicNetwork.get_network_methods_comparison_dir(n_id)\
                         + eval_utils.generate_methods_comparison_key(m_ids) + '/images/'
            for im in utils.get_file_list_from_dir(
                images_dir + 'timeseries-n_changepoints-score/',
                full_path=False
            ):
                sc_m_id = im.replace('.' + settings.CHART_EXT, '')
                # to avoid comparing according to too many scoring methods
                if key_label == 'scoring_method' and sc_m_id not in methods_ids_dict:
                    continue
                scoring_methods_dict[sc_m_id].add(k)

    if len(scoring_methods_dict)==0:
        logger.warning("No methods for score vs number of changepoints analysis")
        return

    if section_level is not None:
        report.add_section(
            section_name="Score depending on the given number of changepoints",
            level=section_level,
            clearPage=kwargs.get('clearPage', False),
        )
    else:
        section_level = -1

    for scoring_method_id in sorted(scoring_methods_dict.keys()):
        ks_set = scoring_methods_dict[scoring_method_id]
        report.add_section(
            section_name=scoring_method_id,
            level=section_level + 1,
            clearPage=False,
        )
        images_list = []
        rows_captions_list = []
        for i_n,n_id in enumerate(networks_ids):
            networks_has_images = False
            for k in sorted(methods_ids_dict.keys()):
                m_ids = methods_ids_dict[k]
                if k not in ks_set:
                    continue
                images_dir = DynamicNetwork.get_network_methods_comparison_dir(n_id)\
                             + eval_utils.generate_methods_comparison_key(m_ids) + '/images/'
                images_list.append(images_dir + 'timeseries-n_changepoints-score/' + scoring_method_id)
                networks_has_images = True
            if networks_has_images:
                rows_captions_list.append("{n_id} network".format(n_id=n_id))

        if len(methods_ids_dict.keys()) == 1:
            columns_captions_list = None
            columns_count = 3
            images_captions_list = rows_captions_list
            rows_captions_list = None
            rows_descr = ""
        else:
            columns_captions_list = sorted(list(ks_set))
            columns_count = len(ks_set)
            images_captions_list = None
            rows_descr = ", rows correspond to networks"

        report.add_group_figure(
            title="Score depending on the given number of changepoints "
                  "(x-axis - given number of changepoints, y-axis - {scoring_method}, "
                  "highlighted values correspond to the methods' actual solutions, "
                  "stars correspond to the scores of ground truth methods)"
                  "{rows_descr}{sep}{cols_descr}"
            .format(
                scoring_method=scoring_methods_utils.get_scoring_method_title(scoring_method_id),
                rows_descr=rows_descr,
                sep="" if rows_descr == "" else ", ",
                cols_descr="columns - to %s" % key_label.replace('_', ' ') if key_label is not None else "",
            ),
            images_list=images_list,
            images_captions_list=images_captions_list,
            rows_captions_list=rows_captions_list,
            columns_captions_list=columns_captions_list,
            columns_count=columns_count,
            max_rows_per_figure=4,
        )


def add_section_partition_qualities_timeseries(report, methods_ids_dict, networks_ids, key_label=None,
                                               section_level=None, **kwargs):
    if section_level is not None:
        report.add_section(
            section_name="Timeseries of partition qualities over time",
            level=section_level,
            clearPage=kwargs.get('clearPage', False),
        )
    else:
        section_level = -1

    for metric in settings.DEFAULT_PARTITION_QUALITY_METRICS:
        metric_title = static_clustering.PARTITION_QUALITY_METRICS[metric]
        report.add_section(
            section_name=metric_title,
            level=section_level + 1,
            clearPage=False,
        )
        images_list = []
        rows_captions_list = []
        for i_n,n_id in enumerate(networks_ids):
            for k in sorted(methods_ids_dict.keys()):
                m_ids = methods_ids_dict[k]
                images_dir = DynamicNetwork.get_network_methods_comparison_dir(n_id)\
                             + eval_utils.generate_methods_comparison_key(m_ids) + '/images/'
                image_name = images_dir + 'timeseries-snapshot-partition_quality/{metric}'.format(
                    dir=images_dir,
                    metric=metric,
                )
                if not os.path.exists(image_name + '.' + settings.CHART_EXT):
                    continue
                images_list.append(image_name)
            rows_captions_list.append("{n_id} network".format(n_id=n_id))

        if len(methods_ids_dict.keys()) == 1:
            columns_captions_list = None
            columns_count = 3
            images_captions_list = rows_captions_list
            rows_captions_list = None
            rows_descr = ""
        else:
            columns_captions_list = sorted(methods_ids_dict.keys())
            columns_count = len(methods_ids_dict.keys())
            images_captions_list = None
            rows_descr = ", rows correspond to networks"

        report.add_group_figure(
            title="Timeseries of partition qualities over time (x-axis - time, y-axis - {metric})"
              "{rows_descr}{sep}{cols_descr}"
            .format(
                metric=metric_title,
                rows_descr=rows_descr,
                sep="" if rows_descr == "" else ", ",
                cols_descr="columns - to %s" % key_label.replace('_', ' ') if key_label is not None else "",
            ),
            images_list=images_list,
            images_captions_list=images_captions_list,
            rows_captions_list=rows_captions_list,
            columns_captions_list=columns_captions_list,
            columns_count=columns_count,
            max_rows_per_figure=4,
        )

