from ..latex_builder import LatexBuilder
import reports_utils
from .. import settings
from ..evaluation import eval_utils
from collections import defaultdict
from .. import utils
from ..log import logger
from ..methods import methods_utils
from collections import OrderedDict
import os
from ..evaluation import method_eval_across_networks
from ..scoring_methods import scoring_methods_utils
from ..com_evs_similarity_methods import com_evs_similarity_methods_utils

def create_compare_methods_networks_dict_report(networks_dict, methods_ids, **kwargs):

    if kwargs.get('run_eval', True):
        # create the required images, as necessary
        # one figure per method (each group in figure - network key, series - measures)
        if kwargs.get('draw_per_method', True):
            for m_id in methods_ids:
                method_eval_across_networks.evaluate_method_across_networks_dict(
                    method_id=m_id,
                    networks_dict=networks_dict,
                    **kwargs
                )

        # one figure per network key (each group in figure - measure, series - methods)
        # across network comparison evaluation
        # e.g., to get results with avg/std across networks
        if kwargs.get('draw_per_network_key', True):
            for n_key, networks_ids in networks_dict.iteritems():
                method_eval_across_networks.evaluate_methods_across_networks(
                    methods_ids=methods_ids,
                    networks_ids=networks_ids,
                    networks_list_key_name_in_dict=n_key.replace(' ', '_'),
                    **kwargs
                )

        # one figure per measure (each group in figure - network key, series - methods)
        # i.e. show results for all networks keys and methods in one figure
        if kwargs.get('draw_per_measure', True):
            method_eval_across_networks.evaluate_methods_across_networks_dict(
                methods_ids=methods_ids,
                networks_dict=networks_dict,
                **kwargs
            )

    report = LatexBuilder(
        report_name='eval',
        title="Basic analysis for {methods} methods in {networks} networks keys".format(
            methods=kwargs.get('methods_key_name') or ", ".join(methods_ids),
            networks=kwargs.get('networks_key_name') or ", ".join(networks_dict.keys()),
        ),
    )

    if kwargs.get('draw_com_evs_scores', True):
        add_section_scores_bar(
            report,
            methods_ids=methods_ids,
            networks_dict=networks_dict,
            section_level=0,
            **kwargs
        )

    if kwargs.get('draw_run_time', True):
        add_section_run_times(
            report,
            methods_ids=methods_ids,
            networks_dict=networks_dict,
            section_level=0,
            **kwargs
        )

    if kwargs.get('draw_against_other', True):
        comparison_methods_key = kwargs.get('methods_key_name') or eval_utils.generate_methods_comparison_key(methods_ids)
        comparison_networks_dict_key = kwargs.get('networks_key_name') or eval_utils.generate_networks_comparison_key(networks_dict)

        # first, collect all available other_methods_ids
        images_dirs = []
        if kwargs.get('draw_per_network_key', True):
            for n_k, networks_ids in networks_dict.iteritems():
                images_dirs.append(settings.EVAL_DIR + eval_utils.get_netwoks_list_comparison_key(
                        networks_ids, networks_list_key_name_in_dict=n_k.replace(' ', '_'), **kwargs
                    ) + '/' \
                                   + comparison_methods_key + '/images/' \
                                   + 'evaluation_against_other/')
        if kwargs.get('draw_per_method', True):
            for m_id in methods_ids:
                images_dirs.append(settings.EVAL_DIR + comparison_networks_dict_key + '/'\
                             + eval_utils.generate_methods_comparison_key([m_id]) + '/images/'\
                             + 'evaluation_against_other/')
        if kwargs.get('draw_per_measure', True):
            images_dirs.append(settings.EVAL_DIR + comparison_networks_dict_key + '/'\
                + comparison_methods_key + '/images/evaluation_against_other/')

        temp_other_methods_ids = []
        for images_dir in images_dirs:
            sub_dirs = utils.get_subdir_list_from_dir(images_dir)
            for sub_dir in sub_dirs:
                other_method_id = os.path.basename(sub_dir)
                temp_other_methods_ids.append(other_method_id)

        other_methods_ids = set(temp_other_methods_ids)
        for other_method_id in other_methods_ids:
            report.add_section(
                section_name="Against %s" % methods_utils.get_method_title(other_method_id),
                level=0,
                clearPage=True,
            )

            add_section_com_evs_similarity(
                report,
                methods_ids=methods_ids,
                networks_dict=networks_dict,
                other_method_id=other_method_id,
                section_level=1,
                clearPage=False,
                **kwargs
            )

            if kwargs.get('draw_changepoints_pr_roc', False):
                add_section_pr_roc_results(
                    report,
                    methods_ids=methods_ids,
                    networks_dict=networks_dict,
                    other_method_id=other_method_id,
                    classification_type='changepoints',
                    section_level=1,
                    clearPage=True,
                    **kwargs
                )


    report.compile(
        output_directory=settings.REPORTS_DIR + 'methods/' + kwargs.get('report_subdir_name', '')
                         + '/' + reports_utils.get_methods_networks_key(methods_ids, networks_dict.keys(), **kwargs) + '/',
        pdflatex=True
    )
    if kwargs.get('copy_to_tmp_dir', True):
        reports_utils.copy_report_to_tmp_dir(
            report,
            tmp_subdir='eval/',
            tmp_report_name=report.get_name() + '_' + kwargs.get('report_subdir_name', ''),
        )

def add_section_scores_bar(report, methods_ids, networks_dict, section_level=None,
                           **kwargs):
    if section_level is not None:
        report.add_section(
            section_name="Scores of the solutions",
            level=0,
            clearPage=True,
        )

    figure_types = []
    if kwargs.get('draw_per_method', True):
        figure_types.append('per_method')
    if kwargs.get('draw_per_network_key', True):
        figure_types.append('per_network_key')
    if kwargs.get('draw_per_measure', True):
        figure_types.append('per_measure')
    if len(figure_types) == 0:
        logger.warning("No per_network_key/per_method/per_measure options selected for run times section")
        return

    comparison_methods_key = kwargs.get('methods_key_name') or eval_utils.generate_methods_comparison_key(methods_ids)
    comparison_networks_key = kwargs.get('networks_key_name') or eval_utils.generate_networks_comparison_key(networks_dict)

    for figure_type in figure_types:
        report.add_section(
            section_name=figure_type.capitalize().replace("_", " "),
            level=section_level + 1,
            clearPage=kwargs.get('clearPage', False),
        )
        # dict - score methods list type => list of comparisons
        # first construct this dict and the use it to add figures
        # (types can be e.g., default - using default fixed set of scoring methods, or custom - selected based on methods)
        scoring_methods_type_dict = defaultdict(set)
        if figure_type == 'per_network_key':
            for k, n_ids in networks_dict.iteritems():
                images_dir = settings.EVAL_DIR\
                    + eval_utils.get_netwoks_list_comparison_key(
                        n_ids, networks_list_key_name_in_dict=k.replace(' ', '_'), **kwargs
                    ) + '/' + comparison_methods_key + '/images/bar-score/'
                for im in utils.get_file_list_from_dir(images_dir, full_path=False):
                    sc_m_type_id = im.replace('.' + settings.CHART_EXT, '')
                    # to avoid comparing according to too many scoring methods
                    scoring_methods_type_dict[sc_m_type_id].add(k)
        elif figure_type == 'per_method':
            for m_id in methods_ids:
                images_dir = settings.EVAL_DIR + comparison_networks_key + '/'\
                             + eval_utils.generate_methods_comparison_key([m_id]) + '/images/bar-score/'
                for im in utils.get_file_list_from_dir(images_dir, full_path=False):
                    sc_m_type_id = im.replace('.' + settings.CHART_EXT, '')
                    # to avoid comparing according to too many scoring methods
                    scoring_methods_type_dict[sc_m_type_id].add(m_id)
        elif figure_type == 'per_measure':
            images_dir = settings.EVAL_DIR + comparison_networks_key + '/'\
                         + comparison_methods_key + '/images/bar-score/'
            for im in utils.get_file_list_from_dir(images_dir, full_path=False):
                sc_m_type_id = im.replace('.' + settings.CHART_EXT, '')
                # to avoid comparing according to too many sim pairs
                scoring_methods_type_dict[sc_m_type_id].add('all')

        if len(scoring_methods_type_dict)==0:
            logger.warning("No scoring methods for solutions score analysis")
            continue

        images_list = []
        images_captions_list = []
        corr_images_list = []
        corr_images_captions_list = []
        for sc_m_type, ks_set in scoring_methods_type_dict.iteritems():

            if figure_type != 'per_measure':
                images_list = []
                images_captions_list = []
                corr_images_list = []
                corr_images_captions_list = []

            if figure_type == 'per_network_key':
                for k in networks_dict.keys():
                    if k not in ks_set:
                        continue
                    n_ids = networks_dict[k]
                    images_dir = settings.EVAL_DIR + eval_utils.get_netwoks_list_comparison_key(
                        n_ids, networks_list_key_name_in_dict=k.replace(' ', '_'), **kwargs
                    ) + '/' + comparison_methods_key + '/images/bar-score/'
                    images_list.append(images_dir + sc_m_type)
                    images_captions_list.append(k)

                    corr_matr_images_dir = images_dir + 'pairwise_corr_matr/'
                    if os.path.exists(corr_matr_images_dir):
                        for im in utils.get_file_list_from_dir(corr_matr_images_dir, full_path=False):
                            im_no_ext = im.replace('.' + settings.CHART_EXT, '')
                            if not im_no_ext.startswith(sc_m_type + '-'):
                                continue
                            corr_images_list.append(corr_matr_images_dir + im_no_ext)
                            corr_images_captions_list.append(
                            '%s - %s' % (k, im_no_ext.replace(sc_m_type + '-', '').capitalize())
                           )
            elif figure_type == 'per_method':
                for m_id in methods_ids:
                    if m_id not in ks_set:
                        continue
                    images_dir = settings.EVAL_DIR + comparison_networks_key + '/'\
                                 + eval_utils.generate_methods_comparison_key([m_id]) + '/images/bar-score/'
                    images_list.append(images_dir + sc_m_type)
                    images_captions_list.append(methods_utils.get_method_title(m_id))

                    corr_matr_images_dir = images_dir + 'pairwise_corr_matr/'
                    if os.path.exists(corr_matr_images_dir):
                        for im in utils.get_file_list_from_dir(corr_matr_images_dir, full_path=False):
                            im_no_ext = im.replace('.' + settings.CHART_EXT, '')
                            if not im_no_ext.startswith(sc_m_type + '-'):
                                continue
                            corr_images_list.append(corr_matr_images_dir + im_no_ext)
                            corr_images_captions_list.append(
                            '%s - %s' % (methods_utils.get_method_title(m_id),
                                         im_no_ext.replace(sc_m_type + '-', '').capitalize())
                           )
            elif figure_type == 'per_measure':
                    images_dir = settings.EVAL_DIR + comparison_networks_key + '/'\
                         + comparison_methods_key + '/images/bar-score/'
                    images_list.append(images_dir + sc_m_type)
                    images_captions_list.append(scoring_methods_utils.get_scoring_method_title(sc_m_type))

            if len(images_list) == 0:
                logger.debug("No images for scoring method type {sc_m_type}".format(
                    sc_m_type=sc_m_type,
                ))
                continue

            if figure_type != 'per_measure':
                report.add_group_figure(
                    title="Scores of the solutions according to {sc_m_type} scoring methods "
                          "with avg/std"
                        .format(
                        sc_m_type=sc_m_type,
                    ),
                    images_list=images_list,
                    images_captions_list=images_captions_list,
                    columns_count=3,
                    max_rows_per_figure=4,
                )

                if len(corr_images_list) > 0:
                    report.add_group_figure(
                        title="Pairwise correlations for {sc_m_type} scoring methods".format(
                            sc_m_type=sc_m_type,
                        ),
                        images_list=corr_images_list,
                        images_captions_list=corr_images_captions_list,
                        columns_count=4,
                        max_rows_per_figure=6,
                    )


        if figure_type == 'per_measure' and len(images_list) > 0:
            report.add_group_figure(
                title="Scores of the solutions with avg/std",
                images_list=images_list,
                images_captions_list=images_captions_list,
                columns_count=3,
                max_rows_per_figure=4,
            )

            slopegraph_rank_images_list = []
            slopegraph_rank_images_captions_list = []
            images_dir = settings.EVAL_DIR + comparison_networks_key + '/'\
                 + comparison_methods_key + '/images/bar-score/'
            slopegraph_rank_images_dir = images_dir + 'slopegraph-ranking/'
            if os.path.exists(slopegraph_rank_images_dir):
                for sc_m_type, ks_set in scoring_methods_type_dict.iteritems():
                    slopegraph_rank_images_list.append(slopegraph_rank_images_dir + sc_m_type)
                    slopegraph_rank_images_captions_list.append(sc_m_type)
            if len(slopegraph_rank_images_list) > 0:
                report.add_group_figure(
                    title="Score rankings",
                    images_list=slopegraph_rank_images_list,
                    images_captions_list=slopegraph_rank_images_captions_list,
                    columns_count=2,
                    max_rows_per_figure=4,
                )

def add_section_run_times(report, methods_ids, networks_dict,section_level=None,
                                            **kwargs):
    if section_level is not None:
        report.add_section(
            section_name="Running times",
            level=section_level,
            clearPage=kwargs.get('clearPage', False),
        )
    else:
        section_level = -1

    figure_types = []
    if kwargs.get('draw_per_method', True):
        figure_types.append('per_method')
    if kwargs.get('draw_per_network_key', True):
        figure_types.append('per_network_key')
    if kwargs.get('draw_per_measure', True):
        figure_types.append('per_measure')
    if len(figure_types) == 0:
        logger.warning("No per_network_key/per_method/per_measure options selected for run times section")
        return

    comparison_methods_key = kwargs.get('methods_key_name') or eval_utils.generate_methods_comparison_key(methods_ids)
    comparison_networks_key = kwargs.get('networks_key_name') or eval_utils.generate_networks_comparison_key(networks_dict)

    for figure_type in figure_types:
        report.add_section(
            section_name=figure_type.capitalize().replace("_", " "),
            level=section_level + 1,
            clearPage=kwargs.get('clearPage', False),
        )

        images_list = []
        images_captions_list = []
        if figure_type == 'per_network_key':
            for k, networks_ids in networks_dict.iteritems():
                images_dir = settings.EVAL_DIR + eval_utils.get_netwoks_list_comparison_key(
                        networks_ids, networks_list_key_name_in_dict=k.replace(' ', '_'), **kwargs
                    ) + '/' + comparison_methods_key + '/images/'
                images_list.append(images_dir + 'bar-run_time')
                images_captions_list.append(k)
        elif figure_type == 'per_method':
            for m_id in methods_ids:
                images_dir = settings.EVAL_DIR + comparison_networks_key + '/'\
                             + eval_utils.generate_methods_comparison_key([m_id]) + '/images/'
                images_list.append(images_dir + 'bar-run_time')
                images_captions_list.append(methods_utils.get_method_title(m_id))
        elif figure_type == 'per_measure':
            images_dir = settings.EVAL_DIR + comparison_networks_key + '/'\
                         + comparison_methods_key + '/images/'
            images_list.append(images_dir + 'bar-run_time')
            images_captions_list = None

        report.add_group_figure(
            title="Running times of the methods with avg/std",
            images_list=images_list,
            images_captions_list=images_captions_list,
            columns_count=3,
            max_rows_per_figure=4,
        )

        if figure_type == 'per_measure':
            slopegraph_rank_images_list = []
            images_dir = settings.EVAL_DIR + comparison_networks_key + '/'\
                             + comparison_methods_key + '/images/'
            slopegraph_rank_images_dir = images_dir + 'slopegraph-ranking/'
            if os.path.exists(slopegraph_rank_images_dir):
                slopegraph_rank_images_list.append(slopegraph_rank_images_dir + 'run_time')
            if len(slopegraph_rank_images_list) > 0:
                report.add_group_figure(
                    title="Running time rankings",
                    images_list=slopegraph_rank_images_list,
                    images_captions_list=None,
                    columns_count=1,
                    max_rows_per_figure=4,
                )

def add_section_pr_roc_results(report, methods_ids, networks_dict,
                                               other_method_id, classification_type, key_label=None, section_level=None, **kwargs):
    if section_level is not None:
        report.add_section(
            section_name="Classification results for changepoints",
            level=1,
            clearPage=kwargs.get('clearPage', False),
        )
    else:
        section_level = -1
        
    figure_types = []
    if kwargs.get('draw_per_method', True):
        figure_types.append('per_method')
    if kwargs.get('draw_per_network_key', True):
        figure_types.append('per_network_key')
    if kwargs.get('draw_per_measure', True):
        figure_types.append('per_measure')
    if len(figure_types) == 0:
        logger.warning("No per_network_key/per_method/per_measure options selected for PR/ROC section")
        return

    comparison_methods_key = kwargs.get('methods_key_name') or eval_utils.generate_methods_comparison_key(methods_ids)
    comparison_networks_key = kwargs.get('networks_key_name') or eval_utils.generate_networks_comparison_key(networks_dict)

    for figure_type in figure_types:
        report.add_section(
            section_name=figure_type.capitalize().replace("_", " "),
            level=section_level + 1,
            clearPage=False,
        )
        chart_types = OrderedDict()
        chart_types['bars'] = "Bars"
        for chart_type, chart_type_title in chart_types.iteritems():
            images_list = []
            images_captions_list = []
            if figure_type == 'per_network_key':
                for k, networks_ids in networks_dict.iteritems():
                    images_dir = settings.EVAL_DIR + eval_utils.get_netwoks_list_comparison_key(
                            networks_ids, networks_list_key_name_in_dict=k.replace(' ', '_'), **kwargs
                        ) + '/' + comparison_methods_key + '/images/'\
                              + 'evaluation_against_other/' + other_method_id + '/'
                    images_dir += 'pr_roc/' + classification_type + '/'
                    if not os.path.exists(images_dir):
                        continue
                    images_list.append(images_dir + chart_type)
                    images_captions_list.append(k)
            elif figure_type == 'per_method':
                 for m_id in methods_ids:
                    images_dir = settings.EVAL_DIR + comparison_networks_key + '/'\
                             + eval_utils.generate_methods_comparison_key([m_id]) + '/images/'\
                              + 'evaluation_against_other/' + other_method_id + '/'
                    images_dir += 'pr_roc/' + classification_type + '/'
                    if not os.path.exists(images_dir):
                        continue
                    images_list.append(images_dir + chart_type)
                    images_captions_list.append(methods_utils.get_method_title(m_id))
            elif figure_type == 'per_measure':
                images_dir = settings.EVAL_DIR + comparison_networks_key + '/'\
                         + comparison_methods_key + '/images/'\
                          + 'evaluation_against_other/' + other_method_id + '/'
                images_dir += 'pr_roc/' + classification_type + '/'
                if not os.path.exists(images_dir):
                    continue
                for im in sorted(utils.get_file_list_from_dir(images_dir, full_path=False)):
                    cl_value_type = im.replace('.' + settings.CHART_EXT, '')
                    images_list.append(images_dir + cl_value_type)
                    images_captions_list.append(cl_value_type.replace('_',' ').upper())

            if len(images_list) == 0:
                logger.warning("No images of type '%s'" % chart_type)
                continue

            report.add_group_figure(
                title=chart_type_title,
                images_list=images_list,
                images_captions_list=images_captions_list,
                columns_count=3,
                max_rows_per_figure=4,
            )

            if figure_type == 'per_measure':
                slopegraph_rank_images_list = []
                slopegraph_rank_images_captions_list = []
                images_dir = settings.EVAL_DIR + comparison_networks_key + '/'\
                         + comparison_methods_key + '/images/'\
                          + 'evaluation_against_other/' + other_method_id + '/'
                images_dir += 'pr_roc/' + classification_type + '/'
                slopegraph_rank_images_dir = images_dir + 'slopegraph-ranking/'
                if os.path.exists(slopegraph_rank_images_dir):
                    for im in sorted(utils.get_file_list_from_dir(slopegraph_rank_images_dir, full_path=False)):
                        im_no_ext = im.replace('.' + settings.CHART_EXT, '')
                        slopegraph_rank_images_list.append(slopegraph_rank_images_dir + im_no_ext)
                        slopegraph_rank_images_captions_list.append(
                            im_no_ext.replace('_',' ').upper()
                        )
                if len(slopegraph_rank_images_list) > 0:
                    report.add_group_figure(
                        title="Classification result rankings",
                        images_list=slopegraph_rank_images_list,
                        images_captions_list=slopegraph_rank_images_captions_list,
                        columns_count=2,
                        max_rows_per_figure=4,
                    )

# TODO: refactor similar code inside functions (e.g., for different figure types) and across functions (e.g., scores and similarities)
def add_section_com_evs_similarity(report, methods_ids, networks_dict,
                                   other_method_id,section_level=None, **kwargs):
    if section_level is not None:
        report.add_section(
            section_name="Community evolutions similarities",
            level=section_level,
            clearPage=kwargs.get('clearPage', False),
        )
        
    figure_types = []
    if kwargs.get('draw_per_method', True):
        figure_types.append('per_method')
    if kwargs.get('draw_per_network_key', True):
        figure_types.append('per_network_key')
    if kwargs.get('draw_per_measure', True):
        figure_types.append('per_measure')
    if len(figure_types) == 0:
        logger.warning("No per_network_key/per_method/per_measure options selected for com evs similarity section")
        return

    comparison_methods_key = kwargs.get('methods_key_name') or eval_utils.generate_methods_comparison_key(methods_ids)
    comparison_networks_key = kwargs.get('networks_key_name') or eval_utils.generate_networks_comparison_key(networks_dict)

    for figure_type in figure_types:
        report.add_section(
            section_name=figure_type.capitalize().replace("_", " "),
            level=section_level + 1,
            clearPage=False,
        )
        # sim_m_type_dict - similarity methods pair => list of comparisons
        # first construct this dict and then use it to add figures
        # e.g., for per method, it may happen that not all methods have comparison of all types
        # method 1 - measures 1,2, method 2 - measures 1,3, method 3 - measures 2,3 =>
        # produce group figure for measure 1 with figures for methods 1 and 2,
        # measure 2 - methods 1 and 3 and measure 3- methods 2 and 3
        for chart_type in ['bar', 'scatter']:
            sim_m_type_dict = defaultdict(set)
            if figure_type == 'per_network_key':
                for k, networks_ids in networks_dict.iteritems():
                    images_dir = settings.EVAL_DIR + eval_utils.get_netwoks_list_comparison_key(
                        networks_ids, networks_list_key_name_in_dict=k.replace(' ', '_'), **kwargs
                    ) + '/' + comparison_methods_key + '/images/'\
                                  + 'evaluation_against_other/' + other_method_id + '/%s-com_evs_similarity/' % chart_type
                    for im in utils.get_file_list_from_dir(images_dir, full_path=False):
                        sim_m_pair_id = im.replace('.' + settings.CHART_EXT, '')
                        # to avoid comparing according to too many sim pairs
                        sim_m_type_dict[sim_m_pair_id].add(k)
            elif figure_type == 'per_method':
                for m_id in methods_ids:
                    images_dir = settings.EVAL_DIR + comparison_networks_key + '/'\
                                 + eval_utils.generate_methods_comparison_key([m_id]) + '/images/'\
                                  + 'evaluation_against_other/' + other_method_id + '/%s-com_evs_similarity/' % chart_type
                    for im in utils.get_file_list_from_dir(images_dir, full_path=False):
                        sim_m_pair_id = im.replace('.' + settings.CHART_EXT, '')
                        # to avoid comparing according to too many sim pairs
                        sim_m_type_dict[sim_m_pair_id].add(m_id)
            elif figure_type == 'per_measure':
                images_dir = settings.EVAL_DIR + comparison_networks_key + '/'\
                    + comparison_methods_key + '/images/'\
                    + 'evaluation_against_other/' + other_method_id + '/%s-com_evs_similarity/' % chart_type
                for im in utils.get_file_list_from_dir(images_dir, full_path=False):
                    sim_m_pair_id = im.replace('.' + settings.CHART_EXT, '')
                    # to avoid comparing according to too many sim pairs
                    sim_m_type_dict[sim_m_pair_id].add('all')
            if figure_type != 'per_measure' and len(sim_m_type_dict)==0:
                logger.warning("No similarity methods pairs for %s plots for %s" % (chart_type, figure_type))
                continue

            images_list = []
            images_captions_list = []
            corr_images_list = []
            corr_images_captions_list = []
            for sim_m_type, ks_set in sim_m_type_dict.iteritems():
                if figure_type != 'per_measure':
                    images_list = []
                    images_captions_list = []
                    corr_images_list = []
                    corr_images_captions_list = []

                if figure_type == 'per_network_key':
                    for k in networks_dict.keys():
                        if k not in ks_set:
                            continue
                        networks_ids = networks_dict[k]
                        images_dir = settings.EVAL_DIR + eval_utils.get_netwoks_list_comparison_key(
                                networks_ids, networks_list_key_name_in_dict=k.replace(' ', '_'), **kwargs
                            ) + '/' + comparison_methods_key + '/images/'\
                            + 'evaluation_against_other/' + other_method_id + '/%s-com_evs_similarity/' % chart_type
                        images_list.append(images_dir + sim_m_type)
                        images_captions_list.append(k)

                        corr_matr_images_dir = images_dir + 'pairwise_corr_matr/'
                        if os.path.exists(corr_matr_images_dir):
                            for im in utils.get_file_list_from_dir(corr_matr_images_dir, full_path=False):
                                im_no_ext = im.replace('.' + settings.CHART_EXT, '')
                                if not im_no_ext.startswith(sim_m_type + '-'):
                                    continue
                                corr_images_list.append(corr_matr_images_dir + im_no_ext)
                                corr_images_captions_list.append(
                                '%s - %s' % (k, im_no_ext.replace(sim_m_type + '-', '').capitalize())
                               )
                elif figure_type == 'per_method':
                    for m_id in methods_ids:
                        if m_id not in ks_set:
                            continue
                        images_dir = settings.EVAL_DIR + comparison_networks_key + '/'\
                             + eval_utils.generate_methods_comparison_key([m_id]) + '/images/'\
                              + 'evaluation_against_other/' + other_method_id + '/%s-com_evs_similarity/' % chart_type
                        images_list.append(images_dir + sim_m_type)
                        images_captions_list.append(methods_utils.get_method_title(m_id))

                        corr_matr_images_dir = images_dir + 'pairwise_corr_matr/'
                        if os.path.exists(corr_matr_images_dir):
                            for im in utils.get_file_list_from_dir(corr_matr_images_dir, full_path=False):
                                im_no_ext = im.replace('.' + settings.CHART_EXT, '')
                                if not im_no_ext.startswith(sim_m_type + '-'):
                                    continue
                                corr_images_list.append(corr_matr_images_dir + im_no_ext)
                                corr_images_captions_list.append(
                                '%s - %s' % (methods_utils.get_method_title(m_id),
                                             im_no_ext.replace(sim_m_type + '-', '').capitalize())
                               )
                elif figure_type == 'per_measure':
                    images_dir = settings.EVAL_DIR + comparison_networks_key + '/'\
                         + comparison_methods_key + '/images/'\
                          + 'evaluation_against_other/' + other_method_id + '/%s-com_evs_similarity/' % chart_type
                    images_list.append(images_dir + sim_m_type)
                    images_captions_list.append(com_evs_similarity_methods_utils.get_com_evs_similarity_title(sim_m_type))

                if len(images_list) == 0:
                    logger.debug("No images for {chart_type} and similarity {sim_m_type}".format(
                        chart_type=chart_type,
                        sim_m_type=sim_m_type,
                    ))
                    continue

                if figure_type != 'per_measure':
                    report.add_group_figure(
                        title="Similarity {chart_type} plot scores with avg/std according to {sim}".format(
                            chart_type=chart_type,
                            sim=sim_m_type.replace('-vs-',' vs ') if chart_type == 'scatter' else sim_m_type
                        ),
                        images_list=images_list,
                        images_captions_list=images_captions_list,
                        columns_count=3,
                        max_rows_per_figure=4,
                    )

                    if len(corr_images_list) > 0:
                        report.add_group_figure(
                            title="Pairwise correlations for {sim} similarity methods".format(
                                sim=sim_m_type,
                            ),
                            images_list=corr_images_list,
                            images_captions_list=corr_images_captions_list,
                            columns_count=4,
                            max_rows_per_figure=6,
                        )

            if figure_type == 'per_measure' and len(images_list) > 0:
                report.add_group_figure(
                    title="Similarity {chart_type} plot scores with avg/std".format(
                        chart_type=chart_type,
                    ),
                    images_list=images_list,
                    images_captions_list=images_captions_list,
                    columns_count=3,
                    max_rows_per_figure=4,
                )

                slopegraph_rank_images_list = []
                slopegraph_rank_images_captions_list = []
                images_dir = settings.EVAL_DIR + comparison_networks_key + '/'\
                    + comparison_methods_key + '/images/'\
                    + 'evaluation_against_other/' + other_method_id + '/%s-com_evs_similarity/' % chart_type
                slopegraph_rank_images_dir = images_dir + 'slopegraph-ranking/'
                if os.path.exists(slopegraph_rank_images_dir):
                    for sim_m_type, ks_set in sim_m_type_dict.iteritems():
                        slopegraph_rank_images_list.append(slopegraph_rank_images_dir + sim_m_type)
                        slopegraph_rank_images_captions_list.append(sim_m_type)
                if len(slopegraph_rank_images_list) > 0:
                    report.add_group_figure(
                        title="Similarity rankings",
                        images_list=slopegraph_rank_images_list,
                        images_captions_list=slopegraph_rank_images_captions_list,
                        columns_count=2,
                        max_rows_per_figure=4,
                    )