import os
from ..log import logger
from ..latex_builder import LatexBuilder
from ..community_evolution import CommunityEvolution
from ..partition import Partition
from ..affinity_matrix import AffinityMatrix
from ..dynamic_network import DynamicNetwork
from ..method import Method
import reports_utils
from .. import utils
from .. import static_clustering
from ..methods import methods_utils
from .. import settings
from ..evaluation import method_eval

def create_network_methods_description_reports(dynamic_network, methods_ids_list=None, **kwargs):
    logger.debug("Creating description report for computed methods of '{network_id}' network "
                 "(except ground truths)".format(
        network_id=dynamic_network.get_id()
    ))
    methods_ids_list = methods_ids_list or dynamic_network.get_computed_methods_ids_list()
    for i_c, method_id in enumerate(methods_ids_list):
        com_ev = CommunityEvolution.load(dynamic_network.get_id(), method_id)
        logger.info("{i_c}/{n_c}. Generating report for '{method_title}' method".format(
            i_c=i_c+1,
            n_c=len(methods_ids_list),
            method_title=com_ev.get_method_title(),
        ))

        # create the required images, as necessary
        method_eval.evaluate_method(
            Method.load_state(com_ev.get_network_id(), com_ev.get_method_id()),
            overwrite=kwargs.get('overwrite', False),
        )

        report = LatexBuilder(
            report_name='summary',
            title="Summary report for '{network_id}' network and '{method_title}' method".format(
                network_id=dynamic_network.get_id(),
                method_title=com_ev.get_method_title(),
            ),
        )

        add_section_method_summary(report, com_ev)

        report.compile(
            output_directory=dynamic_network.get_reports_dir() + 'methods/' + com_ev.get_method_id() + '/',
            pdflatex=True
        )

        # for convenience, copy all reports to tmp dir
        if kwargs.get('copy_to_tmp_dir', True):
            tmp_output_dir = reports_utils.copy_report_to_tmp_dir(
                report,
                tmp_subdir='{network_id}/methods/'.format(
                    network_id=dynamic_network.get_id().replace('/','-'),
                ),
                tmp_report_name='{method_id}'.format(
                    method_id=com_ev.get_method_id(),
                ),
            )
            reports_utils.copy_sankey_partitions_to_tmp_dir(com_ev, tmp_output_dir)

def add_section_method_summary(report, c_ev, section_level=None):
    """
    Add section summarizing method
    :param report:
    :param CommunityEvolution c_ev:
    :param section_level:
    :return:
    """
    if section_level is not None:
        report.add_section(
            section_name=c_ev.get_method_title(),
            level=section_level,
            clearPage=True,
        )
    else:
        section_level = -1
    if c_ev.get_description() is not None:
        report.add_section(
            "Description",
            level=section_level + 1
        )
        report.add_text(c_ev.get_description())

    add_figure_segmentation(
        report,
        input_dir=c_ev.get_images_dir(),
        image_caption="Time interval segmentation for %s" % c_ev.get_method_title(),
        section_level=section_level + 1,
        clearPage=False,
    )

    add_section_method_analysis(
        report,
        images_dir=Method.get_method_images_dir(c_ev.get_network_id(), c_ev.get_method_id()),
        section_level=section_level + 1,
        clearPage=True,
    )

    add_figure_timeseries_partitions_qualities(
        report,
        images_dir=c_ev.get_images_dir(),
        section_level=section_level + 1,
    )
    add_section_heatmaps_partitions_qualities(
        report,
        images_dir=c_ev.get_images_dir(),
        section_level=section_level + 1,
    )

    add_figure_heatmaps_partitions_similarities(
        report,
        images_dir=c_ev.get_images_dir(),
        section_level=section_level + 1,
        is_per_t=False,
    )

    # show precomputed individual one vs one comparison for other methods
    for comp_type in c_ev.get_comparison_types_list(only_pairwise=True):
        comp_methods_ids = DynamicNetwork.get_compared_methods_ids_by_key(
            c_ev.get_network_id(),
            comp_type
        )

        c_ev_other_method_id = comp_methods_ids[0] if comp_methods_ids[0] != c_ev.get_method_id()\
            else comp_methods_ids[1]

        comparison_dir = DynamicNetwork.get_network_methods_comparison_dir(c_ev.get_network_id())\
                         + comp_type + '/images/'

        report.add_section(
            section_name="Comparison vs %s" % methods_utils.get_method_title(c_ev_other_method_id),
            level=section_level + 1,
            clearPage=False,
        )

        add_figure_segmentation(
            report,
            input_dir=comparison_dir,
            image_caption="Time interval segmentation for {methods_ids_list}".format(
                methods_ids_list=", ".join(comp_methods_ids)
            ),
            section_level=section_level + 2,
        )

        add_section_method_analysis(
            report,
            images_dir=comparison_dir,
            section_level=section_level + 2,
        )

        add_figure_timeseries_partitions_qualities(
            report,
            images_dir=comparison_dir,
            section_level=section_level + 2,
        )

        # only for one-vs-one comparison
        add_figure_timeseries_partitions_similaties(
            report,
            comparison_dir,
            section_level=section_level + 2,
        )
        for is_per_t in [False, True]:
            add_figure_heatmaps_partitions_similarities(
                report,
                comparison_dir,
                section_level=section_level + 2,
                is_per_t=is_per_t,
            )
        add_section_pr_roc_results(
            report,
            comparison_dir,
            classification_type='changepoints',
            section_level=section_level + 2,
        )
        add_section_com_evs_similarity_bars(
            report,
            comparison_dir,
            section_level=section_level + 2,
        )


def add_section_heatmaps_partitions_qualities(report, images_dir, section_level=0):
    report.add_section(
        section_name="Heatmaps of partition qualities when applied to different snapshots",
        level=section_level,
        clearPage=True,
    )
    figures = []
    captions = []
    images_dir = images_dir + 'heatmaps-partitions-qualities-per-t/'
    for metric in utils.get_file_list_from_dir(images_dir, ext=settings.CHART_EXT, full_path=False):
        metric = metric.replace('.' + settings.CHART_EXT, '')
        metric_title = static_clustering.PARTITION_QUALITY_METRICS[metric]
        figures.append(
            "{dir}{metric}".format(
                dir=images_dir + 'heatmaps-partitions-qualities-per-t/',
                metric=metric,
            )
        )
        captions.append(metric_title)
    report.add_group_figure(
        title="Heatmaps of partition qualities when applied to different snapshots (vertical ticks - partitions, "
              "horizontal - snapshots, dark vertical dashed lines - time segmentation)",
        images_list=figures,
        images_captions_list=captions,
        columns_count=min(2, len(figures)),
    )


def add_figure_segmentation(report, input_dir, section_level=None, image_caption='', **kwargs):
    if section_level is not None:
        report.add_section(
            section_name="Time interval segmentation",
            level=section_level,
            clearPage=kwargs.get('clearPage', False),
        )
    report.add_group_figure(
        title="Time interval segmentation (x-axis - timepoints, y-axis - methods)",
        images_list=[input_dir + 'segmentation'],
        images_captions_list=[image_caption],
        columns_count=1,
    )


def add_section_pr_roc_results(report, images_dir, classification_type, section_level=None):
    images_dir += 'pr_roc/' + classification_type + '/'
    if not os.path.exists(images_dir):
        logger.warning("No classification results: %s directory does not exists" % images_dir)
        return

    if section_level is not None:
        report.add_section(
            section_name="Classification results for %s" % classification_type,
            level=section_level,
            clearPage=False,
        )

    report.add_group_figure(
        title="Curves",
        images_list=[images_dir + 'pr', images_dir + 'roc'],
        images_captions_list=["Precision-Recall", "Receiver Operating Characteristic"],
        columns_count=2,
    )

    report.add_group_figure(
        title="Bars",
        images_list=[images_dir + 'bars'],
        images_captions_list=["AUPR, AUROC and Max F1-score"],
    )

def add_section_com_evs_similarity_bars(report, images_dir, section_level=None):
    if not os.path.exists(images_dir + 'bar-com_evs_similarity.' + settings.CHART_EXT):
        logger.warning("No community evolutions similarity comparison bars: %s does not exists"
                       % (images_dir + 'bar-com_evs_similarity.' + settings.CHART_EXT))
        return

    if section_level is not None:
        report.add_section(
            section_name="Community evolutions similarities",
            level=section_level,
            clearPage=False,
        )
    report.add_group_figure(
        title="Similarity scores",
        images_list=[images_dir + 'bar-com_evs_similarity'],
        images_captions_list=["Scores according to different metrics"],
    )

def add_figure_timeseries_partitions_similaties(report, images_dir, section_level=None):
    if section_level is not None:
        report.add_section(
            section_name="Timeseries of partitions similarities",
            level=section_level,
            clearPage=False,
        )

    figures = []
    captions = []
    images_dir = images_dir + 'timeseries-snapshot-partitions_similarity/'
    for metric in utils.get_file_list_from_dir(images_dir, ext=settings.CHART_EXT, full_path=False):
        metric = metric.replace('.' + settings.CHART_EXT, '')
        metric_title = Partition.SIMILARITY_METRICS.get(metric, AffinityMatrix.PARTITION_SIMILARITY_METRICS.get(metric))
        figures.append(
            "{dir}{metric}".format(
                dir=images_dir,
                metric=metric,
            )
        )
        captions.append(metric_title)

    report.add_group_figure(
        title="Timeseries of snapshot partition similarities (x-axis - time, y-axis - similarity)",
        images_list=figures,
        images_captions_list=captions,
        columns_count=min(2, len(figures)),
    )

def add_figure_heatmaps_partitions_similarities(report, images_dir, section_level=None, is_per_t=False):
    if section_level is not None:
        report.add_section(
            section_name="Heatmaps of partition similarities per " + ("timepoint" if is_per_t else "segment"),
            level=section_level,
            clearPage=False,
        )
    figures = []
    captions = []
    images_dir = '{input_dir}/heatmaps-partitions'.format(input_dir=images_dir) + ('-per-t' if is_per_t else '') + '/'
    for metric in utils.get_file_list_from_dir(images_dir, ext=settings.CHART_EXT, full_path=False):
        metric = metric.replace('.' + settings.CHART_EXT, '')
        metric_title = Partition.SIMILARITY_METRICS[metric]
        figures.append(
            "{dir}{metric}".format(
                dir=images_dir,
                metric=metric,
            )
        )
        captions.append(metric_title)

    report.add_group_figure(
        title="Heatmaps of partition similarities (ticks correspond to %s)" % ("timepoints" if is_per_t else "segments"),
        images_list=figures,
        images_captions_list=captions,
        columns_count=min(2, len(figures)),
    )

def add_section_method_analysis(report, images_dir, section_level=0, **kwargs):
    has_analysis = os.path.exists(images_dir + 'timeseries-n_changepoints-score/')\
                   or os.path.exists(images_dir + 'timeseries-t-score.' + settings.CHART_EXT)\
                   or os.path.exists(images_dir + 'segmentation-n_changepoints.' + settings.CHART_EXT)
    if not has_analysis:
        # note: comparison against ground truth does not include analysis
        logger.warning("No methods analysis in %s" % images_dir)
        return
    report.add_section(
        section_name="Analysis",
        level=section_level,
        clearPage=kwargs.get('clearPage', False),
    )

    images_list = []
    images_captions_list = []
    for im in utils.get_file_list_from_dir(
        images_dir + 'timeseries-n_changepoints-score/',
        full_path=True
    ):
        im = im.replace('.' + settings.CHART_EXT,'')
        images_list.append(im)
        images_captions_list.append(os.path.basename(im))

    report.add_group_figure(
        title="Score depending on the given number of changepoints "
              "(x-axis - given number of changepoints, y-axis - score, "
              "highlighted values correspond to the methods' actual solutions, "
              "stars correspond to the scores of ground truth methods)",
        images_list=images_list,
        images_captions_list=images_captions_list,
        columns_count=min(2, len(images_list)),
    )

    if os.path.exists(images_dir + 'segmentation-n_changepoints.' + settings.CHART_EXT):
        report.add_group_figure(
            title="Comparison of segmentations for different numbers of changepoints",
            images_list=[images_dir + 'segmentation-n_changepoints'],
            images_captions_list=['Segmentations'],
            columns_count=1,
        )
    else:
        logger.debug('No comparison of segmentations for different fixed numbers of changepoints')

    report.add_group_figure(
        title="Scores summarizing how changepoint-like is each timepoint"
              "(x-axis - given timepoint, y-axis - score)",
        images_list=[images_dir + 'timeseries-t-score'],
        images_captions_list=['Timeseries of scores'],
        columns_count=1,
    )

def add_figure_timeseries_partitions_qualities(report, images_dir, section_level=0):
    report.add_section(
        section_name="Timeseries of partition qualities over time",
        level=section_level,
        clearPage=True,
    )
    figures = []
    captions = []
    images_dir = "{dir}timeseries-snapshot-partition_quality/".format(dir=images_dir)
    for metric in utils.get_file_list_from_dir(images_dir, ext=settings.CHART_EXT, full_path=False):
        metric = metric.replace('.' + settings.CHART_EXT, '')
        metric_title = static_clustering.PARTITION_QUALITY_METRICS[metric]
        figures.append(
            "{dir}/{metric}".format(
                dir=images_dir,
                metric=metric,
            )
        )
        captions.append(metric_title)
    report.add_group_figure(
        title="Timeseries of partition qualities over time (x-axis - time, "
              "y-axis - quality)",
        images_list=figures,
        images_captions_list=captions,
        columns_count=min(2, len(figures)),
    )