import os
import re
from collections import defaultdict

import reports_utils
import compare_methods_external_criteria_report_utils
import compare_methods_against_other_report_utils
from .. import settings
from .. import utils
from ..dynamic_network import DynamicNetwork
from ..evaluation import eval_utils
from ..evaluation import method_eval
from ..evaluation import method_eval_across_networks
from ..latex_builder import LatexBuilder
from ..log import logger
from ..methods import methods_utils
from ..methods.naive import NaiveMethod
from ..parallel import CustomProcess, run_process_pool
from ..scoring_methods import scoring_methods_utils


def create_naive_method_analysis_reports(networks_ids, **kwargs):
    logger.debug("Creating analysis report for Naive methods in {networks_ids} networks".format(
        networks_ids=kwargs.get('networks_combined_name') or ", ".join(networks_ids)
    ))

    m_ids_lists = []
    for n_id in networks_ids:
       m_ids_lists.append([
        m_id for m_id in DynamicNetwork.get_network_computed_methods_ids_list(n_id)
        if NaiveMethod.is_corresponding_method_id(m_id)
    ])
    # find common methods
    m_ids_list = utils.get_intersection(*m_ids_lists)


    """
    Each Naive methods contains a scoring method or consensus clustering methods
    Methods are grouped by scoring method or consensus clustering methods
    (e.g., group by one to study the effect of another)
    If there are 5 scoring methods and 3 consensus clustering methods,
    and we, for example, group by scoring method,
    then each group figure has a row for each network and a column for each of 5 scoring methods,
    each subplot contains comparison of 3 methods that share scoring methods
    and have different consensus clustering methods.
    As it may happen that we have large numbers of parameters, to avoid group figure with large number of columns and
    subplots with large number of compared plots, both keys and values can be subdivided.
    E.g., If there are 50 scoring methods and 30 consensus clustering methods,
    and we, for example, group by scoring method,
    we can split keys into 5 groups and values into 3 groups - having 5*3=15 reports
    instead of one with each subfigure having 50 columns with subplots comparing 30 methods each.
    """
    naive_dict_temp = {
        'scoring_method': defaultdict(list),
        'consensus_clustering_method': defaultdict(list),
    }
    for i_m, m_id in enumerate(m_ids_list):
        m = methods_utils.create_method_by_id(m_id)
        naive_dict_temp['scoring_method'][m.get_scoring_method_id()].append(m_id)
        naive_dict_temp['consensus_clustering_method'][m.get_consensus_clustering_method_id()].append(m_id)

    naive_dict = {
        'scoring_method': defaultdict(list),
        'consensus_clustering_method': defaultdict(list),
    }
    for t,d in naive_dict_temp.iteritems():
        naive_dict[t] = {}
        for k,v in naive_dict_temp[t].iteritems():
            if len(v) < 2:
                continue
            # in case there are too many methods, one can subdivide them by based on some criteria
            # t + '_values_group_by_regexes' - when grouping by t, how to subdivide the methods for [t][k]
            values_group_by_options = kwargs.get(t + '_values_group_by_regexes', ['all'])
            if 'scoring_method_type' in values_group_by_options:
                types = set(type(methods_utils.create_method_by_id(x).get_scoring_method()).get_id_prefix() for x in v)
                values_group_by_options += ['scoring_method_type_' + tx for tx in types]
                values_group_by_options.remove('scoring_method_type')
            for v_reg in values_group_by_options:
                if v_reg == 'all':
                    v_selected = v
                elif v_reg.startswith('scoring_method_type_'):
                    v_selected = [
                        x for x in v
                        if type(methods_utils.create_method_by_id(x).get_scoring_method()).get_id_prefix()
                        == v_reg.replace('scoring_method_type_','')
                    ]
                else:
                    regex = re.compile(v_reg)
                    v_selected = [x for x in v if regex.search(x)]
                if len(v_selected) == 0:
                    continue
                if v_reg not in naive_dict[t]:
                    naive_dict[t][v_reg] = defaultdict(list)
                naive_dict[t][v_reg][k] = sorted(v_selected)
        if len(naive_dict[t]) == 0:
            del naive_dict[t]

    # create the required images, as necessary
    evaluation_processes = []
    for comparison_type, methods_ids_dict_grouped in naive_dict.iteritems():
        logger.debug("Evaluating methods for %s comparison type" % comparison_type)

        for group_key, methods_ids_dict in methods_ids_dict_grouped.iteritems():
            logger.debug("Using group key %s" % group_key)
            extra_scoring_methods = None
            # when fixing scoring method - evaluate against others as well (e.g., for score vs n_changepoints plots)
            # note: as we do filtering by key later on, it may result in too many scoring methods,
            # so we do additional filtering at the report level
            if comparison_type == 'scoring_method':
                extra_scoring_methods =[
                    scoring_methods_utils.create_scoring_method_by_id(s_id) for s_id in methods_ids_dict.keys()
                ]
            for k,m_ids in methods_ids_dict.iteritems():
                logger.info("Comparing {n_m} methods for '{k}': {m_ids}".format(
                    n_m=len(m_ids),
                    k=k,
                    m_ids=", ".join(m_ids),
                ))
                for i_n,n_id in enumerate(networks_ids):
                    logger.info("{i_n}/{n_n}. Computing methods comparison for '{n_id}' network".format(
                        i_n=i_n+1,
                        n_n=len(networks_ids),
                        n_id=n_id,
                    ))
                    p = CustomProcess(
                        name="EvalProc-%s" % (len(evaluation_processes) + 1),
                        target=method_eval.compare_methods,
                        kwargs={
                            'methods_ids': m_ids,
                            'network_id': n_id,
                            'comparison_methods_key': None,
                            'overwrite': kwargs.get('overwrite', False),
                            'extra_scoring_methods': extra_scoring_methods,
                        },
                    )
                    evaluation_processes.append(p)

                # across network comparison evaluation
                # e.g., to get results with avg/std across networks
                if len(networks_ids) > 1:
                    logger.info("Comparing {m_ids} methods for '{k}' across networks {n_n} networks".format(
                        m_ids=", ".join(m_ids),
                        k=k,
                        n_n=len(networks_ids),
                    ))
                    p = CustomProcess(
                        name="EvalProc-%s" % (len(evaluation_processes) + 1),
                        target=method_eval_across_networks.evaluate_methods_across_networks,
                        kwargs={
                            'methods_ids': m_ids,
                            'networks_ids': networks_ids,
                            'overwrite': kwargs.get('overwrite', False),
                        },
                    )
                    evaluation_processes.append(p)
    run_process_pool(evaluation_processes)

    for comparison_type, methods_ids_dict_grouped in naive_dict.iteritems():
        logger.debug("Creating reports for %s comparison type" % comparison_type)
        for v_group_key, methods_ids_dict in methods_ids_dict_grouped.iteritems():
            # in case there are too many methods, one can subdivide them by based on some criteria
            # t + '_keys_group_by_regexes' - when grouping by t, how to subdivide the keys for [t]
            keys_dict = {}
            keys_group_by_options = kwargs.get(comparison_type + '_keys_group_by_regexes', ['all'])
            if 'scoring_method_type' in keys_group_by_options:
                id_prefixes = set(
                    scoring_methods_utils.get_scoring_method_class_by_id(s_id).get_id_prefix()
                    for s_id in methods_ids_dict.keys()
                )
                keys_group_by_options += ['scoring_method_type_' + tx for tx in id_prefixes]
                keys_group_by_options.remove('scoring_method_type')
            for k_reg in keys_group_by_options:
                if k_reg == 'all':
                    k_selected = methods_ids_dict.keys()
                elif k_reg.startswith('scoring_method_type_'):
                    k_selected = [
                        x for x in methods_ids_dict.keys()
                        if scoring_methods_utils.get_scoring_method_class_by_id(x).get_id_prefix()
                        == k_reg.replace('scoring_method_type_','')
                    ]
                else:
                    regex = re.compile(k_reg)
                    k_selected = [x for x in methods_ids_dict.keys() if regex.search(x)]
                if len(k_selected) == 0:
                    continue
                keys_dict[k_reg] = sorted(k_selected)

            for k_group_key, ks in keys_dict.iteritems():
                methods_ids_dict_k_selected = { k: methods_ids_dict[k] for k in ks }
                for func in [
                    create_report_evaluation_external_criteria,
                    create_report_evaluation_against_others,
                ]:
                    func(
                        methods_ids_dict_k_selected,
                        networks_ids,
                        key_label=comparison_type,
                        copy_to_tmp_dir=kwargs.get('copy_to_tmp_dir', True),
                        networks_combined_name=kwargs.get('networks_combined_name'),
                        report_subdir_name=kwargs.get('report_subdir_name') or '%s_%s/' % (
                            len(networks_ids), utils.hash_string("|".join(networks_ids), size=5)
                        ),
                        k_group_key=k_group_key,
                        v_group_key=v_group_key,
                    )


def create_report_evaluation_external_criteria(methods_ids_dict, networks_ids, key_label, **kwargs):
    report = LatexBuilder(
        report_name='basic_evaluation',
        title="Basic analysis for Naive methods by {key_label} in {networks_combined_name} networks "
              "(keys - {k_group_key}, values - {v_group_key})".format(
            networks_combined_name=kwargs.get('networks_combined_name') or ", ".join(networks_ids),
            key_label=key_label.replace('_',' '),
            k_group_key=kwargs.get('k_group_key',' '),
            v_group_key=kwargs.get('v_group_key',' '),
        ),
    )

    compare_methods_external_criteria_report_utils.add_section_methods_comparison(
        report,
        networks_ids=networks_ids,
        methods_ids_dict=methods_ids_dict,
        key_label=key_label,
        **kwargs
    )

    report.compile(
        output_directory=settings.REPORTS_DIR + 'naive/' + kwargs.get('report_subdir_name', '')
                         + '/by_' + key_label
                         + '/k_as_' + kwargs.get('k_group_key',' ') + '/v_as_' + kwargs.get('v_group_key',' ') + '/',
        pdflatex=True
    )
    if kwargs.get('copy_to_tmp_dir'):
        reports_utils.copy_report_to_tmp_dir(
            report,
            tmp_subdir='naive/' + kwargs.get('report_subdir_name', '')+ '/by_' + key_label + '/',
            tmp_report_name=report.get_name()
                + '-k_as_' + kwargs.get('k_group_key',' ') + '-v_as_' + kwargs.get('v_group_key',' '),
        )

def create_report_evaluation_against_others(methods_ids_dict, networks_ids, key_label, **kwargs):
    # dict - other_method_id => list of comparisons for which it was evaluated against
    evaluation_against_others_dict = defaultdict(set)
    for i_n,n_id in enumerate(networks_ids):
        for k, m_ids in methods_ids_dict.iteritems():
            images_dir = DynamicNetwork.get_network_methods_comparison_dir(n_id)\
                         + eval_utils.generate_methods_comparison_key(m_ids) + '/images/'
            sub_dirs = utils.get_subdir_list_from_dir(images_dir + 'evaluation_against_other/')
            for sub_dir in sub_dirs:
                other_method_id = os.path.basename(sub_dir)
                evaluation_against_others_dict[other_method_id].add(k)
    if len(evaluation_against_others_dict.keys()) == 0:
        logger.warning("No evaluation against others")
        return

    for other_method_id, ks_set in evaluation_against_others_dict.iteritems():
        report = LatexBuilder(
            report_name='against-%s' % other_method_id,
            title="Evaluation against {other_method_id} for Naive methods by {key_label} "
                  "in {networks_combined_name} networks "
                  "(keys - {k_group_key}, values - {v_group_key})".format(
                networks_combined_name=kwargs.get('networks_combined_name') or ", ".join(networks_ids),
                other_method_id=methods_utils.get_method_title(other_method_id),
                key_label=key_label.replace('_',' '),
                k_group_key=kwargs.get('k_group_key',' '),
                v_group_key=kwargs.get('v_group_key',' '),
                include_per_network=kwargs.get('include_per_network', True),
                include_across_networks=kwargs.get('include_across_networks', True),
            ),
        )

        compare_methods_against_other_report_utils.add_section_evaluation_against_other(
            report=report,
            networks_ids=networks_ids,
            methods_ids_dict=methods_ids_dict,
            other_method_id=other_method_id,
            key_label=key_label,
            **kwargs
        )

        report.compile(
            output_directory=settings.REPORTS_DIR + 'naive/' + kwargs.get('report_subdir_name', '')
                             + '/by_' + key_label
                             + '/k_as_' + kwargs.get('k_group_key',' ') + '/v_as_' + kwargs.get('v_group_key',' ') + '/',
            pdflatex=True
        )
        if kwargs.get('copy_to_tmp_dir'):
            reports_utils.copy_report_to_tmp_dir(
                report,
                tmp_subdir='naive/' + kwargs.get('report_subdir_name', '')+ '/by_' + key_label + '/',
                tmp_report_name=report.get_name()
                                + '-k_as_' + kwargs.get('k_group_key',' ') + '-v_as_' + kwargs.get('v_group_key',' '),
            )