import os
from ..log import logger
from ..network import Network
import method_summary_report_utils
from ..latex_builder import LatexBuilder
from ..draw import dyn_network_draw
from ..evaluation import com_ev_eval
import reports_utils

def create_description_report(dynamic_network, **kwargs):
    logger.debug("Creating description report for '{network_id}' network".format(
        network_id=dynamic_network.get_id()
    ))

    # create the required images, as necessary
    dyn_network_draw.draw_dyn_network(dynamic_network, overwrite=kwargs.get('overwrite', False))
    for gt in dynamic_network.get_ground_truths_list():
        com_ev_eval.evaluate_community_evolution(gt, overwrite=kwargs.get('overwrite', False))

    report = LatexBuilder(
        report_name='summary',
        title="Summary report for '{network_id}' network".format(network_id=dynamic_network.get_id()),
    )
    add_sections_network_description(
        report,
        dynamic_network,
    )
    report.compile(output_directory=dynamic_network.get_reports_dir(), pdflatex=True)

    # for convenience, copy all reports to tmp dir
    if kwargs.get('copy_to_tmp_dir', True):
        tmp_output_dir = reports_utils.copy_report_to_tmp_dir(
            report,
            tmp_subdir='{network_id}/'.format(
                network_id=dynamic_network.get_id().replace('/','-'),
            )
        )
        # copy sankey diagrams as well
        for com_ev in dynamic_network.get_ground_truths_list():
            reports_utils.copy_sankey_partitions_to_tmp_dir(com_ev, tmp_output_dir)


def add_sections_network_description(report, dynamic_network, section_level=0, include_ground_truths=True):
    if dynamic_network.get_description() is not None:
        report.add_section("Description", level=section_level)
        report.add_text(dynamic_network.get_description())

    add_section_network_properties(report, dynamic_network)
    add_section_snapshot_images(report, dynamic_network)

    if include_ground_truths:
        for com_ev in dynamic_network.get_ground_truths_list():
            method_summary_report_utils.add_section_method_summary(
                report,
                com_ev,
                section_level=section_level or 0,
            )

def add_section_network_properties(report, dynamic_network, section_level=0):
    report.add_section(
        section_name="Network properties",
        level=section_level,
        clearPage=True,
    )
    figures = []
    captions = []
    for prop, prop_title in Network.NETWORK_PROPERTIES.iteritems():
        figures.append(
            "{dir}timeseries-snapshot-{prop}".format(
                dir=dynamic_network.get_images_dir() + 'network_properties/',
                prop=prop,
            )
        )
        captions.append(prop_title)
    report.add_group_figure(
        title="Network properties per snapshot (x-axis - time, y-axis - property)",
        images_list=figures,
        images_captions_list=captions,
        columns_count=min(3, len(figures)),
        )

def add_section_snapshot_images(report, dynamic_network, section_level=0):
    network_images_dir = dynamic_network.get_images_dir() + 'networks/'
    if os.path.exists(network_images_dir):
        report.add_section(
            section_name="Network images",
            level=section_level,
            clearPage=True,
        )
        figures = []
        captions = []
        for i in range(len(dynamic_network.get_timepoints())):
            figures.append(
                "{dir}{snapshot}".format(
                    dir=network_images_dir,
                    snapshot=i,
                )
            )
            captions.append("Snapshot at t=%s" % i)
        report.add_group_figure(
            title="Network images",
            images_list=figures,
            images_captions_list=captions,
            columns_count=min(3, len(figures)),
            max_images_per_figure=12,
        )
    else:
        logger.warning("Network '%s' does not have snapshot images" % dynamic_network.get_id())