import numpy as np
from ..scoring_method import ScoringMethod
from .. import consensus_clustering
from .. import settings


class AvgSegmentPartitionQuality(ScoringMethod):
    """
    Compute community evolution quality as average of consensus partition qualities on segments
    (i.e., compute quality of a consensus partition for each segment and take averages)
    """

    def __init__(self, consensus_partition_quality_metric='avg#modularity'):
        self._consensus_partition_quality_metric = consensus_partition_quality_metric

    def get_id(self):
        return '({prefix})({consensus_partition_quality_metric})'.format(
            prefix=AvgSegmentPartitionQuality.get_id_prefix(),
            consensus_partition_quality_metric=self.get_consensus_partition_quality_metric(),
        )

    @staticmethod
    def get_id_prefix():
        return 'avg_seg'

    @staticmethod
    def get_scoring_method_title(values):
        return "avg (weighted by length) of segment partition qualities according to {metric_title}".format(
            metric_title=consensus_clustering.get_consensus_partition_metric_description(values[1])
        )

    def get_limits(self):
        return consensus_clustering.get_consensus_partition_quality_metric_limits(
            self.get_consensus_partition_quality_metric()
        )

    def get_consensus_partition_quality_metric(self):
        return self._consensus_partition_quality_metric

    def evaluate_community_evolution(self, community_evolution, dynamic_network=None):
        if not community_evolution.has_partitions():
            return None
        return ScoringMethod.evaluate_community_evolution(self, community_evolution, dynamic_network)

    def evaluate(self, networks_list, segments_partitions, changepoints):
        if segments_partitions is None:
            return None
        changepoints_with_start_end = [0] + changepoints + [len(networks_list)]
        qualities = []
        weights = []
        for i in xrange(1, len(changepoints_with_start_end)):
            segment_networks = networks_list[changepoints_with_start_end[i-1]:changepoints_with_start_end[i]]
            qualities.append(
                consensus_clustering.compute_consensus_partition_quality(
                    networks_list=segment_networks,
                    partition=segments_partitions[i-1],
                    metric=self.get_consensus_partition_quality_metric(),
                )
            )
            weights.append(changepoints_with_start_end[i] - changepoints_with_start_end[i-1])
        return np.average(qualities, weights=weights)


