import numpy as np

from ..scoring_method import ScoringMethod
from .. import static_clustering
from .. import settings
from ..log import logger

class AvgSnapshotPartitionQuality(ScoringMethod):
    """
    Compute community evolution quality as average of snapshot partition qualities
    (i.e. for each snapshot, compute the quality of the corresponding segment partition
    and take average)
    """

    def __init__(self, partition_quality_metric='mod'):
        self._partition_quality_metric = partition_quality_metric

    def get_id(self):
        return '({prefix})({partition_quality_metric})'.format(
            prefix=AvgSnapshotPartitionQuality.get_id_prefix(),
            partition_quality_metric=self.get_partition_quality_metric(),
        )

    @staticmethod
    def get_id_prefix():
        return 'avg_sn'

    @staticmethod
    def get_scoring_method_title(values):
        return "$Q_P$ ({partition_quality_metric})".format(
            partition_quality_metric=static_clustering.PARTITION_QUALITY_METRICS[values[1]].lower().capitalize(),
        )

    def get_title(self):
        return AvgSnapshotPartitionQuality.get_scoring_method_title()

    def get_limits(self):
        return static_clustering.get_partition_quality_metric_limits(
            self.get_partition_quality_metric()
        )

    def get_partition_quality_metric(self):
        return self._partition_quality_metric

    def evaluate_community_evolution(self, community_evolution, dynamic_network=None):
        if not community_evolution.has_partitions():
            return None
        # if community_evolution.get_precomputed_score(self.get_id()):
        #     v = community_evolution.get_precomputed_score(self.get_id())
        #     logger.debug("Using precomputed %s score for %s: %s" % (self.get_id(), community_evolution.get_method_id(), v))
        #     #print "%s - %s" % (v, ScoringMethod.evaluate_community_evolution(self, community_evolution, dynamic_network))
        #     return v
        return ScoringMethod.evaluate_community_evolution(self, community_evolution, dynamic_network)

    def evaluate(self, networks_list, segments_partitions, changepoints):
        if len(segments_partitions) == 0:
            return None
        changepoints_with_start_end = [0] + changepoints + [len(networks_list)]
        qualities = []
        for i in xrange(1, len(changepoints_with_start_end)):
            segment_networks = networks_list[changepoints_with_start_end[i-1]:changepoints_with_start_end[i]]
            for s in segment_networks:
                qualities.append(
                    static_clustering.compute_partition_quality(
                        network=s,
                        partition=segments_partitions[i-1],
                        metric=self.get_partition_quality_metric(),
                    )
                )
        return np.average([q for q in qualities if q is not None])

