import numpy as np

from ..scoring_method import ScoringMethod
from .. import static_clustering
from .. import settings
from ..classification_result import ClassificationResult

class ClassificationResultAdapter(ScoringMethod):
    """
    An adapter method that computes Precision/Recall for ground truth
    """

    METRIC_TO_FUNC = {
        'aupr': ClassificationResult.get_aupr,
        'auroc': ClassificationResult.get_auroc,
        'max_f1_score': ClassificationResult.get_max_f1score,
    }
    def __init__(self, metric, ground_truth_method_id):
        if metric not in ClassificationResultAdapter.METRIC_TO_FUNC:
            raise Exception('Unknown classification metric: %s ' % metric)
        self._metric = metric
        self._ground_truth_method_id = ground_truth_method_id

    def get_id(self):
        return '({prefix})({metric})({ground_truth_method_id})'.format(
            prefix=ClassificationResultAdapter.get_id_prefix(),
            metric=self.get_metric(),
            ground_truth_method_id=self.get_ground_truth_method_id(),
        )

    @staticmethod
    def get_id_prefix():
        return 'cl_res'

    @staticmethod
    def get_scoring_method_title(values):
        return "community evolution Precision/Recall to {ground_truth_method_id} via {metric}".format(
            ground_truth_method_id=values[2],
            metric=values[1],
        )

    def get_limits(self):
        return (0,1)
    def get_metric(self):
        return self._metric

    def get_ground_truth_method_id(self):
        return self._ground_truth_method_id

    def evaluate_method(self, method, dynamic_network=None):
        dyn_network = dynamic_network or method.get_dynamic_network()
        gt_com_ev = dyn_network.get_ground_truth(self.get_ground_truth_method_id())
        gt_scores = [
                (j in gt_com_ev.get_changepoints())
                for j in dyn_network.get_timepoints()
        ]
        c = ClassificationResult(
            ground_truth_labels=gt_scores[1:],
            predicted_scores=method.get_timepoints_scores()[1:],
        )
        return ClassificationResultAdapter.METRIC_TO_FUNC[self.get_metric()](c)


