from partition import Partition
from other.louvain import modularity, best_partition
import numpy as np
import networkx as nx
from other.networkx_2_community_quality import coverage, performance

PARTITION_METHODS = {
    'opt_mod': "Optimal modularity",
    'fastgreedy': "Fast Modularity",
    'label_prop': "Label Propagation",
    'lead_eigenv': "Leading Eigenvector",
    'infomap': "Infomap",
    'walktrap': "Walktrap",
    'louvain': "Louvain",
    'stab_louvain': "Stabilized Louvain",
}
PARTITION_QUALITY_METRICS = { # higher better
    'mod': "Modularity",
    'log_l': "Log Likelihood for SBM",
    'cond_q': "1-Conductance",
    'norm_cut_q': "1-Normalized Cut",
    'avg_odf_q': "1-Average ODF", # ODF - Out Degree Fraction
    'perf_q': "1-Performance",
    'cov_q': "1-Coverage",
}
CLUSTER_SCORE_METRICS = { # lower better
    'cond': "Conductance",
    'norm_cut': "Normalized Cut",
    'expan': "Expansion",
    'avg_odf': "Average ODF", # ODF - Out Degree Fraction
}

def get_partition_quality_metric_limits(metric):
    if metric == 'log_l':
        return (None, 0)
    return (0, 1)

def compute_partition(network, method='louvain', **kwargs):
    if method not in PARTITION_METHODS:
        raise Exception("Unknown method {method}, should be one of {metrics}".format(
            method=method,
            metrics=PARTITION_METHODS.keys()
        )
    )
    if method == 'stab_louvain': # from http://hal.upmc.fr/file/index/docid/492058/filename/p508-aynaud.pdf
        init_partition = kwargs.get('init_partition', None)
        if init_partition is not None:
            cl_count = init_partition.get_clusters_count()
            i_cl = cl_count
            init_partition = init_partition.get_objects_clusters()
            for v in network.get_nodes():
                if v not in init_partition:
                    init_partition[v] = i_cl
                    i_cl += 1
        return Partition(
            best_partition(
                graph=network.get_nx_graph(),
                partition=init_partition
            )
        )

    iG = network.as_igraph()
    vertex_clustering = None
    edges_weights = 'weight' if 'weight' in iG.es.attribute_names() else None
    if method == 'opt_mod':
        vertex_clustering = iG.community_optimal_modularity(
            weights=edges_weights
        )
    elif method == 'fastgreedy':
        vertex_clustering = iG.community_fastgreedy(
            weights=edges_weights
        ).as_clustering(n=None)
    elif method == 'label_prop':
        vertex_clustering = iG.community_label_propagation(
            weights=edges_weights
        )
    elif method == 'lead_eigenv':
        vertex_clustering = iG.community_leading_eigenvector(
            clusters=None,
            weights=edges_weights
        )
    elif method == 'infomap':
        vertex_clustering = iG.community_infomap(
            edge_weights=edges_weights,
            vertex_weights=None,
            trials=10
        )
    elif method == 'walktrap':
        vertex_clustering = iG.community_walktrap(
            weights=edges_weights,
            steps=4
        ).as_clustering(n=None)
    elif method == 'louvain':
        vertex_clustering = iG.community_multilevel(
            weights=edges_weights,
            return_levels=False
        )

    if vertex_clustering is not None:
        return Partition(dict(zip(network.get_nx_graph().nodes(), vertex_clustering.membership)))
    return None


def compute_partition_quality(network, partition, metric='mod', **kwargs):
    if metric not in PARTITION_QUALITY_METRICS:
        raise Exception("Unknown metric {metric}, should be one of {metrics}".format(
            metric=metric,
            metrics=PARTITION_QUALITY_METRICS.keys()
        )
    )

    # if partition does not include all network nodes - using a subgraph
    if kwargs.get('use_partition_nodes_subgraph', True)\
            and any(v not in partition.get_objects_names() for v in network.get_nodes()):
        nx_graph = network.get_nx_graph().subgraph(partition.get_objects_names())
    else:
        nx_graph = network.get_nx_graph()

    if nx_graph.order() < 2 or nx_graph.size() < 1:
        return None

    if metric == 'mod': # modularity
        return modularity(partition.get_objects_clusters(), nx_graph, weighted=False)
        # using igraph's implementation (note: slower due to converting network format)
        # this implementation doesn't support edge weights (note: good for real-world networks, as they contain weights although methods ignore them)
        # return network.nx_graph_as_igraph(nx_graph).modularity(
        #     [int(partition.get_object_cluster(v)) for v in network.get_nodes()]
        # )
        # return modularity(partition.get_objects_clusters(), nx_graph)
    elif metric == 'log_l': # stochastic blockmodel log likelihood
        if partition.get_clusters_count() == 1:
            n_count = nx_graph.order()
            e_count = nx_graph.size()
            max_e_count = (n_count * (n_count - 1)) / 2
            p_edge = (1.0 * e_count) / max_e_count
            if e_count == 0 or e_count == max_e_count:
                if e_count == 0:
                    raise Exception("test")
                return 0.0
            return np.log(p_edge) * e_count + np.log(1.0 - p_edge) * (max_e_count - e_count)

        # based on http://tuvalu.santafe.edu/~aaronc/courses/5352/fall2013/csci5352_2013_L17.pdf
        mod_nx_graph = nx.Graph(e for e in nx_graph.edges(data=False) if e[0]!=e[1]) # unweighted graph without self-loops
        block_graph = nx.blockmodel(
            G=mod_nx_graph,
            partitions=partition.get_clusters_objects().values(),
            multigraph=False,
        )
        blocks = block_graph.nodes(data=True)
        res = 0.0
        for i, i_attr in blocks:
            for j, j_attr in blocks:
                if j > i:
                    break
                # number of edges between blocks
                if i==j:
                    e_ij = i_attr['nedges']
                else:
                    e_ij = block_graph[i][j].get('weight', 1.0) if block_graph.has_edge(i, j) else 0.0

                # max possible number of edges between blocks
                if i==j:
                    if not nx_graph.is_directed():
                        n_ij = i_attr['nnodes'] * (i_attr['nnodes'] - 1.0) / 2
                    else:
                        n_ij = i_attr['nnodes'] * (i_attr['nnodes'] - 1.0)  # no self-loops
                else:
                    n_ij = i_attr['nnodes'] * j_attr['nnodes']

                # Note: assume 0^0=1
                res += (e_ij * np.log(e_ij) if e_ij!=0 else 0.0) + \
                   ((n_ij - e_ij) * np.log(n_ij - e_ij) if n_ij - e_ij!=0 else 0.0) - \
                   (n_ij * np.log(n_ij) if n_ij!=0 else 0.0)
        return res
    elif metric == 'perf_q' or metric == 'cov_q': # see http://arxiv.org/pdf/0906.0612v2.pdf and http://arxiv.org/pdf/0906.0612v2.pdf
        if partition.get_clusters_count() < 2 or nx_graph.size() < 1:
            return 0.0
        func = performance if metric == 'perf_q' else coverage
        return 1.0 - func(nx_graph, partition.to_list_of_sets())
    elif metric.replace('_q', '') in CLUSTER_SCORE_METRICS:
        return 1.0 - _get_avg_partition_cluster_score(nx_graph, partition, metric.replace('_q', ''))

    return None

def _get_avg_partition_cluster_score(nx_graph, partition, metric='cond'):
    if partition.get_clusters_count() <= 1:
        return 1.0
    res = 0.0
    for c, c_nodes in partition.get_clusters_objects().iteritems():
        res += get_cluster_score(nx_graph, c_nodes, metric)
    res /= partition.get_clusters_count()
    return res

def get_cluster_score(nx_graph, cluster_nodes, metric='cond'):
    if metric not in CLUSTER_SCORE_METRICS:
        raise Exception("Unknown metric {metric}, should be one of {metrics}".format(
            metric=metric,
            metrics=CLUSTER_SCORE_METRICS.values()
        )
    )
    # http://cs.stanford.edu/people/jure/pubs/comscore-icdm12.pdf
    res = 0.0
    m = nx_graph.size()
    n_s = len(cluster_nodes)
    m_s = nx_graph.subgraph(cluster_nodes).size(weight ='weight')  # intra-cluster edges
    e_s = sum(nx_graph.degree(cluster_nodes, weight ='weight').values()) - m_s # all edges (note: intra edges are counted twice here)
    c_s = e_s - m_s  # inter-cluster edges
    if e_s < 1:
        return 1.0
    if metric == 'cond' or metric == 'norm_cut':
        # conductance is also here:
        # Community Structure of Complex Networks book or http://arxiv.org/pdf/0911.2308.pdf
        # network conductance - average departure probability
        res += c_s / (2 * m_s + c_s)
        if 2 * (m - m_s) + c_s == 0.0:
            return 1.0
        if metric == 'norm_cut':
            res += c_s / (2 * (m - m_s) + c_s)
    elif metric == 'expan':
        res += c_s / n_s # note: can be higher than 1
    elif metric == 'avg_odf':
        temp = 0.0
        for v in cluster_nodes:
            v_degree = nx_graph.degree(v)
            # note: if v not in network, v_degree is {}
            if v_degree == 0 or isinstance(v_degree, dict):
                continue
            v_edges_out = sum(
                    e[2].get("weight", 1.0) for e in nx_graph.edges(v, data=True)
                    if e[0] not in cluster_nodes or e[1] not in cluster_nodes
            )
            temp += (1.0 * v_edges_out) / v_degree
        temp /= n_s
        res += temp
    return res


def get_segment_log_l(segment_snapshots, partition):
    block_graphs = []
    for i in xrange(len(segment_snapshots)):
        block_graphs.append(_nx_blockmodel(
            G=segment_snapshots[i].get_nx_graph(),
            partitions=partition.get_clusters_objects().values(),
            multigraph=False,
            use_g_edge_weights=False,
            use_self_loops=False,
        )
    )

    # first compute MLE of stochastic block matrix theta over all snapshots in the segment
    thetas_m = np.zeros((partition.get_clusters_count(), partition.get_clusters_count()))
    thetas_n = np.zeros((partition.get_clusters_count(), partition.get_clusters_count()))
    for block_graph in block_graphs:
        blocks = block_graph.nodes(data=True)
        for i, i_attr in blocks:
            for j, j_attr in blocks:
                if j > i:
                    break
                # number of edges between blocks
                if i==j:
                    thetas_m[i,j] += i_attr['nedges']
                else:
                    thetas_m[i,j] += block_graph[i][j]['weight'] if block_graph.has_edge(i, j) else 0.0

                # max possible number of edges between blocks
                if i==j: #  assuming undirected graph
                   thetas_n[i,j] += i_attr['nnodes'] * (i_attr['nnodes'] - 1.0) / 2
                else:
                   thetas_n[i,j] += i_attr['nnodes'] * j_attr['nnodes']

    theta = np.zeros((partition.get_clusters_count(), partition.get_clusters_count()))
    for i in xrange(thetas_m.shape[0]):
        for j in xrange(0, i+1):
            theta[i,j] = theta[j,i] = (thetas_m[i,j] / thetas_n[i,j] if thetas_n[i,j] > 0 else 0.0)

    # sum log_l of snapshots using theta
    res = 0.0
    for block_graph in block_graphs:
        blocks = block_graph.nodes(data=True)
        for i, i_attr in blocks:
            for j, j_attr in blocks:
                if j > i:
                    break
                # number of edges between blocks
                if i==j:
                    e_ij = i_attr['nedges']
                else:
                    e_ij = block_graph[i][j]['weight'] if block_graph.has_edge(i, j) else 0.0

                # max possible number of edges between blocks
                if i==j: #  assuming undirected graph
                    n_ij = i_attr['nnodes'] * (i_attr['nnodes'] - 1.0) / 2
                else:
                    n_ij = i_attr['nnodes'] * j_attr['nnodes']

                # Note: assume 0^0=1
                res += e_ij * (np.log(theta[i, j]) if theta[i, j] != 0 else 0) + (n_ij - e_ij) * (np.log(1 - theta[i, j]) if theta[i, j] != 1 else 0)
    return res


def get_snapshot_log_l(snapshot, partition, theta):
    """
    Compute Log-Likelihood of snapshot given partition and stochastic block matrix theta
    :param snapshot:
    :param partition:
    :param theta:
    :return:
    """
    block_graph = _nx_blockmodel(
        G=snapshot.get_nx_graph(),
        partitions=partition.get_clusters_objects().values(),
        multigraph=False,
        use_g_edge_weights=False,
        use_self_loops=False,
    )
    blocks = block_graph.nodes(data=True)
    res = 0.0
    for i, i_attr in blocks:
        for j, j_attr in blocks:
            if j > i:
                break
            # number of edges between blocks
            if i==j:
                e_ij = i_attr['nedges']
            else:
                e_ij = block_graph[i][j].get('weight', 1.0) if block_graph.has_edge(i, j) else 0.0

            # max possible number of edges between blocks
            if i==j: #  assuming undirected graph
                n_ij = i_attr['nnodes'] * (i_attr['nnodes'] - 1.0) / 2
            else:
                n_ij = i_attr['nnodes'] * j_attr['nnodes']

            # Note: assume 0^0=1
            res += e_ij * (np.log(theta[i, j]) if theta[i, j] != 0 else 0) + (n_ij - e_ij) * (np.log(1 - theta[i, j]) if theta[i, j] != 1 else 0)
    return res


# from networkx -- added parameter "use_g_edge_weights" to ignore weights in G (otherwise, need to create an unweighted copy of G)
# also added "use_self_loops" to ignore self loops
def _nx_blockmodel(G,partitions,multigraph=False, use_g_edge_weights=True, use_self_loops=True):
    """
    NOTE: modified function blockmodel from networkx to add option for ignoring weights in G

    Returns a reduced graph constructed using the generalized block modeling
    technique.

    The blockmodel technique collapses nodes into blocks based on a
    given partitioning of the node set.  Each partition of nodes
    (block) is represented as a single node in the reduced graph.

    Edges between nodes in the block graph are added according to the
    edges in the original graph.  If the parameter multigraph is False
    (the default) a single edge is added with a weight equal to the
    sum of the edge weights between nodes in the original graph
    The default is a weight of 1 if weights are not specified.  If the
    parameter multigraph is True then multiple edges are added each
    with the edge data from the original graph.

    Parameters
    ----------
    G : graph
        A networkx Graph or DiGraph
    partitions : list of lists, or list of sets
        The partition of the nodes.  Must be non-overlapping.
    multigraph : bool, optional
        If True return a MultiGraph with the edge data of the original
        graph applied to each corresponding edge in the new graph.
        If False return a Graph with the sum of the edge weights, or a
        count of the edges if the original graph is unweighted.

    Returns
    -------
    blockmodel : a Networkx graph object

    Examples
    --------
    >>> G=nx.path_graph(6)
    >>> partition=[[0,1],[2,3],[4,5]]
    >>> M=nx.blockmodel(G,partition)

    References
    ----------
    .. [1] Patrick Doreian, Vladimir Batagelj, and Anuska Ferligoj
    	"Generalized Blockmodeling",Cambridge University Press, 2004.
    """
    # Create sets of node partitions
    part=list(map(set,partitions))

    # Check for overlapping node partitions
    u=set()
    for p1,p2 in zip(part[:-1],part[1:]):
        u.update(p1)
        #if not u.isdisjoint(p2):  # Python 2.6 required
        if len (u.intersection(p2))>0:
            raise nx.NetworkXException("Overlapping node partitions.")

    # Initialize blockmodel graph
    if multigraph:
        if G.is_directed():
            M=nx.MultiDiGraph()
        else:
            M=nx.MultiGraph()
    else:
        if G.is_directed():
            M=nx.DiGraph()
        else:
            M=nx.Graph()

    # Add nodes and properties to blockmodel
    # The blockmodel nodes are node-induced subgraphs of G
    # Label them with integers starting at 0
    for i,p in enumerate(part):
        M.add_node(i)
        # The node-induced subgraph is stored as the node 'graph' attribute
        SG=G.subgraph(p)
        M.node[i]['graph']=SG
        M.node[i]['nnodes']=SG.number_of_nodes()
        M.node[i]['nedges']=SG.number_of_edges()
        M.node[i]['density']=nx.density(SG)

        # ADDED
        if not use_self_loops:
            M.node[i]['nedges'] -= SG.number_of_selfloops()


    # Create mapping between original node labels and new blockmodel node labels
    block_mapping={}
    for n in M:
        nodes_in_block=M.node[n]['graph'].nodes()
        block_mapping.update(dict.fromkeys(nodes_in_block,n))

    # Add edges to block graph
    for u,v,d in G.edges(data=True):
        # ADDED
        if not use_g_edge_weights:
            d['weight'] = 1.0

        bmu=block_mapping[u]
        bmv=block_mapping[v]
        if bmu==bmv: # no self loops
            continue
        if multigraph:
            # For multigraphs add an edge for each edge in original graph
            M.add_edge(bmu,bmv,attr_dict=d)
        else:
            # For graphs and digraphs add single weighted edge
            weight=d.get('weight',1.0) # default to 1 if no weight specified
            if M.has_edge(bmu,bmv):
                M[bmu][bmv]['weight']+=weight
            else:
                M.add_edge(bmu,bmv,weight=weight)
    return M