import argparse
from core.log import logger, log_func_progress
from core.utils import parse_user_args_string_into_values_list
from collections import OrderedDict
from core.dynamic_network import DynamicNetwork
from core.reports import (
    network_summary_report_utils,
    method_summary_report_utils,
    compare_methods_report_utils,
    naive_analysis_report_utils,
    compare_methods_networks_dict_report_utils,
)
from core.report_params import (
    REPORT_METHODS_BY_KEY,
    REPORT_NETWORKS_DICT_BY_KEY
)

@log_func_progress
def main():
    parser = argparse.ArgumentParser(description="Create network analysis report")
    parser.add_argument('report_type', help="Report type")
    parser.add_argument('network_id', help="Network id", nargs='?', default=None)
    parser.add_argument('--overwrite', action='store_true', help="Overwrite existing files during evaluation")
    args = parser.parse_args()
    report_types_list = parse_user_args_string_into_values_list(args.report_type)
    network_ids_list = parse_user_args_string_into_values_list(args.network_id) if args.network_id else None

    processed_report_types_list = []
    for i_r, report_type in enumerate(report_types_list):
        if report_type.startswith('n_'):
            logger.debug("Processing report type %s" % report_type)
            values = report_type.split('_')
            n_key = values[1]
            m_key = values[2]
            if m_key not in REPORT_METHODS_BY_KEY:
                logger.warning("Unknown methods key: %s" % m_key)
                continue
            if n_key not in REPORT_NETWORKS_DICT_BY_KEY:
                logger.warning("Unknown networks dict key: %s" % n_key)
                continue
            networks_dict = OrderedDict()
            for n_k, n_ids_str in REPORT_NETWORKS_DICT_BY_KEY.get(n_key).iteritems():
                networks_dict[n_k] = parse_user_args_string_into_values_list(n_ids_str)

            methods_ids = []
            for temp_m in REPORT_METHODS_BY_KEY.get(m_key):
                methods_ids += parse_user_args_string_into_values_list(temp_m)
            compare_methods_networks_dict_report_utils.create_compare_methods_networks_dict_report(
                networks_dict=networks_dict,
                methods_ids=methods_ids,
                report_subdir_name='n_%s-m_%s' % (n_key, m_key),
                methods_key_name=m_key,
                networks_key_name=n_key,
                overwrite=args.overwrite,
                run_eval=True,
                show_ground_truth=False,
                draw_com_evs_scores=True,
                draw_run_time=True,
                draw_against_other=True,
                draw_com_evs_sim=False,
                draw_changepoints_pr_roc=True,
                draw_per_method=False,
                draw_per_network_key=False,
                draw_per_measure=True,
            )
            processed_report_types_list.append(report_type)

    report_types_list =[r for r in report_types_list if r not in processed_report_types_list]
    if len(report_types_list) > 0 and network_ids_list is None:
        raise Exception("No networks id for analysis")

    if 'naive' in report_types_list:
        naive_analysis_report_utils.create_naive_method_analysis_reports(
            networks_ids=network_ids_list,
            networks_combined_name=args.network_id,
            report_subdir_name=args.network_id.replace('/','-') if len(args.network_id) < 50 else None,
            # note: values and keys grouped by regex are opposite
            consensus_clustering_method_keys_group_by_regexes=['all'],
            scoring_method_keys_group_by_regexes=['scoring_method_type'],
            consensus_clustering_method_values_group_by_regexes=['scoring_method_type'],
            scoring_method_values_group_by_regexes=['all'],
            overwrite=args.overwrite,
        )
        report_types_list.remove('naive')
    for i_r, report_type in enumerate(report_types_list):
        if report_type.startswith('k_'):
            k = report_type[2:]
            if k not in REPORT_METHODS_BY_KEY:
                logger.warning("Unknown methods key: %s" % k)
                report_types_list.remove(report_type)
                continue

            methods_ids = []
            for temp_m in REPORT_METHODS_BY_KEY.get(k):
                methods_ids += parse_user_args_string_into_values_list(temp_m)

            compare_methods_report_utils.create_compare_methods_report(
                networks_ids=network_ids_list,
                methods_ids=methods_ids,
                report_subdir_name=k,
                methods_key_name=k,
                networks_key_name=args.network_id.replace('/','-') if len(args.network_id) < 50 else None,
                overwrite=args.overwrite,
                include_per_network=True,
                include_across_networks=True,
            )
            report_types_list.remove(report_type)

    if len(report_types_list) == 0:
        return

    for i_n, network_id in enumerate(network_ids_list):
        logger.info("{i_n}/{n_n}. Generating report for '{network_id}' network".format(
            i_n=i_n+1,
            n_n=len(network_ids_list),
            network_id=network_id.upper(),
        ))
        dyn_network = DynamicNetwork.load(network_id)
        logger.info(dyn_network.get_summary())
        for i_r, report_type in enumerate(report_types_list):
            logger.info("{i_r}/{n_r}. Generating '{report_type}' ".format(
                i_r=i_r+1,
                n_r=len(report_types_list),
                report_type=report_type.upper(),
            ))
            if report_type == 'network':
                network_summary_report_utils.create_description_report(
                    dyn_network,
                    copy_to_tmp_dir=True,
                    overwrite=args.overwrite,
                )
            elif report_type == 'methods':
                method_summary_report_utils.create_network_methods_description_reports(
                    dyn_network,
                    copy_to_tmp_dir=True,
                    overwrite=args.overwrite,
                )
            elif report_type.startswith('m_'):
                method_summary_report_utils.create_network_methods_description_reports(
                    dyn_network,
                    methods_ids_list=[report_type.replace('m_', '')],
                    copy_to_tmp_dir=True,
                    overwrite=args.overwrite,
                )
            elif report_type.startswith('r_'):
                r = report_type[2:]
                compare_methods_report_utils.create_compare_methods_by_regex_report(
                    dyn_network,
                    method_id_regex=report_type[r],
                    methods_key_name=r,
                    overwrite=args.overwrite,
                )
            else:
                logger.warning("Unknown report type: %s" % report_type)

if __name__ == '__main__':
    main()