import argparse

from core.log import logger, log_func_progress
from core.methods.methods_utils import create_method_by_id
from core.dynamic_network import DynamicNetwork
from core.utils import parse_user_args_string_into_values_list
from core.evaluation import method_eval

@log_func_progress
def main():
    parser = argparse.ArgumentParser(description="Run method on a sequence of networks")
    parser.add_argument('network_id', help="Network id")
    parser.add_argument('method_id', help="Method id")
    parser.add_argument('--overwrite', action='store_true', help="Overwrite existing files during computation")
    parser.add_argument('--evaluate', action='store_true', help="Run methos evaluation")
    args = parser.parse_args()

    network_id_list = parse_user_args_string_into_values_list(args.network_id)
    method_id_list = parse_user_args_string_into_values_list(args.method_id)

    logger.info("Using {n_n} network{plural}: {networks}".format(
        n_n=len(network_id_list),
        plural="s" if len(network_id_list)!=1 else "",
        networks=', '.join(network_id_list),
    ))
    logger.info("Running {n_m} method{plural}: {methods}".format(
        n_m=len(method_id_list),
        plural="s" if len(method_id_list)!=1 else "",
        methods=', '.join(method_id_list),
    ))

    for i_n, network_id in enumerate(network_id_list):
        logger.info("{i_n}/{n_n}. Processing '{network_id}' network".format(
            i_n=i_n+1,
            n_n=len(network_id_list),
            network_id=network_id.upper(),
        ))
        dynamic_network = DynamicNetwork.load(network_id)
        logger.info(dynamic_network.get_summary())

        for i_m, method_id in enumerate(method_id_list):
            logger.info(
                "{i_m}/{n_m}. Applying '{method_id}' method to '{network_id}' network".format(
                    i_m=i_m+1,
                    n_m=len(method_id_list),
                    method_id=method_id.upper(),
                    network_id=network_id.upper(),
                )
            )

            # temporary
            if "multistep" in method_id and not args.overwrite:
                import shutil, os
                if os.path.exists(DynamicNetwork.get_network_methods_dir(network_id) + '_shared/multistep'):
                    shutil.rmtree(DynamicNetwork.get_network_methods_dir(network_id) + '_shared/multistep')
                if os.path.exists(DynamicNetwork.get_network_methods_dir(network_id) + method_id):
                    shutil.rmtree(DynamicNetwork.get_network_methods_dir(network_id) + method_id)
                logger.warning("Deleting multistep folder")
            if ("cons_matr" in method_id or "avg_mod" in method_id) and not args.overwrite:
                import shutil, os
                if os.path.exists(DynamicNetwork.get_network_methods_dir(network_id) + 'aux'):
                    shutil.rmtree(DynamicNetwork.get_network_methods_dir(network_id) + 'aux')
                logger.warning("Deleting aux folder")

            method = create_method_by_id(method_id)
            method.run(
                dynamic_network,
                overwrite=args.overwrite,
            )
            method.save_run_stats()

            if args.evaluate:
                method_eval.evaluate_method(method, overwrite=args.overwrite)

if __name__ == '__main__':
    main()