import numpy as np
import networkx as nx
from core.network import Network
from core.partition import Partition
from core import static_clustering
from core.log import logger


# check log likelihood partition quality
# use example from http://tuvalu.santafe.edu/~aaronc/courses/5352/fall2013/csci5352_2013_L17.pdf
G = nx.Graph()
G.add_edges_from([
    (0, 1),
    (0, 2),
    (1, 2),
    (1, 3),
    (3, 4),
    (3, 5),
    (4, 5),
])
logger.info("Testing stochastic blockmodel log likelihood computation")
network = Network(nx_graph=G)
test_cases = {
    'log_l': [
        ({0: [0, 1, 2], 1: [3, 4, 5]}, -3.1395),
        ({0: [0, 1, 2, 3], 1: [4, 5]}, -8.3178),
    ],
    'cond_q': [
        ({0: [0, 1, 2], 1: [3, 4, 5]}, 0.8571),
        ({0: [0, 1, 2, 3], 1: [4, 5]}, 0.6500),
    ],
}
for metric, t in test_cases.iteritems():
    logger.info("Testing metric %s" % metric)
    for i,c in enumerate(t):
        logger.info("Testing case %s" % i)
        q_test = static_clustering.compute_partition_quality(
            network=network,
            partition=Partition.from_clusters_objects(c[0]),
            metric=metric
        )
        logger.debug("Test result for partition %s: %.4f" % (c[0], q_test))
        assert(round(q_test, 4) == c[1])


# alternative log_l
for i,c in enumerate(test_cases['log_l']):
    p = Partition.from_clusters_objects(c[0])
    q_test = static_clustering.get_segment_log_l([network], p)
    logger.debug("Test result for partition %s: %.4f" % (c[0], q_test))
    assert(round(q_test, 4) == c[1])


# compare naive python vs cpp
from core.scoring_methods import scoring_methods_utils
from core.utils import parse_user_args_string_into_values_list
from core.method import Method
from core.com_evs_similarity_methods import com_evs_similarity_methods_utils
sc =  scoring_methods_utils.create_scoring_method_by_id("(bic)")
sim = com_evs_similarity_methods_utils.create_com_evs_similarity_method_by_id("(gl_p)(nmi)")
errors = []
for n_id in parse_user_args_string_into_values_list(
        'synthetic_16_50/sbm-rand.[1|2|4|8|16].16.[1-10]_50_0-20-4-0-0,'
        'synthetic/sbm-rand.[1|2|4|8|16].16.[1-10]_100_0-20-4-0-0,'
        'synthetic_16_500/sbm-rand.[1|2|4|8|16].16.[1-10]_500_0-20-4-0-0,'
        'synthetic/sbm-rand.[1|2|4|8|16].16.[1-10]_1000_0-20-4-0-0'
    ):
    m1 = Method.load_state(n_id, 'cppnaive-walktrap-bic')
    m2 = Method.load_state(n_id, 'naive-u^w^walktrap-(bic)-bu')
    print n_id
    c1 = m1.get_community_evolution()
    c2 = m2.get_community_evolution()
    sc1 = sc.evaluate_community_evolution(c1)
    sc2 = sc.evaluate_community_evolution(c2)
    print "Solution: %s - %s" % (sc1, sc2)
    print "Similarity: %s" % (sim.compute_similarity(c1, c2))
    try:
        np.testing.assert_almost_equal(sc1, sc2)
    except:
        errors.append(
            "%s (final solution): %s vs %s (similarity - %s)" % (n_id, sc1, sc2, sim.compute_similarity(c1, c2))
        )
        logger.error(errors[-1])
    for i in xrange(16):
        c1 = m1.get_solution_for_n_changepoints(i)
        c2 = m2.get_solution_for_n_changepoints(i)
        sc1 = sc.evaluate_community_evolution(c1)
        sc2 = sc.evaluate_community_evolution(c2)
        print "%s: %s - %s" % (i, sc1, sc2)
        print "Similarity: %s" % (sim.compute_similarity(c1, c2))
        try:
            np.testing.assert_almost_equal(sc1, sc2)
        except:
            errors.append(
                "%s (%s changepoints): %s vs %s (similarity - %s)" % (n_id, i, sc1, sc2, sim.compute_similarity(c1, c2))
            )
            logger.error(errors[-1])
if len(errors) == 0:
    logger.info("No errors")
else:
    logger.warning("Errors: %s" % len(errors))
    for er in errors:
        logger.warning(er)